const PATHTYPE_ANNE = 1;
const PATHTYPE_UDF = 2;
const RENDERTYPE_BASE = 1;
const RENDERTYPE_STANDARD = 2; 
const RENDERTYPE_FLOATER = 3;
const RENDERTYPE_TAB = 4;
const RENDERTYPE_MODAL = 5; 
const RENDERTYPE_INJECT = 6;

const TYPE_NAME = -1;

const TYPE_SYS = 0;             //Sytem info for neuron (sponsored_by, annother, owned_by).  Builds extra indexes.
const TYPE_BE = 1;              //semantic shothand for BE DIMENSION.  CLASS/INST RELNS create a DAG.
const TYPE_PARAM = 2;           //semantic shorthand for param (class attribute/property)
const TYPE_HAS = 3;             //semantic shorthand for param fulfilling has value (attribute/property)
const TYPE_AWARENESS = 4;       //for awareness space (emotions/feelz)

const TYPE_EXPERIENCE = 5;      //for experience (temporal) neuron space

//CLUMPS can function as a TYPE_BE, TYPE_HAS, OR TYPE_SYS.
//those Types are actually short-hand for clumps but they can get away with
//only using one TX to specify the triplet info.
//A full clump is for specifying more than the standard 3 pieces of semantic info
//in a single ANNE TX.
//** For ease of use we can put a TYPE_XXX in the clump
//to clue our interpreter to what dimension the clump lives in.
//The other CLUMPS of episodic, BRAIN and EXPERIENCE are already their own dimension
//so this is only relevant for semantic clumps that are handled in semantic memory space.
const TYPE_CLUMP = 6;

const TYPE_SYS_MSG = 7;        //only relevant to coms, not the neuron
const TYPE_BRAIN = 8;          //for neural-net space wirings a->wire->b
const TYPE_ROLE = 9;           //for specifiying  semantic roles inside of a clump

const TYPE_METAN = 10;         //for specifying meta information about roles used in a clump

const TYPE_EPISODIC = 11;      //for documenting a self-sensory/vitals data at run-time in a clump.
const TYPE_INTERNAL_PAY = 17;  //keyless neurons paying out.
const TYPE_BROADCAST_PAY = 18; //standard broadcast payment.

const TYPE_BRAIN0 = 20;        //20+ are for sequence in the brain dimension

//100+ are available for brain type/custom use.. assumed to be order if more than one in a neuron's outs.
const TYPE_BRAIN120 = 120;     //used in ASDF-> annode relons to store trigger pats. (can use how u like in other neurons)
const TYPE_BRAIN121 = 121;     //myelination measure (a->b->[to_type_number]#c)

const TO_TYPE_SEED = 0;        //relon to is an encrypted seed (all seed are encrypted)
const TO_TYPE_ANNE = 1;        //relon to is a .nid, the idenfier of another neuron.  This is a legit graph triplet.
const TO_TYPE_TXID = 2;        //relon to is a txid.
const TO_TYPE_URL = 3;         //relon to is a url.

//relon to is a number.  Note: our ANNODE stores numerics in a LONG field
//if a non-long/int is needed use the TO_TYPE_STRING

const TO_TYPE_NUMBER = 4;       
const TO_TYPE_DATE = 5;         //relon to is a date in the ANNE date format.    
const TO_TYPE_DATETIME = 6;     //relon to is a datetime in the ANNE datetime format.
const TO_TYPE_STRING = 7;       //relon to is a string

const TO_TYPE_TEXT = 7;         //alias in case u conflate

const TO_TYPE_BLOB = 8;         //relon to is a blob (bad donkey, not welcome in ANNE)
const TO_TYPE_TIME = 9;         //relon to is a time in the ANNE time format.
const TO_TYPE_BOOLEAN = 10;     //relon to is a boolean string which will be converted to true/false at runtime
const TO_TYPE_FILEHASH = 11;    //relon to is a filehash
const TO_TYPE_IPV4 = 12;    //relon to is a IPV4 address
const TO_TYPE_IPV6 = 13;    //relon to is a IPV6 address

//encrypted versions of the above + 30 in value from its respective type.
//expansion to 30 base types possible (don't foresee this as needed
//in a perfect-brain world, all are references to either neurons or other TXIDs

const TO_TYPE_ENCR_ANNE = 31;                
const TO_TYPE_ENCR_TXID = 32;
const TO_TYPE_ENCR_URL = 33;        
const TO_TYPE_ENCR_NUMBER = 34;
const TO_TYPE_ENCR_DATE = 35;
const TO_TYPE_ENCR_DATETIME = 36;
const TO_TYPE_ENCR_STRING = 37; 
const TO_TYPE_ENCR_BLOB = 38;
const TO_TYPE_ENCR_TIME = 39;
const TO_TYPE_ENCR_BOOLEAN = 40;
const TO_TYPE_ENCR_FILEHASH = 41;
const TO_TYPE_ENCR_IPV4 = 42;
const TO_TYPE_ENCR_IPV6 = 43;

const NTYPE_CLASS = 0;
const NTYPE_INST = 1;
const NTYPE_UNKNOWN = 2;
const NTYPE_CLUMP = 3;

const ORIGIN_CHAIN = 0;         //a real TX on-chain
const ORIGIN_NODE = 1;          //an emulated neuron created by the node
const ORIGIN_NODE_PRELOAD = 2;  //an emulated neuron preloaded by the node (with a real seed)
const ORIGIN_OWNERPROOF = 3;    
const ORIGIN_CONTRACTS = 4;     //emulated from contracts
const ORIGIN_KEYLESS = 5;
const ORIGIN_GLOBAL = 6;
const ORIGIN_CUSTODIAL  = 8;
const ORIGIN_EXTERNAL_ML = 12;  //if a ML is joined w/ the DB, that data is ML.

const RORIGIN_ANNODE = 0;         //origination made on the ANNODE in ANNE hypergraph only
const RORIGIN_BROADCAST = 1;    //origination broadcast as a TX
const RORIGIN_COLLAB_BROADCAST = 2;        //collab broadcast (waiting approval)
const RORIGIN_GHOST = 3;        //origination user local

const ANNE_ACCTID = "7777770163980908667";    
const ANNE_ADMIN_ACCTID = "7777771543345157118";
const ANNE_ADMIN_BACKUP_ACCTID = "7777044340734581180";

const TARGET_NID = 1;
const TARGET_TXID = 2;

const TXMODE_GHOST = 0;             //store to local DB only
const TXMODE_ANNE = 1;              //instant send

const CTYPE_COLLAB_ONLY = 0;
const CTYPE_COLLAB_HANAI = 1;
const CTYPE_PENDING_ANJ = 2;
const CTYPE_HANAI_ONLY = 3;
const CTYPE_MERGE_REQUEST = 4;

//Keyless neurons allow for easy creation of child relons w/o the management of another key
//They can be collaborated w/ by request or easy delegation rights set to other approved authors.
//Note: ANNE globals (date, numbers, locations, keywords) are keyless neurons and are created on-demand.
const NEWNEURONMODE_KEYLESS = 0;  

//Key'd neurons provide real a real/key that is only known by the identity that created it.
//Key'd neurons can be collaborated w/ by request but they cannot be delegated since only the
//Key holder can broadcast transactions from the key'd neuron.
//Note: ANNE preloaded neurons are key'd as they need pre-known neuron IDs to reference them in the base-protocol.
const NEWNEURONMODE_KEYED = 1;        

const NEURON_NOT_FROZE = 0;
const NEURON_FROZE = 1;

const LINKS_NEURON_MODAL = 1;            //quick neuron summary page
const LINKS_NEURON_EMBED = 2;            //full neuron summary page
const LINKS_NEURON_DS = 3;               //data-source view of neuron's BE dimension
const LINKS_NEURON_TBL_VIEW = 4;         //table view of neuron's relons
const LINKS_NEURON_TREE = 5;             //tree view of a neuron's BE dimension
const LINKS_NEURON_ANNET = 6;            //graph explore of neurons
const LINKS_NEURON_ANDY = 7;             //ADMIN AREA
const LINKS_NEURON_ALEXANNDRIA = 8;      //library of your neurons  (can also find others)

const LINKS_NEURON_USER = 10;            //user profile page.
const LINKS_NEURON_INJECT_CB = 11;       //injection link w/ callback
const LINKS_NEURON_BLOCKCHAIN = 12;      //a link to the 3d viewer of blockchain data.
const LINKS_NEURON_DOSSIER = 14;         //detailed dossier view
const LINKS_NEURON_BRAIN = 16;           //decide if rememberz/forgetz is on/off
const LINKS_NEURON_ANNULET_OF_TRUST = 17;

const LINKS_RELON_ANNE = 20;             //an ANNE relon of somekind.
const LINKS_TX_BLOCKEXPLORER = 21;       //not an ANNE relon, a standard blockchain TX

//these mirror the value sin the node L1 code.
const BLOCK_HEADER_LENGTH = 232; //TODO: this may change if we add sm in the block...
const MAX_BALANCE_ANNE = 2200000000000; 
const ONE_ANNE = 100;
const MIN_ANNIES_SEND = 1000;                   //not enforced.               
const DELETE_WHEN_MINER_FIXED = ONE_ANNE;
const MAX_BALANCE_NQT = MAX_BALANCE_ANNE * ONE_ANNE;
const INITIAL_BASE_TARGET = 38650387592;        //irrelevant
const BLOCK_REWARD_START = 1000_000;
const BLOCK_REWARD_CYCLE = 21240;
const BLOCK_REWARD_CYCLE_PERCENTAGE = 99;
const BLOCK_REWARD_18500_HEIGHT = 8453521;    //mining reward sunset mode 
const BLOCK_REWARD_SUNSET_AMOUNT = 18500;
const BLOCK_REWARD_LIMIT_HEIGHT = 14655601;   //mining reward end block 
const BLOCK_REWARD_LIMIT_AMOUNT = 0;            //0 reward after that
const SUPER_CYCLE = BLOCK_REWARD_CYCLE * 13;    //every 13 cycles is a super block.
const BLOCK_REWARD_SOLO_CUT = 51;
const BLOCK_REWARD_SHARE_CUT = 0.07;             //7% of the block reward.
const BLOCK_REWARD_SHARE_BROADCAST_FEE = 0.02;   //2% of the 7% cut.

const MAINNET = "mainnet";
const TESTNET = "testnet";

const MEDIA_AUDIT_NOTRUN = 0;
const MEDIA_AUDIT_PASSED = 1;
const MEDIA_AUDIT_PASSED_W_NEWDATA = 2;
const MEDIA_AUDIT_MISSING_DATA = -2; //missing data
const MEDIA_AUDIT_CONFLICTING_TYPE = -3; 
const MEDIA_AUDIT_CONFLICTING_SIZE = -4;
const MEDIA_AUDIT_FINGER_DOES_NOT_MATCH = -5;

const FILEDATASTATUS_NOT_SET = -1;
const FILEDATASTATUS_REQUEST_ERROR = 6;
const FILEDATASTATUS_ACQUIRED = 7;
const FILEDATASTATUS_PENDING_ANTOR = 277;
const FILEDATASTATUS_CANNOT_INIT_ANTOR = 287;
const FILEDATASTATUS_ANTOR_OFF = 297;

const QT_PERCENT = 0;
const QT_SPECIFIC = 1;
const SRC_FUND = 0;
const SRC_OWNER = 1;
const SP_MIN_PAYMENT = 1_000_000;  //new min, 1m annies (10k annecoin)


/************** (from file ./js/base/consts.js)**/ 

class anne_singleton {
    constructor(){
        this._bINIT = false;
    }
    set_init = function ($REF) {
        this._bINIT=true;
        Object.freeze($REF);
    }
    is_init = function () {
        if (this._bINIT==true) return true;
        return false;
    }
}

/************** (from file ./js/base/anne_singleton.js)**/ 

/**
 * Basically, an array of objects, and a corresponding map to help manage their
 * addition/tracking/grouping
 */
class arraymap {
    constructor($mapKey="") {
        this.arr = [];
        this.map = new Object();
        this.mapKey = $mapKey;
        this.arr_other = [];

        this.secondary_mapKey = "";
        this.secondary_map = null;
        this.secondary_mapKey_forceLowerCase = false;
    }

    //good enough...
    copy = function($arraymap) {
        this.mapKey = $arraymap.mapKey;
        this.secondary_mapKey = $arraymap.secondary_mapKey;

        var keys = Object.keys($arraymap.arr);
        for (var k=0; k < keys.length; k++) {
            this.arr[keys[k]] = $arraymap.arr[ keys[k]]
        }
        if ($arraymap.secondary_map!=null) {
            var keys2 = Object.getKeys($arraymap.arr_other);
            for (var k=0; k < keys2.length; k++) {
                this.arr_other[keys2[k]] = $arraymap.arr_other[ keys2[k]]
            }
        }
    }


    setMapKey = function($mk) { this.mapKey = $mk;}
    setSecondaryIndex = function($2ndkey, $bMAKELOWER=false) {
        this.secondary_mapKey = $2ndkey;
        this.secondary_mapKey_forceLowerCase = $bMAKELOWER;
        this.secondary_map = new Object();
    }

    //aliases
    setPrimaryKey = function($mk) { this.setMapKey($mk); }
    setSecondaryKey = function($2ndkey, $bMAKELOWER=false) { this.setSecondaryIndex($2ndkey, $bMAKELOWER);}


    addArray = function($inputArr) {
        for(var i=0; i<$inputArr.length; i++) {
            this.addU($inputArr[i]);
        }
    }

    contains = function($mapKey) {
        if (this.map[$mapKey]==undefined) return false;
        return true;
    }


    //keep obj so we have ptr in mem still in other arrays (like tbls.)
    updateU_oldkey = function($oldObj, $newObj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $oldObj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];
        delete this.map[key];
        var newkey = $newObj[this.mapKey];
        
        var keys = Object.keys($newObj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if (keys[k]==this.mapKey) continue;
            var key = keys[k];  
            objToUpdate[key] = $newObj[key];
        }
        this.map[newkey]=objToUpdate;
    }


    getSlice=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get(f);
            if (obj==null) {
                console.log("no f:"+f+" obj in arr map");
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    getSlice2=function($from_index, $to_index) {
        var arrx = [];
        for (var f=$from_index; f < $to_index; f++) {
            var obj = this.get2(f);
            if (obj==null) {
                //console.log("no f:"+f+" obj in arr map");
                //console.warn("big problem");
                continue;
            }
            arrx.push (obj);
        }

        //we could just try to slice it 
        return arrx;
    }


    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    updateU = function($obj, $skip_keys=[]) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }

        var key = $obj[this.mapKey];
        if (this.map[key]==undefined) {
            console.log("no mapped object at key '"+key+"' to update")
            return;
        }

        var objToUpdate = this.map[key];

        var keys = Object.keys($obj)
        
        //loop through the new object we want to use to update our existing.
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;

            var key = keys[k];

            objToUpdate[key] = $obj[key];
        }
    }

    
    updateU2 = function($obj, $skip_keys=[]) {
        if ($obj==undefined) {
            console.warn("dude, what are you doing?");
            return;
        }
        if (this.secondary_mapKey=="") {
            console.warn("must have map key set to use updateU");
            return;
        }
        
        var key2 = $obj[this.secondary_mapKey];
        if (this.secondary_mapKey_forceLowerCase==true) {
            key2 = key2.toLowerCase();
        }

        if (this.secondary_map[key2]==undefined) return;
        
        var objToUpdate = this.secondary_map[key2];

        var keys = Object.keys($obj)
        for (var k=0; k < keys.length; k++) {
            if ($skip_keys.indexOf( keys[k])!=-1) continue;
            var keyx = keys[k];
            objToUpdate[keyx]=$obj[keyx];
        }
    }


    getKeys = function() {
        return Object.getKeys(this.map);
    }

    getU = function($obj) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        var key = $obj[this.mapKey];
        return this.get(key);
    }

    get = function($mapKey) {
        if (this.mapKey==null || this.mapKey=="") {
            console.log("warn must have a mapKey set to use getU");
            return null;
        }
        if (this.map[$mapKey]!=undefined) return this.map[$mapKey];
        return null;
    }

    get2 = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
            return null;
        }
        if (this.secondary_map[$mapKey2]==undefined) return null;
        return this.secondary_map[$mapKey2];
    }

    get2_partial = function($mapKey2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use get2 without a secondary index set!");
        }
        var txt =$mapKey2;
        if (this.secondary_mapKey_forceLowerCase==true) txt = $mapKey2.toLowerCase();
        var matches = [];
        var keys = Object.keys(this.secondary_map);
        for (var k=0; k < keys.length; k++) {
            if (keys[k].indexOf(txt)==0 && keys[k].length>txt.length){ 
                matches.push( this.secondary_map[ keys[k]]);
            }
        }
        return matches;
    }


    addIndex2 = function($pkOrObj, $key2) {
        if (this.secondary_mapKey=="") {
            console.warn("can't use addIndex2 without a secondary index set!");
        }

        //if first param string, use it as a primary keyvalue of the .map
        //to get the obj. Use that object's key2 to set the secondary map.
        if (typeof $pkOrObj==="string") {
            var obj = this.get($pkOrObj);
            if (obj==null) return null;
            this.secondary_map[$key2]=obj;
        }

        //standard, the pkOrObj is an object, and the $key2 is the value of the secondary mapKey
        //for that object.
        else {

            //audit.
            if ($key2!=$pkOrObj[this.secondary_mapKey]) {
                console.log("houston we have a problem!");
                return;
            }
            this.secondary_map[$key2]=$pkOrObj;
        }
    }

    quickAddMK = function($obj) {
        if (this.mapKey=="") {
            console.warn("U can't use quick add without a quickAddMK set.");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
        }
    }

    quickAdd = function($obj) {
        this.arr.push($obj);
    }
    
    addUnique = function($obj, $mapKeyVal) {
        if ($mapKeyVal==undefined) {
            console.warn("fix this!")
            return;
        }
        if (this.map[$mapKeyVal]==undefined) {
            this.map[$mapKeyVal]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }
        else this.addU2($obj);
    
        return false;
    }

    /**
     * Use this when mapKey is set.
     * @param {*} $obj 
     */
    addU = function($obj) {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addU");
            return;
        }
        if (this.map[$obj[this.mapKey]]==undefined) {
            this.map[$obj[this.mapKey]]=$obj;
            this.arr.push($obj);
            this.addU2($obj);
            return true;
        }

        //still add secondary key if relevant.
        else this.addU2($obj);
    
        return false;
    }


    addU2=function($obj) {
        if (this.secondary_map!=null) {
            var key2 = $obj[this.secondary_mapKey];
            if (key2==undefined) return true;
            if (this.secondary_mapKey_forceLowerCase==true) key2 = key2.toLowerCase();  
            var val2 = $obj[key2];
            if (val2==undefined) this.addIndex2($obj, key2);
        }
    }

    
    removeNoKey = function($index) {
        this.arr.splice($index, 1); 
    }

    remove = function($mapKey, $prop="") {
        if ($prop!="") this.prop = $prop;
        if (this.map[$mapKey]!=undefined) {
            delete this.map[$mapKey];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.prop]==$mapKey) {
                    this.arr.splice(i, 1);
                    return;
                }
            }
        }
    }

    getArrayFromAltProp = function($altprop, $altpropvalue) {
        var matches = [];
        for (var i=0; i < this.arr.length; i++) {
            if (this.arr[i][$altprop]==$altpropvalue) {
                matches.push(this.arr[i]);
            }
        }
        return matches;
    }

    removeU = function($mapKeyValue) {
        if (this.map[$mapKeyValue]!=undefined) {
            delete this.map[$mapKeyValue];
            for (var i=0; i < this.arr.length; i++) {
                if (this.arr[i][this.mapKey]==$mapKeyValue) {
                    this.arr.splice(i, 1);
                    return true;
                }
            }
        }
        return false;
    }


    /**
     * 
     * @param {*} $obj 
     * @param {*} $objkey 
     * @returns 
     */
    addArrAM = function($obj, $objkey="txid") {
        if (this.mapKey=="") {
            console.warn("must have map key set to use addArrAM");
            return;
        }
        var mk = $obj[this.mapKey];
        if (this.map[mk]==undefined) {
            this.map[mk]=new arraymap();
            if (this.map[mk].addUnique($obj, $objkey)==true) {
                this.arr.push($obj);
                return true;
            }
            return false;
        }
        if (this.map[mk].addUnique($obj, $objkey)==true) {
            this.arr.push($obj);
            return true;
        }
        return false;
    }


    addArrObj = function($obj, $mapKey) {
        if (this.map[$mapKey]==undefined) {
            this.map[$mapKey]=[$obj];
            this.arr.push($obj);
            return true;
        }
        this.map[$mapKey].push($obj);
        this.arr.push($obj);
        return false;
    }

    sortArrDesc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return 1;
            else if ($a[$prop] > $b[$prop]) return -1;
            return 0;
        })
    }

    sortArrAsc = function($prop) {
        if (this.arr.length==0) return;
        this.arr.sort(function($a, $b) {
            if ($a[$prop] < $b[$prop]) return -11;
            else if ($a[$prop] > $b[$prop]) return 1;
            return 0;
        })
    }

}

/************** (from file ./js/base/utils/arraymap.js)**/ 

class ANNE_NET {
    constructor($base_uri="http://localhost:9116") {
		this._nodeConnectedCallback = null;
		this._consecutive_network_errors = 0;
		this.bSSEINIT=false;
		this.last_msg_id = 0;
		this.TODOS = [];
		this.active_coms = false;
		//this.bDOINGROUTERPP=false;
		this.setBaseURI($base_uri);
    }


	/**
	 * This is a RAM-only relon that will only be in RAM if the neuron is in RAM
	 * and was pending a change in the mempool.
	 * 
	 * Since this app is "open" it will receive said pending change and thus we 
	 * can count on being able to remove the pending RAM-only relon here.
	 * @param {*} $nid 
	 */
	cleanupIntraPending = function($nid) {
		console.log("(cleanupIntraPending)"+$nid);

		if ($nid==undefined || $nid==null || $nid=="") return;
		if (parseInt($nid)==0) return;
		
		var nOBJ = MEMG.getOrMakeIfNew($nid);
		if (nOBJ._aplex!=null) nOBJ._aplex.has_pending=false;
		
		var intrablockpending_relon = MEMG.getRelonFromWith(nOBJ.outs, TYPE_SYS_MSG, ANNE.MEM.INTRA_BLOCK_PENDING.nid);
		if (intrablockpending_relon!=null) MEMP.localOnlyArborPrune_RAM(intrablockpending_relon);
		return nOBJ;
	}


	/**
	 * 
	 * @param {*} $uri 
	 */
	setBaseURI=function($uri="") {
		console.log("setting BaseURI:"+$uri);
		this.base_uri = $uri;
		if (this.base_uri!="", this.base_uri[this.base_uri.length-1]=="/") {
			this.base_uri = this.base_uri.substring(0, this.base_uri.length-1);
		}
	}


	/**
	 * 
	 * @param {*} $f 
	 */
	setNodeConnectedCallback=function($f) { 
		this._nodeConnectedCallback = $f; 
	}


	/** 
	 * 
	*/
	setNodeConnected=function($bNODECONNECTED=true) {
		if (this._nodeConnectedCallback!=null) {
			this._nodeConnectedCallback($bNODECONNECTED);
		}
	}



	//***************************************************** */
	//////  TRANACTION CREATION / SIGNING / BROADCAST ////////
	//***************************************************** */

	/**
	 * @param {String} $functName 
	 * @param {Object} $params 
	 * @param {String} $uriOverride - the node uri to use, if not the current node ur connected to. 
	 * @returns 
	 */
	makeOnNodeSignOnJSThenBroadcast=async($functName, $params=null, $uriOverride="")=> {
		var bCHASE_FROM=false;
		var bDO_EXTRAS=false;

		if ($params!=null) {
			if ($params.secretPhrase!=undefined) delete $params.secretPhrase;
			
			//chase the froms in the broadcastTransaction so we can get
			//crucial return data that we need in the processing of the next TXs in 
			//the queue.
			if ($params.chase_from!=undefined) {
				bCHASE_FROM = $params.chase_from;
				delete $params.chase_from;
			}
			if ($params.do_extras!=undefined) {
				bDO_EXTRAS = $params.do_extras;
				delete $params.do_extras;
			}

			//we need to make the TX first (using the annode)
			//then sign it here in the EXT,
			//then broadcast it.
			if ($params.makeSignSend!=undefined && $params.makeSignSend==true) {
				$params.publicKey = USER.pubKey;
				delete $params.makeSignSend;
			}

			//this wasn't set up right and it is just a normal query/send.
			else {
				return await this.query($functName, $params);
			}

			try {
			
				var response = await this.post($functName, $params, $uriOverride);
				console.log('%capi call for '+$functName, TALK.bggreenbig);
				
				this._consecutive_network_errors = 0;

				if (response && response.height!=undefined) {
					HISTORY._GREATEST_LAST_KNOWN_BLOCK = response.height;
				}

				if (response.status!=undefined && response.status!="good") {
					console.log(response);
					var retObj =  {status:"fail", reason:response.status};
					if (response.recourse!=undefined) retObj.recourse = response.recourse;
					if (response.result!=undefined) retObj.result = response.result;
					if (response.results!=undefined) retObj.results = response.results;					
					return retObj;
				}

				if (response.unsignedTransactionBytes==undefined || response.unsignedTransactionBytes==null) {
					return {status:"fail", reason:"problem with unsignedTransactionBytes"};
				}
				var finalResp = await this.signAndBroadcastTransaction(response, bCHASE_FROM, bDO_EXTRAS, $uriOverride);
				
				//don't do this from RELONS!!!
				//we want to remove relons that are beyond this point..
				//for rollback situations.
				if (finalResp.block_height!=undefined) {
					ANNE.HISTORY.updateGLKB_SOLID(finalResp.block_height);
				}
				ANNE.API.setNodeConnected(true);

 				return finalResp;
			}

			catch(e) {
				console.warn(e);
				if (e.data=="Network Error") ANNE.API.setNodeConnected(false);
				var reason = "problem with transaction";
				if (e.data!=undefined && e.data.errorCode!=undefined) {
					if(e.reason!=undefined && e.message==undefined) e.message = e.reason;
					e.data.status="bad";
					return e.data;
				}
				if (e.message!=undefined) reason = e.message;
				var retObj ={status:"bad", reason:reason};
				if (e.recourse!=undefined) retObj.recourse = e.recourse;
				return retObj;
			}
		}

		//this was just a simple query (without params)
		return await this.query($functName);
	}



	/**
	 * 
	 * @param {*} $resp 
	 * @param {*} $bCHASEFROM 
	 * @param {*} $bDOEXTRAS 
	 * @param {*} $uriOverride 
	 * @returns 
	 */
	signAndBroadcastTransaction=async($resp, $bCHASEFROM=false, $bDOEXTRAS=false, $uriOverride="")=> {
		this.active_coms = true;
		if (USER.privKey=="" || USER.privKey==undefined) {
			if (USER.seedObj==null) {
				var aO = await USER.getPrivateKeyHex();
			}
			else var aO = USER.prepANNEAuthObjFromSeed(USER.seedObj.seed);
			USER.privKey = aO.sign_key;
		}

		var sig=null;
		
		//gen the signature
		try{
			sig = ANNE.SIGNA.crypto.generateSignature($resp.unsignedTransactionBytes, USER.privKey);
		}
		catch(e) {
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignature) problem"};
		}

		//verify the signature
		try{
			var rez = ANNE.SIGNA.crypto.verifySignature(sig, $resp.unsignedTransactionBytes, USER.pubKey);
			if (rez==false) {
				this.active_coms = false;
				return {status:"fail", reason:"(verifySignature) did not pass"};
			}
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(verifySignature) problem"};
		}

		//generate signed bytes
		var signedTransactionBytes=null;
		try {
			signedTransactionBytes = ANNE.SIGNA.crypto.generateSignedTransactionBytes($resp.unsignedTransactionBytes, sig);
		}
		catch(e){
			console.log(e);
			this.active_coms = false;
			return {status:"fail", reason:"(generateSignedTransactionBytes) problem"};
		}
		
		//broadcast the transcation.
		var broadcastParams = {
			transactionBytes:signedTransactionBytes,
			transactionJSON:$resp.transactionJSON
		}
		if ($bCHASEFROM==true) {
			broadcastParams.chase_from = true;
		}
		else if ($bDOEXTRAS==true) {
			broadcastParams.do_extras = true;
		}

		try {
			var response2 = await this.broadcastSigned(broadcastParams, $uriOverride);
			if (response2.status!=undefined && response.status!="good") {
				console.log(response2);
			}
			this.active_coms = false;
			return response2;
		}	

		catch(e) {
			console.log('ERROR: api call for broadcastTransaction', e);
			if (e.data=="Network Error") {
				this._consecutive_network_errors+=1;
			}

			var rez = {status:"fail", msg:""};
			
			if (e.data!=undefined && e.data.message!=undefined) {
				rez.msg = e.data.message;
			}
			else {
				rez.msg = e.message;
			}
			if (e.data!=undefined) {
				if (e.data.error!=undefined) rez.error = e.data.error;
				if (e.data.errorCode!=undefined) rez.errorCode = e.data.errorCode;
				if (e.data.errorDescription!=undefined) rez.errorDescription = e.data.errorDescription;
				if (e.data.errorRecourse!=undefined) rez.errorRecourse = e.data.errorRecourse;
				if (e.data.result) rez.result=e.data.result;
			}
			
			this.active_coms = false;
			return rez;
		}
	}



	/**
	 * Last step of process to gen a TX on the annode, sign it on the edge device
	 * and broadcast transaction bytes back to the annode
	 * @param {Object} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	broadcastSigned = async($params, $uriOverride="")=>{
		ANNE.playSound("broadcast");
		return await this.post("broadcastTransaction", $params, $uriOverride);
	}



	/**
	 * Post data (used for TX broadcasts)
	 * @param {*} $functName 
	 * @param {*} $params
	 * @param {String} $uriOverride  
	 * @returns 
	 */
	post = async($functName, $params, $uriOverride="")=>{
		this.active_coms = true;

		var uri = this.base_uri;

		if (uri==undefined) {
			ANNE.TALK.printf_console(ANNE.TALK.RED, "(post) NET.base_uri not set, setting to defaults");
			uri = ANNE.setAnnodeByURI();
		}

		if ($uriOverride!="") {
			ANNE.TALK.printf_console(ANNE.TALK.BLUE, "(post) using URL override", ANNE.TALK.GREEN, $uriOverride);
			uri = $uriOverride;
		}

		var cs = ANNE.SIGNA.sig.ChainService;
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		var response = null;
		if ($params!=null) response = await service.send($functName, $params);
		else response = await service.send($functName);
	

		
		this.active_coms = false;
		return response;
	}

	
	proxyRouter=async($uri, $functName, $params=null)=>{
		var uri = $uri+"?requestType="+$functName;
		if ($params != null) {
			const queryStringComponents = [];
			for (const key in $params) {
				if ($params.hasOwnProperty(key)) {
					const encodedKey = encodeURIComponent(key);
					const encodedValue = encodeURIComponent($params[key]);
					queryStringComponents.push(`${encodedKey}=${encodedValue}`);
				}
			}
			uri += '&' + queryStringComponents.join('&');
		}

		console.log(uri); 

		let resp =  await fetch(uri).catch(function($e){
			console.log("error:"+$e);
			return null;
		});

		if (!resp) {
            console.warn("HTTP error! -- no response");
			return null;
        }

        else if (!resp.ok) {
            console.warn(`HTTP error! status: ${resp}`);
			return null;
        }

        let data = await resp.json();
        console.log(data); 

		return data;
	}


	apiHTTPNOJSON=async($functName, $params)=> {
		this.active_coms = true;
		
		var uri = this.base_uri;
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}

		
		var cs = ANNE.SIGNA.sig.ChainService;
		const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
		if ($params!=null) {
			var rez = await service.send($functName, $params);
			this.active_coms = false;
			return rez;
		}
		else {
			var rez = await service.send($functName);
			this.active_coms = false;
			return rez;
		}
		

	}


    /**
     * Perhaps we lack here too from oversimplification of the signa method...
	 * may have to revert back to using the chainService here too.
	 * 
     * @param {String} $functName 
     * @param {Object} $params 
	 * @param {String} $uriOverride
     * @returns 
     */
    query = async($functName, $params=null, $uriOverride="")=>{
		//if ($functName!="getLatestMsgs") {
			if ($params!=null) console.log("%cAPI.query."+$functName, ANNE.TALK.bgpurple, $params);
			else console.log("%cAPI.query."+$functName, ANNE.TALK.bgpurple);
		//}
		//if ($params!=null && $params.height!=undefined) console.log("for height:"+$params.height);
		
		this.active_coms = true;

		var uri = this.base_uri;
		
		if (uri==undefined) {
			uri = ANNE.setAnnodeByURI();
		}
		if ($uriOverride!="" && $uriOverride!=null) uri = $uriOverride;
		
	
		var cs = ANNE.SIGNA.sig.ChainService;
		if (cs==undefined) console.throw("Must have ANNE.SIGNA.sig.Chainservice");
		
		try {
			const service = new cs({ nodeHost: uri, apiRootUrl:"/anne"}, "anne");
			//console.log('api call for '+$functName);

			if ($params!=null) var response = await service.send($functName, $params);
			else var response = await service.send($functName);
		}
			
		catch(e) {
			ANNE.TALK.printf(TALK.RED, " connection to annode error (no annode) ");
			this.active_coms = false;
			if ($functName=="annodeDetails") {
				if (ANNE && ANNE.badNode!=undefined) ANNE.badNode();
			}
			return this.makeCatchReturn(e);
		}
		

		//get height if we passed it.. helpful for other processes.
		if (response!=null && response.result!=undefined && response.result.block_height!=undefined) {
			ANNE.HISTORY.updateGLKB_SOLID(response.result.block_height);
		}

		if ($functName=="getNeuronBalance" && response.status!=undefined && response.status!="bad") {
			await this.__balanceHack(response, $params);
		}
		this.active_coms = false;
		return response;
    }


	/**
	 * 
	 * @param {*} $e 
	 * @returns 
	 */
	makeCatchReturn = function($e) {
		if ($e==null || $e==undefined) return {status:"fail", reasons:"totally hosed everything"};

		console.log($e);
		var reason = "problem with net query";
		if ($e!=null && $e.status==405) {
			ANNE.API.setNodeConnected(false);
			reason = "no connected annoded.";
		}
		
		else if ($e.data!=undefined && $e.data!=null && $e.data=="Network Error") {
			ANNE.API.setNodeConnected(false); //are we sure we wnat to do this here????
		}

		else if ($e.data!=undefined && $e.data!=null && $e.data.errorCode!=undefined && $e.data.errorCode=="777.2.0") {
			var retObj = {status:"good-but", message:"taken"};
			if ($e.data.result!=undefined) retObj.result = $e.data.result;
			if ($e.data.results!=undefined) retObj.results = $e.data.results;
			return retObj;
		}
		
		return {status:"fail", reason:reason};
	}


	/**
	 * TODO: review this.
	 * @param {*} $response 
	 * @param {*} $params 
	 * @returns 
	 */
	__balanceHack=async($response, $params)=> {
		if ($response==undefined) { return;}
		
		//this will be active sooooon.
		if ($response.activated_in_mempool!=undefined) {
            if (this._userNObj!=null) this._userNObj.activated_in_mempool=true;
        }

		if ($response.activated=="y") {
			if ($params.nid==USER._userNObj.nid && ANNE.USER.isActivated()==false) {
				await ANNE.USER.updateLSActivated(true); 
			}
		}

		//somehow we have this as activated (on this network) but it isn't.
		//correct that.
		else if ($response.activated=="n" && ANNE.USER.isActivated()==true) {
			if ($params.nid==USER._userNObj.nid) {
				await ANNE.USER.updateLSActivated(false); 
			}
		}
		return $response;
	}


	fetchJSON = async (uri ="https://www.biconomy.com/api/v1/depth?symbol=ANNE_USDT") => {
		try {
			const response = await fetch(uri);
			if (!response.ok) {
				throw new Error(`HTTP error! Status: ${response.status}`);
			}
			const data = await response.json();
			return data;
		}
		catch (error) {
			console.error("Error fetching JSON:", error);
			throw error; 
		}
	};
	
}

/************** (from file ./js/network/net.js)**/ 

class anne_alert{
    constructor($title, $body, $cb_confirm=null, $cb_no=null, $cb_close=null){
        var el_alert_header = document.getElementById("alerta_header");
        el_alert_header.innerHTML = $title;

        var el_alert_text = document.getElementById("alerta_text");
        el_alert_text.innerHTML = $body;

        var el_alert_left_btns =  ANNE.A.empty("alerta_left_btns")
        var el_alert_right_btns =  ANNE.A.empty("alerta_right_btns")
        
        var el_alert_close = document.getElementById("alerta_header_close");
        el_alert_close.addEventListener("click", function($e){
            if ($cb_close!=null) $cb_close();
            ANNE.A.hide("alerta");
        })

        var btns = [];

        //confirm callback.  (if exists make confirm/yes button)
        if ($cb_confirm!=null) {
            var btnconf = document.createElement("button");
            btns.push(btnconf);
            btnconf.classList = "alerta-btn-confirm";
            btnconf.innerHTML = "yes";
            el_alert_left_btns.append(btnconf);
            el_alert_left_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_confirm();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        //no callback (if exists make no button)
        if ($cb_no!=null) {
            var btnno = document.createElement("button");
            btns.push(btnno);
            btnno.classList = "alerta-btn-no";
            btnno.innerHTML = "no";
            el_alert_right_btns.append(btnno);
            el_alert_right_btns.addEventListener("click", function($e){
                $e.stopPropagation();
                $e.preventDefault();
                for (var b=0; b<btns.length;b++){
                    ANNE.A.removeSelf(btns[b]);
                }
                el_alert_text.innerHTML="";
                el_alert_header.innerHTML="";
                $cb_no();
                ANNE.A.hide("alerta");
            });
            ANNE.A.show(el_alert_left_btns);
        }

        ANNE.A.fadeIn("alerta");
    }
}


class anne_toggle_obj {
    constructor($divObj) {
        var divObj = $divObj.div;
        var btnObj = $divObj.btn;

        this.div_el = null; //will be set to the el that is made for this anne_toggle_obj

        if (divObj.id!=undefined) this.id = divObj.id;
        else this.id="";

        if (divObj.title!=undefined) this.title = divObj.title;
        else this.title="";
        
        if (divObj.style!=undefined) this.style = divObj.style;
        else this.style="";

        if (divObj.text!=undefined) this.text = divObj.text;
        else this.text="";
        
        if (divObj.div_class!=undefined) this.div_class = divObj.div_class;
        else this.divclass="";
        
        if (divObj.cb!=undefined && typeof divObj.cb==="function") this.cb = divObj.cb;
        else this.cb=null;

        //btn for the toggling.
        this.btn_el = null;
        this.btn_toggle_id = "";
        this.btn_text = "";
        this.btn_class = "";
        this.btn_tnid = "";

        if (btnObj!=undefined) {
            if (btnObj.id!=undefined) this.btn_toggle_id = btnObj.id;
            if (btnObj.text!=undefined) this.btn_text = btnObj.text;
            if (btnObj.btn_class!=undefined) this.btn_class = btnObj.btn_class;
            if (btnObj.btn_tnid!=undefined) this.btn_tnid = btnObj.btn_tnid;   
        }
    }
}


/**
 * 
 */
class anne_toggle_set {
    constructor($objs) {
        this.atos = [];
        if ($objs.length) {
            for (var a=0; a < $objs.length; a++) {
                var ato = new anne_toggle_obj($objs[a]);
                this.atos.push(ato);
            }
        }
    }

    /**
     * 
     */
    makeToggleSetFromAwarePol=function(){
    }

    /**
     * 
     */
    makeToggleSetFromAwareBucket=function(){
    }

}


/**
 * 
 */
class anne_select_option {
    constructor($selVal="", $selDisplayVal="", $titleHTML="", $bodyHTML="", $val2=""){
        this.selval = $selVal;
        this.selDispVal = $selDisplayVal;
        this.titleHTML = $titleHTML;
        this.bodyHTML = $bodyHTML;
        this.val2 = $val2;
        this.gui_id = UTILS.randHex(8);
        this.el_opt = null;
    }
}

class anne_select {
    constructor($w="400px", $el, $options, $misc) {
        this.width = $w;
        this.el = ANNE.A.__getEl($el);
        
        if (this.el == null) return null;
        
        this.options = $options;

        if ($misc.defaultLbl!=undefined) this.defaultLbl = $misc.defaultLbl;
        else this.deafultLbl = "";

        if ($misc.addNewCb!=undefined) this.addNewCb=$misc.addNewCb;
        else this.addNewCb = null;

        if ($misc.cb!=undefined) this.cb = $misc.cb;
        else this.cb=null;

        if ($misc.cb_ddopen!=undefined) this.cb_ddopen = $misc.cb_ddopen;
        else this.cb_ddopen = null;

        if ($misc.cb_ddclose!=undefined) this.cb_ddclose = $misc.cb_ddclose;
        else this.cb_ddclose = null;

        if ($misc.ddclass!=undefined) this.ddclass = $misc.ddclass;
        else this.ddclass = "";

        if ($misc.max_height!=undefined) this.max_height=$misc.max_height;
        //default 660px
    }


}


/******************* anne_dom - a vanilla GUI/DOM stuff *********************
*           lighter, faster, and simpler for my style of coding.            *
****************************************************************************/
class anne_dom extends anne_singleton {
    constructor() {
        super();
    }

    /**
     * Create a button that launches a dialogue w/ various details and callbacks.
     * @param {*} $el 
     * @param {*} $btnClass 
     * @param {*} $title 
     * @param {*} $msg 
     * @param {*} $cbYes 
     * @param {*} $cbNo 
     * @param {*} $cbOk 
     */
    btnToLaunchConfirm=async($el, $btnClass="anne-btn2", $title, $msg, $cbYes=null, $cbNo=null, $cbOk=null)=>{
        this.ezButton($el, "", $title, $btnClass, async($e)=>{
            new anne_alert($title, $msg, $cbYes, $cbNo, $cbOk);
        });
    }


    /**
     * 
     * @param {*} $wPx 
     * @param {function} $clickCB
     * @returns 
     */
    memeframer = function($wPx="900px", $clickCB=null, $name="testjpg") {
        const el_meme = document.createElement("div");
        el_meme.classList="meme-container";
        el_meme.style['max-width']=$wPx;
        el_meme.style['width']=$wPx;
        
        var row = document.createElement("div");
        row.style['min-width']='100%';
        row.classList = "meme-row";
        el_meme.appendChild(row);

        el_meme.setAttribute("file_name", $name);

        var img = document.createElement("img");
        img.src="./imgs/anne_logo.png";
        img.setAttribute("alt", "logo");
        img.classList="meme-logo";
        row.appendChild(img);

        const btn = document.createElement("button");
        btn.classList="meme-btn-make";
        btn.innerHTML="make meme!"
        row.append(btn);
        
        var row2 = document.createElement("div");
        row2.style['min-width']='100%';
        row2.classList = "meme-body";
        el_meme.appendChild(row2);


        if ($clickCB!=null && typeof $clickCB==="function") {
            btn.addEventListener("click", (e)=>{
                ANNE.A.hide(btn);
                var name = el_meme.getAttribute("file_name");
				$clickCB(el_meme, name);
			});
        }

		return {el:el_meme, body:row2};
    }


    buildGridNeuron = function($el, $test, $class, $afile=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-neuron-holder");
        var neuron = document.createElement("div");
        neuron.classList.add("grid-neuron");
        
        if ($afile!=null) {
            ANNE.A.doImageB64(neuron, "", $afile, 0, 99, "", 99);
        }
        else {
			neuron.innerHTML = $test;
			neuron.classList.add($class);
		}
        el.appendChild(neuron);
    }


    buildGridArrow=function($el, $arrowClass, $lbl="test") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("grid-"+$arrowClass); 
        var lbl = document.createElement("div");
        lbl.innerHTML = $lbl;
        lbl.classList.add("grid-arrow-lbl");
        lbl.classList.add($arrowClass+"-text");
        el.appendChild(lbl);
        return;
    }


    /**
     * [x][->][x][<-][x]
     * @param {*} $el 
     */
    build1con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        var grid = this.buildDOMGrid(el, 5, 1);
        console.log(grid);
        
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow", $m0[0].afile);
        this.buildGridArrow(grid["i1"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m0[1].text, "ggreen", $m0[1].afile);
        this.buildGridArrow(grid["i3"], "left", $m1[0].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue", $m1[0].afile);
        grid.parent_el = el;
        return grid;
    }
    

    build1con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "ne", "");
        this.buildGridArrow(grid["i6"], "right", $m0[0].reln_text);
        this.buildGridNeuron(grid["i7"], $m0[1].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");

        this.buildGridArrow(grid["i13"], "horizontal", $m1[1].reln_text);
        this.buildGridArrow(grid["i12"], "left-up", "");
        
        grid.parent_el = el;
        return grid;
    }



    build2con1con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gblue");
        this.buildGridArrow(grid["i11"], "horizontal", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-up", "");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "nw", $m1[0].reln_text);
        this.buildGridArrow(grid["i8"], "left", "");
        this.buildGridNeuron(grid["i7"], $m0[2].text, "ggreen");
        
        grid.parent_el = el;
        return grid;
    }

    /**
     * [a]     [b]
     *  |       |
     * [c]     [d]
     *  |       |   
     *   ->[e]<-
     * @param {*} $el 
     * @param {*} $w 
     * @returns 
     */
    build2con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "ne", "");
        this.buildGridArrow(grid["i16"], "right", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "nw", "");
        this.buildGridArrow(grid["i18"], "left", $m0[1].reln_text);
        

        grid.parent_el = el;
        return grid;
    }

    build3con2con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "horizontal", $m0[2].reln_text);
        this.buildGridArrow(grid["i22"], "right-up", "");
        this.buildGridNeuron(grid["i17"], $m0[3].text, "ggreen");

        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i13"], "left", "");
        this.buildGridArrow(grid["i12"], "left-down", "");
        
        
        grid.parent_el = el;
        return grid;
    }


    build2con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i11"], "right", $m0[1].reln_text);
        this.buildGridArrow(grid["i12"], "right-down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i17"], $m0[2].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
        this.buildGridArrow(grid["i22"], "left-up", "");
    
        grid.parent_el = el;
        return grid;
    }


    build3con3con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "ggreen");
        
        this.buildGridNeuron(grid["i4"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[0].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[1].reln_text);
        this.buildGridNeuron(grid["i24"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i23"], "left", "");
    
        grid.parent_el = el;
        return grid;
    }


    /**
     * [x0]   [y1]-->[y3]
     *  |      |      |
     * [x1]   [y0]   [y4]
     *  |             |
     * [x2]-->[x3]->[match]  
     * @param {*} $el 
     * @param {*} $m0 
     * @param {*} $m1 
     * @param {*} $w 
     * @returns 
     */
    build4con4con=function($el, $m0, $m1, $w="495px") {
        var el = null;
        if ($el==null) {
            el = document.createElement("div");
            //so we can call this 
        }
        else {
            el = this.__getEl($el);
            if (el == null) return null;
        }

        //this will determine the actual rows.
        el.style['max-width']=$w;
        el.style['width']=$w;
        
        var grid = this.buildDOMGrid(el, 5, 4);
        console.log(grid);
        this.buildGridNeuron(grid["i0"], $m0[0].text, "gyellow");
        this.buildGridArrow(grid["i5"], "down", $m0[0].reln_text);
        this.buildGridNeuron(grid["i10"], $m0[1].text, "gyellow");
        this.buildGridArrow(grid["i15"], "down", $m0[1].reln_text);
        this.buildGridNeuron(grid["i20"], $m0[2].text, "gyellow");
        this.buildGridArrow(grid["i21"], "right", $m0[2].reln_text);
        this.buildGridNeuron(grid["i22"], $m0[3].text, "gyellow");
        this.buildGridArrow(grid["i23"], "right", $m0[3].reln_text);
        this.buildGridNeuron(grid["i24"], $m0[4].text, "ggreen");

        
        this.buildGridNeuron(grid["i12"], $m1[0].text, "gblue");
        this.buildGridArrow(grid["i7"], "up", $m1[0].reln_text);
        this.buildGridNeuron(grid["i2"], $m1[1].text, "gblue");
        this.buildGridArrow(grid["i3"], "right", $m1[1].reln_text);
        this.buildGridNeuron(grid["i4"], $m1[2].text, "gblue");
        this.buildGridArrow(grid["i9"], "down", $m1[2].reln_text);
        this.buildGridNeuron(grid["i14"], $m1[3].text, "gblue");
        this.buildGridArrow(grid["i19"], "down", $m1[3].reln_text);

        grid.parent_el = el;
        return grid;
    }



    buildDOMGrid=function($el, $flexWidth = 5, $rows = 5, $pre="gridx") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var grid = new Object();
        el.classList.add("grid-container");
        for (let i = 0; i < $flexWidth * $rows; i++) {
            const gridSquare = document.createElement('div');
            gridSquare.classList.add('gridSquare');
            gridSquare.style.setProperty('--flexWidth', $flexWidth);
            el.appendChild(gridSquare);
            var r = Math.floor((i+$flexWidth)/$flexWidth)-1;
            var row = grid["r"+r];
            if (row==undefined) {
                row = new Object();
                grid["r"+r]=row;
            }
            var c = (i+$flexWidth) % $flexWidth;
            el.setAttribute("id",$pre+"_"+r+"_"+c)
            row["c"+c] = gridSquare;
            grid["i"+i]=gridSquare;
        }

        return grid;
    }


    
    /**
     * Make a nifty little HTML DD for the users DD.
     * //works with .text/.nickname and .nid/.public_nid
     * @param {*} $el 
     * @param {*} $users 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromUserObjs = function($w="400px", $el, $users, $configs) {
        var options = [];
        for (var n=0; n < $users.length; n++) {
            var user = $users[n];
            var nid = user.public_nid;
            if (nid==undefined) nid = user.nid;
            if (nid==undefined || nid==null) {
                continue;
            }
            var userNOBJ = MEMG.getOrMakeIfNew(nid);
            var lastKnownAnnies = user.last_known_annies;
            if (lastKnownAnnies==undefined || lastKnownAnnies==null) lastKnownAnnies = "0";
            //else 
            var titleHTML = "";
            var bodyHTML = "";
            var sel_disp_val = "";
            var text = user.text;
            if (text==undefined || text==null || text == "") {
                text = user.nickname;
                if (text==undefined || text==null) text="";
            }
            if (text!="") {
                titleHTML ='<span class="acct-name">'+text+'</span>';
                sel_disp_val = text;
            }
            else {
                titleHTML ="<span class='acct-name'>"+nid+"</span>";
                sel_disp_val = nid;
            }

            bodyHTML = "<span class='bal-small bal-"+userNOBJ.temp_nid+"'>"+UTILS.displayAnneValue(lastKnownAnnies)+"<span style='font-size:1.6rem;'>&#x1d110;</span></span>";
            
            var opt = new anne_select_option(nid, sel_disp_val, titleHTML, bodyHTML);
            options.push(opt);
            user.gui_id = opt.gui_id;
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
    }


    
    /**
     * make a dropdown of annodes
     */
    makeANNODEDD = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:ANNE._annodesLS.last_uri,
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item")
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
                if ($annode_nid!=undefined && $annode_nid!=null && $uri!=undefined && $uri!=null) {
                    if (typeof ANNE.setNewAnnodeURI==="function") {
                        ANNE.setNewAnnodeURI($annode_nid, $uri);
                    }
                }
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };
        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);
    }

    removeAllButTopOption = function($el) {
        var el = this.__getEl($el);
        if (el==null) return;
        const optionsCount = el.options.length;
        for (let i = optionsCount - 1; i > 0; i--) {
            el.remove(i);
        }
    }

    getDDSelectedValue=function($el) {
        var el = this.__getEl($el);
        if (el==null) return null;
        if (el.options==undefined) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.value;
    }
    getDDSelectedAttr=function($el, $attr) {
        var el = this.__getEl($el);
        if (el==null) return null;
        var opt = el.options[ el.selectedIndex];
        if (opt==null) return null;
        if (opt==undefined) return null;
        return opt.getAttribute($attr);
    }

    setDDToIndex=function($el, $index=0) {
        var el = this.__getEl($el);
        if (el==null) return;
        el.selectedIndex = $index;
    }
    setDDToValue=function($el, $value) {
        var el = this.__getEl($el);
        if (el == null) return;
        for (var i = 0; i < el.options.length; i++) {
            if (el.options[i].value == $value) {
                el.selectedIndex = i;
                return;
            }
        }    
    }


    spinnerInject = function($el, $bWIPE=false) {
        var el = this.__getEl($el);
        if (el==null) return;
        if (el.id==null || el.id=="") el.id = ANNE.UTILS.randHex(4);
        var el_spinner = document.createElement("span");
        el_spinner.id = el.id+"_s";
        el_spinner.classList = "inlinespinner";
        if ($bWIPE==true) ANNE.A.emptySelf(el);
        el.appendChild(el_spinner);
        ANNE.A.show(el);
    }
    
    spinnerRemove = function($el, $bREMOVEPARENT=false) {
		var el = this.__getEl($el);
        if (el==null) return;
        var el_spinner = document.getElementById(el.id+"_s");
        if(el_spinner==null) {
            var inlinespinner = el.querySelectorAll(".inlinespinner");
            if (inlinespinner!=null) ANNE.A.removeSelf(inlinespinner);
        }
        else {
            ANNE.A.removeSelf(el_spinner);
        }
        if ($bREMOVEPARENT==true) {
            ANNE.A.removeSelf(el);
        }
    }

    makeANNODEDD_first_time = async($el="main_annodes", $configs=null)=> {
        var configs = $configs;
        if (configs==null) configs = {
            defaultLbl:"select annode",
            ddclass:"annodedd",
            addNewCb:function(){
                console.log("add new item");
            },
            cb:async($annode_nid, $uri)=>{
                console.log("select new annode! "+$annode_nid);
            //app_main
            }, cb_ddopen:function(){
                ANNE.A.smartDefocus(true);
            }, cb_ddclose:function() {
                ANNE.A.smartDefocus(false);
            }
        };

        return ANNE.A.makeHTMLDD_fromAnnodeObjs("300px", $el, ANNE._annodesLS.annodes, configs);

    }



    /**
     * 
     * @param {*} $bDEFOCUS 
     */
    smartDefocus = async($bDEFOCUS=true)=> {
        if ($bDEFOCUS==true) {
            var el = ANNE.A.addClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.setAttribute("disabled",true);
        }
        else {
            var el = ANNE.A.removeClass("scrollable_"+this.ID, "defocus");
			if (el==null) return;
            el.removeAttribute("disabled",false);
        }
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $annodes 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     */
    makeHTMLDD_fromAnnodeObjs = function($w="400px", $el, $annodes, $configs) { 
        var options = [];
        for (var n=0; n < $annodes.length; n++) {
            var annode = $annodes[n];
            var titleHTML = annode.nickname+' - '+annode.nid;
            var bodyHTML = annode.uri;
            var sel_disp_val = annode.nickname+" "+annode.uri;
            var opt = new anne_select_option(annode.nid, sel_disp_val, titleHTML, bodyHTML, annode.uri);
            options.push(opt);
        }
        var ASELECT  = new anne_select($w, $el, options, $configs);
        this.makeHtmlDD(ASELECT);
        return ASELECT;
    }



    /**
     * 
     * @param {*} $el 
     * @param {*} $options
     * @param {*} config $defaultLbl, $cb
     * 
     * @param {*} $defaultLbl 
     * @param {*} $cb 
     * @returns 
     */
    makeHtmlDD=function($ASELECT) {
        if ($ASELECT.el==null) return;
        var sb = null;  //select box

        if ($ASELECT.el.classList.contains("selectbox")==false) {
            sb = document.createElement("div");
            sb.classList = "selectbox";
            $ASELECT.el.appendChild(sb);
        }
        else sb = $ASELECT.el;

        sb.style.width = $ASELECT.width;
    
        if ($ASELECT.addNewCb!=null) {
            //var plusSpan = document.createElement("span");
            //$ASELECT.el.parentNode.appendChild(plusSpan);
            //plusSpan.innerHTML = "[+]";
            //plusSpan.classList = "btn-select-add-option"; 
        }
        
        const optionsContainer = document.createElement("div");
        optionsContainer.classList = "options-containersb";
        sb.appendChild(optionsContainer);

        if ($ASELECT.max_height!=undefined) {
            sb.style['max-height']=$ASELECT.max_height;
            optionsContainer.style['max-height']=$ASELECT.max_height;
            optionsContainer.setAttribute("mh_backup", $ASELECT.max_height);
        }
		else {
			sb.style['max-height']='50vh';
            optionsContainer.style['max-height']='50vh';
            optionsContainer.setAttribute("mh_backup", '50vh');
		}

        const selectedsb = document.createElement("div");
        selectedsb.classList = "selectedsb";

        if ($ASELECT.ddclass!="") {
            selectedsb.classList.add($ASELECT.ddclass);
        }
        selectedsb.style.width = $ASELECT.width;

        sb.appendChild(selectedsb);
        if ($ASELECT.defaultLbl!="") {
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3 pr30";
            lbl_sp3.innerHTML = $ASELECT.defaultLbl;
            selectedsb.append(lbl_sp3);
        }

        //create options
        for (var i=0; i  < $ASELECT.options.length; i++) {
            var opt = $ASELECT.options[i];
            var selval = opt.selval;
            var val2 = opt.val2;
            const divopt = document.createElement("div");
            optionsContainer.appendChild(divopt);
            divopt.classList = "optionsb";
            if (selval!=null) divopt.setAttribute("selval", selval);
            if (val2!=null) divopt.setAttribute("val2", val2);
            if (opt.selDispVal!=null) divopt.setAttribute("dispval", opt.selDispVal);
            
            var divopt_rad = document.createElement("input");
            divopt_rad.setAttribute("type", "radio");
            divopt_rad.classList = "radiosb";
            divopt_rad.id = opt.gui_id;
            divopt.appendChild(divopt_rad);
            
            var divopt_lbl = document.createElement("label");
            divopt_lbl.setAttribute("for", divopt_rad.id);
            divopt_lbl.classList = "opt-lbl";
            divopt.appendChild(divopt_lbl);
            
            var lbl_sp3 = document.createElement("span");
            lbl_sp3.classList = "sbh3";
            lbl_sp3.innerHTML = opt.titleHTML;
            divopt_lbl.append(lbl_sp3);
            
            var lbl_sp5 = document.createElement("span");
            lbl_sp5.classList = "sbh5";
            lbl_sp5.innerHTML = opt.bodyHTML;
            divopt_lbl.append(lbl_sp5);
            opt.el_opt = divopt;

            opt.el_opt.addEventListener("click", ($e) => {
				console.log("(option clicked!)");
                $e.stopPropagation();
                $e.preventDefault();
				var selval = divopt.getAttribute("selval");
                var val2 = divopt.getAttribute("val2");
                if (val2==undefined) val2 = null;
                var dispval =  divopt.getAttribute("dispval");
				//selectedsb.innerHTML = divopt.querySelector("label").innerHTML;
				//selectedsb.innerHTML = selval;
                ANNE.A.empty(selectedsb);
                var lbl_sp3 = document.createElement("span");
                lbl_sp3.classList = "sbh3 pr30";
                if (dispval) lbl_sp3.innerHTML = dispval;
                else lbl_sp3.innerHTML = selval;
                selectedsb.append(lbl_sp3);

                document.removeEventListener("click", clickout);

                optionsContainer.classList.add("hide");
                optionsContainer.style['z-index']=-1;
				optionsContainer.classList.remove("active");
                optionsContainer.style['pointer-events']='none';

                if ($ASELECT.cb!=null) {
                    if ($ASELECT.cb_ddclose!=null) {
                        $ASELECT.cb_ddclose();
                    }
                    $ASELECT.cb(selval, val2);
                }
			});
        }

        const clickout = function(){
            console.log("clickout");
            optionsContainer.style['z-index']=-1;
            optionsContainer.style['max-height']=0;
            optionsContainer.style['pointer-events']='none';
            optionsContainer.classList.add("hide");
            document.removeEventListener("click", clickout);
            optionsContainer.classList.toggle("active");
            if ($ASELECT.cb_ddclose!=null) {
				$ASELECT.cb_ddclose();
	        }
		}

        optionsContainer.addEventListener("mouseout", $e=> {
            if (optionsContainer.classList.contains("active")==true) {
                const fromElement = $e.relatedTarget;
                if (!optionsContainer.contains(fromElement)) {
                    clickout();
                }
            }
        });

        selectedsb.addEventListener("click", ($e) => {
            $e.stopPropagation();
            $e.preventDefault();

            optionsContainer.classList.remove("hide");
            optionsContainer.style['pointer-events']='all';

            //this will add hide.
			if (optionsContainer.classList.contains("active")==true) {
                console.log("oc active, make inactive");
                clickout();
            }

            //this will have no hide.
            else {
                console.log("oc inactive, make active");
                document.addEventListener("click", clickout);
                optionsContainer.style['z-index']=9999;
                optionsContainer.style['opacity']=1;
                var mh = optionsContainer.getAttribute("mh_backup");
                optionsContainer.style['max-height']=mh;
                optionsContainer.classList.toggle("active");
                if ($ASELECT.cb_ddopen!=null) {
					$ASELECT.cb_ddopen();
				}
            }   
           
		});
    }

    

    /**
     * Standardized link color maker based on the neuron (not the relon)
     * This will be used to generated links for neurons.
     */
    getLinkColorClass = function($nObj) {
        if ($nObj==undefined || $nObj==null) return "link-black";
        if ($nObj.ghost==true) return "link-blue";

        //presume ANNE ownership if not set (in the js)
        if (ANNE.isANNE($nObj.owned_by)==true 
            || $nObj.owned_by==ANNE.MEM.ANNE_ACCTID) return "link-yellow";  //anne yellow
        
        else {
            if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
                //possible hanai could be here.  Look for pending hanai.
                if ($nObj.bPENDING_HANAI==undefined) {
                    var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
                    if (pending_hanai!=null) {
                        $nObj.bPENDING_HANAI=true;
                        return "link-pink";  //hanai pink
                    }
                }
                return "link-red";   //red for self owned
            }
            else {
                return "link-grey";  //black for someone else's neuron (not ANNE's).
            }
        }
    }


    /**
     */
    getNeuronColorClass = function($nObj, $pre="title-", $post, $default="other") {
        if ($pre!="" && $pre[ $pre.length-1]!="-") $pre = $pre+"-";
        if ($post!="" && $post[ $post.length-1]!="-") $post = "-"+$post;

        if (ANNE.isANNE($nObj.owned_by)==false && $nObj.ghost==false && $nObj.bPENDING_HANAI==undefined) {
            var pending_hanai = ANNE.HANAI.getPendingHanai($nObj);
            if (pending_hanai!=null) {
                $nObj.bPENDING_HANAI=true;
                return $pre+"hanai"+$post;  
            }
        }
        else if ($nObj.bPENDING_HANAI==true) return $pre+"hanai"+$post; 

        else if ($nObj.ghost==true) {
            return $pre+"ghost"+$post;  
        }
        else if (ANNE.isANNE($nObj.owned_by)==true) {
            return $pre+"anne"+$post;  
        }
        else if ($nObj.owned_by==USER._userNObj.nid) {
            return $pre+"self"+$post; 
        }
        return $pre+$default+$post;  //hanai pink
    }


    /**
     */
    addClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($classname) == false) el.classList.add($classname);
        return el;
    }

    
    /**
     * Quickly add an ai-ico class image (see anne-icos.css).
     * 
     * @param {*} $el 
     * @param {*} $icoClass 
     * @returns 
     */
    addIco = function ($el, $icoClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $icoClass;
        el.appendChild(span);
    }


    addRemoveClass = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }

    addRemoveClasses = function ($el, $addclass, $removeClasses) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var r = 0; r < $removeClasses.length; r++) {
            if (el.classList.contains($removeClasses[r]) == true) el.classList.remove($removeClasses[r]);
        }
        if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        return el;
    }

    /**
     * Same as addRemoveClass but checks if the removeclass is on.
     */
    addRemoveClassIf = function ($el, $addclass, $removeclass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($addclass) == false) {
            if (el.classList.contains($addclass) == false) el.classList.add($addclass);
        }
        if (el.classList.contains($removeclass) == true) el.classList.remove($removeclass);
        return el;
    }


    animateProp = function ($el, $prop, $propVal, $timestring = "0.5s ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        el.style[$prop] = $propVal;
        el.style.transition = $prop + " " + $timestring;
    }
    animateProps = function ($el, $jsonprops, $dur = ".5s", $delay = "0", $ease = "ease-out", $removeClass = "hide") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($removeClass) el.classList.remove($removeClass);
        var keys = Object.keys($jsonprops);
        var propstrs = [];
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            propstrs.push(prop);
        }
        el.style['transition-property'] = propstrs.join(",");
        el.style['transition-duration'] = $dur;
        el.style['transition-timing-function'] = $ease;
        if ($delay != "") el.style['animation-delay'] = $delay;

        //set the props after the transition styling.
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsonprops[prop];
            el.style[prop] = val;
        }
    }





    /**
     * Disambig chases of information from external lookups.
     * @param {*} $neurons 
     * @param {*} $val 
     * @param {*} $ambigObj 
     * @param {*} $el_console 
     * @param {*} $resolveAmbigs 
     * @returns 
     */
    disambigInner=async($neurons, $val, $ambigObj, $el_console, $resolveAmbigs)=> {
        if ($neurons==undefined || $neurons==null || $neurons.length==0) {
            return $resolveAmbigs(); //call the next one.
        }
        var el_dchoice = document.createElement("div");
        var sp = document.createElement("div");
        sp.innerHTML="original text: <b>"+$val+"</b>";
        sp.style.display="block";
        el_dchoice.appendChild(sp);

        //the choice (and all options)
        $el_console.appendChild(el_dchoice);

        for (var n=0; n < $neurons.length; n++) {
            var el_choice = document.createElement("div");
            el_choice.classList = "neuron-choice";
            el_dchoice.appendChild(el_choice);
            
            var el_choice_info = document.createElement("div");
            el_choice_info.classList = "neuron-choice-info";
            
            var nObj = $neurons[n];
            await ANNE.GUI_STUFF.disambigHTMLStr(el_choice_info, nObj, ["sp"]);
            el_choice.appendChild(el_choice_info);
            
            var el_choice_btn = document.createElement("div");
            el_choice_btn.classList = "neuron-choice-btn-holder";  
            el_choice.appendChild(el_choice_btn);
            
            var el_btn = ANNE.A.ezButton(el_choice_btn, "", "choose match "+(n+1), "anne-tiny-btn btn-purple-sq btn-maxH", async(e)=>{
                var nid_x = e.target.getAttribute("nid");
                var nObjx = MEMG.getNeuronByNid(nid_x);
                
                ANNE.A.empty(e.target.parentElement, true);//empty and hide choice div.
                
                //key to set this.
                $ambigObj.jsonobj.to_s = nObjx.nid;
                return await $resolveAmbigs(); //call the next one.
            });
            el_btn.setAttribute("nid", nObj.nid);
            el_btn.setAttribute("val", $val);
            $el_console.appendChild(el_dchoice);
        }

        //skip option.
        var el_skip = document.createElement("div");
        el_skip.classList = "neuron-choice";
        el_dchoice.appendChild(el_skip);

        var el_skip_info = document.createElement("div");
        el_skip_info.classList = "neuron-choice-skip";
        el_skip.appendChild(el_skip_info);     
        var el_skip_btn = document.createElement("div");
        el_skip_btn.classList = "neuron-choice-btn-holder";  
        el_skip.appendChild(el_skip_btn);
        var el_btn = ANNE.A.ezButton(el_skip_btn, "", "SKIP THIS", "anne-tiny-btn btn-red-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true);//empty and hide choice div.
            $ambigObj.jsonobj.to_s = null;
            return await $resolveAmbigs(); //call the next one.
        });
        

        //make new
        var el_new = document.createElement("div");
        el_new.classList = "neuron-choice";
        el_dchoice.appendChild(el_new);
        var el_new_info = document.createElement("div");
        el_new_info.classList = "neuron-choice-new";
        el_new.appendChild(el_new_info);     
        var el_new_btn = document.createElement("div");
        el_new_btn.classList = "neuron-choice-btn-holder";  
        el_new.appendChild(el_choice_btn);
        var el_btn = ANNE.A.ezButton(el_choice_btn, "", "NEW", "anne-tiny-btn btn-blue-sq btn-maxH", async(e)=>{
            ANNE.A.empty(el_dchoice. true); //empty and hide choice div.
            console.log($ambigObj);
            $ambigObj.jsonobj.to_s = null;
            //this is going to need a callback to make a "NEW" correctly!!!!
            return await $resolveAmbigs(); //call the next one.
        });
        $el_console.appendChild(el_dchoice);
    }


    /**
     * Generalized ez annies entry.
     * i_min_val, i_max_val and i_placeholder attrs can be added to the outter button.
     * They will passthrough to the input field.
     * 
     * @param {*} $el 
     * @param {*} $cb_submit - submit callback
     * @param {*} $btnTitle - title of outter button
     * @param {*} $cb_entry - callback on entry of data (if more than just formatting of annies is desired)
     * @param {*} $idpre - Used to create a unique ezAnnecoin_entry if there is more than one in the DOM at once.
     * @returns 
     */
    ezAnnecoin_entry = function($el, $cb_submit=null, $btnTitle="accept available risk", $btn2Title="accept risk", 
                                $cb_entry=null, $idpre="unique", $btnClassStr = "noround btn-someanniesaction") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_holder = ANNE.A.ezEl(el, "div", "flexrow");
        const idpre = $idpre;
        const btn_aar = ANNE.A.ezButton(el_holder, "btn_main_"+$idpre, $btnTitle, $btnClassStr, async(e)=>{
            
            //snag passthrough attrs we put on button AFTER we created here.
            //these will be ready for this event since we created them outside of this funct on the return
            //of the gui creation. (optional attrs)
            var min_val = btn_aar.getAttribute("i_min_val");
            var max_val = btn_aar.getAttribute("i_max_val");
            var placeholder = btn_aar.getAttribute("i_placeholder");

            //create the btn clone so we get the button input button functionality
            var btnClone = btn_aar.cloneNode(true);
            const disp = document.createElement("div");
            btnClone.id = "btn_"+idpre+"_final";
            btnClone.setAttribute("tied_input", "input_"+idpre);
            ANNE.A.hide(btn_aar);
    
            const input_field = document.createElement("input");
            input_field.setAttribute("size", 24);
            input_field.setAttribute("type", "text");
            input_field.id="input_"+idpre;
            input_field.setAttribute("el_display", "annies_"+idpre+"_disp");
            input_field.setAttribute("el_submit", "btn_"+idpre+"_final");
            if (min_val!=undefined) input_field.setAttribute("min_val", min_val);
            if (max_val!=undefined) input_field.setAttribute("max_val", max_val);
            if (placeholder!=undefined) input_field.setAttribute("placeholder", placeholder);
            
            //add th input field after the btn_aar
            btn_aar.after(input_field);

            var max_btn = null;
            var min_btn = null;


            //add max btn
            if (max_val!=undefined) {
                max_btn = document.createElement("button");
                max_btn.innerHTML="max";
                max_btn.classList = 'btn-minmax';
                max_btn.addEventListener("click", function($e){
                    input_field.value = max_val;
                    ANNE.A.hide(max_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(max_btn)
            }

            //add max btn
            if (min_val!=undefined) {
                min_btn = document.createElement("button");
                min_btn.classList = 'btn-minmax';
                min_btn.innerHTML="min";
                min_btn.addEventListener("click", function($e){
                    input_field.value = min_val;
                    ANNE.A.hide(min_btn);
                    ANNE.A.show(btnClone);
                    if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
                })
                input_field.after(min_btn)
            }


            //add input input event
            input_field.addEventListener("input", async($e)=>{
                await ANNE.A.anne_entered($e);
                if (typeof $cb_entry=="function") $cb_entry(input_field.value, disp);
            });


            //a display for the formatted/summary version of what the user typed.
            disp.classList="flexrow";
            disp.id="annies_"+idpre+"_disp";

            //if we have max, add disp after it.  else add disp after input field.
            if (max_btn!=null) max_btn.after(disp);
            else if (min_btn!=null) min_btn.after(disp);
            else input_field.after(disp);

            disp.after(btnClone);
            btnClone.innerHTML=$btn2Title;
            ANNE.A.hide(btnClone);

            //add submit callback.
            if (typeof $cb_submit=="function") {
                btnClone.addEventListener("click", async($e)=>{
                    var val = input_field.value;
                    $cb_submit(val)
                });
            }
        });
        return el_holder;
    }


    /**
     * annies entered
     * This can handle el-display (and display value entered)
     * This can handle el-submit (and show a submit btn after valid entry)
     * @param {*} $e 
     */
    anne_entered = async($e, $target=null)=>{
        var el_input = $target;
        if (el_input==null && $e!=null && $e.target!=null) {
            el_input = $e.target;
        }
        if (el_input==null) return;
        
        var val = el_input.value+"";
        var lastChar = val[val.length-1];

        var el_display = el_input.getAttribute("el_display");
        if (el_display!=null) el_display = document.getElementById(el_display);

        var el_submit = el_input.getAttribute("el_submit");
        if (el_submit!=null) el_submit = document.getElementById(el_submit);

        var minval = el_input.getAttribute("min_val");
        var maxval = el_input.getAttribute("max_val");
        
        var final_val = 0;

        if (lastChar=="k" || lastChar=="K") {         //shorthand for thousands
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000;
            el_input.value = newval;
        }
        else if (lastChar=="m" || lastChar=="M") {    //shorthand for millions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000;
            el_input.value = newval;
        }
        else if (lastChar=="b" || lastChar=="B") {    //shorthand for billions
            var newval = Number.parseFloat(ANNE.A.getValue(el_input)) * 1000000000;
            el_input.value = newval;
        }
        else if ("0123456789.,".indexOf(lastChar)==-1) {
            var newval = "";
            var amt = el_input.value;
            for (var a=0; a < amt.length; a++) {
                if ((a!=(amt.length-1)) && ("kKmMbB".indexOf(amt[a])!=-1)) continue;
                else if ("0123456789.,".indexOf(amt[a])!=-1) continue;
                newval = newval+amt[a];
            }
            el_input.value = newval;
        }

        var val  = el_input.value;
        if (val=="") {
            return;
        }
        final_val = parseInt(val);

        if (minval!=undefined && final_val < minval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='min-msg'>minimum "+minval+" annecoin</span>"); 
        }

        else if (maxval!=undefined && final_val > maxval) {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.setHTML(el_display, "<span class='max-msg'>max "+maxval+" annecoin</span>"); 
        }

        else if (final_val>0) {
            if (el_submit!=null) ANNE.A.show(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, final_val, true); 
        }
        else {
            if (el_submit!=null) ANNE.A.hide(el_submit);
            if (el_display!=null) ANNE.A.ezAnnecoinDisplay(el_display, 0, true); 
        }
    }

    appendHTML = function($el, $html) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = el.innerHTML+$html;
    }

    append = function ($el, $misc, $bSHOW = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.append($misc);
        if ($bSHOW == true) this.show(el);
        return el;
    }

    appendStringList=function($el, $strArr, $size="10px", $class="") {
        var el = this.__getEl($el);
        if (el==null) {
            console.log("missing el");
            return null;
        }
        if ($strArr==undefined || $strArr==null || Array.isArray($strArr)==false || $strArr.length==0) return;
        var spanlist=document.createElement("span");
        spanlist.style['display']="inline-block";
        spanlist.style['font-size']=$size;
        if ($class!="") spanlist.classList.add($class);
        if ($class!="") {
            spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        }
        else spanlist.innerHTML="("+$strArr.join(", ").trim()+")";
        el.append(spanlist);
    }

    addEmojisFromRelTos=function($el, $rels, $title) {
        var el = this.__getEl($el);
        if (el == null) return null;

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var sp = document.createElement("span");
                sp.innerHTML = toNOBJ._uico;
                sp.setAttribute("title", $title+" "+toNOBJ.bestNameHTML());
                sp.setAttribute("txid", rel.txid);
                el.appendChild(sp);
            }
        }
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    getAnum=function($str) {
        let annum = 0;
        const normalizedText = $str.normalize("NFD").replace(/[\u0300-\u036f]/g, "").toLowerCase().trim();
        for (let i = 0; i < normalizedText.length; i++) {
            const c = normalizedText.charAt(i);
            const asciiValue = c.charCodeAt(0);
            if (asciiValue < 49) continue;
            if (asciiValue <= 57) {
                annum += asciiValue - 48;
                continue;
            }
            else if (asciiValue < 97 || asciiValue > 123) {
                continue;
            }
            else annum += asciiValue - 96;
        }
        return annum;
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $text 
     * @param {*} $classpost 
     */
    doAnumWithCalc = function($el, $text, $classpost="") {
        var num = this.getAnum($text+"");
        this.doAnum($el, num+"", $classpost);
        return num;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $num 
     * @param {*} $classpost 
     * @returns 
     */
    doAnum=function($el, $num, $classpost="") {
        var el = this.__getEl($el);
        if (el == null) return null;

        if(el.classList.contains("anumbox"+$classpost)==false) el.classList.add("anumbox"+$classpost);
        el.innerHTML="";

        $num=$num+"";
        
        for (var n=0; n <$num.length; n++) {
            var sp = document.createElement("span");
            sp.classList = "anum"+$classpost;
            sp.innerHTML = $num[n];
            el.appendChild(sp);
        }
        return;
    }


    /**
     * 
     * @param {*} $rels 
     * @param {*} $title 
     * @param {*} $skips 
     * @returns 
     */
    getEmojisFromRelTos_html=function($rels, $title, $skips=[]) {
        if (ANNE.SETTINGS && ANNE.SETTINGS.bUICOS==false) return "";
        var html = "";
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if ($skips.indexOf(rel.txid)!==-1) continue;
            var toNOBJ = MEMG.getNeuronByTempNid(rel.t);
            if (toNOBJ._uico!=undefined) {
                var span = "<span title='"+$title+" "+toNOBJ.bestNameHTML()+"' txid='"+rel.txid+"'>"+toNOBJ._uico+"</span>";
                html+=span;
                $skips.push(rel.txid);
            }
        }
        return html;
    }


    click = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.click();
        return el;
    }

    //this clones the DOM node.
    //DOM node doesn't have to be in the DOM.. it can just be a temp made
    //el that is not yet appended.
    cloneAndAppend = function ($getel, $putel, $bCLONE = true, $bSHOWPUTEL = true) {
        if ($bCLONE == true) {
            var getel = this.__getEl($getel);
            if (getel == null) return null;
            var clone = getel.cloneNode(true);
        }
        else clone = this.__getEl($getel);
        if (clone == null) return null;
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }

    //this clones the .content of the template
    cloneAndAppendTemplate = function ($getel, $putel, $bSHOWPUTEL = true) {
        var getel = this.__getEl($getel);
        if (getel == null) return null;
        var clone = getel.content.cloneNode(true);
        var putel = this.__getEl($putel);
        if (putel == null) return null;
        putel.appendChild(clone);
        if ($bSHOWPUTEL == true) this.show($putel);
        return clone;
    }



    createElement = function ($el, $tag, $class = "", $html = "", $value = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newtag = document.createElement($tag);
        if ($class != "") newtag.classList = $class;
        if ($html != "") newtag.innerHTML = $html;
        if ($value != "") newtag.value = $value;
        el.append(newtag);
        return newtag;
    }

    /**
     */
    doImageDyno = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile - ANNE FILE
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDynoB64 = function ($div, $imgElId = "", $afile=null, $imgClass = "left_dyno_w") {
        
        if ($afile == null) return null;
        if ($afile.b64data==null) return null;

        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2 = function ($div, $imgElId = "", $src = "", $biggest = 200, $imgClass = "left_dyno_w") {
        if ($src == "") return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }
            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $afile 
     * @param {*} $biggest 
     * @param {*} $imgClass 
     * @returns 
     */
    doImageDyno2_B64 = function ($div, $imgElId = "", $afile = null, $biggest = 200, $imgClass = "left_dyno_w") {
        if ($afile == null) return;
        var newImgEl = document.createElement('img');
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        if ($imgClass != "") this.__classListParse(newImgEl, $imgClass);  //need to do this elsewhere
        newImgEl.addEventListener('error', function () {
            console.log('Error loading image');
            return;
        });
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            ANNE.A.append($div, newImgEl);
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            var biggestD = realW;
            var bgd = "w";
            if (realH > realW) {
                biggestD = realH;
                bgd = "h";
            }
            if (biggestD > $biggest) {
                if (bgd == "w") {
                    newImgEl.setAttribute("width", $biggest);
                    var newH = $biggest / ratio;  //(>1 ratio)
                    newImgEl.setAttribute("height", newH);
                }
                else {
                    newImgEl.setAttribute("height", $biggest);
                    var newW = $biggest * ratio; //<1 ratio
                    newImgEl.setAttribute("width", newW);
                }

            }
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;
        if ($afile.b64data.indexOf("data:")!=-1) newImgEl.setAttribute('src', $afile.b64data); 
        else newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
        return newImgEl;
    }


    /**
     */
    doBadImg = function ($div, $imgElId = "", $src = "", $imgClass = "left_dyno_w") {
        var newImgEl = this.ezEl($div, "span", $imgClass + " img-bad-audit", "FAILED<BR>AUDIT");
        if ($imgElId != "") newImgEl.setAttribute("id", $imgElId);
        newImgEl.style.top = "0px";
        if ($src != undefined && $src != null && $src != "") {
            newImgEl.setAttribute('bad-audit-url', $src);
        }
        var attr_fix = newImgEl.getAttribute("ssfixpostload");
        if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        return newImgEl;
    }



    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {*} $src 
     * @param {*} $w 
     * @param {*} $h 
     * @param {*} $imgClassAdd 
     * @param {*} $max 
     * @returns 
     */
    doImage = function ($div, $imgElId, $src, $w = 600, $h = 900, $imgClassAdd = "", $max = 0) {
        console.log("src "+$src);
        
        const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);
        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
        newImgEl.onerror = function(image) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($max && ($w / ratio) > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", $w);
                }
                else {
                    if ($max!=0 && realH > $max) {
                        newImgEl.setAttribute("height", $max);
                    }
                    else newImgEl.setAttribute("width", realW);
                }
            }
            else if ($h) {
                if ($h < realH) {
                    if ($max && (ratio * $h) > $max) {
                        newImgEl.setAttribute("width", $max);
                    }
                    newImgEl.setAttribute("height", $h);
                }
                else newImgEl.setAttribute("height", realH);
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        newImgEl.setAttribute('src', $src); // Loads the image.
        return newImgEl;
    }


    /**
     * 
     * @param {*} $div 
     * @param {*} $imgElId 
     * @param {anne_file} $afile 
     * @param {int} $w 
     * @param {int} $h 
     * @param {string} $imgClassAdd 
     * @param {int} $maxOtherD 
     * @returns 
     */
    doImageB64 = function ($div, $imgElId, $afile=null, $w = 600, $h = 900, $imgClassAdd = "", $maxOtherD = 0) {      
		const newImgEl = document.createElement('img');
        newImgEl.setAttribute("id", $imgElId);

		if ($afile==null) {
			console.log("must have an afile here");
			return newImgEl;
		}

        if ($imgClassAdd != "") this.__classListParse(newImgEl, $imgClassAdd);  //need to do this elsewhere
		
        newImgEl.onerror = function($img) {
            newImgEl.src = "";
            newImgEl.onerror = null; // Prevent infinite loop if fallback image also fails to load
            return;
        }
        
        newImgEl.addEventListener('load', function () {
            newImgEl.style.top = "0px";
            var realW = this.width;
            var realH = this.height;
            var ratio = realW / realH;
            if ($w && $h) {
                newImgEl.setAttribute("width", $w);
                newImgEl.setAttribute("height", $h);
            }
            else if ($w) {
                if ($w < realW) {
                    if ($maxOtherD && ($w / ratio) > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                    } 
                    else {
                        newImgEl.setAttribute("width", $w);
                        newImgEl.setAttribute("height", $w / ratio); 
                    }
                } 
                else {
                    if ($maxOtherD != 0 && realH > $maxOtherD) {
                        newImgEl.setAttribute("height", $maxOtherD);
                        newImgEl.setAttribute("width", $maxOtherD * ratio); 
                    } 
                    else {
                        newImgEl.setAttribute("width", realW);
                        newImgEl.setAttribute("height", realH);
                    }
                }
            } 

            else if ($h) {
                if ($h < realH) {
                    if ($maxOtherD && (ratio * $h) > $maxOtherD) {
                        newImgEl.setAttribute("width", $maxOtherD);
                        newImgEl.setAttribute("height", $maxOtherD / ratio); // Adjust height to maintain aspect ratio
                    } 
                    else {
                        newImgEl.setAttribute("height", $h);
                        newImgEl.setAttribute("width", ratio * $h); // Adjust width to maintain aspect ratio
                    }
                } 
                else {
                    // If the specified height is greater than or equal to the original height
                    newImgEl.setAttribute("height", realH);
                    newImgEl.setAttribute("width", realW);
                }
            }
            ANNE.A.append($div, newImgEl);
            var attr_fix = newImgEl.getAttribute("ssfixpostload");
            if (attr_fix != undefined) ANNE.A.smartScroll(attr_fix);
        });
        $afile.imgel = newImgEl;

		if ($afile.b64data==undefined || $afile.b64data==null) {
			console.log("problem with b64 data");
			return newImgEl;
		}
		
        if ($afile.b64data.indexOf("data:")!=-1) {
			newImgEl.src = $afile.b64data;
		}
        else {
			newImgEl.setAttribute('src', 'data:'+$afile.mimetype+';base64,'+$afile.b64data);
		}
        return newImgEl;
    }

    emptySelfAfterChild=function($el, $childIndex=0, $bHIDE=false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var c=0; c < el.children.length; c++) {
            if (c<=$childIndex) continue;
            el.removeChild( el.children[c]);
            c=c-1;
        }
        if ($bHIDE == true) this.hide($el);
        return el;
    }

    emptySelf = function($el, $bHIDE=false){ return this.empty($el, $bHIDE);}   //i keep thinking it is named this.
    empty = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        while (el.firstChild) el.removeChild(el.firstChild);
        el.innerHTML = "";
        if ($bHIDE == true) this.hide($el);
        return el;
    }
    emptyParent = function ($el, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        if (el_parent) this.empty(el_parent);
        else return el;
        if ($bHIDE == true) this.hide(el_parent);
        return el_parent;
    }

    emptyAllBut=function($el, $skipEl, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_parent = el.parentNode;
        for (var c=0; c < el_parent.children.length; c++) {
            var child = el_parent.children[c];
            if (child==$skipEl) continue;
            el_parent.removeChild(child);
        }
        if ($bHIDE == true) this.hide(el);
        return;
    }


    ezAlert = function ($el, $class, $id_pre, $id, $msg = "", $insertAfter = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        const alert = document.createElement("div");
        var classList = "alert fade show";
        if ($class) classList = classList + " " + $class; //could be multiple ones.
        alert.classList = classList;
        alert.setAttribute("role", "alert");
        if ($id_pre != "" && $id != "") alert.setAttribute("id", $id_pre + "_" + $id);
        else if ($id != "") alert.setAttribute("id", $id);
        if ($msg != "") alert.innerHTML = $msg;
        if ($insertAfter != null) $insertAfter.after(alert);
        else el.append(alert);
        return alert;
    }


    /**
     * HTML is silly.
     * @param {*} $el - the select el or string of the id
     * @param {String} $attr - property to extract 
     * @returns 
     */
    getSelectedProp=function($el, $attr) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectedOption = el.options[el.selectedIndex];
        if (selectedOption==null) return null;
        return selectedOption.getAttribute($attr);
    }

    //var el_lvl = ANNE.A.setHTML("annex_lvl", aplex.price_lvl);
    ezAnniesToAnneDisplay = function ($el, $val = 0, $bdclass="balance-display", $ico="fermatta1rem", $valClass="annecoin-balance-value", $bRESET=true) {
        var el = this.__getEl($el);
        if (el == null) return null;

		if ($bRESET==true) el.innerHTML="";
		
        if ($val == 0) {
            el.classList.add("text-danger-lbl");
            el.innerHTML = "0";
            return;
        }
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var span = ANNE.A.ezSpan(el, "", $bdclass);
        var val = UTILS.formatNumberCommas(annecoin);
        //console.log(el.id + " val:" + val);
        ANNE.A.ezSpan(span, "", $valClass, val + "");
        ANNE.A.ezSpan(span, "", $ico);
    }

    /**
    * annies are never decimal, only whole
    */
    validateAndFormatAnnecoin = function ($el_input, $displayEl, $annecoinstring, $bRESET = true) {
        var str = "";
        var lastchar = $annecoinstring[$annecoinstring.length - 1];
        var multiplier = 0;
        if (lastchar == "k") {
            multiplier = 1000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "m") {
            multiplier = 1000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        else if (lastchar == "b") {
            multiplier = 1000000000;
            $annecoinstring = $annecoinstring.substring(0, $annecoinstring.length - 1)
        }
        for (var s = 0; s < $annecoinstring.length; s++) {
            if ("0123456789".indexOf($annecoinstring[s]) == -1) continue;
            str = str + $annecoinstring[s];
        }
        if (multiplier != 0) str = (parseInt(str) * multiplier) + "";
        $el_input.value = str;

        if (str.length > 0) {
            this.ezAnnecoinDisplay($displayEl, str, $bRESET);
        }
    }

    
    /**
     * 
     */
    ezAddNote = function($el, $noteText, $class="small-note") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.innerHTML = $noteText;
        span.classList = $class;
        el.append(span);
    }

    displayAnnecoinWithCommas = function ($el, $annecoin = 0, $bRESET=false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = ANNE.UTILS.commas($annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + annecoin + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    ezAnnecoinDisplay = function ($el, $annecoin = 0, $bRESET = false, $pre = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($annecoin=="" || $annecoin==undefined) $annecoin=0;
        if ($bRESET == true) ANNE.A.empty(el);
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = Math.round(parseInt($annecoin)*100)/100;
        var acval = UTILS.formatNumber(annecoin);
        if ($pre != "") $pre = "&nbsp;" + $pre;
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", $pre + acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico");
    }

    /**
     * will round for better display
     */
    ezAnneDisplayApprox = function ($el, $val = 0, $bRESET = false, $addon=null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bRESET == true) ANNE.A.empty(el);
        if ($val == 0) {
            var span = ANNE.A.ezSpan(el, "", "balance-display");
            ANNE.A.ezSpan(span, "", "annecoin-balance-value", 0 + "");
            ANNE.A.ezSpan(span, "", "annecoin-small-ico")
            return;
        }
        var span = ANNE.A.ezSpan(el, "", "balance-display");
        var annecoin = $val / ANNE.ANNE_TO_ANNIES;
        var acval = UTILS.formatNumber(annecoin, true);
        ANNE.A.ezSpan(span, "", "annecoin-balance-value", acval + "");
        ANNE.A.ezSpan(span, "", "annecoin-small-ico")
        if ($addon!=null) ANNE.A.ezSpan(span, "", "", $addon);
    }
    


    /**
     * 
     * Get and display our annies balance.
     * 
     * Get and display the annies balance in both ANNE coin and annies.
     */
    ezDisplayAnnecoinWithNeuron = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj!=null) {
            if (balanceObj.annies != undefined) {
                var annies = parseInt(balanceObj.annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    ezDisplayPendingAnnies = async ($el, $nObj, $lbl = "") => {
        var balanceObj = $nObj.balanceObj;
        if (balanceObj == null && $nObj.ghost == false && $nObj.temp == false) {
            balanceObj = await USER.checkBalance($nObj.nid);
        }
        if (balanceObj) {
            if (balanceObj.avail_soon_annies != undefined) {
                var annies = parseInt(balanceObj.avail_soon_annies);
                ANNE.GUI_STUFF.doAnniesToAnneSpanFull($el, $lbl, annies);
            }
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {*} $value 
     * @param {*} $lbl 
     * @param {*} $checked 
     * @param {*} $disabled 
     * @param {*} $truncateClass 
     * @param {*} $cb 
     * @returns 
     */
    ezCheckMark = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, 
                            $truncateClass = "", $cb=null, $chkclass="chk-parent") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($id=="" || $id==null) $id = ANNE.UTILS.randHex(3);
        const div = document.createElement("div");
        if ($chkclass!="") div.classList.add($chkclass);
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "checkbox");
        if ($id!="") chk.setAttribute("id", $id);
        if ($id!="") chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.classList="chk-label";
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);

        //simple callback
        if ($cb) {
            chk.addEventListener("change", $e=>$cb($e));
        }
        return chk;
    }


    /**
     */
    ezInputText = function ($el, $id = "", $classList = "", $placeholder = "", $inputEventCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var el_input = document.createElement("input");
        el_input.setAttribute("type", "text");
        if ($classList != "") this.__classListParse(el_input, $classList);
        if ($id != "") el_input.setAttribute("id", $id);
        if ($placeholder) el_input.setAttribute("placeholder", $placeholder);
        el.append(el_input);
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            el_input.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                $inputEventCallback(val);
            });
        }
        return el_input;
    }


    /** 
     * @param {element or string} - parent element to append to
     * @param {anne_neuron} - class to apply uom to.
     */
    ezInput = function ($el, $classNOBJ = null, $id = "", $class = "",
        $placeholder = "", $inputEventCallback = null, $inputFilterCallback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($class != "") {
            var div = document.createElement("div");
            div.classList = $class;
            el.append(div);
        }
        else div = el;

        const input = document.createElement("input");
        input.setAttribute("type", "text");
        if ($id != "") input.setAttribute("id", $id);
        if ($placeholder) input.setAttribute("placeholder", $placeholder);
        div.append(input);
        var uom = null;
        if ($classNOBJ != null) uom = MEM.unitOfMeasure($classNOBJ);
        if (uom != null) {
            var span = document.createElement("span");
            span.style.display = 'inline-block';
            span.classList = 'uom-symbol';
            div.appendChild(span);
        }
        if ($inputEventCallback != null && typeof $inputEventCallback === "function") {
            div.addEventListener("input", async ($e) => {
                var val = $e.target.value;
                if ($inputFilterCallback != null && typeof $inputFilterCallback === "function") {
                    val = await $inputFilterCallback(val);
                    $e.target.value = val;
                }
                if (uom != null) {
                    var thisconv = null;
                    if (uom.conversions != undefined && uom.conversions["_" + val] != undefined) {
                        thisconv = uom.conversions["_" + val];
                    }
                    else thisconv = selector_number.convert(val, uom);
                    span.innerHTML = thisconv.total + "<span>&nbsp;" + thisconv.post + "</span>";
                }
                $inputEventCallback(val);
            });
        }
        return div;
    }


    //should do radios.
    ezRadio = function ($el, $id, $value, $lbl, $checked = true, $disabled = false, $truncateClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const div = document.createElement("div");
        if ($truncateClass != "") div.classList.add($truncateClass);
        el.append(div);
        const chk = document.createElement("input");
        chk.setAttribute("type", "radio");
        chk.setAttribute("id", $id);
        chk.setAttribute("name", $id);
        chk.setAttribute("value", $value);
        if ($checked) chk.setAttribute("checked", true);
        if ($disabled) chk.setAttribute("disabled", true);
        div.append(chk);
        const lbl = document.createElement("label");
        lbl.setAttribute("for", $id);
        lbl.innerHTML = $lbl;
        div.append(lbl);
        return chk;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @param {*} $cb 
     */
    injectInputAfter=function($el, $wrapperclass, $inputclass, $cb, $bDOCANCEL=true){
        var el = this.__getEl($el);
        if (el == null) return null;
        var wrapperdiv=document.createElement("div");
        if ($wrapperclass!="") wrapperdiv.classList = $wrapperclass;
        var newel = document.createElement("input");
        if ($inputclass!="") newel.classList = $inputclass;
        newel.style['flex-grow']=1;
        newel.style.cursor='auto';

        wrapperdiv.append(newel);
        this.appendAfter(el, wrapperdiv);
        ANNE.A.makeReturnPressEvent(newel, $cb);
        ANNE.A.injectEnterAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        })
        if ($bDOCANCEL==true) ANNE.A.injectCancelAfter(wrapperdiv, async($e)=>{
            $cb(newel);
            ANNE.A.removeSelf(wrapperdiv);
        });
        return newel;
    }

    injectEnterAfter = function($el, $cb, $class="") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var btnEnter = document.createElement("button");
        btnEnter.innerHTML = "enter";
        if ($class!="") btnEnter.classList = $class;
        el.append(btnEnter);
        btnEnter.addEventListener("click", $cb);
    }

    injectCancelAfter = function($el, $cb, $class="btn-red-sq") {
        var el = this.__getEl($el);
        if (el == null) return null;
        const btnCancel = document.createElement("button");  
        if ($class!="") btnCancel.classList=$class;         
        btnCancel.innerHTML = "[x]";
        el.appendChild(btnCancel);
        btnCancel.addEventListener("click", $cb);
    }


    /**
     * Make a return press event to callback the $cb.
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeReturnPressEvent = async($el, $cb)=>{
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($cb!=null) {
            el.addEventListener("keyup", async($e)=>{
                if ($e.keyCode === 13) {
                    $cb(el);
                }
            });    
        }
        return;
    }

    
    /**
     * Make a new element just like the $el and inject it after the $el as a sibbling.
     * @param {*} $el 
     * @returns 
     */
    makeSibAfter=function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var clone = $el.cloneNode(true);
        clone.innerHTML="";
        clone.id="newidclon_"+Math.floor(Math.random()*1000);
        this.appendAfter(el, clone);
        return clone;
    }


    /**
     * 
     * @param {anne_gui_role} $agr - the gui role object 
     */
    ezRoleEdit = function ($el, $clump_nid, $rel, $guid, $bEDIT = false, $bMOD = false, $el_inject = "", $newcb = null, $cancelcb = null, $deletecb = null, $span2 = null) {
        var span2 = $span2;
        if (span2 == null) span2 = this.ezRoleDisplay($el, $rel, $guid);
        else span2 = this.__getEl($span2);

        var el_inject = this.__getEl($el_inject);
        if (el_inject) {
            span2.setAttribute("el_inject", $el_inject.id);
        }

        span2.setAttribute("clump_nid", $clump_nid);
        span2.setAttribute("guid", $guid);

        if ($bEDIT == true) {
            var btnAdd = document.createElement("button");
            btnAdd.innerHTML = "+add";
            btnAdd.id = "add_" + $rel.rid + "_" + $guid;
            span2.append(btnAdd);
            btnAdd.addEventListener("click", e => ANNE.A.editRoleLaunchSelector(span2, $newcb, $cancelcb));
            span2.setAttribute("delete_ok", 1);
            btnAdd.style['margin-left'] = '4px';

            if ($rel.type == TYPE_ROLE) {
                var btndelete = document.createElement("button");
                btndelete.innerHTML = "[remove]";
                btndelete.id = "delete_" + $rel.rid + "_" + $guid;
                span2.append(btndelete);
                btndelete.setAttribute("txid", $rel.txid);
                btndelete.addEventListener("click", e => $deletecb(span2, btndelete));
                btndelete.style['margin-left'] = '4px';
            }

        }

        if ($bMOD == true) {
            var btnmod = document.createElement("button");
            btnmod.innerHTML = "+mod";
            btnmod.id = "mod_" + $rel.rid + "_" + $guid;
            btnmod.style['margin-left'] = '4px';
            btnAdd.addEventListener("click", e => function (e) { alert("todo"); });
            div.append(btnmod);
        }

        var ptxid = "";
        var to_s = "";
        if ($rel != null) {
            ptxid = $rel.txid;
            to_s = $rel.to_s;
        }

        var default_class_nid = null;
        var clumpNOBJ = MEMG.getNeuronByNid($clump_nid);

        if (ptxid != "") default_class_nid = clumpNOBJ.getDefaultClassForParam(ptxid);
        if (default_class_nid != null) span2.setAttribute("classRestrictNid", default_class_nid);
        else if (to_s != "") span2.setAttribute("classRestrictNid", to_s);
        else if ($default_class_nid != null) span2.setAttribute("classRestrictNid", $default_class_nid);
    }


    /**
     * An editable role has been clicked on, launch the selector
     * @param {*} $span2 
     */
    editRoleLaunchSelector = async ($span2, $cb_new = null, $cb_cancel = null, $cb_defocus = null) => {
        var tnid = $span2.getAttribute("role_tnid");
        var el_inject = $span2.getAttribute("el_inject");
        var roleNOBJ = MEMG.getNeuronByTempNid(tnid);
        var misc = {};
        misc.smartDefocus = $cb_defocus;
        misc.bCALLBACKONLY = true;
        var pObj = new page_params($cb_new, $cb_cancel, null, roleNOBJ.nid);
        pObj.misc = misc;
        var contextNOBJ = null;
        var classRestrictNid = $span2.getAttribute("classRestrictNid");
        if (classRestrictNid != undefined) {
            pObj.misc.classRestrictNid = classRestrictNid;
            contextNOBJ = MEMG.getNeuronByNid(classRestrictNid);
        }
        var page = ANNE.MEM_UDF.getDefaultSelectorPageForClass(roleNOBJ, null, contextNOBJ);
        console.log("%cPAGE "+page, TALK.bggreen);
        var clump_nid = $span2.getAttribute("clump_nid");
        pObj.misc.from_nid = clump_nid;   //all TYPE_ROLE relons will be put on the clump.
        pObj.misc.type = TYPE_ROLE;
        pObj.misc.reln_nid = roleNOBJ.nid;
        ANNE.A.empty(el_inject);
        pObj.injection_div = el_inject;
        var guid = $span2.getAttribute("guid");
        if (guid == null) return;
        
        var pgui = ANNE.MEM_UDF.getPguiByGuid(guid);
        if (pgui.INST && typeof pgui.INST.smartDefocus === "function") {
            pgui.INST.smartDefocus(true);
        }
        if (pgui == null) return;
        await pgui.loadPart(page, pObj);
    }


    /**
     * @param {*} $rel 
     */
    ezRoleDisplay = function ($el, $rel, $guid) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var roleNOBJ = MEMG.getNeuronByTempNid($rel.r);
        var _roletname = roleNOBJ.text.replaceAll("_", "-");

        //frienly replacement of role name from affected_thing to who
        //when dealing with persons in the that slot.
        if ($rel.to_type == TO_TYPE_ANNE) {
            if (_roletname == "role-affected-thing") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-affected-who";
            }
            else if (_roletname == "role-actor") {
                var toNOBJ = MEMG.getNeuronByTempNid($rel.t);
                if (toNOBJ.is(MEM.person) == true) _roletname = "role-actor-who";
            }
        }

        var div = document.createElement("div");
        div.classList = "role2 " + _roletname + "-outter";
        el.append(div);
        var span1 = document.createElement("span");
        span1.classList = _roletname + "-" + "lbl";
        div.append(span1);
        var span2 = document.createElement("span");
        span2.classList = "r-inner";
        span2.id = "val_" + roleNOBJ.temp_nid + "_" + $guid;
        span2.setAttribute("role_tnid", roleNOBJ.temp_nid);
        div.append(span2);

        //display only if a role relon
        if ($rel.type == TYPE_ROLE) {
            if ($rel.to_type == TO_TYPE_ANNE) {
                ANNE.MEM_UDF.makeNeuronLink(div, $rel.to_s);
            }
            else if ($rel.to_type == TO_TYPE_TXID) {
                ANNE.MEM_UDF.makeRelonViewLink(div, $rel.to_s);
            }
            else {
                var totxt = MEMG.__MASTER_DISPLAY_RELON_TO($rel);
                ANNE.A.setHTML(div, totxt);
            }
        }

        return span2;
    }


    ezSQDWrapper = function ($el, $class, $display = "", $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var termdelete = document.createElement("div");
        el.append(termdelete);
        if ($class != "") termdelete.classList = $class;
        if ($display != "") termdelete.style.display = $display;
        ANNE.A.ezSquareDelete(termdelete, $callback, $callbackParams, $size, $title);
        return termdelete;
    }

    ezSquareDelete = function ($el, $callback, $callbackParams, $size = "36px", $title = "remove member") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var dsquare = document.createElement("div");
        dsquare.classList = "delete-square";
        el.append(dsquare);
        var boxX = ANNE.A.ezEl(dsquare, "span", "boxX", "<img src='../imgs/boxX.png' height='" + $size + "' width='" + $size + "'>", "", $title);
        var sl1c = this.sl1c;
        if (typeof $callback !== "function") {
            console.log("callback jacked up in ezSquareDelete");
            return;
        }
        boxX.addEventListener("click", e => $callback(...$callbackParams));
    }
    
    /**
     * 
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearestObjByProp=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return $el;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }


    /**
     * Extract the prop of a nearest-ancestor
     * @param {*} $el 
     * @param {*} $prop 
     * @param {*} $maxgens 
     * @returns 
     */
    extractNearest=function($el, $prop="tnid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var attr_val = $el.getAttribute($prop);
            if (attr_val != null) return attr_val;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }



    /**
     * @param {Element} $el
     */
    ez_getNearestParentWithProp = function ($el, $prop="nid", $maxgens = 4) {
        var gens = 0;
        while ($el != null) {
            gens++;
            var nid = $el.getAttribute($prop);
            if (nid != null) return nid;
            $el = $el.parentElement;
            if ($el == null) return null;
            if (gens > $maxgens) return null;
        }
        return null;
    }

    /**
     */
    ez_floatActionIco = async ($nid, $ainame, $titleOverride = "")=> {
        var title = ""
        var actionz = "";

        if ($ainame == "mine") {
            title = "This is your data.  Click on it to delete it.";
            actionz = "delete";
        }

        else if ($ainame == "collab") {
            title = "This is data you can collab on.  Click on it to delete it.";
            actionz = "delete";
        }

        //this is locally ghosted data..
        else if ($ainame == "ghost") {

            if (ANNE._txMode == TXMODE_GHOST) {
                title = "This is your ghost.  Only you see it.";
            }

            //if in ANNE mode...
            else if (ANNE._txMode == TXMODE_ANNE) {

                //if it is upgradable...
                var nObj = MEMG.getNeuronByTempNid($nid);
                if (nObj._govenor == null) await nObj.setGovRights(USER._userNObj, false);
                if (nObj.collab_ACTION >= govenor.COLLAB_REQ_OK) {
                    title = "Save your ghost to ANNE.";
                    $ainame = "ghost-to-anne";
                    actionz = "upgrade-ghost-to-anne";
                }
                else {
                    title = "This class of local ghost data cannot be upgraded to ANNE Ghost data.";
                }
            }
        }

        var newspan = document.createElement("span");
        newspan.classList = "hoverable";
        newspan.setAttribute("title", title);
        var itag = document.createElement("i");
        itag.classList = "ai-" + $ainame + " actions-ico";

        //this will trigger DOM events on class clicks to optimize (minimize) events.
        //the widget calling this helper will have to be listening.
        if (actionz) itag.setAttribute("actionz", actionz);

        newspan.append(itag);
        return newspan;
    }

    /**
     */
    ez_auto_inject_close_btn = function ($el, $class, $hrefid, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var betterEl = el.querySelector(".inject-close-here");

        //always prepend into the 
        if (betterEl) el = betterEl;

        var bPREPEND = true;    //if we didn't find a better el, PREPEND (to injection div)
        return ANNE.A.ezHrefBtn(el, $class, $hrefid, $spanClass, $icoclass, bPREPEND);
    }


    /**
     * EXT - NOT IN USE IN EXT...
     * 
     * This will render a gui ico for deleting or updgrading ur own records.
     */
    ez_floatActionIcos = async ($nObj, $el, $ownStatus = "collab")=> {
        var el = this.__getEl($el);
        if (el == null) return null;
        var bSHOW = false;

        if ($nObj.mine == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "mine");
            floatico.style = "cursor:default";
            el.append(floatico);
            bSHOW = true;
        }

        //only one of these...
        else if ($nObj.ghost == true) {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "ghost");
            el.append(floatico);
            bSHOW = true;
        }
        else if ($ownStatus == "collab") {
            var floatico = await ANNE.A.ez_floatActionIco($nObj.temp_nid, "collab");
            el.append(floatico);
            bSHOW = true;
        }

        if (bSHOW == true) ANNE.A.show(el);
    }


    /**
     * text dd
     */
    ezDD_basic = function ($el, $items, $prompt = "select", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        ANNE.A.empty(selectel);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(selectel.value));
        }
        var opt = document.createElement("option");
        opt.setAttribute("value", "");
        opt.innerHTML = $prompt;
        selectel.append(opt);
        for (var i = 0; i < $items.length; i++) {
            var item = $items[i];
            var opt = document.createElement("option");
            opt.setAttribute("value", item);
            opt.innerHTML = item;
            selectel.append(opt);
        }
        ANNE.A.show(selectel);
        return selectel;
    }


    /** 
     */
    ezDD = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false, $skiptnids = []) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            if ($skiptnids.length && $skiptnids.indexOf(nObj.temp_nid) != -1) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);

            opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }


    /**
     * used in pref dds 
     */
    ezDD_masterDDObj = function ($el, $masterOBJS, $callback = null, $msg = "easy access acct nids", $ddClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else selectel = this.ezEl(selectec, "select");

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);

        for (var n = 0; n < $masterOBJS.length; n++) {
            var masterOBJ = $masterOBJS[n];
            var nObj = masterOBJ.nObj;
            var type = masterOBJ.type;
            var nick = masterOBJ.nick;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            var text = "";
            if (nick == nObj.nid) opt.innerHTML = "(" + type + ") " + nObj.nid;
            else opt.innerHTML = "(" + type + ") " + nick + " - " + nObj.nid;
            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDD_nObjs = function ($el, $nObjs, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $bSHOWNID = false) {
        var el = this.__getEl($el);
        if (el == null) {
            console.log("(ezDD_nObjs) el "+$el+" doesnt exist");
            return null;
        }
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }

        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => {
				$callback(e);
			});
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        //var aNObjs=MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) $nObjs.sort(ANNE.UTILS.sortByTextNoCase);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            if (nObj == null) continue;
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            if (nObj.local_nickname != undefined) opt.innerHTML = nObj.local_nickname;
            else opt.innerHTML = nObj.bestNameHTML("", $bSHOWNID);

            selectel.append(opt);
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    doGoogleMap = function ($el, $lat = -34.397, $long = 150.644) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var map = new google.maps.Map(el, {
            center: { lat: $lat, lng: $long },
            zoom: 8,
        });
        return map;
    }

    /**
     */
    ezDDWithGens = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;
            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];
                //doHeading(nObj);
                var indent = "&#9494;";
                var space = '&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
                //items.push({id: nObj.temp_nid,text: '<div>'+space+nObj.text+' ('+nObj.nid+')</div>',value: nObj.nid});
            }
            //results.push({text: $heading,children: items});
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     */
    ezDDWithGens_ACTIONS = function ($el, $temp_nids, $callback = null, $msg = "select something", $bSORTTEXT = true, $ddClass = "", $bINSTOK = true, $V1ONLY = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var selectel = null;
        if (el.tagName == "SELECT") selectel = el;
        else {
            selectel = document.createElement("select");
            el.appendChild(selectel)
        }
        if ($ddClass != "") selectel.classList.add($ddClass);
        if ($callback != null && typeof $callback === "function") {
            selectel.addEventListener("change", e => $callback(e));
        }
        ANNE.A.empty(selectel);
        var headerO = document.createElement("option");
        headerO.setAttribute("data-divider", true);
        headerO.setAttribute("value", null);
        headerO.innerHTML = $msg;
        selectel.append(headerO);
        var aNObjs = MEMG.getNObjsFromTempNidList($temp_nids);
        if ($bSORTTEXT == true) aNObjs.sort(ANNE.UTILS.sortByTextNoCase);

        //aNOBJS are the are the tops.  
        //We want to dig down and get their children.
        for (var n = 0; n < aNObjs.length; n++) {
            var nObj = aNObjs[n];
            if ($bINSTOK == false && nObj.ntype == NTYPE_INST) continue;

            //we only want V1 vbs.
            if ($V1ONLY == true) {
                var rel_vbForm = MEMG.getBeRelon(nObj.outs, MEM.vb_form.temp_nid);
                if (rel_vbForm != null) {
                    if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                        console.log("skipping non v1");
                        continue;
                    }
                }
            }

            var downline = [];
            nObj.getAllDownlineChildren(downline, 0, false, false, true, false);
            var opt = document.createElement("option");
            opt.setAttribute("value", nObj.nid);
            opt.innerHTML = nObj.bestNameHTML();
            opt.classList = "option-top";
            selectel.append(opt);

            for (var a = 0; a < downline.length; a++) {
                var childNOBJ = downline[a];

                //we only want v1 vbs.
                if ($V1ONLY == true) {
                    var rel_vbForm = MEMG.getBeRelon(childNOBJ.outs, MEM.vb_form.temp_nid);
                    if (rel_vbForm != null) {
                        if (rel_vbForm.t != MEM.vb_v1.temp_nid) {
                            console.log("skipping non v1");
                            continue;
                        }
                    }
                }

                var indent = "&#9494;";
                var space = '&#160;&#160;&#160;';
                if (childNOBJ.be_lvl == 0) space = "";
                else {
                    for (var i = 0; i < childNOBJ.be_lvl - 1; i++) { space += '&#160;&#160;&#160;'; }
                    space = space + indent;
                }
                var opt2 = document.createElement("option");
                opt2.setAttribute("value", childNOBJ.nid);
                opt2.innerHTML = space + childNOBJ.bestNameHTML();
                selectel.append(opt2);
            }
        }
        selectel.setAttribute('selectedIndex', 0);
        ANNE.A.show(selectel);
        return selectel;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $val 
     * @returns 
     */
    setSelectedIndexByValue=function($el, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        for (var i=0; i<el.options.length; i++) {
            if (el.options[i].value==$val) {
                selectel.setAttribute('selectedIndex', i);
                break;
            }
        }
    }


    /**
     * ez row with 2 col
     */
    ez2Col = function ($el, $v1 = "", $v2 = "", $v1extraclass = "", $v2extraclass = "") {
        var v1class = "simple-col";
        if ($v1extraclass != "") v1class = v1class + " " + $v1extraclass;
        var v2class = "simple-col";
        if ($v2extraclass != "") v2class = v2class + " " + $v2extraclass;
        var row = ANNE.A.ezEl($el, "div", "simple-row");
        ANNE.A.ezEl(row, "div", v1class, $v1);
        ANNE.A.ezEl(row, "div", v2class, $v2);
        return row;
    }


    //intended for new divs, spans, and other elements that
    //you want to also givev a class, id and style.
    ezEl = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.append(newEl);
        return newEl;
    }
    ezElPrepend = function ($el, $type, $classList = "", $text = "", $id = "", $title = "", $style = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newEl = document.createElement($type);
        if ($classList != "") this.__classListParse(newEl, $classList);
        if ($text != "") newEl.innerHTML = $text;
        if ($title != "") newEl.setAttribute("title", $title);
        if ($id != "") newEl.setAttribute("id", $id);
        if ($style != null) ANNE.A.setStyle($el, $style);
        el.prepend(newEl);
        return newEl;
    }


    doImgB64 = function($el, $nObj, $icostyle="dossier", $w=84, $h=0) {     
        var el = this.__getEl($el);
        if (el == null) return null;
        var url = ANNE.A.getThumbImgUrl($nObj, $icostyle);
        if (url.length) {
            ANNE.A.doImageB64(el, "", $afile, $w, $h);
        }
    }



    /**
     * 
     * @param {*} $nObj 
     * @returns 
     */
    getThumbImgUrl=function($nObj) {
        var parentNOBJ = $nObj.getParent();
        var neuronImgNOBJ = MEM.__GET_DEFAULT_IMG_NEURON(parentNOBJ, $nObj);
        if (neuronImgNOBJ!=null) {
            var url = IMAGES.getUrl(neuronImgNOBJ);
            if (url!="") return url;
        }

        //we're going to nixx all this drama.
        return ANNE.A.getThumbImgUrlFromRelons($nObj);
    }


    /**
     * 
     * @returns 
     */
    getThumbImgUrlFromRelons = function($nObj, $icostyle=null) {
        var test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_thumbnail.temp_nid);
        if (test==null) {
            test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Skyline.temp_nid);
            if (test==null) {
                test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Map.temp_nid);
                if (test==null) {
                    test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Flag.temp_nid); 
                    if (test==null) test = MEMG.getHasRelon($nObj.outs, MEM.wcImgUriPart_Seal.temp_nid); 
                }
            }
        }
        if (test!=null) { 
            return "http://commons.wikimedia.org/wiki/Special:FilePath/"+test.to_s+"?width=200";
        }

        //console.log("not enough info on "+$nObj.text+" for img or ico");
        return "";
    }


    focusTextFix = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.addEventListener("focus", function () {
            el.setSelectionRange(0, 0);
        });
    }

    
    __classListParse = function ($el, $cl) {
        if ($cl == "") return;

        if ($cl != "") {
            if ($cl.indexOf(" ") != -1) {
                var cl = $cl.split(" ");
                for (var c = 0; c < cl.length; c++) {
                    var clstr = cl[c].trim();
                    if (clstr == "") return;
                    $el.classList.add(clstr);
                }
            }
            else $el.classList = $cl;
        }
    }

    //we can inject a close button into a model with this..
    //and then wire it to its pgui.close function, automatically. :)
    ezHrefBtn = function ($el, $hrefClass, $hrefid=null, $spanClass = "hoverable", $icoclass = "ai-times-circle ai-2x", $bPREPEND = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var href = document.createElement("a");
        if ($hrefid!=null) href.setAttribute("id", $hrefid);
        if ($hrefClass) href.classList = $hrefClass;
        if ($bPREPEND == false) $el.append(href);
        else $el.prepend(href);
        ANNE.A.ezIcoSpan(href, $spanClass, $icoclass);
        return href;
    }

    /**
     * ez ico img using the css ico classes
     */
    ezIcoImgSpan = function ($el, $img_class, $msg, $bEMPTY = true, $bSHOW = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($bEMPTY == true) ANNE.A.empty(el);
        var span = document.createElement("span");
        span.classList = $img_class;
        el.append(span);
        if ($msg != "") {
            var msg_span = document.createElement("span");
            msg_span.innerHTML = $msg;
            el.append(msg_span);
        }
        if ($bSHOW == true) ANNE.A.show(el);
    }


    ezIcoLink = function ($el, $href = "#", $ico_class = "ai-anne") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        el.append(link);
        var link_i = document.createElement("i");
        link_i.classList = $ico_class;
        link.append(link_i);
    }


    ezHref = function ($el, $href = "#", $target = "_blank", $txt = "", $ico_class = "") {
        var el = this.__getEl($el);
        var link = document.createElement("a");
        link.setAttribute("href", $href);
        link.setAttribute("target", $target);
        if ($txt == "") link.innerHTML = $href;
        else link.innerHTML = $txt;
        el.append(link);
        if ($ico_class != "") {
            var link_i = document.createElement("i");
            link_i.classList = $ico_class;
            link.append(link_i);
        }
    }


    ezIcoSpan = function ($el, $spanClass, $icoClass, $spanMsg = "", $stylejson = null) {
        var el = this.__getEl($el);
        //if (el==null) return null;
        var newel = document.createElement("span");
        if ($spanClass) newel.classList = $spanClass;
        if (el != null) el.append(newel); //append if we have it, else we just return the new one
        if ($stylejson != null) el.style = $stylejson;
        var ico = document.createElement("i");
        ico.classList = $icoClass;
        newel.append(ico);
        if ($spanMsg != "") newel.innerHTML = newel.innerHTML + $spanMsg;
        return newel;
    }

    ezlblSpan = function ($el, $lbl, $value = "", $spanclass = "", $valclass = "", $sep = "&nbsp;") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var newel = document.createElement("span");
        if ($spanclass != "") newel.classList = $spanclass;
        if ($valclass == "") newel.innerHTML = $lbl + $sep + $value;
        else newel.innerHTML = $lbl + $sep + '<span class="' + $valclass + '">' + $value + '</span>';
        el.append(newel);
        return newel;
    }

    ezTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $class = "simple-row", $exclude = [TYPE_PARAM], $sep = ", ") {
        var list = TALK.printListByReln($outs, $relnNOBJ.temp_nid, $exclude, $sep);
        if (list.length > 0) {
            var eltextlist = document.createElement("div");
            if ($class != "") eltextlist.classList = $class;
            if ($lbl != "") eltextlist.innerHTML = $lbl + list;   //lbl should have trailing space if it is needed.
            $el.append(eltextlist);
        }
    }

    
    ezTwoColTextList = function ($el, $outs, $relnNOBJ, $lbl = "", $exclude = [TYPE_PARAM], $lblClass = "span-lbl", $dataClass = "span-data") {
        var relons = MEMG.getRelonsByTempR($outs, $relnNOBJ.temp_nid, $exclude);
        if (relons.length == 0) return "";

        var elrow = this.ezEl($el, "div", "simple-row");
        var lblcol = this.ezEl(elrow, "div", "simple-col tal");
        this.ezSpan(lblcol, "", $lblClass, $lbl);

        var datacol = this.ezEl(elrow, "div", "simple-col tal");

        for (var r = 0; r < relons.length; r++) {
            var rel = relons[r];
            var totxt = MEMG.__MASTER_DISPLAY_RELON_TO(rel);
            var el_link = EXT_BTNS.makeRelonViewLink(datacol, rel.txid, false, 8, true, "", totxt);
            el_link.style.display = "block";
        }
    }

    ezSortCol = function ($el, $colObj, $tbldivid, $sort) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var classList = ["sortico", "hoverable-green"];

        if ($colObj.ttype == "number") classList.push("ai-sort-amount-down");
        else if ($colObj.ttype == "string" || $colObj.ttype == "anne") classList.push("ai-sort-alpha-down");
        else if ($colObj.ttype == "date") classList.push("ai-sort-down");

        var span = document.createElement("span");
        span.innerHTML = $colObj.text + "&nbsp;";
        el.append(span);
        var ico = document.createElement("i");
        ico.classList = classList.join(" ");
        ico.setAttribute("prop", $colObj.ttype);
        ico.setAttribute("sort", $sort);
        ico.setAttribute("tbl", $tbldivid);
        span.append(ico);

        return ico; //return the sortico (so we can easily put events on it)
    }


    ezSpan = function ($el, $id, $class, $innermsg, $styleJson = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, $styleJson);
        el.append(span);
        return span;
    }

    //make it even if empty.
    ezSpan2 = function ($el, $id, $class, $innermsg, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = this.ezSpanSolo($id, $class, $innermsg, null);
        el.append(span);
        if ($callback) span.addEventListener("click", $e=>$callback($e));
        return span;
    }

    /**
     * does not append, just creates
     */
    ezSpanSolo = function ($id = "", $class = "", $innermsg = "", $styleJson = null) {
        var newel = document.createElement("span");
        if ($class) newel.classList = $class;
        if ($id != "") newel.setAttribute("id", $id);
        newel.innerHTML = $innermsg;
        if ($styleJson != null) newel.style = $styleJson;   //not working as i thought
        return newel;
    }

    ezButton = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        btn.innerHTML = $text;
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        return btn;
    }

    ezButtonWithIcoClass = function ($el, $id = "", $text = "save", $class = "anne-btn btn-success", $callback = null, $icoClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = $class;
        var icoEl = document.createElement("span");
        icoEl.classList = $icoClass;
        btn.appendChild(icoEl);
        icoEl.style['pointer-events']= 'none';
        var textEl = document.createElement("span");
        textEl.innerHTML = $text;
        textEl.style['pointer-events']= 'none';
        btn.appendChild(textEl);
        if ($id != "") btn.setAttribute("id", $id);
        el.appendChild(btn);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", function ($e) { 
                $e.preventDefault();
                $callback($e); 
            });
        }
        
        return btn;
    }

    ezFbwBtn = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.appendChild(btn);
        return btn;
    }

    ezFbwBtn_prepend = function ($el, $tnid, $guid, $classList = "", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        if ($classList != "") this.__classListParse(btn, $classList);
        if ($tnid) btn.setAttribute("tnid", $tnid);
        if ($tnid && $guid) btn.setAttribute("id", $guid + "_" + $tnid);
        if ($callback != null && typeof $callback === "function") {
            btn.addEventListener("click", $callback);
        }
        el.prepend(btn);
        return btn;
    }

    /**
     */
    getPrintNameIcoSpan = function ($nObj) {
        if ($nObj.temp == true) return "<span class='temp-neuron'>" + $nObj.bestName() + "</span>";
        else if ($nObj.ghost == true) return "<span class='ghost-neuron'><span class='ghost-small-blue-ico'></span>" + $nObj.bestName() + "</span>";
        else if (ANNE.isANNE($nObj.owned_by)==true) return "<span class='real-neuron'><span class='anne-small-black-ico'></span>" + $nObj.bestName() + "</span>";
        else if ($nObj.owned_by == USER._userNObj.nid) return "<span class='real-neuron'><span class='mine-small-ico'></span>" + $nObj.bestName() + "</span>";
        return $nObj.text;
    }


    imgBadAudit = function ($el, $id = "", $imgref = "", $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var spanEl = this.doBadImg(el, $id, $imgref, $imgclass)
        if ($callback != null && typeof $callback === "function") {
            spanEl.addEventListener("click", function (e) { $callback(e); });
        }
        return spanEl;
    }


    imgButton = function ($el, $id = "", $imgref, $imgclass = "search-img-small", $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var imgEl = this.doImageDyno(el, $id, $imgref, $imgclass)
        if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $id 
     * @param {anne_file} $afile 
     * @param {*} $imgclass 
     * @param {function} $callback 
     * @returns {element}
     */
    imgButtonB64 = function ($el, $id = "", $afile=null, $imgclass = "search-img-small", $callback = null) {
        if ($afile==null || $afile==undefined) {
			return null;
		}
        var el = this.__getEl($el);
        if (el == null) {
			return null;
		}
        var imgEl = this.doImageDynoB64(el, $id, $afile, $imgclass)
        if (imgEl==undefined || imgEl==null) {
			return null;
		}
		if ($id != "") imgEl.setAttribute("id", $id);
        el.appendChild(imgEl);
        if ($callback != null && typeof $callback === "function") {
            imgEl.addEventListener("click", function (e) { $callback(e); });
        }
        return imgEl;
    }

    
    /**
     * Insert an element into the DOM after another element.
     * lolz, as this is the same as the native .after, so we should just use that.
     * @param {Element or String} $existingNode 
     * @param {Element} $newNode 
     * @returns 
     */
    insertAfter = function($existingNode, $newNode) {
        var el = this.__getEl($existingNode);
        if (el == null) return;
        //if (el.parentNode==null) return;
        //el.parentNode.insertBefore($newNode, el.nextSibling);
        $existingNode.after($newNode);
        return;
    }

    /**
     * Move an el to be after another el.
     */
    moveElAfter=function($el1, $insertAfterThisEl) {
        var el1 = this.__getEl($el1);
        if (el1 == null) return;
        var insertAfterThisEl = this.__getEl($insertAfterThisEl);
        if (insertAfterThisEl == null) return;
        insertAfterThisEl.after($el1);
    }

    appendAfter = function($existingNode, $newNode) {
        this.insertAfter($existingNode, $newNode);
    }

    //triple toggle,  asc, desc, none
    sortToggle = function ($el, $colNumb, $sort = "desc", 
                            $callback = null, 
                            $sortype="numeric") {

                            
                            //$descending = "bi bi-sort-alpha-down", 
                            //$ascending = "ai-sort-asc") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("span");
        btn.style.cursor = "pointer";
        btn.classList = "ez-sort-btn";
        btn.setAttribute("col_num", $colNumb);
        if ($sort == "desc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
            btn.classList.add("active");
        }
        else if ($sort == "asc") {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
            }
            btn.classList.add("active");
        }

        //none is parked on open desc (not selected);
        else {
            if ($sortype=="numeric") {
                btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
            }
            else if ($sortype=="alpha") {
                btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
            }
        }

        btn.setAttribute("sort_val", $sort);

        el.appendChild(btn);

        if ($callback != null) {
            //CLICK ORDER CYYCLE::
            //NONE->DESC->ASC
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                target = ANNE.A.getFirstParentOf(target, "ez-sort-btn");
                var colNumb = target.getAttribute("col_num");
                var sort_val = target.getAttribute("sort_val");
                if (sort_val == "desc") {
                    target.setAttribute("sort_val", "asc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-up'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-up'></i>";
                    }
                    $callback(colNumb, "asc");
                }

                //asc becomes NONE.. not active.
                //show non-active desc sort.  if clicked, it becomes active to desc sort.
                else if (sort_val == "asc") {
                    target.classList.remove("active");
                    target.setAttribute("sort_val", "none");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "none");
                }

                else if (sort_val == "none") {
                    target.setAttribute("sort_val", "desc");
                    ANNE.A.addClass(target, "active");
                    if ($sortype=="numeric") {
                        btn.innerHTML="<i class='bi bi-sort-numeric-down'></i>";
                    }
                    else if ($sortype=="alpha") {
                        btn.innerHTML="<i class='bi bi-sort-alpha-down'></i>";
                    }
                    $callback(colNumb, "desc");
                }


            });

            //init sort.
            //if ($sort=="desc" || $sort=="asc") $callback( $colNumb, $sort);
        }
    }

    /**
     * simple toggle of a bool attribute with get/set
     */
    toggleElAttributeBool=function($el, $prop) {
        if ($el==null) return;
        if ($el.getAttribute($prop)==true) $el.setAttribute($prop, false);
        else $el.setAttribute($prop, true);
    }

    toggleBtn = function ($el, $nObj, $class = "", $callback = null, $on = true, $lbl = "", $ton = "toggled-on", $toff = "toggled-off") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("button");
        btn.classList = "toggle-btn " + $class;
        if ($on == true) btn.classList.add($ton);
        else btn.classList.add($toff);
        btn.setAttribute("tnid", $nObj.temp_nid);
        if ($lbl.length) btn.innerHTML = $lbl;
        else btn.innerHTML = nObj.bestNameHTML();
        el.appendChild(btn);

        btn.addEventListener("click", async (e)=> {
            var target = e.target;

            //if this is off we are turning it on.
            //if it is on.. it stays on even when clicked again.
            if (target.classList.contains($toff)) {
                target.classList.remove($toff);
                target.classList.add($ton);
            }
            else {
                target.classList.remove($ton);
                target.classList.add($toff);

                //callback.
                if ($callback != null && typeof $callback === "function") $callback(e);
            }
        });

        return btn;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $masterObj 
     * @param {*} $eltype 
     */
    makeToggleSet = function($el, $masterObj, $eltype="div") {
        var ats = new anne_toggle_set($masterObj);
        this.makeToggleSetGUI($el, ats, $eltype);
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $class 
     * @returns 
     */
    geAllSibs = function($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent_el = el.parentNode;
        if (parent_el==null) return null;
        return parent_el.querySelectorAll("."+$class);
    }


    /**
     * Creates divs from array of anne_toggle_objs.
     * Will allow for toggle of each div via a toggle btn.
     * @param {*} $el 
     * @param {anne_toggle_set} ats 
     * @param {*} $eltype 
     * @returns {anne_toggle_set}
     */
    makeToggleSetGUI = function($el, $ats, $eltype="div") {
        var el = this.__getEl($el);
        if (el == null) return;

        var el_btn_row = ANNE.A.ezEl($el, "div", "tg-btns", "", "");
        var el_toggle_divs = ANNE.A.ezEl($el, "div", "tg-toggle-divs", "", "");

        for (var i=0; i < $ats.atos.length; i++) {
            const ato = $ats.atos[i];
            console.log("btn tnid "+ato.btn_tnid);
            
            ato.el_toggle_divs = el_toggle_divs;

            //put new button in btn row
            var btn = ANNE.A.ezButton(el_btn_row, "", ato.btn_text, ato.btn_class);
            ato.btn_el = btn;
            if (ato.btn_tnid!="") btn.setAttribute("tnid", ato.btn_tnid);

            //put new div in the .tg-set div.
            var newEl = ANNE.A.ezEl(el_toggle_divs, $eltype, ato.div_class+" tg-set", ato.text, ato.id, ato.title, ato.style);
            ato.div_el = newEl;
            btn.setAttribute("btn_tnid", ato.btn_tnid);

            
            btn.addEventListener("click", async($e)=>{
                if (ato.div_el.classList.contains("active")==true) {
                    ato.div_el.classList.remove("active");
                    if (ato.cb!=null) ato.cb(ato, false);
                }
                else {
                    //should get all tg-set divs
                    //var els = ANNE.A.geAllSibs(ato.div_el, "tg-set");
                    for (var e = 0; e < el_toggle_divs.children.length; e++) {
                        el_toggle_divs.children[e].remove("active");
                    }
                    ato.div_el.classList.add("active");
            
                    if (ato.cb!=null) ato.cb(ato, true)
                };
            });
        }
        return $ats;
    }


    /**
     * simple toggle set of buttons
     */
    toggleBtnSet = function ($el, $nObjs, $classes, $callback, $defaultNid = null, $lbloverride = []) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($nObjs.length == 0) return;
        var elset = document.createElement("div");
        elset.classList = "toggle-btn-set";
        el.appendChild(elset);
        for (var n = 0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var btnclass = $classes[n];
            var btn = document.createElement("button");
            btn.classList = "toggle-btn " + btnclass;
            if ($defaultNid && nObj.nid == $defaultNid) btn.classList.add("toggled-on");
            else btn.classList.add("toggled-off");
            btn.setAttribute("nid", nObj.nid);
            if ($lbloverride.length) btn.innerHTML = $lbloverride[n];
            else btn.innerHTML = nObj.text;
            elset.appendChild(btn);
            btn.addEventListener("click", async (e)=> {
                var target = e.target;
                //if this is off we are turning it on.
                //if it is on.. it stays on even when clicked again.
                if (target.classList.contains("toggled-off")) {
                    //turn off the curren toggled btn.
                    var btnset = ANNE.A.getFirstParentOf(target.parentNode, "toggle-btn-set", true);
                    if (btnset != null) {
                        var onbtn = btnset.querySelector(".toggled-on");
                        if (onbtn != null) {
                            onbtn.classList.remove("toggled-on");
                            onbtn.classList.add("toggled-off");
                        }
                    }
                    //turn on this one.
                    target.classList.remove("toggled-off");
                    target.classList.add("toggled-on");

                    //callback.
                    if ($callback != null && typeof $callback === "function") $callback(e);
                }
            });
        }
        return elset;
    }

    wireShowHideToggler=function($el_to_click, $el_showhide) {
        var el = this.__getEl($el_to_click);
        if (el == null) return;

		var elshowhide = null;
		if (el.classList.contains("hide")==true) {
			elshowhide = ANNE.A.setHTML($el_showhide, "show"); 
		}
		else {
			elshowhide = ANNE.A.setHTML($el_showhide, "hide"); 
		}

		if (elshowhide && elshowhide.getAttribute("showlink_id")==undefined) {
			elshowhide.setAttribute("showlink_id", el.id);
		}
		
        el.addEventListener("click", (e)=>{
            var newval = ANNE.A.toggle($el_showhide);
            //self fixing if things get whacky
            if (newval=="show") {
                if (el.innerHTML === 'show') el.innerHTML = 'hide';
            }
            else if(newval=="hide") {
                el.innerHTML = 'show';
            }
        });
    }

    toggleShowHideLink=function($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var el = document.getElementById('toggleLink');
        if (el.innerHTML === 'hide') el.innerHTML = 'show';
        else el.innerHTML = 'hide';
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $idToToggle 
     * @param {*} $addText 
     * @param {*} $defaultOpen 
     * @param {*} $ss 
     * @returns 
     */
    toggleMinMaxBtn = function ($el, $idToToggle, $addText = "", $defaultOpen = true, $ss = null, $addClass="") {
        var el = this.__getEl($el);
        if (el == null) return;
        var btn = document.createElement("div");
        btn.setAttribute("toggle_divid", $idToToggle);
        btn.classList = "ezMinMax hover";
        if ($addClass!="") btn.classList.add($addClass)
        if ($defaultOpen == false) {
            btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            ANNE.A.hide($idToToggle);
        }
        else {
            btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
            ANNE.A.fadeIn($idToToggle);
        }
        btn.style.cursor = "pointer";
        btn.style.display = "inline-block";
        el.append(btn);
        const ss = $ss;
        btn.addEventListener("click", async (e)=> {
            var target = e.target;
            target = ANNE.A.getFirstParentOf(target, "ezMinMax");
            var idToToggle = target.getAttribute("toggle_divid");
            if (idToToggle == undefined || idToToggle == null) return;
            var classz = ANNE.A.toggleWithFade(idToToggle);
            if (classz == "hide") {
                btn.innerHTML = "<span class='ai-right-triangle'></span>" + $addText;
            }
            else if (classz == "show") {
                btn.innerHTML = "<span class='ai-down-triangle'></span>" + $addText;
                var ssEl = this.__getEl(ss);

                //var percentagePosition = (el.offsetTop / ssEl.scrollHeight) * 100;
                //ssEl.scrollTop = (ssEl.scrollHeight * percentagePosition) / 100;
                //console.log("show section");
                //var scrollHeight = ssEl.scrollHeight;
                //ssEl.scrollTop = scrollHeight;
            }
        });

        return btn;
    }


    __getEl = function ($el) {
        if (typeof $el === "string") {
            if ($el[0] == ".") return document.querySelector($el);

            //legacy references to jquery will have # in front of them.
            else if ($el[0] == "#") return document.getElementById($el.substring(1));

            return document.getElementById($el);
        }
        return $el;
    }

    delayThen = async ($dataparam, $delayms = 300, $callback = null) => {
        await UTILS.sleep($delayms);
        if ($callback != null) $callback($dataparam);
    }

    //we may add a delay to this..
    fadeIn = function ($el, $ms = 500, $ease = "power1.in", $displayProp = "") {
        const el = this.__getEl($el);
        if (el == null) return null;
		el.style.opacity=0;
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return;
        }
        gsap.to(el, {
            opacity:1,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = $displayProp;
                el.classList.add("show");
                el.classList.remove("hide");
            }
        });

        return el;
    }

    fadeInPop = function ($el, $addglowclass = "", $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = $displayProp;
        el.style.transition = "all 2000ms cubic-bezier(.39,.54,0,1.6);";
        if ($addglowclass != "") el.classList.add($addglowclass);
        return el;
    }


    fadeInDelayFadeOut = async ($el, $delayms, $inms = 500, $intt = "ease-in", $outms = 500, $outtt = "ease-out") => {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.add("show");
        el.classList.remove("hide");
        el.style.opacity = 1;
        el.style.display = "";
        el.style.transition = "all " + $innms + "ms " + $intt + ";";
        await UTILS.sleep($inms + $delayms);
        el.classList.add("hide");
        el.classList.remove("show");
        el.style.opacity = 0;
        el.style.display = "none";
        el.style.transition = "all " + $outms + "ms " + $outtt + ";";
        return el;
    }


    /**
     * This is blocking if it is awaited on!
     * Don't await it!
     */
    fadeOut = async ($el, $ms = 500, $delay = 0, $ease = "power1.out", $onComplete=null) => {
        const el = this.__getEl($el);
        if (el == null) return null;
        
        if (el.style==undefined || el.style==null) {
            console.log("el style problem on el id ="+$el)
            return el;
        }

        //max 2 sec delay to avoid accidents
        if ($delay > 0) await UTILS.sleep();

        gsap.to(el, {
            delay: Math.min(2, $delay/1000),
            opacity:0,
            duration:$ms/1000,
            ease:$ease,
            onComplete:function() {
                el.style.display = "none";
                el.classList.add("hide");
                el.classList.remove("show");
                if ($onComplete!=null) $onComplete();
            }
        });

        return el;
    }


    /**
     * Tired of doing this.. so I will do it here.
     */
    formatSpanStrings = function ($el, $classes, $vars, $str, $bSHOW = true) {
        var el = this.__getEl($el);
        if ($classes.length == 0) return $str;
        if ($classes.length != $vars.length) {
            console.log("(formatSpanStrings) classes array and vars array are different lengths.  This may not do what you think it should do.");
        }
        for (var c = 0; c < $classes.length; c++) {
            var clz = $classes[c];
            var i = $str.indexOf("%");
            if (i >= 0) $str = $str.replace("%", "<span class='" + clz + "'>" + $vars[c] + "</span>");
        }
        el.innerHTML = $str;
        if ($bSHOW == true) ANNE.A.show(el);
    }

    getBottom = function ($dfc = ".smart-scroll-framer") {
        return parseInt(document.querySelector($dfc).getBoundingClientRect().bottom);
    }

    getElIndex = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return -1;
        return Array.prototype.indexOf.call(el.parentElement.children, el);
    }

    //because i forget names.
    getNearestParentOf = function($el, $stopClass, $bINCLUDESELF = true) {
        return this.getFirstParentOf($el, $stopClass, $bINCLUDESELF);
    }

    getFirstParentOf = function ($el, $stopClass, $bINCLUDESELF = true) {
        if ($stopClass == "") return null;
        var el = this.__getEl($el);
        if (el == null) return null;
        var safety = 0;
        if ($bINCLUDESELF == true) {
            if (el.classList.contains($stopClass) == true) return el;
        }
        while (el.parentNode != null) {
            if (safety > 20) return null;
            el = el.parentNode;
            if (el == null || el == undefined) return null;
            if (el.nodeName == "#document") return null;
            if (el.classList.contains($stopClass) == true) return el;
            safety++;
        }
        return null;
    }

    getFirstParentAttr = function ($el, $stopClass, $getAttribute) {
        if ($getAttribute == "") return null;
        var el = this.getFirstParentOf($el, $stopClass);
        if (el == null) return null;
        return el.getAttribute($getAttribute);
    }


    getHTML = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.innerHTML;
    }

    getValue = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.value;
    }

    /**
     * input (w/ inputid) marked with class="rqd" wired to element with mod-el=inputid of rqd class el
     * displays helper msg
     * DOM wires rqd element with a .modxx ptr to "mod-el" span element
     * uses css::before to display color/status in the mod-el span.
     */
    applyRQD = function ($el, $bVALID = true, $msg = "required") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("rqd") == true && el.modxx != undefined) {
            if ($bVALID) {
                el.modxx.style['color'] = "#00db00";
                el.modxx.innerHTML = 'ok';
            }
            else {
                el.modxx.style['color'] = "red";
                el.modxx.innerHTML = $msg;
            }
        }
        return el;
    }


    hasClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        return el.classList.contains($class);
    }


    hide = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;

        if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "show";
			}
		}

        if (el.classList==undefined || el.classList==null) {
            alert("what in the world???");
            return;
        }
        if (el.classList.contains("hide") == false) el.classList.add("hide");
        if (el.classList.contains("show") == true) el.classList.remove("show");
        el.style.display = "none";
        return el;
    }
    hideParent = function ($el, $bHIDESELFTOO = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.hide(parent_el);
        if ($bHIDESELFTOO == true) this.hide(el);
    }

    hideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
        }
        return el;//return the last el
    }
    emptyAndHideMulti = function ($arrDivids) {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            el.classList.add("hide");
            el.classList.remove("show");
            el.style.display = "none";
            el.innerHTML="";
        }
        return el;//return the last el
    }

    showAll = function($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.show($arr[a]);
        }
        return el;
    }

    hideAll = function ($arr) {
        var el = null;
        for (var a=0; a < $arr.length; a++) {
            el = ANNE.A.hide($arr[a]);
        }
        return el;
    }

    jsonSyntaxHighlight = function ($json) {
        if ($json == null) return "";
        if (typeof $json != 'string') {
            $json = JSON.stringify($json, undefined, 2);
        }
        $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }



    masterThumbImg = function ($el, $beNOBJ, $nObj, $imgClass = "", $bNOIMGIMG = false) {
        var el = this.__getEl($el);
        if (el == null) return null;

        var img = document.createElement("img");
        if ($imgClass != "") img.classList = $imgClass;

        //get the thumb for this if we have it,
        //falls back on class pic if we don't have thumb for this.
        var url = MEM.__GET_DEFAULT_IMG_NEURON($beNOBJ, $nObj);
        if (url == "") {
            if ($bNOIMGIMG == true) img.setAttribute("src", "imgs/unknown.png");
        }
        else img.setAttribute("src", url);
        el.append(img);
        return img;
    }


    setElBorder = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;

        if ($pendingHanai != null) el.classList.add("hanai-border");
        else if ($nObj.ghost == true) el.classList.add("ghost-border");
        else if ($nObj.freeze == true) el.classList.add("freeze-border");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-border");
        else el.classList.add("other-owned-border");
    }

    setElBG = function ($el, $nObj, $pendingHanai = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($pendingHanai != null) el.classList.add("hanai-bg");
        else if ($nObj.ghost == true) el.classList.add("ghost-bg");
        else if ($nObj.freeze == true) el.classList.add("freeze-bg");
        else if ($nObj.owned_by == USER._userNObj.nid || $nObj.nid == USER._userNObj.nid) el.classList.add("self-owned-border");
        else if (ANNE.isANNE($nObj.owned_by)==true) el.classList.add("anne-owned-bg");
        else el.classList.add("other-owned-bg");
    }

    getImgBorder = function ($nObj, $postfix="") {
        if ($postfix!="") {
            if ($postfix.substring(0,1)!="-") $postfix="-"+$postfix;
        }
        if ($nObj == null) return "";
        if ($nObj.ghost == true) return "ghost-img"+$postfix;
        else if ($nObj.owned_by == USER._userNObj.nid) return "self-owned-img"+$postfix;
        else if (ANNE.isANNE($nObj.owned_by)==true) return "anne-owned-img"+$postfix;
        else return "other-owned-img"+$postfix;
    }


    injectPartFail = function ($x) {
        console.log("newNeuronFail");
        console.log($x);
    }


    /**
     * Update SIF options from outside
     * @param {*} $el_datalist 
     * @param {*} $nObjs 
     * @param {*} $bEMPTY 
     */
    addSIF_opts_externalupdate=function($el_datalist, $nObjs, $bEMPTY=true) {
        var el_datalist = this.__getEl($el_datalist);
        if ($bEMPTY==true) ANNE.A.empty(el_datalist);
        for (var n=0; n < $nObjs.length; n++) {
            var nObj = $nObjs[n];
            var option = document.createElement("option");
            option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
            option.setAttribute("temp_nid", nObj.temp_nid);
            option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
            el_datalist.append(option);
        }
    }
    

    /**
     * This is intended for lists and lookups that have potentially GINORMOUS result sets.
     * We don't want the entire universe of possibilities in a data object in the gui.
     * A neuron class with 100000 options would have 100k elements in a datalist.  This is silly.
     * 
     * Here we default to what the user has been doing in the app. (context of neurons)
     * 
     * When they click that local memory AND ANNODE memory is searched.
     * 
     * @param {Element or String} $el 
     * @param {Function} $search_cb 
     * @param {int} $seachCharMin 
     * @returns 
     */
    makeANNE_SIF=function($el, $search_locl_only_cb, $search_cb=null, $selected_cb=null, 
                                $defaultNOBJS=[], $searchCharMin=2, $bDODELETE=true, $findMoreClass="") {
        const el = this.__getEl($el);
        const local_cb = $search_locl_only_cb;
        const selected_cb = $selected_cb;

        if (el == null) return;
        if (el.type!="text" && el.nodeName!="INPUT"){
            console.log("to make an ANNE SIF we need to start with an input field!");
            return;
        }

        if(typeof $search_cb!=="function") {
            console.log("to make an ANNE SIF we need a search cb!");
            return;
        }

        el.classList.add("ANNE-SIF");

        const searchCharMin = $searchCharMin;

        el.setAttribute("autocomplete", "off");
        
        var bcr = el.getBoundingClientRect();
        var el_datalist = document.createElement("datalist");
        el_datalist.setAttribute("id", el.id+"__list");
        el.parentElement.append(el_datalist);
        el.setAttribute("list", el.id+"__list");       //tie input to data list.

        const btn = document.createElement("button");  
        btn.classList="btn-yellow-sq";
        if ($findMoreClass!="") btn.classList.add($findMoreClass);
        btn.setAttribute("id", el.id+"_sif");             
        btn.innerHTML = "find more";
        //btn.style.position="absolute";
        //btn.style['margin-left'] = (bcr.right-70)+"px";
        btn.setAttribute("list_id", el.id+"__list");
        el.parentNode.append(btn);
        if (bcr.height>0) {
            btn.style['max-height']=Math.floor(bcr.height)+2+"px";
            btn.style['height']=Math.floor(bcr.height)+2+"px";
        }
        else {
            btn.style['max-height']=32+"px";
            btn.style['height']=32+"px";
        }
        btn.style['max-width']='70px';
        btn.style['width']='70px';
        ANNE.A.hide(btn);
        
        const btn2 = document.createElement("button");  

        if ($bDODELETE) {
            btn2.classList="btn-red-sq";   
            btn2.setAttribute("id", el.id+"_sif_x");             
            btn2.innerHTML = "[x]";
            //btn2.style.position="absolute";
            //btn.style['margin-left'] = (bcr.right-40)+"px";
            btn2.setAttribute("list_id", el.id+"__list");
            el.parentNode.append(btn2);
            if (bcr.height>0) {
                btn2.style['max-height']=Math.floor(bcr.height)+2+"px";
                btn2.style['height']=Math.floor(bcr.height)+2+"px";
            }
            else {
                btn2.style['max-height']=32+"px";
                btn2.style['height']=32+"px";
            }
            btn2.style['max-width']='40px';
            btn2.style['width']='40px';

            //clear the selection.
            btn2.addEventListener("click", function(e){
                el.value="";
                ANNE.A.hide(btn);
                ANNE.A.hide(btn2);
                if (selected_cb!=null) selected_cb(-1);
                return;
            });

            ANNE.A.hide(btn2);
        }


        function addOpts($nObjs) {
            if ($nObjs==undefined || $nObjs==null) return;
            for (var n=0; n < $nObjs.length; n++) {
                var nObj = $nObjs[n];
                var option = document.createElement("option");
                option.setAttribute("value", nObj.bestSelectDisplay("", true, true));
                option.setAttribute("temp_nid", nObj.temp_nid);
                option.setAttribute("text", nObj.text); //for willy wonka datalist using value as lbl 
                el_datalist.append(option);
            }
        }
        //add in any default tnids (like local favs)
        if ($defaultNOBJS.length) addOpts($defaultNOBJS);

        el.setAttribute("search-roots", ""); 
        el.setAttribute("local-search-roots", ""); 

        function isInRoot($val, $which="search-roots") {
            var searchRootsArr = [];
            var search_roots = el.getAttribute($which);
            if (ANNE.UTILS.isset(search_roots)==true) {
                searchRootsArr = search_roots.split(",");
                for (var s=0; s < searchRootsArr.length; s++) {
                    if (searchRootsArr[s]=="") continue;
                    if ($val.indexOf(searchRootsArr[s])==0) {
                        return true;
                    }
                }
            }
            //jam the new search root val back in.  This optimizes lookups so we don't 
            //needlessly chase extended strings of strings we've already looked up!
            searchRootsArr.push($val.trim());
            el.setAttribute($which, searchRootsArr.join(","));
            return false;
        }

        var getActualVal = function() {
            //see if it matches anything we have in the datalist.
            //if so, it is selected -- so call the callback.
            //if there are multiples matching this text the user can use the dd to selecte which one they want.
            var selected_tnid = -1;
            var lowerz = el.value.toLowerCase();
            var actualVal = el.value;

            //find the selected datalist item by its unique name.
            for (var j = 0; j < el_datalist.options.length; j++) {
                if (el_datalist.options[j].value.toLowerCase()==lowerz) {
                    selected_tnid = parseInt(el_datalist.options[j].getAttribute("temp_nid") );
                    actualVal = el_datalist.options[j].getAttribute("text");
                    break;
                }
            }
            return {actualVal:actualVal, selected_tnid:selected_tnid};
        }

        
        //CALLED ON TEXT INPUT.
        var testLocal = async($e)=>{
            //willy wonka data lists only use the value prop as display.
            //if we want a disambig display in the option then that becomes the value.
            //use extra text attr in option for real text value.
            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            var selected_tnid = avObj.selected_tnid;

            //already searched for this and it is in the widget.
            //if (actualVal.length) {//} && isInRoot(actualVal, "local-search-roots")==false) {
            if (actualVal.length>0 && local_cb!=null) {
                ANNE.A.show("loading");
                var nObjs = await local_cb(actualVal);
                ANNE.A.hide("loading");
                if (ANNE.UTILS.isset(el_datalist)==false) {
                    el_datalist = document.getElementById(btn.getAttribute("list_id"));
                    if (ANNE.UTILS.isset(el_datalist)==false) return;
                }
                ANNE.A.empty(el_datalist);
                addOpts(nObjs);
            }
            //}

            if (selected_tnid>-1 && selected_cb!=null) {
                ANNE.A.hide(btn);
                ANNE.A.show(btn2);
                selected_cb(selected_tnid);
                return;
            }
            if (actualVal.length>searchCharMin) {
                ANNE.A.hide(btn2);
                ANNE.A.show(btn);
            }
            else {
                if (selected_tnid==-1 && selected_cb!=null) {
                    ANNE.A.hide(btn2);
                    ANNE.A.hide(btn);
                    selected_cb(-1);
                    return;
                }
            }
        }
    
        
        const returnKeyPress = async($e)=> {
            $e.preventDefault();
            if ($e.keyCode === 13) {
                if (btn.classList.contains("show")==true) {
                    var event = new MouseEvent('click', {
                        'view': window,
                        'bubbles': true,
                        'cancelable': true
                    })
                    btn.dispatchEvent(event);
                }
            }
        }

        //if more than the search min, open up the anne search btn.
        //if we haven't touched this search root yet, do a local ajax style search.
        el.addEventListener("input", $e=>testLocal($e));
        el.addEventListener("keyup", $e=>returnKeyPress($e));

        //find more button
        btn.addEventListener("click", async($e)=>{
            ANNE.A.show("loading");
            ANNE.A.hide(btn);

            var avObj = getActualVal();
            var actualVal = avObj.actualVal;
            
            //already searched for this and it is in the widget.
            if (actualVal.length>0 && isInRoot(actualVal, "search-roots")==true) return;

             //add in to local-search-roots too so data list feature takes over.
            if (actualVal.length>0) isInRoot(actualVal, "local-search-roots");  
        
            var nObjs = await $search_cb($e);
            if (ANNE.UTILS.isset(el_datalist)==false) {
                el_datalist = document.getElementById(btn.getAttribute("list_id"));
            }
            if (ANNE.UTILS.isset(el_datalist)==false) return;
            ANNE.A.empty(el_datalist);
            addOpts(nObjs);
            ANNE.A.hide("loading");
        });

        return btn;
    }


    setAsGhost=function($el, $class="ghost-text") {
        var el = this.__getEl($el);
        if (el == null) return;
        el.classList.add($class);
        var span=document.createElement("span");
        span.classList = "ghost-small-blue-ico";
        el.appendChild(span);
        el.placeholder = "Ghost neuron, must send TX to ANNE to make real.";
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $cb 
     * @returns 
     */
    makeQuickDeleteBtn = function($el, $cb=null) {
        var el = this.__getEl($el);
        if (el == null) return;
        const btn = document.createElement("button");  
        btn.classList="btn-red-sq";                
        btn.innerHTML = "[x]";
        el.append(btn);

        if ($cb!=null && typeof $cb ==="function") {
            btn.addEventListener("click", e=>$cb(e));
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $injid 
     * @param {*} $paramRelon 
     * @param {*} $GUID 
     * @param {*} $page 
     * @param {*} $existingClumpNid 
     * @param {*} $lbl 
     * @param {*} $bNEWOK 
     * @returns 
     */
    makeANTFromParam = function ($el, $injid, $paramRelon, $GUID, $page = "selector_action", $existingClumpNid = "", $lbl = "", $bNEWOK = true) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($paramRelon == undefined || $paramRelon == null) return;
        var relNOBJ = MEMG.getNeuronByTempNid($paramRelon.r);
        var toNOBJ = MEMG.getNeuronByTempNid($paramRelon.t);
        var antag = document.createElement("an");
        antag.setAttribute("role", relNOBJ.text);
        antag.setAttribute("rolereln", relNOBJ.nid);
        antag.setAttribute("newok", +$bNEWOK);
        antag.setAttribute("role_class_nid", toNOBJ.nid);
        if ($lbl == "") {
            var parts2 = relNOBJ.text.split("_");
            parts2.shift();
            $lbl = parts2.join("_");
        }
        antag.setAttribute("lbl", $lbl);
        antag.setAttribute("injid", $injid);
        antag.setAttribute("pguid", $GUID);
        antag.setAttribute("clump_nid", $existingClumpNid);
        antag.setAttribute("page", $page);
        el.append(antag);
    }

    /**
     * Using clone will destroy the existing events. (that's what it is here for)
     */
    moveElement = function ($subj2bmoved, $el_destination, $bCLONE = true) {
        if ($subj2bmoved == null) return;
        var el_parent = $subj2bmoved.parentElement;
        if (el_parent == null) return;
        var el_destination = this.__getEl($el_destination);
        if (el_destination == null) return;
        el_parent.removeChild($subj2bmoved);
        if ($bCLONE == true) {
            return this.cloneAndAppend($subj2bmoved, el_destination, true);
        }
        else el_destination.appendChild($subj2bmoved);
        return $subj2bmoved;
    }

    /**
     * Take a DOM node from its parent and move it to the spot of first child (prepend)
     */
    moveToFirst = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.prepend(el);
    }
    moveToEnd = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return;
        var parent = el.parentNode;
        if (parent == null) return;
        parent.removeChild(el);
        parent.append(el);
    }


    /**
     */
    parseANT_sentence = function ($ant) {
        console.log("%c (parseANT_sentence)", TALK.bgpink)
    }

    /**
     * Parse the ANTag (<an/>) which can show up in a blob.
     */
    parseANTags = function ($el, $bCHASE = false, $vars = null, $extraClass = "") {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");

        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];

            //sentence type is for parsing sentence
            var type = ant.getAttribute("type");
            if (type != undefined) {
                console.log("%c ANT TAG TYPE " + type, TALK.bgpink)
                if (type == "sentence") {
                    this.parseANT_sentence(ant);
                    continue;
                }
            }

            var nid = ant.getAttribute("nid");
            if (nid != undefined && nid != "") {
                var nObj = MEMG.getNeuronByNid(nid);
                var txt = "";
                var dnrt = ant.getAttribute("dnrt");
                if (dnrt != null && dnrt == 1) {
                    txt = ant.innerHTML;
                }

                ant.innerHTML = "";

                if (nObj != null) {
                    EXT_BTNS.makeNeuronLink(ant, nid, $extraClass, "", false, txt);

                }
                else if ($bCHASE == true) {
                    console.log("we need to go get this value");
                }
            }

            // to do. create helpful injectable gui
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }

        }
    }


    /**
      * Parse the ANTag (<an/>) which can show up in a blob.
      * only do vars, not text/html tags.
      */
    parseANTagsVarsOnly = function ($el, $vars = null) {
        var el = this.__getEl($el);
        if (el == null) return;
        var ants = el.getElementsByTagName("an");
        for (var a = 0; a < ants.length; a++) {
            var ant = ants[a];
            if ($vars != null) {
                var varname = ant.getAttribute("varname");
                if (varname != undefined) {
                    var dyno = $vars[varname];
                    if (dyno != undefined) {
                        ant.innerHTML = dyno;
                        ant.classList = "dynamic-value";
                    }
                }
            }
        }
    }

    stopPlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
        el.currentTime = 0;
    }
    pausePlay = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.pause();
    }
    stopSound = function($el) { 
        this.stopPlay($el);
    }
    startSound = function($el) { this.play($el); }
    playSound = function($el) { this.play($el); }
    play = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.play();
    }

    pointerEventsNone = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="none";
        ANNE.TALK.printf_console(TALK.RED, "pointerEventsNone");
        return el;
    }
    pointerEventsAll = function($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.pointerEvents="all";
        ANNE.TALK.printf_console(TALK.GREEN, "pointerEventsAll");
        return el;
    }

    removeSelf = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        el.parentNode.removeChild(el);
        return el;
    }

    /**
     * 
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     * @param {*} $prefName 
     * @param {*} $featureProp
     * @param {*} $fpLbl
     * @param {*} $fpClass
     * 
     * @returns 
     */
    neuronIdentityDisplay = async($nObj, $bEMULATE=false, $prefName="", $featureProp="did_annex", $fpLbl="ANNEX'd neurons", $fpClass="color-green")=> {
        var anneid = $nObj._anneid;
        if (anneid==undefined || anneid==null) {
            if ($bEMULATE==true) {
                anneid = AWARE.emulateANNEID($nObj);
            }
            else {
                anneid = await ANNE.API.getANNEID($nObj, false);
            }
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "anneid-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
        
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        if ($nObj.getParent()==null) {
            await ANNE.CHASE.queryMinBem($nObj);
        }
        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="anneid-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        var name = $prefName;
        if (name=="") name = $nObj.bestNameHTML();
        acn2.innerHTML = name;
        if (acn2.innerText.length>25) acn2.innerText = acn2.innerText.substring(0, 25);
        //if ($prefName[$prefName.length-1]!=")") acn2.innerText+=")";
        acn.appendChild(acn2);

        //anneid
        var fp = UTILS.commas(anneid[$featureProp]);
        var acs = document.createElement("div");
        acs.classList = "acs";
        acs.classList.add("acs-"+$nObj.nid);    //find the smart-reference.. however we update that.
        acs.innerHTML = "<span class='"+$fpClass+"'>"+fp+"</span>&nbsp;"+$fpLbl;
        acn.appendChild(acs);
        
        return el_acf;
    }


    /**
     * 
     * @param {*} $nObj 
     * @param {object} $match
     * @param {object} $arr
     * @returns 
     */
    neuronRevelatianMatchDisplay = async($nObj, $match, $lvlOffsets)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }

        //console.log($match);

        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
        el_acf.setAttribute("overlay", true);
        el_acf.setAttribute("tnid", $nObj.temp_nid);
        el_acf.setAttribute("id", "match_tnid_"+$nObj.temp_nid);
            
            var el_donut = document.createElement("div");
            el_donut.classList = "donutNONE";
            el_acf.appendChild(el_donut);
    
            var p1 = document.createElement("p");
            p1.classList="n-circ-annex";
            el_acf.appendChild(p1);
                var sp = document.createElement("span");
                sp.classList = "REV-ico";
                p1.appendChild(sp);

                var sp2 = document.createElement("span");
                sp2.style="color:color-green font-size:20px; font-weight:900";
                sp2.innerHTML="";
                p1.appendChild(sp2)

            var p2 = document.createElement("p");
            p2.classList = "n-circ-imgfix";
            el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

            var el_revcirc = document.createElement("div");
            el_revcirc.classList="rev-circ-name";
            el_acf.appendChild(el_revcirc);
            
                var acn2 = document.createElement("div");
                acn2.classList = "rev";
                var name = $nObj.bestNameHTML();
                acn2.innerHTML = name;
                el_revcirc.appendChild(acn2);

                var lbl = document.createElement("div");
                lbl.classList="revelatian-sublbl";
                
                if ($nObj.ntype==NTYPE_INST) lbl.innerHTML = beNOBJ.bestNameHTML() + "<span class='rev-tiny-ntype'>inst</span>";
                else if ($nObj.ntype==NTYPE_CLASS) lbl.innerHTML = beNOBJ.bestNameHTML()+ "<span class='rev-tiny-ntype'>class</span>";
                el_revcirc.appendChild(lbl);

                //var lvloffset0 = document.createElement("span");
                //if ($lvlOffsets[0]!=undefined) {
                   // lvloffset0.innerHTML = "<span class='rev-lvl0ffset0'>"+$lvlOffsets[0]+"</span>";
                //}
				//lbl.appendChild(lvloffset0);
                //var lvloffset1 = document.createElement("span");
                //if ($lvlOffsets[1]!=undefined) {
                   // lvloffset1.innerHTML = "<span class='rev-lvl0ffset1'>"+$lvlOffsets[1]+"</span>";
                //}
				//lbl.appendChild(lvloffset1);

                var pathways = document.createElement("div");
                pathways.setAttribute("id", "pathways_"+$nObj.temp_nid);
                pathways.classList="revelatian-pathways";
                el_revcirc.appendChild(pathways);

        // var match = {tnid:$matchPt.tnid, myl:1, paths:paths};
        // paths (arraymap) = {pathid:path, hits:1}
        return {el:el_acf, el_pathways:pathways};
    }


    /**
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronRevelatianDisplay = async($nObj)=> {
        var beNOBJ = $nObj.getParent();
        if (beNOBJ==null) {
            console.log("nid:"+$nObj.nid+" text:"+$nObj.tex+" needs a parent");    
            return;
        }
        var el_acf = document.createElement("div");
        el_acf.classList = "rev-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donutNONE";
        el_acf.appendChild(el_donut);
       
        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "REV-ico";
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-green font-size:20px; font-weight:900";
        sp2.innerHTML="";
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="rev-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "rev";
        var name = $nObj.bestNameHTML();
        if (name.length>25) name = name.substring(0, 25);
        acn2.innerHTML = name;
        acn.appendChild(acn2);

        var lbl = document.createElement("div");
        lbl.classList="revelatian-sublbl";
       
        lbl.innerHTML = "<span>parent&nbsp;<b>"+beNOBJ.bestNameHTML().substring(0, 25)+"</b></span>";
        acn.appendChild(lbl);
        return el_acf;
    }


    /**
     * make and populate an html element to display goodies for an annexable neuron
     * this is short-hand used for the user to click to get to a more in-depth view.
     * @param {*} $nObj 
     * @param {*} $bEMULATE 
     */
    neuronAplexDisplay = async($nObj, $bEMULATE=false, $prefName="", $colorClass="green", $icopost="2")=> {
        var apx = $nObj._aplex;
        if (apx==null) {
            if ($bEMULATE==true) {
                apx = new aplex($nObj, $nObj.nid, null);    //3rd param applex obj
                apx.emulateAplex();
            }
            else {
                apx = await ANNE.CHASE.get_aplex_data($nObj.nid, false, "neuronAplexDisplay");
            }
        }

        var el_acf = document.createElement("div");
        el_acf.classList = "awarez-circ-full";
            
        var el_donut = document.createElement("div");
        el_donut.classList = "donut";
        el_acf.appendChild(el_donut);
        if (apx.posPer==null) apx.initPercents();
        var l1 = Math.round(apx.posPer*360);
        var l2 = Math.round(apx.negPer*360);
        var varstring = "--l1:0deg "+l1+"deg; --l2:"+l1+"deg "+(l1+l2)+"deg; --l3:"+(l1+l2)+"deg 360deg;";
        el_donut.style=varstring;

        var p1 = document.createElement("p");
        p1.classList="n-circ-annex";
        el_acf.appendChild(p1);
        var sp = document.createElement("span");
        sp.classList = "annex-"+$colorClass+"-ico"+$icopost;
        p1.appendChild(sp)
        var sp2 = document.createElement("span");
        sp2.style="color:color-"+$colorClass+" font-size:20px; font-weight:900";
        sp2.classList = "annex-price-lvl";
        sp2.classList.add("apltnid-"+$nObj.temp_nid);
        sp2.innerHTML=apx.price_lvl;
        p1.appendChild(sp2)
        var p2 = document.createElement("p");
        p2.classList = "n-circ-imgfix";
        el_acf.appendChild(p2);

        await ANNE.IMAGES.loadElIntoBG($nObj, p2);

        var acn = document.createElement("div");
        acn.classList="awarez-circ-name";
        el_acf.appendChild(acn);
        var acn2 = document.createElement("div");
        acn2.classList = "acn";
        if ($prefName!="") {
            acn2.innerHTML = $prefName;
        }
        else acn2.innerHTML = $nObj.text;
        acn.appendChild(acn2);

        if ($nObj.sponsored_by!="") {
            var spNOBJ = MEMG.getOrMakeIfNew($nObj.sponsored_by);
            var acs = document.createElement("div");
            acs.classList = "acs";
            acs.innerHTML = spNOBJ.bestName();
            acn.appendChild(acs);
        }

        return el_acf;
    }



    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        return this.removeSelf(parent);
    }

    /**
     * 
     * @param {*} $el 
     * @returns 
     */
    removeParentParent = function ($el) {
        var el = this.__getEl($el);
        if (el == null || el.parentNode == null) return null;
        var parent = el.parentNode;
        var parentParent = parent.parentNode;
        if (parentParent == null) return this.removeSelf(parent);
        return this.removeSelf(parentParent);
    }


    removeAll=function($class) {
        const elements = document.querySelectorAll('.'+$class);
        elements.forEach(element => element.remove());
    }

    
    /**
     * 
     * @param {*} $el 
     * @param {*} $classname 
     * @returns 
     */
    removeClass = function ($el, $classname) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.classList.remove($classname);

        //remove a 2nd, if there.
        if (el.classList.contains($classname)==true) el.classlist.remove($classname);
        
        return el;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $childClass 
     * @returns 
     */
    removeByParentAndClass = function ($el, $childClass) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var els = el.querySelectorAll("." + $childClass);
        for (var e = 0; e < els.length; e++) {
            var elchild = els[e];
            el.removeChild(elchild);
        }
    }

    /**
     * remove the $removeClass from all element children inside of $el
     * that match the selector.  Adds it back to the targetEl.   
     * This is used to turn an active off, then on for the selected item.
     */
    removeAddClassOfAllIn = function ($el, $selector, $removeClass, $targetEl) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($selector[0] != "#" && $selector[0] != ".") $selector = "." + $selector; //presume class.
        el.querySelectorAll($selector).forEach(target => {
            target.classList.remove($removeClass);
        });
        var targetEl = this.__getEl($targetEl);
        if (targetEl == null) return;
        if (targetEl != null) targetEl.classList.add($removeClass);
    }

    setRadioVals = function ($pre, $numMax, $checkedVal) {
        for (var c = 0; c < $numMax; c++) {
            var el = this.__getEl($pre + "_" + c);
            if (el == null) return null;
            if ($checkedVal == c) el.checked = true;
            else el.checked = false;
        }
        return;
    }


    setBodyBackground = function ($url) {
        var newImgEl = document.createElement('img');
        newImgEl.addEventListener('load', function () {
            document.body.style['background-image'] = "url('" + $url + "')";
            ANNE.A.removeSelf(newImgEl);

        });
        newImgEl.setAttribute('src', $url); // Loads the image.
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaOut=function($el, $duration=1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.hide(el);
            el.style.alpha=0;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if (el.classList.contains("hide")==false) el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @returns 
     */
    gs_alphaScaleOut=function($el, $duration=1, $scale=0.1, $ease="power1.out"){
        if (ANNE.gsap==null) ANNE.A.hide($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        var restore = el.style.display;
        if (el.getAttribute("display_restore")==undefined && restore!="" && restore!="none") {
            el.setAttribute("display_restore", restore);
        }
        ANNE.gsap.to(el, {
            ease:$ease,
            duration: $duration,
            transform:'scale('+$scale+')',
            alpha: 0,
            onUpdate: function() {
            },
            onComplete: async()=> {
                el.classList.add("hide");
                el.style.display = "none";
            }
        } );
    }
    

    /**
     * 
     * @param {*} $el 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaIn=function($el, $duration=1, $ease="power1.in", $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");

        if (ANNE.gsap==undefined || ANNE.gsap==null) {
            ANNE.A.show(el);
            el.style.alpha=1;
            return;
        }
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                    el.classList.remove("hide"); //incase another animation
                }
            }
        } );
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $duration 
     * @param {*} $ease 
     * @param {*} $displayProp 
     * @returns 
     */
    gs_alphaScaleIn=function($el, $duration=1, $ease="power1.in", $scale=1, $displayProp="", ){
        if (ANNE.gsap==null) ANNE.A.show($el);
        var el = this.__getEl($el);
        if (el == null) return null;
        el.alpha=0;
        el.classList.remove("hide");
        
        ANNE.gsap.to(el, {
            ease:$ease,
            duration:$duration,
            transform:'scale('+$scale+')',
            alpha: 1,
            onUpdate: function() {
            },
            onComplete: async()=> {
                if ($displayProp!="") {
                    el.style.display=$displayProp;
                }
                else {
                    var restore = el.getAttribute("display_restore");
                    if (restore!=undefined) el.style.display=restore;
                    else el.style.display="";
                }
            }
        } );
    }


	showing = function($el) {
		var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == true) return true;
		if (el.classList.contains("hide") == true) return false; 
		if (el.style.display == "none") return false;
		return true;
	}
	
    /**
     * 
     * @param {*} $el 
     * @param {*} $displayProp 
     * @returns 
     */
    show = function ($el, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show") == false) el.classList.add("show");
        if (el.classList.contains("hide") == true) el.classList.remove("hide");
        if (el.classList.contains("faded-neuron") == true) el.classList.remove("faded-neuron");

		if (el.getAttribute("showlink_id")!=undefined) {
			var showlink = document.getElementById(el.getAttribute("showlink_id"));
			if (showlink!=null) {
				showlink.innerHTML = "hide";
			}
		}
		
        if ($displayProp == "") {
            var restore = el.getAttribute("display_restore");
            if (restore!=null && restore.length>0) {
				el.style.display = restore;
			}
            else if (el.style.display == "none") {
				el.style.display = "";
			}
        }
        else el.style.display = $displayProp;
        return el;
    }


    showParent = function ($el, $bSHOWSELFTOO = true) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent_el = el.parentNode;
        this.show(parent_el);
        if ($bSHOWSELFTOO == true) this.show(el);
        return el;
    }

    showMulti = function ($arrDivids, $displayProp = "") {
        var els = document.querySelectorAll($arrDivids);
        for (var e = 0; e < els.length; e++) {
            var el = els[e];
            if (el == null) continue;
            el.classList.add("show");
            el.classList.remove("hide");
            el.style.display = $displayProp;
            el.style.opacity = 1; //correct in case a fadeOut was used prior
        }
        return el;//return the last el
    }

    setAttribute = function ($el, $attribute, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.setAttribute($attribute, $value);
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setDeactiveWLbl = function ($el, $prepend = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($prepend == true) var dl = ANNE.A.ezElPrepend(el, "span", "dib deactivate-lbl", "");
        else var dl = ANNE.A.ezEl(el, "span", "dib deactivate-lbl", "");
        ANNE.A.ezEl(dl, "span", "dm", "");
        return dl;
    }

    setHTML = function ($el, $html, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.innerHTML = $html;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setJSON = function ($el, $json, $bSHOW = true, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($json == undefined || $json == null) return;
        var str = this.jsonSyntaxHighlight($json);
        el.innerHTML = str;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * Set key up date event.
     * All dates get handled by our doDateFormat handler
     */
    setKeyUpDate = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doDateFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doDateFormat);
    }

    setKeyUpTime = function ($el, $callback = null) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($callback != null) {
            el.addEventListener('keyup', function (e) {
                MEM.doTimeFormat(e);
                $callback(e);
            });
        }
        else el.addEventListener('keyup', MEM.doTimeFormat);
    }


    setStyle = function ($el, $jsoncss, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var keys = Object.keys($jsoncss);
        for (var k = 0; k < keys.length; k++) {
            var prop = keys[k];
            var val = $jsoncss[prop];
            el.style[prop] = val;
        }
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }

    setStyleVar = function ($el, $prop, $val) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style.setProperty($prop, $val);
    }


    setValue = function ($el, $value, $bSHOW = false, $addclass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        if ($value==null) {
            el.selectedIndex=0;
        }
        else el.value = $value;
        if ($bSHOW == true) this.show(el);
        if ($addclass != "") el.classList.add($addclass);
        return el;
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $localRef 
     */
    soundFileLink = function($el, $localRef, $bDOWAVEFORM=true) {
        var el = this.__getEl($el);
        if (el == null) return null;
    
        
        var el_btn = document.createElement("span");
        el_btn.innerHTML = "▶️";
		el_btn.style.cursor = "pointer";
		el_btn.classList = "hoverable";
        el.appendChild(el_btn);

        
        var audio_tag = document.createElement("audio");
        audio_tag.src = $localRef;
        el.append(audio_tag);

        el_btn.addEventListener('click', function() {
            audio_tag.play();
        });


        if ($bDOWAVEFORM==false) return audio_tag;
		
        var el_wf_canvas = document.createElement("canvas");
		el_wf_canvas.style["height"]="90px";
		el_wf_canvas.style["min-width"]="100px";
		
        el.append(el_wf_canvas);
        this.drawWaveform(audio_tag, el_wf_canvas);

        return audio_tag;
    }


    /**
     * 
     * @param {*} $audio_tag 
     * @param {*} $cavasForWaveform 
     */
    drawWaveform=function($audio_tag, $cavasForWaveform) {
        const audioElement = this.__getEl($audio_tag);
        if (audioElement==null) return;

        const canvas = this.__getEl($cavasForWaveform);
        if (canvas==null) return;

        const audioContext = new (window.AudioContext || window.webkitAudioContext)();
        const audioSrc = audioContext.createMediaElementSource(audioElement);
        const analyser = audioContext.createAnalyser();

        audioSrc.connect(analyser);
        analyser.connect(audioContext.destination);

        const bufferLength = analyser.frequencyBinCount;
        const dataArray = new Uint8Array(bufferLength);

        const canvasCtx = canvas.getContext('2d');

        function draw() {
        const WIDTH = canvas.width;
        const HEIGHT = canvas.height;

        analyser.getByteTimeDomainData(dataArray);

        canvasCtx.fillStyle = 'rgb(200, 200, 200)';
        canvasCtx.fillRect(0, 0, WIDTH, HEIGHT);

        canvasCtx.lineWidth = 2;
        canvasCtx.strokeStyle = 'rgb(0, 0, 0)';

        canvasCtx.beginPath();

        const sliceWidth = (WIDTH * 1.0) / bufferLength;
        let x = 0;

        for (let i = 0; i < bufferLength; i++) {
            const v = dataArray[i] / 128.0;
            const y = (v * HEIGHT) / 2;

            if (i === 0) {
            canvasCtx.moveTo(x, y);
            } else {
            canvasCtx.lineTo(x, y);
            }

            x += sliceWidth;
        }

        canvasCtx.lineTo(canvas.width, canvas.height / 2);
        canvasCtx.stroke();

        requestAnimationFrame(draw);
        }

        draw();
    }


    copyIco = function ($el, $val) {
        var el = this.__getEl($el);
        if (el) {
            ANNE.A.empty(el);
            this.addCopyIco(el, $val);
        }
    }


    /**
     * 
     * @param {*} $el 
     * @param {*} $data 
     * @returns 
     */
    addCopyIco=function($el, $data){
        if (ANNE.UTILS==null) return;
        var spanpre=document.createElement("button");
        spanpre.classList="ico-copy";
        spanpre.title = "copy to clipboard";
        spanpre.setAttribute("data-copy", $data);
        var sp = document.createElement("i")
        sp.classList = 'bi bi-clipboard-plus';
        spanpre.append(sp);
        spanpre.addEventListener("click", (e)=>{
            ANNE.playSound("click");
            e.stopPropagation();
            e.preventDefault();
            ANNE.UTILS.copyToClipboard($data); 
        });
        var el = this.__getEl($el)
        if (el==null) return;
        el.append(spanpre);
    }


    setMaxHeights = function ($el, $maxheight = 0, $height = 0) {
        var el = this.__getEl($el);
        if (el == null) return;
        if ($maxheight > 0) el.style["max-height"] = $maxheight + "px";
        if ($height > 0) el.style["height"] = $height + "px";
    }


    //updates any class, ie, 'bal-[temp_nid]' tags in the DOM
    quickBalanceUpdate = function ($class, $html) {
        var bals = document.querySelectorAll(".bal-" + USER._userNObj.temp_nid);
        for (var b = 0; b < bals.length; b++) {
            bals[b].innerHTML = $html;
        }
    }

    //return the span
    span = function ($html, $class = "") {
        var span = document.createElement("span");
        if ($class != "") span.classList = $class;
        span.innerHTML = $html;
        return span;
    }

    /* strip all occurances of a class out from an el or its decendants */
    stripClass = function ($el, $class, $addClass = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var children = $el.querySelectorAll("." + $class);
        if (children == null || children.length == 0) return;
        for (var c = 0; c < children.length; c++) {
            children[c].classList.remove($class);
            if ($addClass != "") children[c].classList.add($addClass);
        }
    }

    toggleWithFade = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.fadeOut($el);
            return "hide";
        }
        else {
            this.fadeIn($el);
            return "show";
        }
    }

    toggle = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains("show")) {
            this.hide($el);
            return "hide";
        }
        else {
            this.show($el);
            return "show";
        }
    }


    toggleClass = function ($el, $class) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class) == true) el.classList.remove($class);
        else el.classList.add($class);
    }

    /**
     * if class1 is turned on (after the fucntion) it is -1
     * if class2 is turned on (after the function) it is 1
     */
    toggleClass2 = function ($el, $class1, $class2) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el.classList.contains($class1) == true) {
            el.classList.remove($class1);
            el.classList.add($class2);
            return 1;
        }
        else if (el.classList.contains($class2) == true) {
            el.classList.remove($class2);
            el.classList.add($class1);
            return -1;
        }
        return 0;
    }

    triggerHover = function ($el) {
        var el = this.__getEl($el);
        if (el == null) return null;
        var event = new MouseEvent('mouseover', {
            'view': window,
            'bubbles': true,
            'cancelable': true
        });
        el.dispatchEvent(event);
    }


    setMaxH = function ($el, $maxH) {
        var el = this.__getEl($el);
        if (el == null) return null;
        el.style['max-height'] = parseInt($maxH) + "px";
        el.style['height'] = parseInt($maxH) + "px"
    }

    toggleByClass = function ($el, $topSelector, $targetClassSelector, $displayProp = "") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var parent = el.closest("." + $topSelector);
        if (parent != null) {
            var toggleTarget = parent.querySelector("." + $targetClassSelector);
            if (toggleTarget != null) {
                if (toggleTarget.classList.contains("show")) this.hide(toggleTarget);
                else this.show(toggleTarget, $displayProp);
            }
        }
    }


    /**
     * Forgot about this one.. should try it again (lol)
     * 
     * @param {*} $el 
     * @param {*} $msg 
     * @param {*} $hovertext 
     * @param {*} $attr 
     * @param {*} $classList 
     * @returns 
     */
    tooltip = function ($el, $msg, $hovertext = "hover over me", $attr = "data-text", $classList = "tooltip") {
        var el = this.__getEl($el);
        if (el == null) return null;
        var span = document.createElement("span");
        span.classList = $classList;
        span.setAttribute($attr, $msg);
        span.innerHTML = "&nbsp;&#9432;"+$hovertext;
        el.append(span);
        return span;
    }

    /**
     * 
     * @param {*} $el 
     * @param {*} $event 
     * @param {*} $bHIDE 
     * @returns 
     */
    trigger = function ($el, $event, $bHIDE = false) {
        var el = this.__getEl($el);
        if (el == null) return null;
        if (el[$event] != undefined && typeof el[$event] === 'function') {
            el[$event]();
        }
        else {
            console.log("(anne_DOM.trigger) cannot trigger event " + $event + " on element " + el.id);
            return;
        }
        if ($bHIDE == true) this.hide(el);
    }


    updateAll = function($el, $html) {
        var els = document.querySelectorAll($el);
        if (els == null || els.length==0) return null;
        for (var e=0; e < els.length; e++) {
            e.innerHTML = $html;
        }
    }


    /**
     * 
     * @param {*} $input_el 
     * @param {*} $showbtn 
     * @param {*} $rNObj 
     * @param {*} $voNOBJ 
     * @param {*} $txt 
     * @param {*} $validc 
     * @param {*} $invalidc 
     * @returns 
     */
    validator = function ($input_el, $showbtn, $rNObj = null, $voNOBJ = null, $txt = "", $validc = "valid", $invalidc = "invalid") {
        if ($txt.length) {
            if ($rNObj != null && $rNObj.format_validator != null) {
                var valid_check_rez = $rNObj.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else if ($voNOBJ != null && $voNOBJ.format_validator != null) {
                var valid_check_rez = $voNOBJ.format_validator($txt);
                if (valid_check_rez != true) {
                    ANNE.A.hide($showbtn);
                    $input_el.classList.remove($validc);
                    $input_el.classList.add($invalidc);
                    return false;
                }
                else {
                    ANNE.A.show($showbtn);
                    $input_el.classList.remove($invalidc);
                    $input_el.classList.add($validc);
                    return true;
                }
            }
            else {
                ANNE.A.show($showbtn);
                $input_el.classList.remove($invalidc);
                $input_el.classList.add($validc);
                return true;
            }
        }
        else {
            ANNE.A.hide($showbtn);
            $input_el.classList.remove($validc);
            $input_el.classList.add($invalidc);
            return false;
        }

        return false;
    }


    /**
     * YES, this is dangerous, if we were loading external scripts.  We aren't.
     * We are loading scripts in our own pages.
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $pgui 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
    */
    load = async ($loadIntoEl, $pgui, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    loadIntoEl.innerHTML=resp;
                    var scripts = loadIntoEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            loadIntoEl.removeChild(scripts[s]);
                        }
                        loadIntoEl.append(newscript);
                        //eval(newscript);
                    }

                    //if this has a guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($pgui._guid && ($pgui.page.multiples == true || $pgui.page.instantiate == true)) {
                        var items = loadIntoEl.querySelectorAll('.dynamic');

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + $pgui._guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + $pgui._guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + $pgui._guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + $pgui._guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }

                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }

                    else {
                        if ($pgui.pObj && $pgui.pObj.noshow != undefined);
                        else ANNE.A.show(loadIntoEl);
                        return resolve(loadIntoEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }


    /**
     * 
     * @param {*} $loadIntoEl 
     * @param {*} $spObj 
     * @param {*} $url 
     * @param {*} $loadClass 
     * @returns 
     */
    load2 = async ($loadIntoEl, $spObj, $url, $loadClass = "load-wrapper") => {
        var loadIntoEl = typeof $loadIntoEl === "string" ? document.getElementById($loadIntoEl) : loadIntoEl = $loadIntoEl;
        this.show(loadIntoEl);
        if (loadIntoEl == null) return;
        var xhr;
        if (window.XMLHttpRequest) xhr = new XMLHttpRequest();
        else xhr = new ActiveXObject("Microsoft.XMLHTTP");

        var tempEl = document.createElement("div");
        if ($loadClass != "") tempEl.classList = $loadClass;

        return new Promise((resolve, reject) => {
            xhr.onreadystatechange = function () {
                if (this.readyState == 4 && this.status == 200) {
                    var resp = this.response;
                    var newEl = document.createElement("div");
                    if ($loadClass != "") newEl.classList = $loadClass;
                    else newEl.classList = "pichachu";
                    newEl.innerHTML = resp;
                    loadIntoEl.append(newEl);

                    //hmm. do we want to do this?  I suppose.
                    var scripts = newEl.getElementsByTagName('script');
                    for (var s = 0; s < scripts.length; s++) {
                        var newscript = document.createElement("script");
                        newscript.type = 'text/javascript';
                        newscript.innerHTML = scripts[s].innerHTML;
                        if (scripts[s].parentNode == newEl) {
                            newEl.removeChild(scripts[s]);
                        }
                        newEl.append(newscript);
                    }

                    //if this has a spObj._guid, clone it.  This is crucial to not
                    //getting DOM ID conflicts when using templates and injecting 
                    //them multiple times.
                    if ($spObj!=null && $spObj._guid!=null && $spObj._guid!="" && $spObj._guid.length>0) {
                        var items = newEl.querySelectorAll('.dynamic');
                        var guid=$spObj._guid;

                        //any divs marked as class=dynamic get their ID altered.
                        //this will allow us to have multiple copies of these templates in
                        //the dom and stil be able to uniquely reference their elements by id.
                        var mod_els = [];
                        var temp_list = [];

                        for (var i = 0; i < items.length; i++) {
                            var item = items[i];
                            if (item.id != null && item.id != "") {
                                temp_list[item.id.substring(0)] = item;
                                item.id = item.id + "_" + guid;
                            }
                            var forattr = item.getAttribute("for");
                            if (forattr != null) item.setAttribute("for", forattr + "_" + guid);
                            var nameattr = item.getAttribute("name");
                            if (nameattr != null) item.setAttribute("name", nameattr + "_" + guid);
                            var mod_el_attr = item.getAttribute("mod-el");
                            if (mod_el_attr != null) {
                                item.setAttribute("mod-el", mod_el_attr + "_" + guid);
                                mod_els.push({ el: item, mod_el_attr: mod_el_attr });
                            }
                        }

                        //create .modxx and tie it to the element that was tagged with mod-el
                        //and the id of the target.
                        for (var m = 0; m < mod_els.length; m++) {
                            var me = mod_els[m];
                            var srcel = temp_list[me.mod_el_attr];
                            if (srcel == undefined) continue;
                            srcel.modxx = me.el;
                        }
                        return resolve(newEl);
                    }

                    else {
                        return resolve(newEl);
                    }
                }
            };

            xhr.open("GET", $url + "?v=" + Math.random() * 10, true);
            xhr.responseType = "";
            xhr.send();
        });
    }
    
}

var A = new anne_dom();
A.set_init(A);

/************** (from file ./js/base/anne_DOM.js)**/ 

class anne_utils extends anne_singleton {
    constructor(){
        super(false);
	}

    formatBitcoinPrice=function($price) {
        if (isNaN($price) || $price < 0) {
            throw new Error("Invalid input: Please provide a non-negative number.");
        }
        return $price.toFixed(8);
    }

    exists = function($val) {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        return true;
    }
    existsNonZero() {
        if ($val==undefined) return false;
        if ($val==null) return false;
        if ($val==="") return false;
        if ($val==0) return false;
    }
    ne($a,$b) {
        if (this.exists($a)==false) {
            if (this.exists($b)==true) return true;
            else return false;
        }
        if (this.exists($b)==false) return true;
        if ($a==$b) return false;
        return true;
     }

    quickAbbreviateName = function($text) {
        $text =  $text.trim();
        if ($text.length<10) return $text;
        const suffix = ["jr", "jr.", "sr", "sr.", "iii", "ii", "iv", "i", "v", "vi", "vii", "viii", "ix"];
        var parts = $text.split(" ");

        if (parts.length==2 && $text.length<17) return $text;

        var abbr = "";
        var first = "";
        var lastpart = parts[parts.length-1];
        var addsuffix = "";
        var pmax = parts.length;

        if (suffix.indexOf(lastpart.toLowerCase())!=-1) {
            addsuffix = lastpart;
            pmax = pmax-1;
        }

        for (var p=0; p < pmax; p++) {
            if (p==0) {
                abbr+=parts[p][0];
                first=parts[p];
            }
            else if (p < pmax-1){
                abbr+=parts[p][0]; //JD Rockfeller Jr
            }
            else if (p==(pmax-1)) abbr+=" "+parts[p];
            
        }
        return abbr;
    }

    stripAllNonAlphaChar=function($str, $charSlot) {
        var char = $str[$charSlot];
        if("abcdefghijklmnopqrstuvwzyz".indexOf(char)==-1) {
            $str = $str.substring(1);
            return this.stripAllNonAlphaChar($str, $charSlot);
        }
        return $str;
    }
    stripNonAlphaNum = function($str) {
        return this.stripNonIn($str, "abcdefghijklmnopqrstuvwzyz1234567890");
    }

    isMixedCase = function($str) {
        return /[a-z]/.test($str) && /[A-Z]/.test($str);
    }

    /**
     * note to self: comment this so i remember what it is.
     * @param {*} $str 
     * @param {*} $ifstr 
     * @returns 
     */
    extractNumberIf = async($str, $ifstr)=>{
        if ($str.indexOf($ifstr)!=-1) {
            if ($afterONLY==true) {
                var i = $str.indexOf($ifstr)+$ifstr.length;
                var newstr = $str.substr(i);
                return this.extractNumber($str);
            }
            else {
                return this.extractNumber($str);
            }
        }
        return null;
    }

    isSameYear = function ($nObj1, $nObj2) {
        var year1 = -1;
        var year2 = -1;
        
        if ($nObj1.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year1 = parseInt($nObj1.text);
        else {
            var year_relon = ANNE.MEM.getHasRelon($nOb1.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year1 = parseInt(yearNOBJ.text);
            }
        }

        if ($nObj2.bes.indexOf(ANNE.MEM.integer.temp_nid)!=-1) year2 = parseInt($nObj2.text);
        else {
            var year_relon = ANNE.MEMG.getHasRelon($nObj2.outs, MEM.year.temp_nid);
            if (year_relon!=null) {
                var yearNOBJ = MEMG.getNeuronByNid(year_relon.to_s);
                if (yearNOBJ!=null) year2 = parseInt(yearNOBJ.text);
            }
        }

        if(year1!=-1 && year2!=-1 && year1==year2) return true;
        return false;
        
    }


    extractNumber2 = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        parts.splice(p, 1);
                        return {nObj:numNOBJ, str:part, remaining:parts.join(" ")};
                    }
                    break;
                }
            }
        }
        return null;
    }


    extractNumber = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    var numNOBJ = await MEMG.getOrMakeIfNewInteger(null, number);
                    if (numNOBJ!=null) {
                        return numNOBJ;
                    }
                    break;
                }
            }
        }
        return null;
    }

    extractNumberStr = async($str, $pmax=-1)=> {
        var parts = $str.split(/[\s-]+/);
        if (parts.length>0) {
            if ($pmax>-1 && parts.length>$pmax) return null;
            for (var p=0; p < parts.length; p++) {
                var part = parts[p];
                if (part=="") continue;
                if (ANNE.UTILS.isNum(part)==true) {
                    const ordinalRegex = /(st|nd|rd|th)$/;
                    let strippedNumber = part.replace(ordinalRegex, '');
                    let number = parseInt(strippedNumber, 10);
                    return number+"";
                }
            }
        }
        return "";
    }


    replaceWeirdLetters = function($str) {
        const regex = /[À-ÖØ-öø-ÿ]/g;
        const AtoZ = {
          À: "A",
          Á: "A",
          Â: "A",
          Ã: "A",
          Ä: "A",
          Å: "A",
          Æ: "AE",
          Ç: "C",
          È: "E",
          É: "E",
          Ê: "E",
          Ë: "E",
          Ì: "I",
          Í: "I",
          Î: "I",
          Ï: "I",
          Ð: "D",
          Ñ: "N",
          Ò: "O",
          Ó: "O",
          Ô: "O",
          Õ: "O",
          Ö: "O",
          Ø: "O",
          Ù: "U",
          Ú: "U",
          Û: "U",
          Ü: "U",
          Ý: "Y",
          Þ: "TH",
          ß: "ss",
          à: "a",
          á: "a",
          â: "a",
          ã: "a",
          ä: "a",
          å: "a",
          æ: "ae",
          ç: "c",
          è: "e",
          é: "e",
          ê: "e",
          ë: "e",
          ì: "i",
          í: "i",
          î: "i",
          ï: "i",
          ð: "d",
          ñ: "n",
          ò: "o",
          ó: "o",
          ô: "o",
          õ: "o",
          ö: "o",
          ø: "o",
          ù: "u",
          ú: "u",
          û: "u",
          ü: "u",
          ý: "y",
          þ: "th",
          ÿ: "y",
        };
        var result = $str.replace(regex, (match) => AtoZ[match]);
        console.log(result); // Output: "AAA EEE III OOO UUU"
        return result;
    }


    containsTXID=function($rels, $txid) {
        for (var r=0; r <$rels.length; r++) {
            if ($rels[r].txid==$txid) return true;
        }
        return false;
    }
    

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    fixHack=function($str=null, $replace_underscores="") {
        if ($str==undefined || $str==null || typeof $str !="string") {
            console.log("don't call this, sir, and non-existant $str");
            return "";
        }

        //dont muck with date looking strings.
        var aParts = $str.split("-");
        if (aParts.length>=2) {
            var numbtest = parseInt($str.replaceAll("-", ""));
            if (numbtest > 0) {
                var parenObj = ANNE.UTILS.stripParens($str);
                console.log(parenObj);
                return parenObj;
            }
        }
        
        $str = $str.replaceAll("-2D","-");
        $str = $str.replaceAll("-28", "(");
        $str = $str.replaceAll("-29", ")");

        //too tired to chase it.  lazy kill
        //we already yank info from wikispooks in a hacky non rdf way..
        //so it comes with drama
        var apostrophe = $str.indexOf("-27");
        if (apostrophe!=-1 && apostrophe!=0 && apostrophe!=$str.length-1) {
            $str = $str.replaceAll("-27", "'");
        }
        
        var fix = $str.indexOf("-28unidentified-29");
        if (fix!=-1 && fix!=0) {
            $str = $str.replaceAll("-28unidentified-29", "");
        }

        $str = $str.replaceAll("-26", "&");
        $str = $str.replaceAll("27-", "");

        var parenObj = ANNE.UTILS.stripParensMulti($str);

        if ($replace_underscores!="") {
            parenObj.str = parenObj.str.replaceAll("_", $replace_underscores).trim();
        }
        else parenObj.str.trim();
        
        //console.log(parenObj);
        return parenObj;
    }

    
    blowOutAmpersands = function($arr) {
        var finalArray = [];
        for (var a=0; a < $arr.length; a++) {
            var val = $arr[a];
            if (val.indexOf("&")==-1) finalArray.push(val);
            else {
                var parts = val.split(" & ");
                for (var p=0; p < parts.length; p++) {
                    var val2 = parts[p].trim();
                    if (finalArray.indexOf(val2)==-1) finalArray.push( parts[p].trim() );
                }
            }
        }
        return finalArray;
    }


    //bob and suzy smith
    //s.b.  "bob smith", "suzy smith"
    extrapAmperStr = function($str) {
        var parts = $str.split(" & ");
        if (parts.length==1) part2 = $str.split(" and ");
        if (parts.length!=2) return null;
        
        var last_part = parts[ parts.length-1];
        var first_part = parts [0];
        if (last_part.split(" ").length > first_part.split(" ").length) {
            var last = last_part.split(" ");
            var last_name = last [ last.length-1];
            var first_person = first_part+" "+last_name;
            return [first_person, last_part];
        }
        return null;
    }

    /**
     * Strips a string of its paren and returns the new string with the stripped value in the context.
     * @param {*} $str 
     * @returns 
     */
    stripParens = function($str){
        const regex = /\s?\(([^)]+)\)/;
        var match = $str.match(regex);
        var newStr = $str.replace(/\s?\([^)]+\)/, '');
        var context = "";
        if (match && match.length>0) context = match[1];
        if (newStr==null || newStr=="" || newStr.length==0) newStr = $str;
        return {str:newStr.trim(), context:context};
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    prepNakedSearch = function($str) {
        var pOBJ = this.stripParens($str);
        pOBJ.inputStr = $str;
        const inputString = pOBJ.str;
        const regexPattern = /(['"])(.*?)\1/;

        let QUOTED = false;
        let extractedString = '';

        const match = inputString.match(regexPattern);
        if (match) {
            QUOTED = true;
            extractedString = match[2]; 
        }

        if (QUOTED==true) {
            pOBJ.bEXACT=true;
            pOBJ.str = extractedString;
        }
        else pOBJ.bEXACT=false;

        console.log("naked search:", pOBJ);
        return pOBJ;
    }


    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    stripParensMulti=function($str) {
        const regex = /\(([^)]+)\)/;
        const match = $str.match(regex);
        let newStr = $str.replace(/\([^)]+\)/, '');
        let context = [];
        if (match && match.length > 0) {
            context = match[1].split(',').map(item => item.trim());
        }
        if (newStr == null || newStr == "" || newStr.length == 0) {
            newStr = $str;
        }
        return {str: newStr.trim(), context: context};
    }


    /**
     * 
     * @param {*} $str 
     * @param {*} $stripifnotthis 
     * @returns 
     */
    stripNonIn=function($str, $stripifnotthis) {
        var newstr = $str.concat();
        for (var a=0; a < newstr.length; a++)  {
            var chara = newstr[a];
            if ($stripifnotthis.indexOf(chara)==-1) {
                newstr.slice(a, 1);
                a--;
            }
        }
        return newstr;
    }


    /**
     * For testing purposes, the firstseen datestamp marks
     * when the annode first saw a relon.
     * 
     * If the relon was in the mempool (time/block=-1) but the annode restarted..
     * and the relon never made it back to the mempool...
     * 
     * We have a faux relon in our localDB that the ANNE network doesn't actually know
     * about.  Our annode may know of it, but the the network does not.
     * 
     * We need to fix our indexDB.
     * 
     * We need to solve this problem in the annode as well.
     * Shutdown mid-mempool will leave neuron indexes created, neurons froze, etc..
     * which will reject the next attempt to create the same neuron indexes
     * (and will actually resolve the nid index conflict by creating a new one)
     * 
     * This will be a HUGE problem as that annode will be out of concensus.
     * @param {*} $sec 
     * @returns 
     */
    shouldBeMined = function($sec) {
        var se = this.timeElapsed($sec);
        if (se > 600) return true;
        return false;
    }

    timeElapsed = function($secs) {
        var currentTimestamp = Date.now();
        var timeElapsed = currentTimestamp - parseInt($secs);
        var secondsElapsed = Math.floor(timeElapsed / 1000);
        console.log("Time elapsed in seconds: " + secondsElapsed);
        return secondsElapsed;
    }


    secsElapsedToDHMS = function($secs) {
        const secmin = 60;
        const minshr = 60;
        const sechr = secmin*60;
        const secsday = sechr*24;
        var hrsday = 24;
        var minutes = Math.floor($secs/60);

        if (minutes > secmin) {
            var hours = Math.floor(minutes/60);
            if (hours > hrsday) {
                var days = Math.floor(hours / hrsday);
                var hrs_left = hours - (days*hrsday);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var dayswd = ANNE.TALK.pluralizeNounIf(days, "day", "s");
                var hrswd = ANNE.TALK.pluralizeNounIf(hrs_left, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
                return days+dayswd+" "+hrs_left+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
            else {
                hours = Math.floor(hours);
                var mins_left = Math.round(minutes % 60);
                var secs_left = Math.round($secs % 60);
                var hrswd = ANNE.TALK.pluralizeNounIf(hours, "hr", "s");
                var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
                var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s");
                return hours+hrswd+" "+mins_left+minswd+" "+secs_left+secsswd;
            }
        }
        else if (minutes > 0) {
            var minswd = ANNE.TALK.pluralizeNounIf(mins_left, "min", "s");
            var secs_left = Math.round($secs % 60);
            var secsswd = ANNE.TALK.pluralizeNounIf(secs_left, "sec", "s")
            return minutes +minswd+" "+secs_left+secsswd;
        }
        else {
            var secsswd = ANNE.TALK.pluralizeNounIf($secs, "sec", "s")
            return $secs+secsswd;
        }
    }

    forceInt = function($int) {
        if ($int==undefined || $int==null || $int=="") return 0;
        var int2 = parseInt($int);
        if (isNaN(int2)==true) return 0;
        return int2;
    }

    capFirstLetters = function($str) {
        var wds = $str.split(" ");
        for (var p=0; p < wds.length; p++) {
            wds[p] = wds[p][0].toUpperCase() + wds[p].substr(1);
        }
        return wds.join(" ");
    }

    concatUniqueNObjs=function($a1, $a2) {
        if ($a1.length==0) return $a2;
        if ($a2.length==0) return $a1;
        var a_temp_nid_list=[];
        for (var a=0; a < $a1.length; a++) {
            a_temp_nid_list.push($a1[a].temp_nid);
        }
        var a2_uniques = [];
        for (var a2=0; a2 < $a2.length; a2++) {
        var a2_temp_nid = $a2[a2].temp_nid;
            if (a_temp_nid_list.indexOf(a2_temp_nid)!=-1) continue;
            a2_uniques.push($a2[a2]);
        }
        return $a1.concat(a2_uniques);
    }


    copyToClipboard=function($str) {
        var el = document.createElement('textarea');
        el.value = $str;
        el.setAttribute('readonly', '');
        el.style = {position: 'absolute', left: '-9999px'};
        document.body.appendChild(el);
        el.select();
        document.execCommand('copy');
        document.body.removeChild(el);
    }

    concatUniqueRels = function ($arr1, $arr2) {
        var txids = [];
        var rels = [];
        for (var t=0; t < $arr1.length; t++) {
            var txid = $arr1[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr1[t]);
            }
        }
        for (var t=0; t < $arr2.length; t++) {
            var txid = $arr2[t].txid;
            if (txids.indexOf(txid)==-1) {
                txids.push(txid);
                rels.push($arr2[t]);
            }
        }
        return rels;
    }
    

    convertToCSV = function ($objArray, $bGENHEADER=false) {
        var array = typeof $objArray != 'object' ? JSON.parse($objArray) : $objArray;
        var str = '';

        if ($bGENHEADER==true) {
            var keys = Object.keys(array[0]);
            var header = new Object();
            for (var k=0; k < keys.length; k++) {
                header[keys[k]]=keys[k];
            }
            array.unshift(header);
        }

        for (var i = 0; i < array.length; i++) {
            var line = '';
            for (var index in array[i]) {
                if (line != '') line += ','

                line += array[i][index];
            }
            str += line + '\r\n';
        }
        return str;
    }

    l33t_checkForce = function($str) {
        var l33ts_num = [];
        var l33ts_alpha = [];
        
        //1234567890
        //1=l
        for (var s=0; s < $str.length; s++) {
            var char = $str[s]=$str[s];
            if (char=="0") {
                l33ts_num.push("0");
                l33ts_alpha.push("O");
            }
            else if (char=="1") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else if (char=="2") {
                l33ts_num.push("1");
                l33ts_alpha.push("l");
            }
            else {
            } 
        }

    }


    /**
     * Alpha num.
     * single spaces ok.
     */
    cleanAnorgName = function($propName) {
        var str = $propName.replace(/ +(?= )/g,'');
        var cleanstr = str.replace(/[^a-z0-9 ]/gi,'');
        return cleanstr;
    }


    cleanPropName = function($propName) {
        var b = $propName.replace(/[^a-z0-9_]/gi,'');
        if (b.match(/^\d/)==true) {
            console.log("(utils.cleanPropname) result: "+"p_"+b);
            return "p_"+b;
        }
        return b;
    }

    roundThenCommas=function($x) {
        var rounded = Math.round($x);
        return this.commas($x);
    }

    //.toFixed(8);
    abbreviateNumber=function($num) {
        $num = new Number($num);
        if ($num >= 1e12) {
            return (Math.round($num / 1e11) / 10) + 't'; 
        } 
        else if ($num >= 1e9) {
            return (Math.round($num / 1e8) / 10) + 'b'; 
        } 
        else if ($num >= 1e6) {
            return (Math.round($num / 1e5) / 10) + 'm'; 
        } 
        else if ($num >= 1e3) {
            return (Math.round($num / 1e2) / 10) + 'k'; 
        } 
        else {
            return $num.toString(); 
        }
    }

    commas=function($x, $bDOSPACES=false) {
        if ($x === undefined || $x === null) {
            console.log("huh?");
            return;
        }

        let parts = $x.toString().split('.');
        let integerPart = parts[0];
        let decimalPart = parts.length > 1 ? '.' + parts[1] : '';
    
        if ($bDOSPACES === true) {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, " ");
        } 
        else {
            integerPart = integerPart.replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        }
    
        return integerPart + decimalPart;
    }


    estimateTimeByBlocks = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        const currentDate = new Date();
        const futureDate = new Date(currentDate.getTime() + secsMore * 1000);
        const options = { month: 'long', day: 'numeric', year: 'numeric' };
        const formattedDate = futureDate.toLocaleDateString('en-US', options);
        console.log(formattedDate);
        return formattedDate;
    }   

    estimateTimeByBlocksLeft = function($blocksMore) {
        var minsMore = $blocksMore*1.31;
        var secsMore = minsMore*60;
        return this.secsElapsedToDHMS(secsMore);
    }

    
    firstCap = function($str, $bALL=true) {
        if ($bALL==true) {
            var newstr = [];
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
            for (var p=0; p < parts.length; p++) {
                var str = parts[p][0].toUpperCase()+parts[p].substring(1);
                newstr.push(str);
            }
            return newstr.join(" ");
        }
        else {
            var parts = $str.split(" ");
            if (parts.length==1) return parts[0][0].toUpperCase()+parts[0].substring(1);
        } 
    }


    //won't get fully spelled numbers like "one"
    isNum = function($n) {
        if (!isNaN(parseFloat($n)) && isFinite($n)==true) return true;
        const regex = /^\d+(st|nd|rd|th)?$/;
        return regex.test($n);
    }
    

    /**
     * 
     * from https://gist.github.com/tralves/9e5de2bd9f582007a52708d7d4209865
     * @param {*} bytes 
     * @returns 
     */
    frenlySize = function (bytes) {
        var thresh = 1024;
        if (Math.abs(bytes) < thresh) return bytes + ' B';
        var units = ['KB','MB','GB','TB','PB','EB','ZB','YB'];
        var u = -1;
        do {
            bytes /= thresh;
            ++u;
        } 
        while(Math.abs(bytes) >= thresh && u < units.length - 1);
        return bytes.toFixed(1)+' '+units[u];
    }



    /**
     * Calculate annex price lvl factor via nth fib.
     * 
     * could be done with matrix math but we won't be getting high lvl numbers here
     * and that optimization is just showing off.
     * this suffices.
     * 
     * DO NOT CHANGE THIS FUNCTION UNLESS YOU ALSO CHANGE THE FUNCTION ON THE ANNODE.
     * 
     * @param {*} $nth 
     * @returns 
     */
    nthFib=function($nth){
        if ($nth==1 || $nth==2) return 1;
        var seq = [1, 1];
        var nextval = 0;
        for (var n=2; n < $nth; n++) {
            nextval = seq[n-1] + seq[n-2];
            seq.push(nextval);
        }
        return nextval;
    }


    /**
     * Only 2 decimal places
     * @param {*} $annies 
     * @param {*} $gap 
     * @returns 
     */
    displayAnneValue = function($annies, $gap=" ") {
        var parts = ((parseInt($annies)/ANNE.ANNE_TO_ANNIES)+"").split(".");
        if (parts.length==2) {
            var whole = UTILS.commas(parts[0]);
            var deci = parts[1];
        }
        else if (parts.length==1) {
            if ($annies<ANNE.ANNE_TO_ANNIES) {
                var whole = 0;
                var deci = parts[0];
            }
            else {
                var whole = UTILS.commas(parts[0]);
                var deci = "0";
            }
        }
        if (deci==undefined) return whole;
        deci = deci+"";
        
        if (deci.length==2) deci = "."+$gap+deci;
        else if (deci.length==1) deci = "."+$gap+deci+"0";
        else if (deci.length==0) return whole;
        
        return whole;//+deci;
    }


    /**
     * 
     * @param {*} $val 
     * @param {*} $minDigits 
     * @returns 
     */
    formatDigits = function($val, $minDigits=4) {
         if ($val==null) {
            console.log("%cERROR val can't be null here", TALK.bgred);
            return "";
        }
        var final = $val+"";
       
        if (final.length<$minDigits) {
            var dif = $minDigits-final.length;
            for (var d=1; d <= dif; d++) { final = "0"+final; }
        }
        return final;
    }

    
    //For calculating edit distance
    editDistance = function (s1, s2) {
        s1 = s1.toLowerCase();
        s2 = s2.toLowerCase();
        var costs = new Array();
        for (var i = 0; i <= s1.length; i++) {
            var lastValue = i;
            for (var j = 0; j <= s2.length; j++) {
                if (i == 0)
                    costs[j] = j;
                else {
                if (j > 0) {
                    var newValue = costs[j - 1];
                    if (s1.charAt(i - 1) != s2.charAt(j - 1)) newValue = Math.min(Math.min(newValue, lastValue), costs[j]) + 1;
                    costs[j - 1] = lastValue;
                    lastValue = newValue;
                }
            }
        }
        if (i > 0)
            costs[s2.length] = lastValue;
        }
        return costs[s2.length];
    }


    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportJSONFileViaID=function($dataEl, $fileTitle, $el="", $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportJSONFile(jsonStringData, $fileTitle, $el, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $object 
     * @param {*} $fileTitle 
     */
    exportJSONFile=function($object, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.json' || 'export.json';
        var jsonstring = JSON.stringify($object);
        const bytes = new TextEncoder().encode(jsonstring);
        const json = new Blob([bytes], {
            type: "application/json;charset=utf-8"
        });

        var link = document.createElement("a");
        if (link.download !== undefined) { // feature detection
            // Browsers that support HTML5 download attribute
            var url = URL.createObjectURL(json);
            link.setAttribute("href", url);
            link.setAttribute("download", exportedFilenmae);
            link.style.visibility = 'hidden';
            var el = null;
            el = document.body;
            el.appendChild(link);
            link.click();
            el.removeChild(link);
        }
    }

    /**
     * 
     * @param {*} $dataEl 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFileViaID=function($dataEl, $fileTitle, $bGENHEADER=true) {
        console.log("'"+$dataEl+"'");
        var jsonStringData = document.getElementById($dataEl).innerHTML;
        if (jsonStringData) {
            this.exportCSVFile(jsonStringData, $fileTitle, $bGENHEADER);
        }
    }

    /**
     * 
     * @param {*} $jsonStringOrObjectArray 
     * @param {*} $fileTitle 
     * @param {*} $el 
     * @param {*} $bGENHEADER 
     */
    exportCSVFile=function($jsonStringOrObjectArray, $fileTitle,  $bGENHEADER=true) {
        var csv = this.convertToCSV($jsonStringOrObjectArray, $bGENHEADER);
        this.exportCSVFile_inner(csv, $fileTitle);
    }

    /**
     * 
     * @param {*} $csv 
     * @param {*} $fileTitle 
     */
    exportCSVFile_inner=function($csv, $fileTitle) {
        var exportedFilenmae = $fileTitle + '.csv' || 'export.csv';
        var blob = new Blob([$csv], { type: 'text/csv;charset=utf-8;' });
        if (navigator.msSaveBlob) { // IE 10+
            navigator.msSaveBlob(blob, exportedFilenmae);
        } 
        else {
            var link = document.createElement("a");
            if (link.download !== undefined) { // feature detection
                // Browsers that support HTML5 download attribute
                var url = URL.createObjectURL(blob);
                link.setAttribute("href", url);
                link.setAttribute("download", exportedFilenmae);
                link.style.visibility = 'hidden';
                var el = null;
                el = document.body;
                el.appendChild(link);
                link.click();
                el.removeChild(link);
            }
        }
    }

    posNegRand = function($fact) {
        return (Math.random()-.5)*$fact;
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValAndRemoveInObjnArr = function ($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return true;
            }
        }
        return false;
    }

    /**
     * 
     * @param {*} $rels 
     * @param {*} $skip_tnids 
     * @returns 
     */
    getLatest = function($rels, $skip_tnids=[]) {
        var lastValid = null;
        for (var sr=0; sr < $rels.length; sr++) {
            var rel = $rels[sr];
            if ($skip_tnids.length && $skip_tnids.indexOf(rel.r)!=-1) continue;
            if (lastValid==null) lastValid = rel;

            //if we have a bigger block height OR a 0 (indicating mempool not updated yet)
            else if ((lastValid.time) > 0 && (rel.time > 0 || rel.time==0)) {
                if (rel.time > lastValid.time) lastValid = rel;
            }

            //this has been firstseen later, meaning it is newer
            else if (lastValid.firstseen < rel.firstseen) {
                lastValid = rel;
            }
        }
        return lastValid;
    }


    /**
     * 
     * @param {*} $rels 
     * @returns 
     */
    getLastSeen = function($rels) {
        if ($rels==null) return null;
        if ($rels.length==0) return null;
        if ($rels.length==1) return $rels[0];
        var last_seen = 0;
        var lastseenRel = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (rel.firstseen!=undefined && rel.firstseen > last_seen) {
                lastseenRel = rel;
                last_seen = rel.firstseen;
            }
        }
        if (lastseenRel!=null) return lastseenRel;
        if ($rels.length) return $rels [ $rels.length-1 ];
    }

        
    /**
     * 
     * @param {*} $rels 
     * @param {*} $annie 
     * @returns 
     */
    getFirst = function($rels, $annie=0) {
        if ($annies>0) return this.getFirstWithGTEAnnies($rel, $annies);
        var first = null;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if (first==null) first=rel;
            else if ((rel.time < first.time) || (rel.time==0 && first.time!=0)) first=rel;
            else if (rel.time==first.time) {
                if (rel.firstseen < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * attemps to handle multi payments if there was underpayment.
     * will use time/first seen from the last payment in the bunch to qualify
     * the payment as being the first seen (winner) of the FCFS 
     * 
     * @param {*} $rels 
     * @param {*} $annies 
     * @returns 
     */
    getFirstWithGTEAnnies = function($rels, $annies) {
        var first = null;
        var from_multi = new Object();

        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var annies = rel.annies; 
            var rel_fs = rel.firstseen;
            var rel_time = rel.time;

            if (annies==undefined || annies==0) continue;
            if (annies < $annies) {
                var fm_this = from_multi[ rel.f ];
                if (fm_this==undefined) {
                    from_multi[ rel.f ] = {annies:annies, time:rel.time, firstseen:rel.firstseen, rel:rel};
                }
                else {
                    fm_this.annies += annies; 
                    if (rel.time > fm_this.time) {
                        fm_this.time = rel_time;
                        fm_this.firstseen = rel_fs;
                        fm_this.rel = rel;
                    }

                    if (fm_this.annies < $annies) continue; //no enough, still!
                }
            }

            if (first==null) first=rel;
            else if ((rel_time < first.time) || (rel_time==0 && first.time!=0)) first=rel;
            else if (rel_time==first.time) {
                if (rel_fs < first.firstseen) first=rel;
            }
        }
        return first;
    }


    /**
     * 
     * @param {*} $txt 
     * @param {*} $bDECIMALOK 
     * @returns 
     */
    inputNumberOnly = function ($txt, $bDECIMALOK=false) {
        var newString = "";
        for (var c=0; c < $txt.length; c++) {
            var charCode = $txt.charCodeAt(c);
            //console.log(charCode);
            if ($bDECIMALOK==false) {
                if (charCode < 48 || charCode> 57) continue;
                newString = newString + $txt[c];
            }
            else if ($bDECIMALOK==true) {
                if (charCode < 48 || (charCode> 57 && charCode!=190)) continue;
                newString = newString + $txt[c]; 
            }  
        }
        return newString;
    }


    /**
     */
    addUniqueRel=function($arr, $rel) {
        UTILS.addUnique($arr, "txid", $rel);
    }


    /**
     * 
     * @param {Array} $objArr 
     * @param {Array} $destination_arr 
     * @param {String} $uniqueObjProp 
     */
    addUniqueObjArray=function($objArr, $destination_arr, $uniqueObjProp="txid") {
        for (var r=0; r < $objArr.length; r++) {
            var newObj = $objArr[r];
            UTILS.addUnique($destination_arr, $uniqueObjProp, newObj);
        }
    }

    /**
     * 
     * @param {*} $arr 
     * @param {*} $tnids 
     */
    addUniqueTnids=function($arr, $tnids) {
        for (var t=0; t < $tnids.length; t++) {
            var tnid=$tnids[t];
            if ($arr.indexOf(tnid)==-1) $arr.push(tnid);
        }
    }



    /**
     * 
     * @param {Array} $arr 
     * @param {String} $uniqueProp 
     * @param {Object} $newRelon 
     * @returns 
     */
    addUnique=function($arr, $uniqueProp="txid", $newRelon) {
        if ($arr==null) return false;
        if ($newRelon==null) return false;

        if (this.findValInObjnArr($arr, $uniqueProp, $newRelon[$uniqueProp])==null) {
            $arr.push($newRelon);
            return true;
        }
        return false;
    }


    /**
     * 
     */
    deepCopy=function($obj) {
        if ($obj === null || typeof $obj !== 'object') {
            return $obj;
        }
    
        let copy = new $obj.constructor();
    
        for (let key in $obj) {
            if ($obj.hasOwnProperty(key)) {
                copy[key] = this.deepCopy($obj[key]);
            }
        }
        return copy;
    }
    
    
    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    findValInObjnArr = function ($arr, $prop, $val) {
        if ($prop!="") {
            for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return $arr[a];
            return null;
        }
        for (var a=0; a < $arr.length; a++) if ($arr[a]==$val) return $arr[a];
        return null;
    }


    /**
     * Does an item exist in the array with the existing prop value set?
     * @param {*} $arr 
     * @param {*} $prop 
     * @returns 
     */
    findValInObjnArr2 = function ($arr, $prop, $val) {
        for (var a=0; a < $arr.length; a++) if ($arr[a][$prop]==$val) return true;
        return false;
    }


    sortByProp = function($arr, $prop) {
        $arr.sort( 
            function($a, $b) {
                var val_a = $a[$prop];
                if (val_a==undefined) return 0;

                var val_b = $b[$prop];
                if (val_b==undefined) return 0;

                if (val_a < val_b) return -1;
                else if (val_a > val_b) return 1;
                return 0;
            }
        );
        return $arr;
    }


    /**
     * 
     * @param {*} $arr 
     * @param {*} $prop 
     * @param {*} $val 
     * @returns 
     */
    removeObjInObjArr = function($arr, $prop, $val) {
        for (var a=0; a<$arr.length; a++) {
            if ($arr[a][$prop]==$val){
                $arr.splice(a, 1);
                return;
            }
        }
    }


    //weak attempt to add .png to end of bad imgur url
    fixImgURL=function($imgUrl) {
        if ($imgUrl.indexOf("imgur")==-1) return $imgUrl;   //do nothing
        var aParts = $imgUrl.split(".");
        aParts.reverse();
        var ext = aParts[0];
        aParts.reverse();
        if (ext.length<3 || ext.length>4) $imgUrl = $imgUrl+".png";
        else if (ext!="png" && ext!="jpg") $imgUrl = $imgUrl+".png";
        return $imgUrl;
    }


    /**
     * try to format the bigger number.
     * if it has a deci and we can't short-hand it with a B or M or K,
     * then just format the number with commas.
     */
    formatNumber=function($num, $bROUND=false) {
        var num_parts = ($num+"").toString().split(".");
        var num = parseInt(num_parts[0]);
        if (num >0 && num_parts.length>1 && $bROUND==false) return this.formatNumberCommas($num);

        var b = num / 1000000000;
        if (b>=1) {
            if ($bROUND==true) {
                var b100=Math.floor(b*100)/100;
                num_parts[0]=((b100+"").split(".").join("."))+" B";
            }
            else num_parts[0]=b+"B";
        }
        else {
            var m = num / 1000000;
            if (m>=1) {
                if ($bROUND==true) {
                    var m100=Math.floor(m*100)/100;
                    num_parts[0]=((m100+"").split(".").join("."))+" M";
                }
                else num_parts[0]=m+"M";
            }
            else {
                var k = num / 1000;
                if (k>=1) {
                    if ($bROUND==true) {
                        var k100=Math.floor(k*100)/100;
                        num_parts[0]=((k100+"").split(".").join("."))+" K";
                    }
                    else num_parts[0]=k+" K";
                }
                else num_parts[0]=num_parts[0];
            }
        }
        if ($bROUND==true) return num_parts[0];
        else if (num_parts.length==1) return num_parts[0];
        return num_parts[0]+"."+num_parts[1];
    }


    formatNumberCommas=function($num) {
        var num_parts = ($num+"").toString().split(".");
        num_parts[0] = num_parts[0].replace(/\B(?=(\d{3})+(?!\d))/g, ",");
        return num_parts.join(".");
    }

    convArr=function($arr) {
        var final_arr = [];
        var keys = Object.keys($arr);
        for (var k=0; k < keys.length; k++) {
            final_arr.push ( $arr [ keys[k]]);
        }
        return final_arr;
    }
    

    getSelectedOptionFromDD=function($dd_el) {
        var sel = document.getElementById($dd_el);
        return this.getSelectedOption(sel);
    }

    getSelectedOption = function($sel) {
        var opt;
        for ( var i = 0, len = $sel.options.length; i < len; i++ ) {
            opt = $sel.options[i];
            if ( opt.selected === true ) {
                break;
            }
        }
        return opt;
    }

    getUrlVars = function () {
        var vars = {};
        var parts = window.location.href.replace(/[?&]+([^=&]+)=([^&]*)/gi, function (m, key, value) {
            vars[key] = value;
        });
        return vars;
    }

    hex_to_ascii=function($str1) {
        var hex  = $str1.toString();
        var str = '';
        for (var n = 0; n < hex.length; n += 2) {
            str += String.fromCharCode(parseInt(hex.substr(n, 2), 16));
        }
        return str;
    }

    htmlEntities=function(str) {
        return String(str).replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;').replace(/"/g, '&quot;').replace(/'/g, '&apos;');
    }

    
    isPrime($x){     
        var d = $x-1;
        while (d > 1){
          if (($x % d) == 0) return false;
          d--;
        }
        return true;
      }


    /**
     * 
     * @param {*} $map 
     * @param {*} $prop 
     * @returns 
     */
    __sortNumObjMapByProp = function($map, $prop="other") {
        var keys = Object.keys($map);
        keys.sort( function($a, $b) {
            var val_a = $map[$a][$prop];
            if (val_a==undefined) return 0;

            var val_b = $map[$b][$prop];
            if (val_b==undefined) return 0;

            if (val_a < val_b) return -1;
            else if (val_a > val_b) return 1;
            return 0;
        });
        
        var sortedMapObj = new Object();
        keys.forEach(key => {
            sortedMapObj[key] = $map[key];
          });
        return sortedMapObj;
    }


    __sortRelnKeyMap = function($relnMAP) {
        var keys = Object.keys($relnMAP);
        keys.sort( function($a, $b) {
            var rels_a = $relnMAP[$a];
            var text_a = MEMG.getNeuronByTempNid(rels_a[0].r).text.toLowerCase();
            var rels_b = $relnMAP[$b];
            var text_b = MEMG.getNeuronByTempNid(rels_b[0].r).text.toLowerCase();
            if (text_a < text_b) return -1;
            else if (text_a > text_b) return 1;
            return 0;
        });
        return keys;
    }

    endsWithStrip = function($str, $char=",") {
        if ($str[$str.length-1]==$char) {
            return $str.substring(0, $str.length-1);
        }
        return $str;
    }
    
    /**
     * 
     * @param {*} $val 
     * @returns 
     */
    isset = function($val) {
        if ($val==undefined || $val==null) return false;
        var _type = typeof $val;
        if (_type =="boolean") return $val;
        
        if (_type =="string") { 
            if($val.length==0) return false;
            return true;
        }
        if (_type=="number") {
            if($val==0) return false;
        }
        if (Array.isArray($val)==true && $val.length==0) return false;
        return true;
    }


    /**
     * move over unique props from one object to another
     * This is helpful when merging objects that are not in a custom class..
     * (basic objs or json being used)
     */
    moveUniquePropsToObj=function($obj_from, $obj_to) {
        if ($obj_from == null) return;
        if ($obj_to == null) return;
        var keys = Object.keys($obj_from);
        for (var k=0; k < keys.length; k++) {
            var prop = keys[k];
            if ($obj_to[prop]!=undefined) continue;
            var val = $obj_from[ prop ];
            $obj_to[prop]=val;
        }
    }


    printunixds=function($unixds) { 
        var d = new Date($unixds); //, timeZone:'UTC', timeZoneName:'short'
        const options = { year: 'numeric', month: '2-digit', day: '2-digit'};
        return d.toLocaleDateString('fr-CA', options)+" "+d.toLocaleTimeString();
    }

    //too tired to right now to implement the listsplitter.
    //this is for lazy sloppy jackass .csv files that include multiple
    //values in the column with a list.
    //we can split them into their own column and then give the user
    //the option to map it to an AKA.
    //for now, it gets nuked.
    processCSV_withHeader=async($csv, $delim, $listsplitter=";")=> {
        var allTextLines = $csv.split(/\r\n|\n/);
        var headers = [];
        var headerRow = allTextLines[0].split($delim);
        for (var h=0; h<headerRow.length; h++) {
            var clean = headerRow[h].replace(/['"]+/g, '');
            headers.push(clean.trim());
        }
        console.log(headers);
        var csvRowObjs = [];
        for (var n=1; n < allTextLines.length-1; n++) {
            var row = allTextLines[n].split($delim);
            var rowObj = [];
            for (var j=0; j<row.length; j++) {
                var clean2 = row[j].replace(/['"]+/g, '');
                var colName = headers[j];
                if (clean2.indexOf($listsplitter)!=-1) {
                    var aParts = clean2.split(";");
                    clean2 = aParts[0].trim();
                }
                if (headers!=null) rowObj[ colName ] = clean2;
                else rowObj.push(clean2.trim());
            }
            rowObj.__ROW_ACTIVE=true;
            csvRowObjs.push(rowObj);
        }
        return {'headers':headers, 'data': csvRowObjs};
    }

    processJsonKeyPairs=async($json, $delim)=> {
        var headers = [];
        var finaldata = [];
        for (var n=0; n < $json.length; n++) {
            var row = $json[n];
            //first loop snag all props from first row and make headers.
            if (n==0) {
                for (var prop in row) {
                    var header = prop;//.replace(/['"]+/g, '');
                    headers.push(header);
                }
            }
            var rowObj = [];
            //loop the headers, snag the objects
            for (var h=0; h<headers.length; h++) {
                var prop = headers[h];
                var value = row[prop];
                if (value=="null") value=="";
                else if (typeof value ==='string') value = value.replace(/['"]+/g, '');
                rowObj[ prop ] = value;
            }
            rowObj.__ROW_ACTIVE=true;
            finaldata.push(rowObj);
        }
        return {'headers':headers, 'data': finaldata};
    }
    
    pushUniqueProp=function($obj, $prop, $val) {
        if (Array.isArray($obj[$prop]) == true) { if ($obj[$prop].indexOf($val)==-1) $obj[$prop].push($val); }
        else {
            $obj[$prop]=[];
            $obj[$prop].push($val);
        }
    }

    fisherYatesShuffle=function($array) {
        for (let i = $array.length - 1; i > 0; i--) {
            const j = Math.floor(Math.random() * (i + 1));
            [$array[i], $array[j]] = [$array[j], $array[i]];
        }
        return $array;
    }

    range = function ($min,$max) {
        return Math.floor(Math.random()*($max-$min+1)+$min);
    }

    inclusive_range($min, $max) {
        return Math.floor(Math.random() * ($max - $min + 1)) + $min;
    }

    randTXID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+txid;
    }

    formatDeci = function($num, $benchmark) {
        var decis = this.decis($benchmark);
        return $num.toFixed(decis);
    }

    getRadioValue=function($radelid) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].checked) {
                return radios[i].value; 
            }
        }
        return null;
    }

    /**
     * 
     * @param {*} $base 
     * @param {*} $per - in the form of a whole number.  4 = 4%. 
     */
    addPercent=function($base, $per) {
        var decis =  this.decis($base);
        var base = $base;
        console.log(typeof base);
        if ((base instanceof Number) == false) base = new Number($base);
        base = base + (base * $per/100);
        return base.toFixed(decis+1);
    }

    setRadioValue=function($radelid, $value) {
        const radios = document.getElementsByName($radelid);
        for (let i = 0; i < radios.length; i++) {
            if (radios[i].value != $value) {
                radios[i].checked=false;
                continue;
            }
            radios[i].checked=true; 
            return;
        }
        return;
    }

    decis = function($num) {
        var numStr = $num;
        if (typeof $num == "Number") numStr = $num.toString();
        else numStr = $num+"";
        if (numStr.includes('.')) {
            return numStr.split('.')[1].length;
        }
        return 0;
    }


    /**
     * ghosts start with z.
     * @param {*} $len 
     * @returns 
     */
    randNID = function($len=20) {
        var txid = "";
        for (var s=1; s <= $len; s++) {
            txid = txid+Math.floor(Math.random()*10);
        }
        return "z"+tx 
    }

    convertDMSLatToDec($degrees, $minutes, $seconds) {
        return $degrees + ($minutes / 60) + ($seconds / 3600);
    }
    convertDMSLongToDec($degrees, $minutes, $seconds, $direction) {
        var dd = $degrees + ($minutes / 60) + ($seconds / 3600);
        if ($direction === 'W') {
            return dd * -1;
        } 
        else if ($direction === 'E') {
            return dd;
        }
    }
    testGeo=function() {
        //giza great pyramid
        var lat = this.convertDMSLatToDD(40, 42, 31.0, "N"); // 40.70861°N
        var long = this.convertDMSLongToDD(74, 0, 32.0, "W"); // -74.00889°W
        console.log(lat+", "+long)
    }
    


    
    getRandHex = function($len) { return UTILS.randomHex($len);}
    randHex = function($len) { return UTILS.randomHex($len); }
    randomHex = function($len) {
        var maxlen = 8;
        var min = Math.pow(16,Math.min($len,maxlen)-1) 
        var max = Math.pow(16,Math.min($len,maxlen)) - 1;
        var n   = Math.floor( Math.random() * (max-min+1) ) + min;
        var r   = n.toString(16);
        while ( r.length < $len ) {
            r = r + this.randomHex( $len - maxlen );
        }
        return r;
    };

    randomSpherePoint=function($x0,$y0,$z0,$radius){
        var u = Math.random()+.01;
        var v = Math.random()+.01;
        var theta = 2 * Math.PI * u;
        var phi = Math.acos(2 * v - 1);
        var pt = new Object();
        pt.x = $x0 + ($radius * Math.sin(phi) * Math.cos(theta));
        pt.y = $y0 + ($radius * Math.sin(phi) * Math.sin(theta));
        pt.z = $z0 + ($radius * Math.cos(phi));
        return pt;
    }

    //nid or tnid array works here
    removeFromNidArray=function($arr, $nid) {
        var index = $arr.indexOf($nid)
        if (index!==-1)  $arr.splice(index,1);
    }

    removeFromTnidsArray=function($tnids_arr, $temp_nid) {
        var index = $tnids_arr.indexOf($temp_nid)
        if (index!==-1)  $tnids_arr.splice(index,1);
    }

    removeFromRelonsArray=function($relons_arr, $r) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.r==$r) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    removeRelonByTxid=function($relons_arr, $txid) {
        for (var r=0; r < $relons_arr.length; r++) {
            var rel = $relons_arr[r];
            if (rel.txid==$txid) {
                $relons_arr.splice(r,1);
                return $relons_arr;
            }
        }
        return $relons_arr;
    }

    
    getHighestSimilarity = function($NOBJs, $lookupval) {
        var highestSim = 0;
        var highestNOBJ = null;
        for (var c=0; c <$NOBJs.length; c++) {
            var nObj = $NOBJs[c];
            var sim = ANNE.UTILS.similarity(nObj.text, $lookupval);
            if (sim>highestSim) {
                highestSim = sim;
                highestNOBJ = nObj;
            }
        }
        return {nObj:highestNOBJ, sim:highestSim};
    }


    //levenshtein distance, altered
    similarity = function(s1, s2, $factor=1) {
        var longer = s1;
        var shorter = s2;
        if (s1==undefined || s1==null) {
            alert("woh");
            return 0;
        }
        if (s2==undefined || s2==null) {
            alert("woh");
            return 0;
        }
        if (s1.length < s2.length) {
            longer = s2;
            shorter = s1;
        }
        var longerLength = longer.length;
        
        if (longerLength == 0) return 1.0; //s1 and s2 have no length? lolz.

        var lenFact = shorter.length/longer.length;
        if ($factor>1) {
            var final = Math.floor($factor * (longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength));
            return (final/$factor)*lenFact;
        }
        else return ((longerLength - this.editDistance(longer, shorter)) / parseFloat(longerLength))*$factor;
    }


    sortByNid=function($a, $b) {
        if ($a==undefined || $a==null || $b.nid==undefined || $b.nid==null) return 0;
        if ($a.nid < $b.nid) return -1;
        else if ($a.nid > $b.nid) return 1;
        return 0;
    }


    //targetting the earliest here..
    sortByBeTime=function($a, $b) {
        if ($a==undefined || $a.be_time==null || $b.be_time==undefined || $b.be_time==null) return 0;
        if ($a.be_time < $b.be_time) return 1;
        else if ($a.be_time > $b.be_time) return -1;
        return 0;
    }


    //sort clump neurons by their WHEN role.
    sortClumpsByWhen=function($events) {
        if ($events==undefined || $events==null || $events.length<=1) return;
        $events.sort( UTILS.sortByWhenRole );
        return;
    }
    


    //these are role_when arrays.  We will only use the first
    //in 99.99% of cases there will only be one of a given role per sentence.
    sortByWhenRole=function($a, $b, $bFWD=true) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.role_when==undefined || $a.role_when==null) {
            if ($b.role_when!=undefined && $b.role_when!=null) return 1; 
            else return 0;
        }
        else if ($b.role_when==undefined && $b.role_when==null) return -1; 

        var a_when = MEMG.getNeuronByTempNid($a.role_when[0].t);
        var b_when = MEMG.getNeuronByTempNid($b.role_when[0].t);
        
        if ($bFWD) return UTILS.compareDates(b_when, a_when);
        else return UTILS.compareDates(a_when, b_when);
    }

    tnidsUnion = function($a, $b) {
        var union = [];
        for (var a=0; a <$a.length; a++) {
            var atnid = $a[a];
            for (var b=0; b <$b.length; b++) {
                if ($b[b]==atnid) {
                    if (union.indexOf(atnid)==-1) {
                        union.push(atnid);
                        break;
                    }
                }
            }   
        }
        return union;
    }


    //we should have an ANNE epoch day number for all dates.
    //that would be a better (more efficient) date compare.

    compareDates = function($a, $b) {
        $a.processAsDateNeuron();
        $b.processAsDateNeuron();
        if ($a.year < $b.year) return -1;
        else if ($a.year > $b.year) return 1;
        else {
            if ($a.month < $b.month) return -1;
            else if ($a.month > $b.month) return 1;
            else {
                if ($a.day < $b.day) return -1;
                else if ($a.day > $b.day) return 1;
                else return 0;
            }
        }
        return 0;
    }

    isUSSuffix=function($test) {
        var suffix = ["JR", "SR", "I", "II", "III", "IV", "V", "VI", "VII", "VIII", "IX", "X"];
        if ($test[0]==".") $test = $test.substring(1);
        $test = $test.toUpperCase();
        if (suffix.indexOf($test)!=-1) return true;
        return false;
    }

    removeOption = function($el, $val) {
        var el = document.getElementById($el);
        if (el==undefined || el==null) return;
        var options = document.querySelector("#"+$el).options;
        if (options==undefined || options==null) return;
        for (var o=0; o < options.length; o++) {
            if (options[o].value==$val) {
                options.remove(o);
                break;
            }
        }
    }


    roughSizeOfObject($object) {
        var objectList = [];
        var stack = [$object];
        var bytes = 0;
        while (stack.length) {
            var value = stack.pop();
            if (typeof value === 'boolean') bytes += 4;
            else if (typeof value === 'string') bytes += value.length * 2;
            else if (typeof value === 'number') bytes += 8;
            else if (typeof value === 'object' && objectList.indexOf( value ) === -1) {
                objectList.push( value );
                for( var i in value ) stack.push( value[ i ] );
            }
        }
        return bytes;
    }

    sleep=function($ms) {
        return new Promise(resolve => setTimeout(resolve, $ms));
    }

    //test this with DOM converstion...
    sortDD = function ($el) {
        var sel = docuement.getElementById($el);
        var selected = sel.value; // cache selected value, before reordering
        var opts_list = sel.querySelector('option');
        opts_list.sort(function (a, b) { return document.getElementById(a).text > document.getElementById(b).text ? 1 : -1; });
        sel.innerHTML="";
        sel.append(opts_list);  //append children back
        sel.value=selected; 
    }

    sortByFire=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire < $b.fire) return -1;
        if ($a.fire > $b.fire) return 1;
        return 0;
    }

    sortByFireThenIns=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null || $a.fire==undefined || $b.fire==undefined) return 0;
        if ($a.fire > $b.fire) return -1;
        if ($a.fire < $b.fire) return 1;
        if ($a.ins.length > $b.ins.length) return -1;
        if ($a.ins.length < $b.ins.length) return 1;
        return 0;
    }

    sortByDimension=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.type < $b.type) return -1;
        if ($a.type > $b.type) return 1;
        return 0;
    }

    sortByText=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text < $b.text) return -1;
        if ($a.text > $b.text) return 1;
        return 0;
    }

    sortTnidsByAlpha=function($tnids) {
        var nObjs = MEMG.getNeuronsFromTnids($tnids);
        nObjs.sort(this.sortByTextNoCase);
        return MEMG.getTnidList(nObjs);
    }

    sortByTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.text==undefined || $b.text==undefined || $a.text==null || $b.text==null) {
            console.log("alsdfkjalsdfjlafjadkfj;alsdfj");
            return 0;
        }
        if ($a.text.toUpperCase() < $b.text.toUpperCase()) return -1;
        if ($a.text.toUpperCase() > $b.text.toUpperCase()) return 1;
        return 0;
    }


    sortRelonsByFromNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.f);
        if (aNOBJ==null || aNOBJ.text==undefined) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.f);
        if (bNOBJ==null || bNOBJ.text==undefined) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortRelonsByToNObjText=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.to_type!=TO_TYPE_ANNE) return -1;
        if ($b.to_type!=TO_TYPE_ANNE) return -1;
        var aNOBJ = MEMG.getNeuronByTempNid($a.t);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.t);
        if (bNOBJ==null) return;
        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }


    sortGhostsLast=function($a, $b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.ghost==true) {
            if ($b.ghost==false) return 1;
            else return 0;
        }
        return 0;
    }

    sortByRelnTextNoCase=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        var aNOBJ = MEMG.getNeuronByTempNid($a.r);
        if (aNOBJ==null) return 0;
        var bNOBJ = MEMG.getNeuronByTempNid($b.r);
        if (bNOBJ==null) return;

        if (aNOBJ.text.toUpperCase() < bNOBJ.text.toUpperCase()) return -1;
        if (aNOBJ.text.toUpperCase() > bNOBJ.text.toUpperCase()) return 1;
        return 0;
    }

    sortByTime=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByTimeAsc=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return 1;
        if ($b.time && $a.emulated==false <=0) return -1;
        if ($a.time < $b.time) return -1;
        if ($a.time > $b.time) return 1;
        return 0;
    }
    sortByTimeASC=function($b,$a) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.time <=0 && $a.emulated==false && $b.time <=0 && $b.emulated==false) return 0;
        if ($a.time <=0 && $a.emulated==false) return -1;
        if ($b.time && $a.emulated==false <=0) return 1;
        if ($a.time < $b.time) return 1;
        if ($a.time > $b.time) return -1;
        return 0;
    }
    sortByFirstSeen=function($a,$b) {
        if ($a==undefined || $a==null || $b==undefined || $b==null) return 0;
        if ($a.firstseen > $b.firstseen) return 1;
        if ($a.firstseen < $b.firstseen) return -1;
        return 0;
    }




    syntaxHighlight=function($json) {
        json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
        return json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
            var cls = 'number';
            if (/^"/.test(match)) {
                if (/:$/.test(match)) {
                    cls = 'key';
                } else {
                    cls = 'string';
                }
            } else if (/true|false/.test(match)) {
                cls = 'boolean';
            } else if (/null/.test(match)) {
                cls = 'null';
            }
            return '<span class="' + cls + '">' + match + '</span>';
        });
    }

        

    /**
     * validate an ipv4 addressed
     * @param {*} $nObj 
     * @returns {Boolean} 
     */
     __ipv4Validation=function($nObj) {
        console.log("(__ipv4Validation) to do!");
        return true;
    }


    /**
     * validate an ipv6 address
     * @param {*} $nObj 
     * @returns {Boolean}
     */
    __ipv6Validation=function($nObj) {
        console.log("(__ipv6Validation) to do!");
        return true;
    }


    /**
     * Validate an email string conforms to known email strings.
     * 
     * @param {anne_neuron} - The neuron with .value OR string we want to test
     * @author {Mr.Scatman} 
     *
     */
    __emailValidation = function($nObj) {
        var mailformat = /(?!.*\.{2})^([a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+(\.[a-z\d!#$%&'*+\-\/=?^_`{|}~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]+)*|"((([ \t]*\r\n)?[ \t]+)?([\x01-\x08\x0b\x0c\x0e-\x1f\x7f\x21\x23-\x5b\x5d-\x7e\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|\\[\x01-\x09\x0b\x0c\x0d-\x7f\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]))*(([ \t]*\r\n)?[ \t]+)?")@(([a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\d\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.)+([a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]|[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF][a-z\d\-._~\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF]*[a-z\u00A0-\uD7FF\uF900-\uFDCF\uFDF0-\uFFEF])\.?$/i
        
        //just a string (not an nObj);
        if (typeof $nObj === "string") {
            var test = mailformat.test($nObj.toLowerCase());
            //console.log(test);
            if(test==undefined || test==null) return "invalid format";
            else if (test==true) return true;
            else return false;
        }

        if ($nObj.value==undefined || $nObj.value==null || $nObj.value=="") {
            $nObj.value = $nObj.text.trim();    //default nObj .value is the .text if none was provided by a .set_value function.
        }
        if(mailformat.test($nObj.value.toLowerCase())==true) {
            $nObj.valid = true;
            return true;
        }
        else {
            $nObj.value = "";
            $nObj.valid = false;
            return false;
        }
    }

    //from devshed
    __urlValidation=function($str) {
        var pattern = new RegExp('^(https?:\\/\\/)?'+ // protocol
        '((([a-z\\d]([a-z\\d-]*[a-z\\d])*)\\.)+[a-z]{2,}|'+ // domain name
        '((\\d{1,3}\\.){3}\\d{1,3}))'+ // OR ip (v4) address
        '(\\:\\d+)?(\\/[-a-z\\d%_.~+]*)*'+ // port and path
        '(\\?[;&a-z\\d%_.~+=-]*)?'+ // query string
        '(\\#[-a-z\\d_]*)?$','i'); // fragment locator
      return !!pattern.test($str);
    }

    
    /**
     * 
     * @param {*} $aMatches 
     * @param {*} $aMatches1 
     * @param {*} $prop 
     * @returns 
     */
    concatUnique=function($aMatches, $aMatches1, $prop="nid") {
        if($aMatches.length==0) return $aMatches1;
        if($aMatches1.length==0) return $aMatches;
        var arr = new arraymap();
        arr.setMapKey($prop);
        arr.addArray($aMatches);
        arr.addArray($aMatches1);
        return arr.arr;
    }


    uniqueConcat=function($a, $b, $prop="nid") {
        if ($a.length==0) return $b.concat();
        var a2 = $a.concat();
        for (var b=0; b < $b.length; b++) { 
            if ($a.indexOf($b[b])==-1) a2.push($b[b]);
        }
        return a2;
    }

    uniqueConcatRels = function($a, $b) {
        var a_rels_txids = [];
        var rels = [];
        for (var a=0; a < $a.length; a++) {
            a_rels_txids.push($a[a].txid);
            rels.push($a[a]);
        }
        for (var b=0; b < $b.length; b++) {
            if (a_rels_txids.indexOf($b[b].txid)!=-1) continue;
            a_rels_txids.push($b[b].txid);
            rels.push($b[b]);
        }
        return rels;
    }

    /**
     * 
     * @param {*} $str 
     * @returns 
     */
    poorMansAbbreviate=function($str) {
        $str = $str.trim();
        const words = $str.split(' ');
        
        if (words.length === 1) {
            return $str; 
        } 
        else {
            var abbr = '';    
            words.forEach((word, index) => {
                if (index === 0 || index === words.length - 1) {
                    abbr += word.charAt(0).toUpperCase() + '.';
                } 
                else {
                    abbr += word.charAt(0).toUpperCase();
                }
            });
            return abbr;
        }
    }

    
    /**
     * Check for a relon in a relon array.
     * If it is unique, add it to the array and return true.
     * if it is not unique, return false.
     * @param {array} $rels 
     * @param {Object} $rel 
     * @returns 
     */
    pushIfUniqueRel = function($rels, $rel) {
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            if(rel.txid==$rel.txid) {
                return false;
            }
        }
        $rels.push($rel);
        return true;
    }

    seededRandom=function() {
        var x = Math.sin(Math.random()*1000) * 10000;
        return x - Math.floor(x);
    }

    getScatModSeedArrItem=function($seed, $arr) {
        var per = UTILS.getScatModSeedPer($seed);
        var max_usable = Math.round($arr.length * per);
        return $arr [ Math.floor(Math.random()*max_usable)];
    }
    

    getScatModSeedPer=function($seed) {
        $seed = Math.round($seed);
        var per = .5;
        if ($seed % 3 !=0) per +.1;
        if ($seed % 2 !=0) per +.15;
        if ($seed % 7 ==0) per +.25;
        return per;
    }
    
    isWholeNumber=function($number) {
        return $number % 1 === 0;
    }

    HEXtoRGB = function($HEX) {
        var isNumb = Number.isInteger($HEX);
        if (isNumb) var bigint = $HEX;
        else var bigint = parseInt($HEX.slice(1), 16);
        const r = (bigint >> 16) & 255;
        const g = (bigint >> 8) & 255;
        const b = bigint & 255;
        return { r, g, b };
    }

    RGBtoHEX = function($rgb) {
        const { r, g, b } = $rgb;
        const hex = `0x${((1 << 24) + (r << 16) + (g << 8) + b).toString(16).slice(1)}`;
        return hex;
    };

    shiftRGB = function($RGB, $incr) {
        const shiftedColor = {
          r: $RGB.r + $incr,
          g: $RGB.g + $incr,
          b: $RGB.b + $incr,
        };
        shiftedColor.r = Math.max(0, Math.min(255, shiftedColor.r));
        shiftedColor.g = Math.max(0, Math.min(255, shiftedColor.g));
        shiftedColor.b = Math.max(0, Math.min(255, shiftedColor.b));
        return shiftedColor;
    };
}


const UTILS = new anne_utils ();
UTILS.set_init(UTILS);

/************** (from file ./js/base/utils.js)**/ 

//extras

class annode_auth  {
    constructor($resultObj){
        if ($resultObj==undefined || $resultObj==null) return;
        this.annode_time = $resultObj.annode_time;
        this.annode_pubKey = $resultObj.annode_pubKey;
        this.annode_nid = $resultObj.annode_nid;
        this.annode_drl = $resultObj.annode_drl;
        this.next_proof_id = $resultObj.next_proof_id;
        if(this.annode_drl==undefined) this.annode_drl=1000;

        //we can show these in a gui of some-kind.
        this.network = $resultObj.network;
        if (this.network==undefined) this.network = false;

		this.ANNDLE_active = $resultObj.ANNDLE_active;
        if (this.ANNDLE_active==undefined) this.ANNDLE_active = false;
        
		this.ASDF_active = $resultObj.ASDF_active;
		if (this.ASDF_active==undefined) this.ASDF_active = false;
        
        this.BOUNTY_active = $resultObj.BOUNTY_active;
        if (this.BOUNTY_active==undefined) this.BOUNTY_active = false;
        
        this.PDNS_active = $resultObj.PDNS_active;
        if (this.PDNS_active==undefined) this.PDNS_active = false;
        
        this._nonce = "";
        this._encrypted_sk = "";

        //used to prove annode identity.
        this._nonce2 = "";
        this._encrypted_proof = "";
    }


    //These are hex strings from the ANNODE we are handshook with.
    //NOTE: We will neeed to change the LS user storage object to allow also for multiple
    //annodes for the same origin/app/user if we want to store more than the "last"
    //annode handshook with for the origin/app/user.
    //At present this is location in the annode code folder html/ui/annewallet/js/localstorage.js.

    //We already decrypted the shared key from these values and 
    //then encrypted them with a simpler encryption for "us" in our LS User object.
    //This is a backup for testing.
    set encrypted_sk ($esk) { this._encrypted_sk=$esk};
    get encrypted_sk(){ return this.encrypted_sk;}
    set encrypted_proof ($ep) { this._encrypted_proof=$ep};
    get encrypted_proof(){ return this._encrypted_proof;}
    set nonce ($nonce) { this._nonce=$nonce};
    get nonce () { return this._nonce; }
    set nonce2 ($nonce2) { this._nonce2=$nonce2};
    get nonce2 () { return this._nonce2; }
    get time() { return this.annode_time; }
    get pubKey() { return this.annode_pubKey; }
    get id() { return this.annode_nid; }
    drl=function() { 
        return this.annode_drl; 
    }
}

/************** (from file ./js/network/ANNODE_AUTH.js)**/ 

class ANNODE_API extends ANNE_NET {
    constructor($base_uri){	//expects url:port like http://localhost:9116
		super($base_uri);	//converthing ANNODE_API to use ANNE_NET (standalone networking class)
		this.client = null;
		this.db = null;
		this.anon_user_id = null;	//for use when AUTHing anon externally.
		this._AAUTH = null;
		this._bNODECONNECTED=false;
		this._consecutive_network_errors = 0;
		this.privateFolderAbsolute = "";
		this._annodeDetailsObj = null;
		this.OPT_CHECKER = [];

	}

	opwipe = function() {
		this.OPT_CHECKER=[];
	}
    opcheckerPass = function($funct, $secs=10) {
		if (this.OPT_CHECKER[$funct]!=undefined) {
			var lasttime = this.OPT_CHECKER[$funct];
			let currentTime = Math.floor(Date.now() / 1000);
			var diff = currentTime - lasttime;
			if (diff < $secs) {
				console.log("skip, "+diff+" "+$funct+" is too soon. Find out who is calling this.");
				return false; 
			}
		}
		return true;
	}

	//tracking the seconds marker of last uses of functions.
	//avoiding hitting them too frequently for no good reason.
	opcheckerTrackIt = function($funct) {
		this.OPT_CHECKER[$funct] = Math.floor(Date.now() / 1000);
	}

	getAnnodeProp=function($prop) {
		var val = this._annodeDetailsObj[$prop];
		if (val==undefined) return null;
		return val;
	}

	
	/**
	 * 
	 * @param {*} $start_index 
	 * @param {*} $end_index 
	 * @param {*} $bDOTRANSACTIONS 
	 * @returns 
	 */
	getLastBlocks=async($start_index, $end_index, $bDOTRANSACTIONS)=> {
		var rez = await this.query("getLastBlocks", {
			from_height:$start_index,
			to_height:$end_index,
			includeTransactions:$bDOTRANSACTIONS
		});

		return rez;
	}

	getBlockByHeight = async($height)=>{
		var rez = await this.query("getBlockByHeightOrID", {
			height:$height
		});
		return rez;
	}

	getBlockById = async($blockId)=>{
		var rez = await this.query("getBlockByHeightOrID", {
			blockId:$blockId
		});
		return rez;
	}



	/**
	 * 
	 * @param {*} $prop 
	 * @returns 
	 */
	getAnnodeDetail=async($prop="circulating_supply", $bREFRESHSPECIAL=true)=> {
		if (this._annodeDetailsObj==null) {
			await this.annodeDetails();
			if (this._annodeDetailsObj==null) return;
		}
		var val = this._annodeDetailsObj[$prop];
		if (val==undefined) return null;

		if ($prop=="annode_mode" && val=="SYNC" && $bREFRESHSPECIAL==true) {
			this._annodeDetailsObj = null;
			return await this.getAnnodeDetail($prop, false);
		}
		return val;
	}


	/**
	 * Used when getting new annode details.
	 * We potentially have new _active vars.
	 * @param {*} $obj 
	 */
	processAuxAAUTHData=async($obj)=>{
		var bCHANGE = false;
		this._annodeDetailsObj = $obj;
		//console.log(this._annodeDetailsObj);

		if ($obj==null) return;
		if ($obj.height!=undefined) {
			if (ANNE.HISTORY!=null) ANNE.HISTORY.updateGLKB_SOLID($obj.height);
		}
		
        if (ANNE.USER==null || ANNE.USER._userLS==null || ANNE.USER._userLS.aauth==null) return;
		
		if (USER._userLS.aauth.ANNDLE_active!=$obj.anndle_active) {
            ANNE.USER._userLS.aauth.ANNDLE_active = $obj.anndle_active;
			ANNE.API._AAUTH.ANNDLE_active = $obj.anndle_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.ANNDLE_active = $obj.anndle_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.ASDF_active!=$obj.asdf_active) {
            ANNE.USER._userLS.aauth.ASDF_active = $obj.asdf_active;
			ANNE.API._AAUTH.ASDF_active = $obj.asdf_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.ASDF_active = $obj.asdf_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.PDNS_active!=$obj.pdn_active) {
            ANNE.USER._userLS.aauth.PDNS_active = $obj.pdn_active;
			ANNE.API._AAUTH.PDNS_active = $obj.pdn_active;
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.PDNS_active = $obj.pdn_active;
            bCHANGE=true;
        }
        if (ANNE.API._AAUTH.BOUNTY_active!=$obj.bounty_active) {
            ANNE.USER._userLS.aauth.BOUNTY_active = $obj.bounty_active;
			ANNE.API._AAUTH.BOUNTY_active = $obj.bounty_active; 
			if (ANNE.SETTINGS!=null) ANNE.SETTINGS.BOUNTY_active = $obj.bounty_active;
            bCHANGE=true;
        }
        if (bCHANGE==true) {	//will save updates to _AAUTH.
			if (USER.seedObj!=null) {
				ANNE.API._AAUTH = ANNE.USER._userLS.aauth;
				await ANNE.USER.__saveANNE_JO(ANNE.USER._userLS);
			}
		}
	}


	setAAUTHFromLS = function($resp) {
		this._AAUTH = new annode_auth($resp);
		this._bNODECONNECTED=true;
		USER.aauth = this._AAUTH;
		
	}

	getNextProofId = function() {
		if (this._AAUTH==null) return -1;
		return this._AAUTH.next_proof_id;
	}

	updateAndSaveNextProofId = async($id)=> {
		if (this._AAUTH==null) return -1;
		this._AAUTH.next_proof_id = $id;
		USER._userLS.aauth = this._AAUTH;
		console.log("%c (updateAndSaveNextProofId) id:"+$id, TALK.bgbigred);
		return await ANNE.USER.updateNextProofId($id);
	}

	getRelonLimit = function() {
		if (this._AAUTH == null) return 0;
		return this._AAUTH.drl();
	}


	/**
	 * 
	 * @param {*} $functName 
	 * @param {*} $params 
	 * @returns 
	 */
	apiSimpleQueryWithParams=async($functName, $params, $bDOLOADER=true)=> {
		var response = await this.query($functName, $params);
		return response;
	}
	apiSimpleQueryWithParamsNS=async($functName, $params)=> {
		var response = await this.query($functName, $params);
		return response;
	}
	apiSimpleQuery=async($functName)=> {
		var response = await this.query($functName, null);
		return response;
	}


	//***************************************************** */
	//////  TRANACTION CREATION / SIGNING / BROADCAST ////////
	//***************************************************** */



	//***************************************************** */
	////////////////  TRANSACT WITH AN ANNDOE ////////////////
	//***************************************************** */
	

	/**
	 * Add boosties to de-centivize annexes against your sponsored neuron!
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	addBoosties = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=>{
		 console.log("(adddBoosties) sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}
		else if ($relon.sys_ref_txid!=undefined) {
			alert("THIS SHOULDNT BE NEEDED HERE");
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}
		
		var json = await this.makeOnNodeSignOnJSThenBroadcast("addBoosties",params);
		return json;
	}
	

	/**
	 * Send a relon where the FROM is the sender (you).
	 * 
	 * If the .sys_ref_txid is set in the sending relon the API will lookup 
	 * the full hash of the transaction.  When broadcast as part of the TX it serves
	 * as a dependency to mine first.
	 * 
	 * In our ANNE GRAPH we also use this variable after-the-fact to link TXs to other
	 * TXs they are in reference (or response) to.
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	sendSelfRelon = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;

		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};

		if ($relon.r==ANNE.MEM.annex.temp_nid) {
			params.chase_from=true;
		}

		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		//with annexes this is already stored and set as part of the annex chain.
		//just pass it through here.
		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("sendSelfRelon",params);
		console.log(json);
		return json;
	}

	
	/**
	 * Call for a distribution vote for an annorg (technically this will work on ANY keyless neuron)
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anorgVoteCall = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;

		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
			else if ($relon.r==MEM.distro_vote.nid) {
				alert("error, couldn't get ref sys txid's hash for DISTRO VOTE (must tie to FREEZE)");
				return;
			}
		}
		
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anorgVoteCall",params);
		return json;
	}
	

	/**
	 * Claim a bounty.
	 * This will invoke an ACHAT session on the annode it is broadcast through.
	 * The bounty will specify the annode that it wants to ACHAT on.
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	bountyClaim = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeBountyClaim",params);
		return json;
	}

	/**
	 * Accept some risk being offered by an ASDF.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	asdfAcceptRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFacceptRisk",params);
		return json;
	}


	/**
	 * Add more risk to an ASDF you own.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	asdfAddRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFaddRisk",params);
		return json;
	}


	/**
	 * A way to stop accepting risk takers for an ASDF.
	 * AKin to no-more-bets or the real-world event has started.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	asdfNomoreRisk = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("ASDFnoMoreRisk",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anndleSellback = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleSellback",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	anndleRenew = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleRenew",params);
		return json;
	}
	

	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	annexNeuron = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("annexNeuron",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleBuysideOffer",params);
		return json;
	}



	
	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	acceptAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_froms:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleOfferAccept",params);
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	revokeAnndleOffer = async($relon, $annies=400, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("anndleOfferRevoke",params);
		return json;
	}

	/**
	 * Make an ANNE admin for a specific class you have access to grant rights to.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeAnneAdmin = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnneAdmin",params);
		return json;
	}


	/**
	 * Add to the data of an ANNDLE you have sponsored.
	 * Only class Params are relns that will be accepted by the ANNODE.
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeAnndleEdit = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnndleEdit",params);
		return json;
	}
	

	/**
	 * Make a subanndle under an anndle that you control (are the sponsor of).
	 * @param {Object} $relon 
	 * @param {int} $annies 
	 * @param {int} $minerFee 
	 * @returns 
	 */
	makeSubAnndle = async($relon,  $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeSubAnndle",params);
		return json;
	}


	/**
	 * Accept some amount of an ANNE OTC offer.
	 * This is specific to the SIGNA integration since we can 
	 * broadcast the SIGNA transaction from the ANNODE to the SIGNA node.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	acceptAOTCOffer = async($relon,  $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		if (ANNE.UTILS.isset($relon.signedTransactionBytes)==false) {
			alert("we must have the signed transaction bytes from the AOE CHAIN TX!")
			return;
		}
		else params.signedAOETransactionBytes = $relon.signedTransactionBytes;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("AOTCacceptOfferSIGNA",params);
		return json;
	}

	

	/**
	 * Response to a collab request.
	 * 
	 * NOTE: This must be a TO_TYPE_TXID with the collab request as the to
	 * 
	 * It also must reference the full hash of the collab request (even if only in blockchain and not in hypergraph).
	 * This ensures payment for admin is available prior to the admin's response being mined.
     * 
	 * The only time amountNQT is relevant is when approving a merge that has a required annex to settle it.
     * No other collab response patterns require annies to respond.  (in fact, admins gain annies for responding)
	 */
	 makeCollabResp = async($relon, $annies=0, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCollabResp", params);

		return json;
	}

	/**
	 * Respond to the invitation to be an Anjucator.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeAnjInviteResp = async($relon, $annies=0, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if (MEM.__isMalformedRelon($relon)==true) return false;
		
		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$minerFee,
			deadline:1440
		};
		
		//the api needs the referencedTransactionFullHash
		if ($relon.sys_ref_txid!=undefined) {
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeAnjInviteResp", params);
		return json;
	}


	/**
	 * Make a keyless class
	 * 
	 * Keyless class makers are the owner of the neuron class.
	 * 
	 * The RELN should be the standard CLASS or INST to denote the nType (neuron type).
	 * The TO should be the parent class.  This is a standard TYPE_BE parent relon,
	 * We send it through as a TYPE_SYS_MSG but on the annode a TYPE BE will be created.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	 makeKeylessNeuron = async($relon,  $annies=900, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeKeylessNeuron",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		//console.log(json);
		return json;
	}


	/**
	 * Make a new Private Data Neuron.
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makePDN = async($relon,  $annies=2000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makePDN",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}


	/**
	 * Make a custodial Neuron that this particular ANNODE will manage.
	 * 
	 * DO EXTRAS FLAG HERE IS CRUCIAL!!!!!
	 * WE MUST RETURN THE CUSTODIAL NEW NID AND KEY AFTER SIGNING AND BROADCAST!!!
	 * @param {*} $relon 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	makeCustodialNeuron = async($relon,  $annies=800, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCustodialNeuron",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			do_extras:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}


	/**
	 * We default to 1 coin in the custodial neuron.
	 * THere are NO fees so these remain in the faucet as an initial bump.
	 * @param {*} $relon 
	 * @param {*} $recipientPublicKey 
	 * @param {*} $annies 
	 * @param {*} $minerFee 
	 * @returns 
	 */
	announceCustodial = async($relon, $recipientPublicKey, $referencedTransactionFullHash, $recipient, $annies=100, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if ($recipientPublicKey=="") {
			console.log("MUST HAVE A RECIPIENT PUBLIC KEY!");
			return false;
		}
		if ($recipient=="") {
			console.log("WE MUST HAVE RECIPIENT HERE!");
			return false;
		}
		if (MEM.__isMalformedRelon($relon)==true) return false;


		var params = {
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440,
			recipientPublicKey:$recipientPublicKey,
			recipient:$recipient
		}

		if ($referencedTransactionFullHash!="") params.referencedTransactionFullHash = $referencedTransactionFullHash;
		else if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}
		else if ($relon.sys_ref_txid!=undefined) {
			alert("THIS SHOULDNT BE NEEDED HERE");
			var json = await this.getBlockchainTransaction($relon.sys_ref_txid);
			console.log(json);
			if (json!=null && json.fullHash!=undefined) {
				params.referencedTransactionFullHash = json.fullHash;
			}
			else if ($relon.r==MEM.action_delegate.temp_nid) {
				alert("error, couldn't get ref sys txid's hash for delegation request");
				return;
			}
		}

		var json = await this.makeOnNodeSignOnJSThenBroadcast("announceCustodialNeuron", params);

		console.log(json);
		return json;
	}

	

	/**
	 * Make (sponsor) a new global neuron.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	 makeNewGlobal = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeGlobal",{
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}

	/**
	 * Make (sponsor) a new global neuron. (location)
	 * 
	 * @param {Object}
	 * @param {string}
	 * @param {string}
	 * @param {integer}
	 * @author Mr.Scatman
	 */
	makeGlobalLocation = async($relon, $annies=1000, $minerFee=ANNE.ANNE_TO_ANNIES)=> {
		if (MEM.__isMalformedRelon($relon)==true) return false;
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeGlobalLocation",{
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			chase_from:true,
			feeNQT:$minerFee,
			amountNQT:$annies,
			deadline:1440
		});
		console.log(json);
		return json;
	}	


	/**
	 * 
	 * @param {Object} $relon 
	 * @param {int} $amount   (2,147,483,647 billion annies)  going to likely have to do a sys overhaul of all amounts and use bigint.
	 * @param {int} $fee 
	 * @returns 
	 */
	 sendChildRelon = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
    	if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			console.log($relon);
			return {status:"fail", reason:"missing relon info"};
		}
		//console.log("(sendChildRelon) sending");
		ANNE.TALK.printRelon($relon);

		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var params = {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			flathexint:fhi,
			deadline:1440
		};

		//we want the relons back on a freeze.
		if ($relon.nid==MEM.CMD_FREEZE.nid) {
			params.chase_from=true;
		}

		if ($relon.referencedTransactionFullHash!=undefined) {
			params.referencedTransactionFullHash = $relon.referencedTransactionFullHash;
		}

		else {
			//if future devs use this API this will ensure any NEW uses of custodials will
			//not throw any errors on the annodes.
			var fromNOBJ = MEMG.getNeuronByNid($relon.from_nid);
			if (fromNOBJ!=null && ANNE.UTILS.isset(fromNOBJ.origin)==true && fromNOBJ.origin==ORIGIN_CUSTODIAL) {
				//we need to make sure all child relon TXs of custodials refer back to their
				//TXTYPE_ANNOUNCE   reln:MAKE_CUSTODIAL TO_TYPE_TXID transaction.

				//we can find our make custodial request as an in of the from (the cusodial)
				//pubkey was broadcast on MAKE_CUSTODIAL tx.. we can get that txid
				//and lookup the pubky if we don't have it in local mem.

				var rel = null;
				var the_txid_we_want_the_fullhash_of = "";
				var fullHash = "";

				for (var o=0; o < fromNOBJ.outs.length; o++) {
					if (fromNOBJ.outs[o].r==MEM.class.temp_nid || fromNOBJ.outs[o].r==MEM.inst.temp_nid) {
						the_txid_we_want_the_fullhash_of = fromNOBJ.outs[o].sys_ref_txid;
						rel  = MEMG.lookupByTXID(fromNOBJ.outs[o].sys_ref_txid);
						if (rel!=null && rel.fullHash!=undefined) fullHash = rel.fullHash;
						break;
					}
				}

				if (fullHash.length==0) {

					//we need to get the txid.. otherwise we can't continue.
					if (the_txid_we_want_the_fullhash_of=="") {
						await CHASE.lookupMinNid(fromNOBJ.nid);
						for (var o=0; o < fromNOBJ.outs.length; o++) {
							if (fromNOBJ.outs[o].r==MEM.class.temp_nid || fromNOBJ.outs[o].r==MEM.inst.temp_nid) {
								the_txid_we_want_the_fullhash_of = fromNOBJ.outs[o].sys_ref_txid;
								break;
							}
						}
					}

					//if we don't have the txid, we can't continue.
					if (the_txid_we_want_the_fullhash_of=="") {
						console.log("problem with finding custoidal reference.");
						return;
					}


					//go get the hash.  We must reference it.
					if (the_txid_we_want_the_fullhash_of.length>0) {	//this is a BE relon.. just do min chase
						var json = await this.getBlockchainTransaction(the_txid_we_want_the_fullhash_of);
						console.log(json);
						if (json!=null && json.fullHash!=undefined) {
							params.referencedTransactionFullHash = json.fullHash;
						}
						else if ($relon.r==MEM.action_delegate.temp_nid) {
							alert("error, couldn't get ref sys txid's hash for delegation request");
							return;
						}
					}
				}

				else params.referencedTransactionFullHash = fullHash;
			}
		}


		var json = await this.makeOnNodeSignOnJSThenBroadcast("sendChildRelon", params);
		return json;
	}

	
	
	/**
	 * Send a viewPDN relon transaction.
	 * This is the same pattern as child relon and it 
	 * will use the child relon TX in the ANNODE.
	 * 
	 * @param {Object}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	viewPDN = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
		console.log("sending viewPDN relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		console.log("(viewPDN) sending");
		ANNE.TALK.printRelon($relon);

		var json = await this.makeOnNodeSignOnJSThenBroadcast("viewPDN", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440
		});
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $amount 
	 * @param {*} $fee 
	 * @param {*} $bEXPECTED_DELEGATION 
	 * @author Mr.Scatman
	 */
	makeCollabReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES, $bEXPECTED_DELEGATION=false)=> {
    	console.log("sending child relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to_s);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);

		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeCollabReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440,
			flathexint:fhi,
			expect_delegation:$bEXPECTED_DELEGATION
		});
		return json;
	}


	/**
	 * 
	 * @param {*} $relon 
	 * @param {*} $amount 
	 * @param {*} $fee 
	 * @param {*} $bEXPECTED_DELEGATION 
	 * @author Mr.Scatman
	 */
	makeHanaiCollabReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES, $bEXPECTED_DELEGATION=false)=> {
    	console.log("sending child relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);
		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeHanaiCollabReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true, 
			feeNQT:$fee,
			amountNQT:$amount,
			deadline:1440
		});
		return json;
	}	


	/**
	 */
	makeHanaiOutReq = async($relon, $amount=400, $fee=ANNE.ANNE_TO_ANNIES)=> {
    	console.log("sending hanai out relon from:"+$relon.from_nid+" type:"+$relon.type+" reln:"+$relon.reln, "to_type:"+$relon.to_type+" to:"+$relon.to);
		if ($relon==undefined || $relon.from_nid==undefined || $relon.type==undefined || $relon.reln==undefined || $relon.to_type==undefined || $relon.to_s==undefined ) {
			console.warn("FIX THIS!!! this should NEVER happen.");
			return {status:"fail", reason:"missing relon info"};
		}
		ANNE.TALK.printRelon($relon);
		
		var fhi=0;
		if ($relon.flathexint!=undefined) fhi = $relon.flathexint;

		var json = await this.makeOnNodeSignOnJSThenBroadcast("makeHanaiOutReq", {
			from:$relon.from_nid,
			type:$relon.type,
			reln:$relon.reln,
			to_type:$relon.to_type,
			to:$relon.to_s,
			makeSignSend:true,
			feeNQT:$fee,
			amountNQT:$amount,
			flathexint:fhi,
			deadline:1440
		});
		return json;
	}	



	
	/**
	 * Send a standard payment TX to another Account.
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	 sendAnne = async($recipient, $recipientPublicKey, $amount, $fee=ANNE.ANNE_TO_ANNIES, $deadline=1440, $msg="", $encrypted=false, $edata=null, $ndata=null)=> {
		console.log("(sendAnne) recipient:"+$recipient+" amount:"+$amount+" fee:"+$fee);
		var paramsObj = {
			recipient:$recipient, 
			amountNQT:$amount, 
			feeNQT:$fee,
			deadline:$deadline,
			makeSignSend:true,
		} 

		//onboarding a pubKey with send..
		//this should be auto with the sendAnne functionality and the PublicKeyAnnouncement appendix
		//also required when sending encrypted message to recipient (even when their key is known)
		if (ANNE.UTILS.isset($recipientPublicKey)==true) paramsObj.recipientPublicKey=$recipientPublicKey;
		else if ($encrypted==true) return {status:"bad", reason:"must have recipientPublicKey to encrypt"};
		 
		if ($msg!="") {
			console.log("%c (sendAnne) with plaintext msg", TALK.bgpink);
			paramsObj.message = $msg;
			paramsObj.messageIsText = true;
		}
		else {
			if ($encrypted==true) {
				if (ANNE.UTILS.isset($edata)==false || ANNE.UTILS.isset($ndata)==false) {
					return {status:"bad", reason:"user wants to send encrypted data but the provided encrypted data is missing."};
				}
				console.log("%c (sendAnne) with encrypted msg", TALK.bgpink);
				paramsObj.encryptedMessageData = $edata;
		        paramsObj.encryptedMessageNonce = $ndata;
			}
		}
		var rez = await this.makeOnNodeSignOnJSThenBroadcast("sendAnne", paramsObj); 
		if (rez.broadcasted==undefined) rez.status = "bad";
		else rez.status = "good";
		return rez;
	}


	//***************************************************** */
	///////////////////  PING & HANDSHAKE ////////////////////
	//***************************************************** */

	/**
	 * Initialize the ANNODE auth handshake.
	 * 
	 * If already "logged" in, just ping node w/ ANNODEAuthHandshake and set the js _AAUTH object.
	 * This will be required to get the nodeId for certain types of transactions that payout the annode that processes the TX.
	*/
	ANNODEPING=async()=> {

		var uri = ANNE._annodesLS.last_uri;
		if (ANNE._annodesLS.last_uri[ ANNE._annodesLS.last_uri.length-1]=="/") {
			uri = ANNE._annodesLS.last_uri.substring(0, ANNE._annodesLS.last_uri.length-1);
		}

		var response = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthHandshake", {
			annode_claimed_uri:uri} 
			);//"https://www.anne.network:9116"
		if (response.status=="good") {
			var AAUTHOBJ = new annode_auth(response.result);
			this._AAUTH = AAUTHOBJ;
		}
		else {
			console.warn("we have a problem houston");
			return;
		}
	}


	/**
	 * Handshake with an annode.
	 * 
	 * Set up the active aauth object.
	 * This does not persist session to session.
	 * If another handshake is made w/ the same or different node
	 * the API uses it.
	 * 
	 * @param {*} $aO 
	 * @param {*} $uniqueReqRandId 
	 * @param {*} $height 
	 * @returns 
	 */
	ANNODEHANDSHAKE=async($aO, $uniqueReqRandId, $height=0) => {
		//handshake with annode
		var uri = ANNE._annodesLS.last_uri;
		if (ANNE._annodesLS.last_uri[ ANNE._annodesLS.last_uri.length-1]=="/") {
			uri = ANNE._annodesLS.last_uri.substring(0, ANNE._annodesLS.last_uri.length-1);
		}
		var response = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthHandshake", {
			annode_claimed_uri:uri}); //"https://www.anne.network:9116"


		if (response.status=="good") {
			var AAUTHOBJ = new annode_auth(response.result);
			this._AAUTH = AAUTHOBJ;
			if (USER._userNObj!=null) USER._userNObj.aauth = this._AAUTH;
		}
		else {
			//roll back to last successful node.
			console.warn("we have a problem houston");
			return;
		}

		//returned from the NODE.
		var annode_time = this._AAUTH.time;
		var annode_pubKey = this._AAUTH.pubKey;

		//We will encrypt a string of the user's public nid and the annode time
		//The handling node will verify this time is within a reasonable ping range 
		//of milliseconds.
		var encryptThis = USER.public_nid+"::"+annode_time;

        var eObj = await ANNE.USER.encryptMsgForPUBKEY(encryptThis, annode_pubKey, $aO.priv_key, true);
        //console.log(eObj);

        
		//we send info back to the server to close the ANNODEAUTH
        var result = await ANNE.API.apiSimpleQueryWithParams("ANNODEAuthFinish", {
												user_nid:USER.public_nid, 
												user_pubKey:$aO.pub_key, 
												annode_time:annode_time, 
												user_sig_data:eObj.data_string,
												user_sig_nonce:eObj.nonce_string,
												height:$height,
												uid:$uniqueReqRandId}
												);

		//console.log(result);
		result.original_uid = $uniqueReqRandId;
		if (result.lnots!=undefined) {
			if (result.lnots.length) {
				console.log("%c lnots present", TALK.bgbigred);
				console.log(result.lnots);
			}
		}
		return result;
	}
	
	//***************************************************** */
	//////////////////  QUERY THE l1 ANNODE //////////////////
	//***************************************************** */
	


	/**
	 * Match condition and do join to chase values to prevent further lookups
	 * 
	 * normal space relons contain awareness relons so they must be filtered
	 * for type to avoid clutter.				
	 * @param {string}
	 * @param {array}
	 * @param {array} 
	 * @param {boolean}
	 * @param {array}  
	 * @param {array}   
	 * 
	 * @author Mr.Scatman								
	 */
	chaseInsGWV = async($nid, $adopt, $bMUSTPULL, $types=[1,3], $chasetypes=[0,1,2])=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.chaseInsGWV ", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {	//means it is a one-time query per session.
			console.log("super optimizer");
			return [];
		}

		var response = await this.apiSimpleQueryWithParams("chaseInsGWV", 
				{
					nid:$nid, 
					adopt:$adopt,
					types:$types,
					chase_types:$chasetypes,
					height:lastBlock

				});
		console.log(response);
		phObj.bDONETHISSESSION=true;
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.chaseInsGWV");
		}
		return aRelons;
	}
	
	
	/**
	 * Match condition and do join to chase values to prevent further lookups
	 * 
	 * normal space relons contain awareness relons so they must be filtere
	 * for type to avoid clutter.				
	 * @param {string}
	 * @param {string}
	 * @param {array}
	 * @param {array} 
	 * @param {boolean}
	 * @param {array}  
	 * @param {array}   
	 * @author Mr.Scatman								
	 */
	chaseInsWithRelnGWV = async($nid, $reln, $adopt, $types=[1,3], $chasetypes=[0,1,2], $parentNOBJ)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.chaseInsWithRelnGWV ", $nid);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("chaseInsWithRelnGWV", 
			{
				nid:$nid, 
				reln:$reln,
				adopt:$adopt,
				types:$types,
				chase_types:$chasetypes,
				be:$parentNOBJ.nid,
				height:lastBlock

			});

		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.chaseInsWithRelnGWV");
		}
		return aRelons;
	}
	
		
	/**
	 * Get account by accountId (ANNE neurons in the memory handler call this nid).
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	getAccount = async($accountId)=> {
		var json = await this.apiSimpleQueryWithParams("getAccount",{account:$accountId});
		if (json.status!=undefined && json.status=="fail") return null;
		return json;
	}


	/**
	 * 
	 * @param {*} $account 
	 * @param {*} $ts 
	 * @param {*} $firstIndex 
	 * @param {*} $lastIndex 
	 * @returns 
	 */
	getAccountBlocks = async($account, $firstIndex=0, $lastIndex=100000000)=>{
		var response = await this.apiSimpleQueryWithParams("getAccountBlocks", 
		{	account:$account,
			timestamp:0,
			firstIndex:$firstIndex,
			lastIndex:$lastIndex
		});
		if (response.blocks!=undefined) {
			console.log(response.blocks);
			return response.blocks;
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $adtype 
	 * @param {*} $adsubtype 
	 * @param {*} $adpayloadjsonstr 
	 * @param {*} $adpayload_sig 
	 * @returns 
	 */
	sendAltData = async($adtype, $adsubtype, $adpayloadjsonstr="", $adpayload_sig="")=>{
		var response = await this.apiSimpleQueryWithParams("sendAltData", 
		{	adtype:$adtype,
			adsubtype:$adsubtype,
			adpayload:$adpayloadjsonstr,
			adpayloadsig:$adpayload_sig
		});
		return response;
	}



	
	/**
	 * 
	 * @param {*} $alt_id_reln 
	 * @param {*} $alt_id_value 
	 * @param {*} $wvs 
	 * @returns 
	 */
	queryForAltID=async($alt_id_reln, $alt_id_value, $wvs=null)=>{
		if ($alt_id_reln==undefined || $alt_id_reln==null) return;
		if ($alt_id_value==undefined || $alt_id_value==null) return;
		
		if ($wvs==null) {
			$wvs = [ANNE.MEM.ANNE_ACCTID];
			if (ANNE.USER!=null && ANNE.USER._userNObj!=null) {
				if ($wvs.indexOf(ANNE.USER._userNObj.nid)==-1) $wvs.push(ANNE.USER._userNObj.nid);
			}
		}
		var response = await this.apiSimpleQueryWithParams("queryForAltID", {
			alt_id_reln:$alt_id_reln,
			alt_id_value:$alt_id_value,
			wvs:$wvs
		});
		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForAltID");
				return aRelons;
			}
		}
		return [];
	}

	
	/**
	 * firstIndex is the literal list index of the TXs for that account.
	 * 1st TX is 0, 2nd is 1, etc.
	 * 
	 * TO DO: we can store these to opimize.. OR just accept the last 500 and 
	 * do a more robust transaction area later.
	 * 
	 * @param {string} - account we want to query TXs for.
	 * 
	 * @author Mr.Scatman								
	 */
	getAccountTransactions = async($account, $firstIndex=0)=> {

		//todo first the first index.

		var response = await this.apiSimpleQueryWithParams("getAccountTransactions", 
			{	account:$account,
				timestamp:0,
				type:-1,
				subtype:-1,
				firstIndex:0,
				lastIndex:10000000,
				numberOfConfirmations:0,
				includeIndirect:true
			});
		console.log(response);
		if (response.transactions!=undefined) {
			console.log(response.transactions);
			console.log(response.transactions.length);//153
			return response.transactions;
		}
		return [];
	}


	/**
	 * Get all preloaded relons from the ANNODE.
     * 
     * This is a one-time preload done to establish a base in the browser's indexeddb.
	 * This is NOW (updated) stored with a .preload flag for optimized preload.
	 * 
     * The data is retained as long as the indexeddb is not wiped.
     * When it is wiped this function will run again the next time the app loads.
     * 
	 * @author Mr.Scatman
	 */
	 getAllPreloads=async()=> {
		console.log("%c GET ALL PRELOADS", ANNE.TALK.bgbigred);
		ANNE.A.setHTML("loader_msg", "getting all preloads (one-time)");
		if (ANNE._annodesLS.last_uri=="") {
			ANNE._bNEED_TO_PRELOAD_DATA=true;
			console.log("Can't continue with preload until node is set.");
			return;
		}
		var aMinBootRelons = [];
		const LIMIT = 9000;
		var offset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllPreloads", {
			limit:LIMIT,
			offset:offset
		});

		var safety=4;

		while (response.status=="good" && --safety>=0) {
			aMinBootRelons = aMinBootRelons.concat(response.result);
			ANNE.A.setHTML("loader_msg", ANNE.UTILS.commas(aMinBootRelons.length)+" relons acquired");
			if (response.result.length==LIMIT) {
				offset+=LIMIT;
				response = await this.apiSimpleQueryWithParams("getAllPreloads", {
					limit:LIMIT,
					offset:offset
				});
			}
		}

		//switching to preload=1 for preload vars.
		console.log("final process of all relons: "+aMinBootRelons.length);
		for (var r=0; r < aMinBootRelons.length;r++) {
			aMinBootRelons[r].preload=1;
		}
		return await MEMP.processNeuronsFromRelons(aMinBootRelons, true, true, "relons", "API.get@llPreloads");
	 }



	/**
	 * We have payment relons from sender->receiver with .annies.
	 * @param {*} $nid 
	 * @returns 
	 */
	getAllPaymentsIn=async($nid)=>{
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllPaymentsIn", $nid);
		lastBlock = phObj.lastBlock;
		console.log("%c(getAllPaymentsIn) with lastBlock:"+lastBlock, TALK.bgred);
        var aNewRelons = [];
		var response = await this.apiSimpleQueryWithParams("getAllPaymentsIn", {
			nid:$nid,
			height:lastBlock,
		});
		if (response.status=="good") {
			var relons_to_process = response.result;
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllPaymentsIn");
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
        }
        return aNewRelons;
	}

	
	/**
	 * TYPE_BROADCAST and TYPE_INTERNAL_PAY are now in relons_payment tbl.
	 * The API will pull from them.  Once in this app they are all treated the same
	 * as they are still just relons.
	 * @param {*} $nid 
	 * @param {*} $fromMsg 
	 * @returns  
	 */
	getAllPayments=async($nid, $fromMsg="")=>{
		if (this.opcheckerPass("getAllPayments", 30)==false) {
			return [];
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("getAllPayments");
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllPayments", $nid);
		var aNewRelons = [];
		var lastHeight = phObj.lastBlock;

		var response = await this.apiSimpleQueryWithParams("getAllPayments", {
			nid:$nid,
			height:lastHeight
		});
		
		if (response.status=="good") {
			var relons_to_process = response.result;

			if (relons_to_process.length==0) {
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
				return [];
			}
			
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllPayments <-"+$fromMsg);
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
        }

		this.opcheckerTrackIt("getAllPayments");
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
        return aNewRelons;
	}


	/**
	 * 
	 * @param {*} $el 
	 * @param {*} $annecoin 
	 * @param {*} $fee 
	 * @param {*} $msg 
	 * @returns 
	 */
	generateSendTransactionQRCode=async($el, $annies, $fee=100, $msg="")=>{
		const elImg = document.createElement("img");
		$el.appendChild(elImg);
        elImg.classList.add("qrsquare");
		var uri = this.base_uri;
		if (uri==undefined) {
			alert("?????");
			console.log("error, fix this generateSendTransactionQRCode")
			return null;
		}
		elImg.src = uri+"/anne?requestType=generateSendTransactionQRCode&immutable=true&receiverId="+USER._userNObj.nid+"&amountNQT="+$annies+"&feeNQT="+$fee+"&message="+$msg;
		return {requestType:'generateSendTransactionQRCode',
				immutable:true,
				receiverId:USER._userNObj.nid,
				amountNQT:$annies,
				feeNQT:$fee,
				message:$msg
		}
	}


	/**
	 * 
	 * @returns 
	 */
	getREDLIST=async()=>{
		var response = await this.apiSimpleQueryWithParams("getRedlist", {});
		if (response.status=="good") {
			return response;
        }
		return null;
	}
	


	/**
	 * Get All my neurons
	 */
	getAllNeuronsFor=async($nid, $bSPONSOREDTOO=true)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllNeuronsFor", $nid+$bSPONSOREDTOO);
		lastBlock = phObj.lastBlock;

		console.log("%c(getAllNeuronsFor) with lastBlock:"+lastBlock, TALK.bgred);
        var aNewRelons = [];
		var response = await this.apiSimpleQueryWithParams("getAllNeuronsFor", {
			nid:$nid,
			height:lastBlock,
			sponsored:$bSPONSOREDTOO
		});
	
		console.time("processganf");
		if (response.status=="good") {
			var relons_to_process = response.result;
            aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getAllNeuronsFor");
			await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
			console.log(aNewRelons.length);
        }
		console.timeEnd("processganf");

        return aNewRelons;
	}


	/**
	 * 
	 * @param {*} $txids_to_check 
	 */
	getTXIDSThatDontExistInMEMPOOOL=async($txids_to_check)=>{
		return;
		console.log("%c(getTXIDSThatDontExistInMEMPOOOL) with "+$txids_to_check.length+" txids", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("getTXIDSThatDontExistInMEMPOOOL", {
			txids:$txids_to_check,
		});
	
		if (response.status=="good") {
			console.log(response.IDS_NOT_IN_MEMPOOL);
			return response.IDS_NOT_IN_MEMPOOL;
        }
        return [];
	}

	
	verifyPayloadSig=async($user_nid, $sighex, $jsonstring)=>{
		console.log("%c(verifyPayloadSig)", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("verifyPayloadSig", {
			user_nid:$user_nid,
			sighex:$sighex,
			jsonstring:$jsonstring
		});
		console.log(response);
		if (response.status=="good") {
			return response.verified;
        }
        return false;
	}


	verifyLogin=async($user_nid, $user_pub_key, $proof_sig, $msg_that_was_signed)=>{
		console.log("%c(verifyLogin_EX)", TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("verifyLogin", {
			user_nid:$user_nid,
			user_pub_key:$user_pub_key,
			proof_sig:$proof_sig,
			msg_that_was_signed:$msg_that_was_signed
		});
	
		if (response.status=="good") {
			console.log(response.IDS_NOT_IN_MEMPOOL);
			return response.IDS_NOT_IN_MEMPOOL;
        }
        return [];
	}

	/**
	 * An example how to use this QUERY LANGUAGE.
	 * 
	 * from:{
	 * 		val:"{}" || "somevalue", 
	 * 		op: "=" | "!="" | "LIKE", 
	 * 		use_results_from_pool:"unique_name_of_pool_to_use_for_filter"
	 * 		slot: "toA" || "from"
	 * 		},
	 * type: TYPE,
	 * reln: {val:nid, op:"="},
	 * to_type: TO_TYPE,
	 * to: {
	 * 		val:"{}" || "somevalue", 
	 * 		op: "=" | "!="" | "LIKE" | "DOWNLINE" | "DATE_GREATERTHAN" | "DATE_LESSTHAN" | "DATE_EQUAL" | "DATE_BETWEEN", 
	 * 		use_results_from_pool:"unique_name_of_pool_to_use_for_filter" 
	 * 		slot: "toA" || "from" (defaults to "from if not set",
	 * 		reln_down: "(set if using op "DOWNLINE") nid of reln if using DOWNLINE",
	 *      max_gens: "(optional) numbber of gens to chase down building the pool"
	 * 		},
	 * pool_type: "match" || "lookup_pool"
	 * results_pool_name:"unique_name_for_results_in_this_pool",
	 * debug: "somenameforprinting"
	 * 
	 * @returns 
	 */
	test_queryByRelonJSON_archerPool = async()=>{

		//test dynamic lang
		//person named trump
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:ANNE.MEM.person.nid, op:"="},
				pool_type:"match"
			},
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.name.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"%archer%", op:"LIKE"},
				results_pool_name:"archerpool",
				pool_type:"match",
				debug:"people with name archer in their name"
			},

			//do a lookup pool of all TEXAS locations.
			{
				from:{val:"", op:""},
				type:-1,
				reln:{val:"", op:""},
				to_type:-1,
				to:{val:"9008074888561760466", op:"DOWNLINE", reln_down:MEM.constituent.nid, max_gens:4},
				results_pool_name:"texas locations",
				pool_type:"lookup_pool",
				debug:"texas locations"
			},

			//find a relon to ANY LOCATION IN THE ABOVE POOL.
			//birthplace IN Texas (anywhere)
			{
				from:{val:"", op:"", use_results_from_pool:"archerpool"},
				type:-1,
				reln:{val:"7777813598381855599", op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"{}", op:"=", use_results_from_pool:"texas locations", slot:"toA"},
				pool_type:"match",
				debug:"match users to birth place in texas"
			},
		];

		return await this.queryByRelonJSON(relons);
	}



	test_queryByRelonJSON_dates = async()=>{
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:ANNE.MEM.person.nid, op:"="},
				pool_type:"match",
				debug:"all people pool (not efficient, lol)"
			},
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.name.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"%trump%", op:"LIKE"},
				results_pool_name:"trumppool",
				pool_type:"match",
				debug:"people with name trump in their name"
			},

			//do a lookup pool of all TEXAS locations.
			{
				from:{val:"", op:""},
				type:-1,
				reln:{val:"", op:""},
				to_type:-1,
				to:{val:"", op:"DATE_BETWEEN", date_min:"4655170", date_max:"4655190"},
				results_pool_name:"datepool",
				pool_type:"lookup_pool",
				debug:"dates equal to 1946-06-14"
			},

			//find a relon to ANY LOCATION IN THE ABOVE POOL.
			//birthplace IN Texas (anywhere)
			{
				from:{val:"", op:"", use_results_from_pool:"trumppool"},
				type:-1,
				reln:{val:MEM.birthday.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"{}", op:"=", use_results_from_pool:"datepool", slot:"from"},
				pool_type:"match",
				debug:"persons born on date 4655179"
			},
		];

		return await this.queryByRelonJSON(relons);
	}


	/**
	 * find a clump by its roles (however many you like.)  Daisy chain of pools..
	 * @param {*} $action_nid 
	 * @param {*} $roles 
	 */
	findClump=async($action_nid, $roles)=> {
		var role_cmds = [];
		var actionNOBJ = MEMG.getNeuronByNid($action_nid);
		if (actionNOBJ.ghost==true) {
			console.warn("ghost neuron detected in to "+actionNOBJ.text);
			return [];
		}

		//action nid is the BE/parent.
		//this will build a pool of ALL clumps with this action.
		var action_role = {
			from:{val:"", op:""},
			type:TYPE_BE,
			reln:{val:ANNE.MEM.inst.nid, op:"="},
			to_type:TO_TYPE_ANNE,
			to:{val:$action_nid, op:"="},
			pool_type:"match"
		};
		console.log("audit clump lookup: [BE] "+actionNOBJ.text);
		role_cmds.push(action_role);

		//{role:$roleNid, role_value:$roleValueNid}
		//these cmds will serve as additional filters for the pool of 
		//actions that match the initial pool.
		for (var r=0; r < $roles.length; r++) {
			var role = $roles[r];
			
			var toNOBJ = ANNE.MEMG.getNeuronByNid(role.role_value);
			if (toNOBJ.ghost==true) {
				console.warn("ghost neuron detected in to "+toNOBJ.text);
				return [];	
			}
			
			var rcmd = 
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:role.role, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:role.role_value, op:"="},
				pool_type:"match"
			};
			
			if (r==0) {
				rcmd.results_pool_name="r0";
			}
			else {
				if (r!=$roles.length-1) {
					rcmd.results_pool_name="r"+r;
				}
				else rcmd.use_results_from_pool = "r"+(r-1);
			}
			
			var roleNOBJ = ANNE.MEMG.getNeuronByNid(role.role);
			console.log("audit clump lookup: ["+roleNOBJ.text.toLowerCase()+"] "+toNOBJ.text);
			role_cmds.push(rcmd);
		}
		
		var relons = await this.queryByRelonJSON(role_cmds);
		
		if (relons.length>0) {
			var froms = [];
			for (var r=0; r < relons.length; r++) {
				if (froms.indexOf(relons[r].from_nid)==-1) {
					froms.push(relons[r].from_nid);
				}
			}
			return ANNE.MEMG.getNObjsFromList(froms);
		}
		
		else return [];
	}


	test_queryByRelonJSON_clump = async()=>{

		//test dynamic lang
		//person named trump
		var relons = [
			{
				from:{val:"", op:""},
				type:TYPE_BE,
				reln:{val:ANNE.MEM.inst.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"7618523810988349729", op:"="},
				pool_type:"match"
			},
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:ANNE.MEM.role_when.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"2579424936603799988", op:"="},
				pool_type:"match",
				debug:"when role to 2579424936603799988"
			},
			{
				from:{val:"", op:""},
				type:TYPE_ROLE,
				reln:{val:ANNE.MEM.role_actor.nid, op:"="},
				to_type:TO_TYPE_ANNE,
				to:{val:"3482471539687014039", op:"="},
				pool_type:"match",
				debug:"role actor should be x.com"
			}
		];

		console.log("debug");
		console.log(relons);
		return await this.queryByRelonJSON(relons);
	}


	/**
	 * 
	 * @param {*} $rObjs 
	 * @returns 
	 */
	queryByRelonJSON = async($rObjs)=>{
		console.log($rObjs);
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryByRelonJSON");
		lastBlock = 0;// phObj.lastBlock;
		console.log("%c(queryByRelonJSON) with lastBlock:"+lastBlock, TALK.bgred);
		var response = await this.apiSimpleQueryWithParams("queryByRelonJSON", {
			jsonstring:JSON.stringify($rObjs),
			height:lastBlock
		});

		console.log(response);
		var aRelons = [];

		if (response.status=="good") {
			if(response.result.length) {
				await HISTORY.processResultsForNewPullHistory(response.result, phObj);
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryByRelonJSON");
			}
		}
		return aRelons;
	}


	/**
	 * 
	 * @param {*} $nid 
	 * @param {*} $forceNEW 
	 * @returns 
	 */
	getAllMyStuff_lite=async($nid, $forceNEW=false)=> {
		if (USER._userNObj.nid==MEM.ANNE_ACCTID) return [];
		if (USER._userNObj.ghost==true) return [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllMyStuff_lite", $nid);
		
		lastBlock = Math.max(0, phObj.lastBlock-10);
		
		if ($forceNEW==true) lastBlock=0;
		console.log("%c(getAllMyStuff_lite) with lastBlock:"+lastBlock, TALK.bgred);
		
		//moving to an IDB helper to keep it cleaner.
		if (ANNE.SETTINGS && ANNE.SETTINGS.bPRELOADALLDATA==true) {
			await ANNE.IDB.getAllMyStuff(phObj, $nid, true);
		}
		
		var aRelons = [];
		var LIMIT = 8000;
		var currentOffset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllMyStuff_lite", {
			nid:$nid,
			height:lastBlock,
			limit:LIMIT,
			offset:currentOffset
		});
		
		//abandoned the limit/offset as we have multiple functions (AAAAR)
		if (response.status=="good") {
			if (response.result.length>0) {
				await HISTORY.processResultsForNewPullHistory(response.result, phObj);
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllMyStuff");
			}
		}
		return;
	}


	/**
	 * Get all neurons, outbound payments and inbound payments
	 * First load data from IDB so we have it in RAM.
	 * @param {*} $nid 
	 * @param {*} $bSPONSOREDTOO 
	 * @returns 
	 */
	getAllMyStuff=async($nid, $bSPONSOREDTOO=true, $forceNEW=false)=> {
		if (USER._userNObj.nid==MEM.ANNE_ACCTID) return [];
		if (USER._userNObj.ghost==true) return [];
	
		if (this.opcheckerPass("getAllMyStuff")==false) {
			return []; 
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("getAllMyStuff");
		
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getAllMyStuff", $nid+$bSPONSOREDTOO);
		
		lastBlock = Math.max(0, phObj.lastBlock-10);
		
		if ($forceNEW==true) lastBlock=0;
		console.log("%c(getAllMyStuff) with lastBlock:"+lastBlock, TALK.bgred);

		//moving to an IDB helper to keep it cleaner.
		//if we didn't preload all data, we need to get all our stuff from IDB atleast once
		//NOTE:  our stuff should be marked as preload already, but .. this is fine to do.
		if (ANNE.SETTINGS && ANNE.SETTINGS.bPRELOADALLDATA==false) {
			await ANNE.IDB.getAllMyStuff(phObj, $nid, $bSPONSOREDTOO);
		}

		var LIMIT = 8000;
		var currentOffset = 0;
		var response = await this.apiSimpleQueryWithParams("getAllMyStuff", {
			nid:$nid,
			height:lastBlock,
			sponsored:$bSPONSOREDTOO,
			limit:LIMIT,
			offset:currentOffset
		});

		
		//abandoned the limit/offset as we have multiple functions (AAAAR)
		if (response.status=="good") {
			var rels = [];
			if (response.narrs && response.narrs.result && response.narrs.result.length>0) {
				//console.log("we have nars!")
				if (response.narrs.result) {
					rels = response.narrs.result;
				}
			}
			if (response.awareness && response.awareness.result && response.awareness.result.length>0) {
				if (response.awareness.result) {
					rels = rels.concat(response.awareness.result);
				}
			}
			if (rels.length>0) {
				await MEMP.processNeuronsFromRelons(rels, true, true, "relons", "API.getAllMyStuff");
			}
				
			await HISTORY.processResultsForNewPullHistory(rels, phObj);
		}
		
		this.opcheckerTrackIt("getAllMyStuff");
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("getAllMyStuff");
		
		return;
	}
	


	/** 
	 * Get the anneid summary level data for an ANNE identity.
	 * this is used for tallies, summaries, of a user's andy_actions, relons_sponsored, hanais, and other metrics.
	 * @author Mr.Scatman
	 */
	getANNEID=async($nObj, $bFORCEREDO=false)=> {
		if ($nObj==null) return;
		if ($bFORCEREDO==false && $nObj._anneid) return $nObj._anneid;

		var response = await this.apiSimpleQueryWithParams("getANNEID", {
			nid:$nObj.nid
		});

		if (response.status=="good") {
			if (response.result!=null && response.results) $nObj._anneid = new anneid($nObj, response.result);
			else $nObj._anneid = null;
		}
		return $nObj._anneid;
	}


	/**
	 * 
	 * @param {*} $nObj 
	 */
	getANNEID_ranks=async($nObj, $bFORCEREDO=false)=>{
		if ($nObj==null) return;
		if ($bFORCEREDO==false && $nObj._anneid!=null && $nObj._anneid.bRANKS==true) return $nObj._anneid;

		var response = await this.apiSimpleQueryWithParams("getANNEID_ranks", {
			nid:$nObj.nid
		});

		if (response.status=="good") {
			if (response.result!=null && response.results) {
				$nObj._anneid = new anneid($nObj);
				$nObj._anneid.processFromObj(response.result[0]);
			}
			else $nObj._anneid = null;
			
			if (response.extra_relons!=undefined) {
				await MEMP.processNeuronsFromRelons(response.extra_relons, false, true, "relons", "");
			}
		}
		return $nObj._anneid;
	}


	/**
	 * Get the annode pubKey from the annode auth object stored from the handshake.
	 */
	getAnnodePk = function() {
		if (this._AAUTH==null) return null;
		else return this._AAUTH.pubKey;
	}


	/**
	 * Get balance by nid.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	getBalance = async($nid)=> {
		var json = await this.apiSimpleQueryWithParams("getNeuronBalance",{nid:$nid});
		if (json.status!=undefined && json.status=="fail") return null;
		return json;
	}

	
	/**
	 * Get a merge analysis object for helpful info.
	 * @param {*} $nid1 - nid of a neuron.
	 * @param {*} $nid2 - nid of another neuron.
	 * @returns 
	 */
	getMergeReconByNids = async($nid1, $nid2)=> {
		if ($nid1==$nid2) return {status:"bad", reason:"can't merge a neuron into itself"};
		var json = await this.apiSimpleQueryWithParams("getMergeReconByNids",{nid1:$nid1, nid2:$nid2});
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		return json;
	}

	getMergeReconByTXID = async($txid)=> {
		var json = await this.apiSimpleQueryWithParams("getMergeReconByTXID",{txid:$txid});
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		return json;
	}


	/**
	 * query the NAPLEX of a given sponsor and class.
	 * Build the aplex objects so we have access to their data in our viz.
	 * @param {*} $sponsored_by_nid 
	 * @param {*} $class_nid 
	 * @returns 
	 */
	queryNAPLEX=async($sponsored_by, $class_nid, $offset=0, $nObjs=[])=> {
		const LIMIT = 2000;
		var json = await this.apiSimpleQueryWithParams("getNaplexByClass", 
				{sponsored_by:$sponsored_by, class_nid:$class_nid, limit:LIMIT, offset:$offset}
		);
		if (json.status!=undefined && json.status=="fail") return null;
		console.log(json);
		var results = json.result;
		var nObjs = $nObjs;

		for (var r=0; r < results.length; r ++) {
			var rez = results[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;

			if ($sponsored_by!=undefined && $sponsored_by!=null) {
				nObj.sponsored_by = $sponsored_by;
			}
			
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here

			if (results.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(results.pending, false, true, "relons", "API.queryNAPLEX");
            }

			nObjs.push(nObj);
		}

		if (nObjs.length==LIMIT+$offset) {
			return await this.queryNAPLEX($sponsored_by, $class_nid, $offset+LIMIT, nObjs);
		}
		return nObjs;
	}


	/**
	 * We get a set of aplex objects here.
	 * @returns {Array} nObjs - an array of neurons that are the top by level.
	 */
	getTopXAnnexByLevel = async($limit=20)=>{
		var response = await this.apiSimpleQueryWithParams("getTopXAnnexByLevel", {limit:$limit});
		//if (json.status!=undefined && json.status=="fail") return null;
		//console.log(json);
		//
		return await this.processNaplexOptimized(response, "API.queryForRecents");

		var results = json.result;
		var nObjs = [];
		for (var r=0; r < results.length; r ++) {
			var rez = results[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);

			if (results.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(results.pending, false, true, "relons", "API.getTopXAnnexByLevel");
            }
			nObjs.push(nObj);
			if (nObjs.length==$limit) break;
		}
		return nObjs;
	}



	
	getAwarvuSummary = async($since_height=0)=>{
		var response = await this.apiSimpleQueryWithParams("getAwarvuSummary", {height:$since_height});
		var annexerWhales = [];
		if (response.status=="good") {
			var results = response.result;
			for (var r=0; r < results.length; r++) {
				var anneid_obj = results[r];
				var nid = anneid_obj.nid;
				var nObj = MEMG.getOrMakeIfNew(nid);

				nObj._anneid = new anneid(nObj, anneid_obj);
				
				annexerWhales.push(nObj);
			}
		}
		return annexerWhales;
	}


	/**
	 * we get an array of anneid records here.
	 * @param {*} $cnt 
	 * @returns {array} annexWhalers - an array of nObjs with thie respective ._anneid objects set.
	 */
	getAnnexerWhales = async($limit=10)=>{
		var response = await this.apiSimpleQueryWithParams("getANNEXWhalesNaplex", {limit:$limit});
		var anneids = response.anneids;
		var naplex = response.naplex;

		var nObjs = [];

		if (anneids && anneids.status=="good") {
			var results = anneids.result;
			for (var r=0; r < results.length; r++) {
				var anneid_obj = results[r];
				var nid = anneid_obj.nid;
				var nObj = MEMG.getOrMakeIfNew(nid);
				console.log(nObj.nid+" "+nObj.text+" (has _anneid)");
				nObj._anneid = new anneid(nObj, anneid_obj);
				if (nObj._anneid==null) {
					console.warn("WHAAAAAAT?");
				}
				nObjs.push(nObj);
			}
		}

		if (naplex==undefined || naplex==null) return nObjs;
		console.log(naplex);
		var result = naplex.result;
        for (var r=0; r < result.length; r ++) {
			var rez = result[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here
			console.log(nObj.nid+" "+nObj.text+" (has _aplex)");
			if (rez.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(rez.pending, false, true, "relons", "API.queryForNakedText");
            }
		}

		return nObjs;
	}


	/**
	 * 
	 * @param {*} $nObj 
	 * @param {*} $height 
	 * @returns 
	 */
	getAwarvuSummaryFor = async($nObj, $since_height=0) => {
		if ($nObj==null) return null;
		if ($nObj.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAwarvuSummaryFor", {
			nid:$nObj.nid,
			since_height:$since_height
		});
		if (response.status=="good") {
			var apxs = new aplex_sum($nObj, APLEX_SUMMARY_SPECIFIC_NEURON, response);
			$nObj._aplex_sum = apxs;
			return apxs;
        }
		return null;
	}


	/**
	 * Will summarize ALL AWAREZ over the entire system.
	 * @param {*} $since_height 
	 * @returns 
	 */
	getAwarvuMetaSummary = async($since_height=0) => {
		var response = await this.apiSimpleQueryWithParams("getAwarvuMetaSummary", {
			since_height:$since_height
		});
		if (response.status=="good") {
			if (response.result!=null && parseInt(response.results)>0) {
				var apxs = new aplex_sum(MEM.___SELF___, APLEX_SUMMARY_NEURON_LINE, response);
				return apxs;
			}
			else {
				var apxs = new aplex_sum(MEM.___SELF___, APLEX_SUMMARY_NEURON_LINE, null);
				return apxs;
			}
        }
		return null;
	}

	
	/**
	 * Get the aplex data for neuron or hypergraph edge (relon).
	 * 
	 * This will get a feelz summary for any neuron or edge.
	 */
	getAplex=async($nObj=null, $bFORCENEW=false)=> {
		if ($bFORCENEW==false && $nObj._aplex!=undefined && $nObj._aplex!=null) return $nObj._aplex;

		if ($nObj==null) return null;
		if ($nObj.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			id:$nObj.nid,
			id_type:0
		});
		if (response.status=="good") {
			if (response.result!=null && response.results!=0 && response.results!="") {
				if (response.result.text!="" && $nObj.text=="") $nObj.text = response.result.text;
				
				var extra_out = null;
				
				if (response.result.pending!=undefined) {
					if (response.result.pending.extra_outs!=undefined) {
						var extras = response.result.pending.extra_outs;
						extra_out = extras[0];
						ANNE.TALK.printRelons(extras);
						await MEMP.processNeuronsFromRelons(extras, false, true, "relons", "API.getAplex pending.extra_outs");
					}
				}
				else if (response.result && response.result.sponsored_by && response.result.sponsored_by!=$nObj.sponsored_by) {
					//update neuron sponsored by
					console.log("update neuron sponsored by.");
					$nObj.sponsored_by = response.result.sponsored_by;
				}

				var updates = [];
				var ghost_updates = [];

				if($nObj.sponsored_by!=null && $nObj.sponsored_by!=undefined) {
					for (var r=0; r < $nObj.outs.length; r++) {
						if ($nObj.outs[r].sponsored_by!=$nObj.sponsored_by) {
							$nObj.outs[r].sponsored_by = $nObj.sponsored_by;
							if ($nObj.ghost==true || $nObj.outs[r].ghost==true || $nObj.outs[r].txid[0]=="z") {
								ghost_updates.push($nObj.outs[r]);
							}
							else {
								updates.push($nObj.outs[r]);
							}
						}
					}
				}
				
				if (updates.length) {
					await ANNE.DB.___updateAll(updates);
				}
				if (ghost_updates.length) {
					await ANNE.DB_local.___updateAll(ghost_updates, "ghost_relons");
				}

				$nObj._aplex = new aplex($nObj, $nObj.nid, response.result);
			}
			else {
				$nObj._aplex = new aplex($nObj, $nObj.nid, null);
			}
        }

        //augment with our ghost tally. (????)
        ANNE.AWARE.augmentGhosts($nObj);

        return $nObj._aplex;
	}


	/**
	 * 
	 * @param {*} $yearNid 
	 * @param {*} $monthNid 
	 * @param {*} $dayNid 
	 */
	getDates = async($yearNid, $monthNid, $dayNid)=>{
		var searchRels = [];
		var target="from_nid";

		if ($yearNid!=0) {
			var yrRel = {target:target, be:[ANNE.MEM.date_gregorian.nid], type:[TYPE_HAS], reln:ANNE.MEM.year.nid, to_a:$yearNid};
			searchRels.push(yrRel);
			target="pool";
		}

		if ($monthNid!=0) {
			var monthRel = {target:target, type:[TYPE_HAS], reln:ANNE.MEM.month.nid, to_a:$monthNid};
			if (searchRels.length==0) monthRel.be=[ANNE.MEM.date_gregorian.nid];
			searchRels.push(monthRel);
			target="pool";
		}

		if ($dayNid!=0) {
			var dayRel = {target:target, type:[TYPE_HAS], reln:ANNE.MEM.day.nid, to_a:$dayNid};
			if (searchRels.length==0) dayRel.be=[ANNE.MEM.date_gregorian.nid];
			searchRels.push(dayRel);
		}

		
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			json_relon:searchRels
		});


		console.log(response);
		return;
	}

	
	/**
	 * for relons (pass id_type 1)
	 */
	 getAplexRelon=async($relon=null)=> {
		if ($relon==null) return null;
		if ($relon.ghost==true) return null;
		var response = await this.apiSimpleQueryWithParams("getAplex", {
			id:$relon.txid,
			id_type:1
		});
		if (response.status=="good") {
			if (response.result!=null && response.results!=0) new aplex($relon, $relon.txid, response.result);
			else new aplex($relon, $relon.txid, null);
        }

		//augment with our ghost tally.
		ANNE.AWARE.augmentGhosts($relon);

        return $relon._aplex;
	}


	/** 
	 * Get the blockchain status info. 
     */
	getBlockchainStatus = async()=> {
		var json = await this.apiSimpleQuery("getBlockchainStatus");
		console.log(json);
		return json;
	}


	/**
	 * @param {String} $nid - nid longstring 
	 * @author Mr.Scatman
	 */
	getLNOTS = async($nid)=> {
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getLnots", $nid);
		var lastBlock = 0;//phObj.lastBlock;	//optimize this later after testing.
		var json = await this.apiSimpleQueryWithParams("getLnots",{nid:$nid});// height:lastBlock});
		if (json.status!=undefined && json.status=="fail") return null;
		await HISTORY.processResultsForNewPullHistory(json.lnots, phObj);
		return json.lnots;
	}

	
	/**
	 * 
	 * @param {*} $reln 
	 * @param {*} $to 
	 * @param {*} $to_type 
	 * @returns 
	 */
	getRelonsWithTo = async($reln, $to, $to_type)=> { 
		var response = await this.apiSimpleQueryWithParams("getRelonsWithTo",{reln:$reln, to:$to, to_type:$to_type});// height:lastBlock});
		if (response.status=="good") return response.result;
		return [];
	}
	


	/**
	 * 
	 * @param {String} $parentNid - nid longstring 
	 * @param {*} $value - proposed new subanndle under the parent anndle/subanndle
	 * @returns 
	 */
	getSubAnndle = async($parentNid, $value)=> {
		var json = await this.apiSimpleQueryWithParams("getSubAnndle",{parent_nid:$parentNid, value:$value});
		return json;
	}

	/**
	 * 
	 * @param {String} $anndle_str 
	 * @returns 
	 */
	getAnndlePricing = async($anndle_str)=>{
		var response = await this.apiSimpleQueryWithParams("getAnndlePricing", {anndle:$anndle_str});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return null;
	}

	/**
	 * 
	 * @param {*} $nid 
	 * @returns 
	 */
	calculateAnnexPrice = async($nid)=>{
		var response = await this.apiSimpleQueryWithParams("calculateAnnexPrice", {nid:$nid});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return null;
	}


	/**
	 * 
	 * @param {string} - account we want to query TXs for.
	 * 
	 * @author Mr.Scatman								
	 */
	getUnconfirmedTransactions = async($account)=> {
		var response = await this.apiSimpleQueryWithParams("getUnconfirmedTransactions", 
			{	
				account:$account
			}
			);
		console.log(response);
		if (response.unconfirmedTransactions!=undefined
			&& response.unconfirmedTransactions.length) {
			console.log(response.unconfirmedTransactions);
			return response.unconfirmedTransactions;
		}
		return [];
	}




	/**
	 * @param {*} $txid 
	 * @returns 
	 */
	getRelon = async($txid)=> {
    	var response = await this.apiSimpleQueryWithParams("getRelon",{
			txid:$txid
		});
		if (response.status=="good") {
			if (response.result!=null) {
				response.result.bMUSTINSERT=true;
				await MEMP.processNeuronsFromRelons([response.result], false, true, "relons", "API.getRelon");
			}
		}
		return ANNE.MEMG.lookupByTXID($txid);
	}

	getRelonsBySYSREFTXID = async($txid)=> {
    	var response = await this.apiSimpleQueryWithParams("getRelonsWithSysRefTxid",{
			txid:$txid
		});
		if (response.status=="good") {
			if (response.result.length) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getRelonsBySYSREFTXID");
				return aRelons;
			}
		}
		return [];
	}

	//accepts blockid or height
	getTransactionIdsInBlock = async($blockid)=> {
		var json = await this.apiSimpleQueryWithParams("getTransactionIdsInBlock",{
			blockId:$blockid,
			height:0
		});
		return json;
	}
	getTransactionIdsInBlockByHeight = async($height=0)=> {
		var json = await this.apiSimpleQueryWithParams("getTransactionIdsInBlock",{
			height:$height
		});
		return json;
	}


	/**
	 * Get the ANNODE tranasction (traditional bockchain format), not the hypergraph relon.
	 * @param {*} $txid 
	 * @returns 
	 */
	 getBlockchainTransaction = async($txid)=> {
    	var json = await this.apiSimpleQueryWithParams("getBlockchainTransaction",{
			transaction:$txid
		});

		console.log(json);

		if(ANNE.UTILS.isset(json.attachment)==true) {
			if (json.attachment.reln!=undefined) {
				var from_nid = "";
				var attachment = json.attachment;
				if (json.type==40) from_nid=json.sender;
				else from_nid = attachment.from;
	
				var temp_rel = {
					txid:json.transaction,
					from_nid:from_nid,
					type:attachment.type,
					reln:attachment.reln,
					to_type:attachment.to_type,
					to_s:attachment.to,
					height:json.height,
					annies:json.amountNQT,
					force_update:true
				}
	
				//put this in mem (or update existing relon in mem)
				await MEMP.processNeuronsFromRelons([temp_rel], false, true, "relons", "API.getBlockchainTransaction");
			}
		}

		return json;
	}


	/**
	 * Get the public key for a neuron.
	 * This is only relevant with key'd neurons, not keyless.
	 * @param {*} $nid 
	 */
	getAccountPublicKey = async($nid, $uri="")=>{
		var nObj = MEMG.getOrMakeIfNew($nid);
		if (nObj.pubKey!=undefined) return nObj.pubKey;	
		
		var json = await this.apiSimpleQueryWithParams("getAccountPublicKey",{account:$nid}, $uri);
		console.log("json:",json);
		
		if (json.publicKey!=undefined && json.publicKey.length>0) {

			//this is a keyless neuron, not onboardable.
			if (json.non_zero_pub_key!=undefined && json.non_zero_pub_key==false) {
				nObj.non_zero_pub_key = false;
				return "";
			}

			//this is the pubkey onchain for this neuron.
			nObj.non_zero_pub_key = true;
			nObj.pubKey = json.publicKey; 
			return json.publicKey;
		}

		//no acct on record.
		else if (json.error_wtf!=undefined) {
			return null;
		}

		return null;
	}


	getNeuronRecon = async($nid)=> {
		console.log("API.getNeuronRecon");
		var json = await this.apiSimpleQueryWithParams("getNeuron",{nid:$nid});
		console.log("json:",json);
		if (json && json.result) {
			var rez = json.result;
			if (rez.acct_public_key!=undefined) {
				var nObj = MEMG.getOrMakeIfNew($nid);
				nObj.pubKey = json.acct_public_key; 
				return rez;
			}
		}
		return null;
	}


	/**
	 * No saving in local db here.. just a quick pull
	 */
	 getDirectFundsOf = async($account)=>{
		var response = await this.apiSimpleQueryWithParams("getDirectFundsOf", {nid:$account,});
		console.log(response);
		if (response.status=="good") {
			return response.result;
		}
		return [];
	}

	
	//get a specific feelz/opinion reln (from a list) that points into a nid.
	getSpecificAwarenesses = async($to_nid, $relns)=> {
		var lastBlock = 0;
		if ($to_nid[ $to_nid.length-1]=="z") {
			console.log("DONT CALL THIS WITH TEMP NEURONS!");
			return [];
		}
		var hash = ANNE.other.sha256($to_nid+$relns.join());
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.getSpecificAwarenesses ", hash);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("getSpecificAwarenesses", 
			{
				nid:$to_nid, 
				relns:$relns,
				height:lastBlock

			});
		console.log(response);
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.getSpecificAwarenesses");
		}
		return aRelons;
	}
	
	

	/**
	 * 
	 */
	getVoteData = async ($nid)=> {
		var lastBlock = 0;
		var response = await this.apiSimpleQueryWithParams("getVoteData", 
				{
					nid:$nid,
					height:0
				});
		console.log(response);
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.getVoteData");
		}
		return aRelons;					
	}


	/**
	 * We pull this everytime on this app, and also get the aplex records for display.
	 */
	queryForRecents=async($classNids, $max_results=200)=> {
		var lastBlock = ANNE.HISTORY._GREATEST_LAST_KNOWN_BLOCK;
		if ($max_results==true) $max_results=100;
		var response = await this.apiSimpleQueryWithParams("queryForRecentsNaplex", {
			class_nids:$classNids,
			height:lastBlock-100000,
			max_results:$max_results
		});
		
		//make sure we have updated sponsored_by info here.
		if (response && response.narr && response.narr.result) {
			ANNE.TALK.printRelons(response.narr.result);
		}
		else return [];

		return await this.processNaplexOptimized(response, "API.queryForRecents");
	}



	/**
	 * Query for awareness relons.
	 * There could potentially be alot of these for any given $nid.
	 * These are outbound awareness relons, or "feelz" and "opinionz" expressed.
	 * @param {*} $nid 
	 * @param {*} $lastBlock 
	 * @param {*} $offset 
	 * @returns {Array}
	 */
	queryForAwareness=async($nid, $lastBlock, $offset)=> {
	   var aNewRelons = [];
	   var response = await this.apiSimpleQueryWithParams("queryForAwareness", {
		   nid:$nid,
		   height:$lastBlock,
		   offset:$offset
	   });

	   console.log("(queryForAwareness)");
	   if (response.status=="good") {
		   aNewRelons = response.result;
	   }

	   return aNewRelons;
   }


	/**
	 * get the global for this class and value
	 * 
	 * @param {String} $classNid - nid longstring
 	 * @param {*} $value
	 * @author Mr.Scatman
	 */
	queryForGlobal = async($classNid, $value)=> {
		var json = await this.apiSimpleQueryWithParams("getGlobal",{class_nid:$classNid, value:$value});
		return json;
	}


	/**
	 * Query for a neuron of a given class with a given value.
	 * Optionally provide a required reln and to_s value of that reln.
	 * The ANNODE will first match the class/value and then filter for reln_nid/to_s (if provided).
	 * 
	 * This does not check BEM line (be line class).  The neuron must have a be value of the value searched for 
	 * in classNid.  This is intended for flat value classes like persons, company, thing etc.
	 * @param {*} $classNid 
	 * @param {*} $value 
	 * @param {*} $reln_nid 
	 * @param {*} $to_type 
	 * @param {*} $to_s 
	 * @returns 
	 */
	queryFoNeuronByClassAndRelon = async($classNid, $value, $reln_nid, $to_type=TO_TYPE_ANNE, $to_s)=> {
		var json = await this.apiSimpleQueryWithParams("getByClassAndRelon",
			{class_nid:$classNid, 
				value:$value,
				reln:$reln_nid,
				to_type:$to_type,
				to:$to_s});
		return json;
	}


	queryFoNeuronByAltBeTexAndBe = async($classNid, $text, $alt_be_reln, $alt_be_to)=> {
		var json = await this.apiSimpleQueryWithParams("getByAltBeTextAndBe",
			{class_nid:$classNid, 
				text:$text,
				alt_be_reln:$alt_be_reln,
				alt_be_to:$alt_be_to});
		return json;
	}


	/**
	 * get a person by name and bday
	 * 
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForPersonWithBday=async($bday, $namestr, $alias="")=> {
		var json = await this.apiSimpleQueryWithParams("getPersonWithBday",
				{	
					bday:$bday, 
					name:$namestr,
					alias:$alias
				});
		return json;
	}

	getPerson=async($namestr)=> {
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("getPerson");
		var json = await this.apiSimpleQueryWithParams("getPerson",{name:$namestr});
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return json;
	}


	/**
	 * 
	 * @param {*} $foundedOnStr 
	 * @param {*} $namestr 
	 * @param {*} $foundedAtNID 
	 * @returns 
	 */
	getCompany=async($namestr, $registratonPolityNid=null)=> {
		var json = await this.apiSimpleQueryWithParams("getCompany",
		{	 
			name:$namestr,
			registration_polity_nid:$registratonPolityNid
		});
		return json;
	}


	/**
	 * Query for a global location.
	 * We pass along the sender nid so we can check for user-owned neurons too..
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForLocation = async($name, $constituent_of_nid, $alt_be_nid)=> {
		var json = await this.apiSimpleQueryWithParams("getLocation",{
					name:$name, 
					constituent_of_nid:$constituent_of_nid,
					alt_be_nid:$alt_be_nid,
					sender_nid:ANNE.USER._userNObj.nid
				});
		return json;
	}


	/**
	 * Recon on the parts of a location that we are looking to potentially resolve.
	 * @param {*} $name 
	 * @returns 
	 */
	getLocationParts = async($name)=> {
		var response = await this.apiSimpleQueryWithParams("getLocationParts",{
			name:$name
		});
		
		if (response.status=="good") {
			/************************************************************** */
			//OPTIMIZED and far cleaner.
			//this should be done, everywhere we pull in new data.
			var relons_to_process = response.result;
			//var am = new arraymap();
			//am.setMapKey("txid");
			//for(var r=0; r < relons_to_process.length; r++) {
				//am.addU( relons_to_process[r]);
			//}
			//await ANNE.IDB.checkForNewTxids(am);
			//those left in the relons map are new.
            if(relons_to_process.length>0) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.getLocationParts");
				return aNewRelons;
			}
			/************************************************************** */
		}
		return;
	}


	/**
	 * 
	 * @param {*} $anums 
	 * @returns 
	 */
	getNidsByAnums = async($anums)=> {
		var json = await this.apiSimpleQueryWithParams("getNidsByAnums",{
			anums:$anums
		});
		if (json && json.status=="good") {
			for (var a=0; a < json.result.length; a++) {
				var nid = json.result[a].nid;
				var nObj = ANNE.MEMG.getOrMakeIfNew(nid);
				if (nObj.text=="") nObj.text = json.result[a].anblstr;
			}
			return json.result.concat();
		}
		return null;
	}

	

	/**
	 * 
	 * @param {*} $name 
	 * @returns 
	 */
	queryForNamedLocation = async($name)=> {
		console.log("queryForNamedLocation: "+$name);
		var response = await this.apiSimpleQueryWithParams("getNamedLocation",{
					name:$name
				});

		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				aRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForNamedLocation");
				//await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
				return aRelons;
			}
		}
		return [];
	}

	/**
	 * Query for a hanai.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForHanai = async($nid, $height=0)=> {
		$height = Math.max(0, $height);
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForHanai", $nid);
		//lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("queryForHanai", {
			nid:$nid, 
			height:$height
		});
		console.log(response);
		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				var aRelons = await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForHanai");
				//await HISTORY.processResultsForNewPullHistory(relons_to_process, phObj);
				return aRelons;
			}
		}
		return [];
	}


	/**
	 * User is an admin for an open hanai request that doesnt have a sys_ref_txid set.
	 * That is set when the repsonse is officially processed.
	 * Look for an admin response to the request that is pending mining.
	 */
	queryForHanaiResponseTo = async($txid)=>{
		var response = await this.apiSimpleQueryWithParams("queryForHanaiResponseTo", {
			txid:$txid
		});
		console.log(response);
		if (response.status=="good" && response.result.length>0) {
			var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForHanaiResponseTo");
			return aRelons[0];
		}
		return null;
	}
	

	/**
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	 queryForAndy=async($nid, $nids_i_am_delegated_admin_for=[])=> {
		//var lastBlock = 0;	//optimize later
		if ($nid==null) return;
		if ($nids_i_am_delegated_admin_for==null) $nids_i_am_delegated_admin_for=[];

		var height = 0;
		if (ANNE.MEM_UDF!=null) height = ANNE.MEM_UDF._collab_hlc;
		 
height = 0;

		//de-optimize until we get this right....
		var response = await this.apiSimpleQueryWithParams("queryForAndy", {
				from:$nid, 
				nids:$nids_i_am_delegated_admin_for, 
				height:height
			});

		//_collab_hlc is managed in MEM_UDF _processCollabs.
		//when a user is "unlocked" it preloads from local DB mem.
		//run-time it updates through here.
		console.log(response);
		if (response.status=="good") {
			if (response.result.length) {
				await MEM_UDF._processCollabs(response.result);
			}
			return MEM_UDF._collab_relons;
		}
		return MEM_UDF._collab_relons;
	}


	/**
	 * Query the node for this text name.
	 * 
	 * @param {string}
	 * @author Mr.Scatman
	 */
	queryForNakedText = async ($searchString, $bEXACT=false) => {
		if ($searchString=="") return;
		var lastBlock = 0;
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForNakedTextNaplex", $searchString);
		//lastBlock = 0;//phObj.lastBlock;
		var exact = "no";
		if ($bEXACT==true) exact = "yes";

		var response = await this.apiSimpleQueryWithParams("queryForNakedTextNaplex", {
				text:$searchString,
				exact:exact, 
				height:lastBlock
		});

		//now returns an array of 
		return await this.processNaplexOptimized(response, "API.queryForNakedText");
	}


	/**
	 * we get annex sys relons back here of the most recent annex relons pointing to TO nids (the annexed neuron)
	 * @param {*} $cnt 
	 * @returns {Array} an array of {.nObj, .cnt}
	 */
	getLatestXAnnexes = async($limit=20)=>{
		var response = await this.apiSimpleQueryWithParams("queryForLatestXANNEXNaplex", {limit:$limit});
		return await this.processNaplexOptimized(response, "API.getLatestXAnnexes");
	}


	/**
	 * not optimizing for height, as we want all the latest napplex data.
	 * When using sponsored_by index, we get all relons unless we filter them out on the L1 query,
	 * as in the NARR view the .sponsored_by index is on all relons coming from the from_nid join.
	 * @param {*} $nid 
	 * @returns 
	 */
	getSponsored = async ($nid) => {
		var lastBlock = 0;
		var response = await this.apiSimpleQueryWithParams("getMySponsoredNaplex", {
				nid:$nid,				
				height:lastBlock
		});
		//now returns an array of 
		return await this.processNaplexOptimized(response, "API.getSponsored");
	}


	/**
	 * 
	 * @param {*} $response 
	 * @returns 
	 */
	processNaplexOptimized = async($response, $debug="")=>{
        var narr = $response.narr;
		if (narr && narr.status=="good") {
			if (narr.result.length>0) {
				await MEMP.processNeuronsFromRelons(narr.result, true, true, "relons", $debug);
			}
		}
		var naplex = $response.naplex;
		var nObjs = [];
        if (naplex==undefined || naplex==null) return nObjs;
		console.log(naplex);
		var result = naplex.result;
		for (var r=0; r < result.length; r ++) {
			var rez = result[r];
			var nObj = MEMG.getOrMakeIfNew(rez.id);
			if (rez.text!="" && nObj.text=="") nObj.text = rez.text;
			nObj._aplex = new aplex(nObj, nObj.nid, rez);	//here
			nObjs.push(nObj);
			if (rez.pending!=undefined) {
                await MEMP.processNeuronsFromRelons(rez.pending, false, false, "relons", "API.queryForNakedText");
            }
		}
		return nObjs;
	}


		
	/**
	 * 
	 * @param {*} $searchString 
	 * @param {*} $wvs 
	 * @param {*} $bes 
	 * @param {*} $bEXACT 
	 * @param {*} $bAPPLYTOBELINE ---must be true to invoke enforement of $bes array OR bem line via inbeline
	 * @param {*} $inbeline  --- left at 0 (or null) when not wanting a BEM LINE seearch of that class nid
	 * @returns 
	 */
	queryForText = async ($searchString, $wvs=[], $bes=[], $bEXACT=false, $bAPPLYTOBELINE=false, $inbeline=0) => {
		if ($searchString=="") return;
		$searchString = $searchString+"";
		$searchString = $searchString.trim();
		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText", $searchString+$bEXACT+$bAPPLYTOBELINE);
		lastBlock = 0;//phObj.lastBlock;

		console.log("(API.queryForText) "+$searchString);
		var response = await this.apiSimpleQueryWithParams("queryForText", {
				text:$searchString, 
				wvs:$wvs, 
				bes:$bes,
				applytobeline:$bAPPLYTOBELINE,
				inbeline:$inbeline, 
				height:lastBlock
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				if(response.result.length>100) {
					console.log("what is this??  so many results for "+$searchString);
				}
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForText");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				return aRelons;
			}
		}
		return [];
	}


	/**
	 * Much more efficient to use this when we know we are looking for a valid identity (by text string)
	 * @param {*} $searchString 
	 * @returns 
	 */
	queryForIdentity = async ($searchString) => {
		if ($searchString=="") return;
		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForIdentity", $searchString);
		lastBlock = 0;//phObj.lastBlock;

		console.log("(API.queryForIdentity)");
		var response = await this.apiSimpleQueryWithParams("queryForIdentity", {
				text:$searchString, 
				height:lastBlock
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForIdentity");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				return aRelons;
			}
		}
		return [];
	}
	


	/**
	 * Test rollback handling.
	 * REMOVE THIS IN MAINNET -- AS WELL AS ALTER THE APIKEY (if we want to use it at all!)
	 * @returns 
	 */
	popOffMocknet = async()=> {
		return await this.apiSimpleQueryWithParams("popOff",{numBlocks:1, height:0, apiKey:"aaaa"}); 
	}

	
	/**
	 * 
	 * @returns 
	 */
	mempoolOk = async()=>{
		var rez = await this.annodeDetails();
		if (rez.mempool_ok==true) return true;
		return false;
	}


	/**
	 * reset any warings.. give it a go.
	 * @returns 
	 */
	annodeDetails = async($uri_override="", $frenlyname="")=> {

		//if ($frenlyname!="") console.log("frenlyname:"+$frenlyname);
		//else console.log("no frenly name");
		
		var el = document.getElementById("current_uri");
		if (el!=null) el.classList.remove("bg-disconnect-warning");

		var rez = await this.query("annodeDetails", null, $uri_override);
		//console.log(rez);
		if (rez==null) {
			ANNE.API._bNODECONNECTED=false;
			ANNE.A.setHTML("current_uri", "?? no response", true);	
			return {};
		}

		//this was a test..
		if ($uri_override!="") {
			if (rez.status=="fail") {
				return rez;
			}
			rez.result.status="good";
			await this.processAuxAAUTHData(rez.result);
			await ANNE.setAnnodeByDetails(rez.result);
			return rez.result;
		}
		
		if (rez.status=="good") {
			if (rez.result.network=="ANNE") ANNE.setMainnet();
			else if (rez.result.network=="ANNE-TESTNET") ANNE.setTestnet();

			if ($frenlyname=="") $frenlyname = this.base_uri;
			else $frenlyname = $frenlyname+" - "+this.base_uri;

			ANNE.A.setHTML("current_uri", this.base_uri, true); 
			ANNE.API._bNODECONNECTED=true;
			await this.processAuxAAUTHData(rez.result);
			await ANNE.setAnnodeByDetails(rez.result);
			rez.result.status="good";
			return rez.result;
		}
		
		return {};
	}
	

	/** 
	 * Get outs for an nid, and types w/ filters.
	 * Note: This gets the BEM outs on the ANNODE (the entire be-line)
	 *
 	 * @param {string}
	 * @param {Boolean}
	 * @param {Array}
	 * @author Mr.Scatman
	*/
	queryForOuts_not = async($nid, $bMUSTPULL=false, $types=[4])=> {
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}
		
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.openSubHunt("queryForOuts_not");
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		await ANNE.IDB.queryForOuts_not(phObj, $nid, $types);

		var response = await this.apiSimpleQueryWithParams("queryForOuts_not", {
				from:$nid, 
				type:$types, 
				height:lastBlock
		});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOuts_not");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
				return aRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
	}


	getANNEBasicDemogrpahics = async($nids, $bDOLOCALONLY=false, $bEXTERNALONLY=false)=>{
		return await this.queryTosFromRelns($nids, 0, null, $bDOLOCALONLY, $bEXTERNALONLY);
        //await ANNE.API.queryTosFromRelns(MEM.tribal.nid, 0, null, bDOLOCALONLY);
	}


	/**
	 * 
	 * @param {*} $reln_nid 
	 * @param {*} $offset 
	 * @param {*} $phObj 
	 * @param {*} $bDOLOCALONLY 
	 * @param {*} $bEXTERNALONLY 
	 * @returns 
	 */
	queryTosFromRelns=async($relns, $offset=0, $phObj=null, $bDOLOCALONLY=false, $bEXTERNALONLY=false)=>{
		var nObjs = [];
		var nids = [];
		for (var r=0; r < $relns.length; r++) {
			var relNOBJ = MEMG.getOrMakeIfNew($relns[r]);
			if (relNOBJ.ghost==true) {
				continue;
			}
			nObjs.push(relNOBJ);
			nids.push(relNOBJ.nid);
		}
			
		var lastBlock = 0;
		var hash = ANNE.other.sha256(nids.join(""));
		
		var phObj = $phObj;
		if (phObj==null && $bEXTERNALONLY==false) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryTosFromRelns", hash);
			await ANNE.IDB.queryTosFromRelns(phObj, nids);
		}

		if ($bDOLOCALONLY==true) return;

		const LIMIT = 2000;

		var response = await this.apiSimpleQueryWithParams("queryTosFromRelns", {
				nids:nids, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
		});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryTosFromRelns");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.getTosFromRelns(nids, $offset+LIMIT, phObj);
			}
		}

		if (response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}

		return;
	}


	//this will NOT get awareness relons..
    getIns = async($nid, $bMUSTPULL=false)=>{
        return await this.queryForIns($nid, $bMUSTPULL);
    }


	/**
	 * THIS WILL NOT GET AWARENESS RELONS.
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @returns 
	 */
    queryForIns = async($nid, $bMUSTPULL=false)=>{
        var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForIns");

		var aRelons = [];
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForIns", $nid);
		lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForIns(phObj, $nid);

		if ($bMUSTPULL==true) lastBlock = 0; //phObj.lastBlock;

		var response = await this.apiSimpleQueryWithParams("queryForIns", {
				to:$nid, 
				height:0
			});

		if (response.status=="good") {
			if (response.result.length>0) {
				var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForIns");
				await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();

				return aRelons;
			}
		}
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
    }

	
	/** 
	 * We want to chase updated sponsored by values to prevent us drama on the annode
	 * by submitting the wrong TXs.
	 * 
	 * 
	 * @author Mr.Scatman
	 */
	queryForGhostPrep = async($nids)=> {
		var hash = ANNE.other.sha256($nids.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForGhostPrep", hash);
		
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nid = $nids[n];
			var nOBJ = null;
			if ((nid instanceof anne_neuron)==true) {
				console.warn("FIX THIS, s.b. a nid, not a neuron");
				nOBJ = nid;
			}
			else nOBJ = MEMG.getOrMakeIfNew($nids[n]);
			
			if (nOBJ.ghost==true) continue;
			nids.push(nOBJ.nid);
		}

		await ANNE.IDB.queryForOutsFroms_not(phObj, nids, [4]);
		var nids2 = [];
		for (var n=0; n < nids.length; n++) {
			var nObj=MEMG.getNeuronByNid(nids[n]);
			if (nObj.sponsored_by=="" || nObj.sponsored_by==ANNE.MEM.USER._userNObj.nid) {
				nids2.push(nObj.nid);
			}
		}
		nids = nids2;
		

		if (nids.length==0) return []; //we're good now with the IDB pull.

		var aRelons = [];
		var lastBlock = phObj.lastBlock;
		
		//too many nids..
		const nid_SHARD=30;
		if (nids.length>nid_SHARD) {
			var loops = Math.ceil(nids/nid_SHARD);
			var aRelsAll = [];
			for (var n=0; n < loops; n++) {
				var nids2 = nids.slice(n*nid_SHARD, Math.min( (n*nid_SHARD)+nid_SHARD, nids.length-1) );
				var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
					froms:nids2, 
					types:$types, 
					height:lastBlock
				});
				if (response.status=="good") {
					if (response.result.length>0) {
						var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForGhostPrep");
						aRelsAll = aRelsAll.concat(aRelons);
					}
				}
			}
			if (aRelsAll.length>0) {
				await HISTORY.processResultsForNewPullHistory(aRelsAll, phObj);
				return aRelsAll;
			}
		}

		else {
			//we only need 0/1 here.
			var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
				froms:nids, 
				types:[2,3,4,5,6,7,8,9,10], 
				height:lastBlock
			});
			if (response.status=="good") {
				if (response.result.length>0) {
					var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForGhostPrep");
					await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
					return aRelons;
				}
			}
		}
		return [];
	}

	
	/** 
	 * Get outs for a list of nids, and types w/ filters.
	 * 
	 * This is helpful when wanting all the data but not relons like feelz/awareness..
	 * as we could eventually get millions of those edges on any given neuron.
	 * 
	 * @author Mr.Scatman
	*/
	queryForOutsFroms_not = async($nids, $bMUSTPULL=false, $types=[4], $bDOIDB=true, $bREBASE01=false)=> {
		if ($nids==undefined || $nids==null) return [];
		
		//use this instead.
		if ($nids.length==1) {
			return await this.queryForOuts_not($nids[0], $bMUSTPULL, $types);
		}
		
		var hash = ANNE.other.sha256($nids.join("")+$types.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_not", hash);
		
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nid = $nids[n];
			var nOBJ = null;
			if ((nid instanceof anne_neuron)==true) {
				console.warn("FIX THIS, s.b. a nid, not a neuron");
				nOBJ = nid;
			}
			else nOBJ = MEMG.getOrMakeIfNew($nids[n]);
			
			if (nOBJ.ghost==true) continue;
			nids.push(nOBJ.nid);
		}

		if ($bDOIDB==true) await ANNE.IDB.queryForOutsFroms_not(phObj, nids, $types);

		//we only are using this function to chase for name and be.
		//if we now have those, don't chase them.
		if ($bREBASE01==true) {
			var nids2 = [];
			for (var n=0; n < nids.length; n++) {
				var nObj=MEMG.getNeuronByNid(nids[n]);
				if (nObj.text.length>0 && nObj.bes.length>0) {
					nids2.push(nObj.nid);
				}
			}
			nids = nids2;
		}

		if (nids.length==0) return []; //we're good now with the IDB pull.

		var aRelons = [];

		//intended for multi nids
		var lastBlock = 0;// phObj.lastBlock;
		
		//too many nids..
		const nid_SHARD=30;
		if (nids.length>nid_SHARD) {
			var loops = Math.ceil(nids/nid_SHARD);
			var aRelsAll = [];
			for (var n=0; n < loops; n++) {
				var nids2 = nids.slice(n*nid_SHARD, Math.min( (n*nid_SHARD)+nid_SHARD, nids.length-1) );
				var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
					froms:nids2, 
					types:$types, 
					height:lastBlock
				});
				if (response.status=="good") {
					if (response.result.length>0) {
						var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsFroms_not");
						aRelsAll = aRelsAll.concat(aRelons);
					}
				}
			}
			if (aRelsAll.length>0) {
				await HISTORY.processResultsForNewPullHistory(aRelsAll, phObj);
				return aRelsAll;
			}
		}

		else {
			var response = await this.apiSimpleQueryWithParams("queryForOutsFroms_not", {
				froms:nids, 
				types:$types, 
				height:lastBlock
			});
			if (response.status=="good") {
				if (response.result.length>0) {
					var aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsFroms_not");
					await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
					return aRelons;
				}
			}
		}
		return [];
	}


	/** 
	 * Get outs for an nid, and types w/ filters.
	 * Note: This gets the BEM outs on the ANNODE (the entire be-line)
	 *
 	 * @param {string} $nid - nidlongstr
	 * @param {Boolean}
	 * @param {Array} $types - passing none means all types
	 * @author Mr.Scatman
	*/
	getAllOuts = async($nid, $bMUSTPULL=false, $types=[])=> {
		if (this.opcheckerPass("getOuts"+$nid)==false) return []
		var rels = await this.queryForOuts($nid, $bMUSTPULL, $types);
		this.opcheckerTrackIt("getOuts"+$nid);
		return rels;
	}

	getOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		return await this.queryForOuts($nid, $bMUSTPULL, $types);
	}

	chaseOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		return await this.queryForOuts($nid, $bMUSTPULL, $types);
	}


	/**
	 * This now returns arrays of 
	 * .narrs
	 * .payments
	 * .awareness
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOuts = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		if ($nid==undefined || $nid==null) return [];
		
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}

		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts", $nid+""+$types.join(":"));
		lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForOuts(phObj, $nid, $types);

		if ($bMUSTPULL==true) lastBlock = 0; //phObj.lastBlock;
		
		var response = await this.apiSimpleQueryWithParams("queryForOuts", {
				from:$nid, 
				type:$types, 
				height:lastBlock
			});

		if (response.status=="good") {
			var rels = [];
			if (response.narrs.result && response.narrs.result.length>0) {
				//console.log("we have nars!")
				if (response.narrs.result) {
					rels = response.narrs.result;
				}
			}
			if (response.awareness && response.awareness.result.length>0) {
				if (response.awareness.result) {
					rels = rels.concat(response.awareness.result);
				}
			}
			if (response.payments && response.payments.result.length>0) {
				if (response.payments.result) {
					rels = rels.concat(response.payments.result);
				}
			}

			if (rels.length>0) {
				await MEMP.processNeuronsFromRelons(rels, true, true, "relons", "API.queryForOuts");
			}
				
			await HISTORY.processResultsForNewPullHistory(rels, phObj);
		}
		return [];
	}
	
	


	/**
	 * Used in emulation when we need more identities
	 * @param {*} $limit 
	 * @returns 
	 */
	getIdentities=async($limit=100, $offset=0)=>{
		return await this.getChildren(MEM.identity.nid, $limit, $offset);
	}
	
	/**
	 * We should do an IDB query here too... dagnab.
	 * @param {*} $parent_nid 
	 * @param {*} $limit 
	 * @param {*} $offset 
	 * @returns 
	 */
	getChildren=async($parent_nid, $limit=100, $offset=0, $phObj=null, $bDOLOCALONLY=false) => {
		if ($parent_nid==undefined || $parent_nid==null) return [];
		
		$limit = Math.min($limit, 2000);

		var phObj = $phObj;
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForInstChildren", $nid);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				return [];
			}
			await ANNE.IDB.getChldren(phObj, $parent_nid, NTYPE_INST);
		}

		if ($bDOLOCALONLY==true) return;

		var response = await this.apiSimpleQueryWithParams("queryForInstChildren", {
			nid:$parent_nid,
			offset:$offset,
			limit:$limit
		});
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				aRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getChildren");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.getChildren($parent_nid, $limit, $offset+$limit, phObj);
			}
		}

		if (response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}
		return;
	}



	/**
	 * USE THIS WISELY.  If done against a big class the return is not desirable.
	 * It should not be used to query large classes.  It is intended for instance neurons.
	 * 
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $type 
	 * @returns 
	 */
	queryForOutsAndIns = async($nid, $bMUSTPULL=false, $types=[0,1,2,3], $offset=0, $phObj=null)=> {
		if ($nid==undefined || $nid==null) return [];
		
		var nOBJ = MEMG.getOrMakeIfNew($nid)
		if (nOBJ.ghost==true) {
			alert("DO NOT QUERY GHOSTS AGAINST THE ANNODE!!!");
			return [];
		}
		if (nOBJ.ntype==NTYPE_CLASS) {
			console.warn("You are going to to query with queryForOutsAndIns with a class neuron.  NOT IDEAL.")
		}

		var lastBlock = 0;
		var phObj = $phObj;
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndIns", $nid);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				return [];
			}
			await ANNE.IDB.queryForOutsAndIns(phObj, $nid, $types);
		}

		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		const LIMIT = 4000;

		console.log("(API.queryForOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndIns", {
				from:$nid, 
				type:$types, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result && response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsAndIns");
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				return await this.queryForOutsAndIns($nid, $bMUSTPULL, $types, $offset+LIMIT, phObj);
			}
		}

		if (response.result && response.result.length && $offset==0) {
			await HISTORY.processResultsForNewPullHistory(response.result, phObj);
		}
	}


	/**
	 * This can get CRAZY big.. if ur not calling it wisely.
	 * 
	 * @param {*} $nids 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOutsAndInsFroms = async($nids, $bMUSTPULL=false, $types=[0,1,2,3], $offset=0, $phObj=null, $bDOIDB=true, $rels=[])=> {
		var nids = [];
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getOrMakeIfNew($nids[n]);
			if (nObj.ghost==true) continue;
			nids.push(nObj.nid);
		}

		if (nids.length==0) return [];

		var aRelons = [];
		var lastBlock = 0;

		var hash = ANNE.other.sha256(nids.join(""));
		var phObj = $phObj;
		
		//only first pass (when phObj is null) is optimizer.. 
		if (phObj==null) {
			phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndInsFroms", hash);
			if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
				console.log("super optimizer");
				return [];
			}
			if($bDOIDB==true) await ANNE.IDB.queryForOutsAndInsFroms(phObj, nids, $types);
		}
		
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		const LIMIT = 4000;
		
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndInsFroms", {
				froms:nids, 
				type:$types, 
				height:lastBlock,
				limit:LIMIT,
				offset:$offset
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				$rels = $rels.concat(response.result);
			}

			//if we got max values, we need to go back in again and get the rest.
			if (response.result.length==LIMIT+$offset) {
				aRelons = await this.queryForOutsAndInsFroms($nids, $bMUSTPULL, $types, $offset+LIMIT, phObj, $bDOIDB, $rels);
			}
		}

		if (aRelons.length>0 && $offset==0) {
			await MEMP.processNeuronsFromRelons(aRelons, true, true, "relons", "API.queryForOutsAndInsFroms");
			await HISTORY.processResultsForNewPullHistory(aRelons, phObj);
			return aRelons;
		}

		return $rels;
	}

	/**
	 * This one does the entire BEM line for the outs (chasing up the parents and getting their outs too)
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOutsAndIns_BEM = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		var lastBlock = 0;
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForOutsAndIns_BEM");

		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOutsAndIns_BEM", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}

		var nObj = MEMG.getNeuronByTempNid($nid);
		var bem_nids = nObj.getBemNids();
		if (bem_nids.length) {
			await ANNE.IDB.queryForOutsAndIns_BEM(phObj, $nid, $types);
		}

		console.log("(API.queryForOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOutsAndIns_BEM", {
				from:$nid, 
				type:$types, 
				height:lastBlock
		});

		var aNewRelons = [];
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOutsAndIns_BEM");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("queryForOutsAndIns_BEM");
				return aNewRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt("queryForOutsAndIns_BEM");
		return [];
	}


	/**
	 * This one does the entire BEM line for the outs (chasing up the parents and getting their outs too)
	 * @param {*} $nid 
	 * @param {*} $bMUSTPULL 
	 * @param {*} $types 
	 * @returns 
	 */
	queryForOuts_BEM = async($nid, $bMUSTPULL=false, $types=[0,1,2,3])=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForOuts_BEM", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}

		var nObj = MEMG.getNeuronByNid($nid);
		if (nObj==null) return [];
		
		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.addSubHunt("queryForOuts_BEM");

		var bem_nids = nObj.getBemNids();
		
		await ANNE.IDB.queryForOuts_BEM(phObj, bem_nids, $types);
		
		  console.log("(API.queryForOuts_BEM) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForOuts_BEM", {
				from:$nid, 
				type:$types, 
				height:lastBlock
			});

		var aNewRelons = [];
		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				try {
					//ANNE.TALK.printRelons(response.result);
					aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForOuts_BEM");
				}
				catch(e) {
					console.log(e);
				}
				
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();

				return aNewRelons;
			}
		}

		if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.closeSubHunt();
		return [];
	}


	/**
	 * 
	 * @param {*} $last_id 
	 */
	getLatestMsgs = async($last_id)=> {
		console.log("getLatestMsgs");
		var response = await this.apiSimpleQueryWithParamsNS("getLatestMsgs", {
			last_id:$last_id
		});
		//this.processBlockCycleMsgs(response);
	}


	/**
	 * graph explore neuron query
	 * Get all outs/ins other than awareness relons
	 */
	queryForCircuitOutsAndIns = async($nid, $bMUSTPULL=false)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForCircuitOutsAndIns", $nid);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;
		if (phObj.bDONETHISSESSION==true && $bMUSTPULL==false) {
			return [];
		}
		
		var aRelonsAudit = await ANNE.IDB.queryForCircuitOutsAndIns(phObj, $nid);
		console.log(aRelonsAudit.length);
		
		console.log("(API.queryForCircuitOutsAndIns) with nid:"+$nid+" and lastBlock:"+lastBlock);
		var response = await this.apiSimpleQueryWithParams("queryForCircuitOutsAndIns", {
				from:$nid, 
				height:lastBlock
			});

		if (response.status=="good") {
			phObj.bDONETHISSESSION=true;
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForCircuitOutsAndIns");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
		return [];
	}


	/**
	 */
	queryForLocationParts = async($aParts)=>{
		var response = await this.apiSimpleQueryWithParams("queryForLocationParts", {
			parts:$aParts
		});
		console.log(response);
		if (response.status=="good") {
			var relons_to_process = response.result;
			if (relons_to_process.length) {
				return await MEMP.processNeuronsFromRelons(relons_to_process, true, true, "relons", "API.queryForLocationParts");
			}
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $text 
	 * @param {*} $be_nid 
	 * @param {*} $WVs 
	 * @param {*} $bMUSTPULL 
	 * @returns 
	 */
	queryForDate=async($text, $WVs=[], $bMUSTPULL=false)=>{
		console.log("(API.queryForDate) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];
		var hash = ANNE.other.sha256($text+$WVs.join(""));
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForTextAndBe", hash);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		//if we have results for date, bail.
		//we're trying to resolve here only.
		var aRels = await ANNE.IDB.queryForTextAndBe(phObj, $text, MEM.date_gregorian.nid, $WVs);
		if (aRels.length>0) {
			return aNewRelons;
		}

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:MEM.date_gregorian.nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForDate");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
			}
		}

		var dateNOBJ = await ANNE.MEMG.getDateNeuron($text);
		if (dateNOBJ!=null) {
			return dateNOBJ;
		}
		
		return null;
	}



	queryForMonthYear=async($text, $WVs=[], $bMUSTPULL=false)=>{
		console.log("(API.queryForMonthYear) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];
		var hash = ANNE.other.sha256($text+$WVs.join(""));
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForMonthYear", hash);
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		//if we have results for date, bail.
		//we're trying to resolve here only.
		var aRels = await ANNE.IDB.queryForTextAndBe(phObj, $text, MEM.monthyear.nid, $WVs);
		if (aRels.length>0) {
			return aNewRelons;
		}

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:MEM.monthyear.nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForMonthYear");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
			}
		}

		//will get dates or date parts.
		var dateNOBJ = await ANNE.MEMG.getDateNeuron($text);
		if (dateNOBJ!=null) {
			return dateNOBJ;
		}
		
		return null;
	}

	
	/** 
 	 * @param {string}
	 * @param {string} - nid long string of the be class
	 * @author Mr.Scatman
	*/
	queryForTextAndBe = async($text, $be_nid, $WVs=[], $bMUSTPULL=false)=> {
		console.log("(API.queryForTextAndBe) with text:"+$text);
		if (ANNE.UTILS.isset($text)==false) return [];

		var hash = ANNE.other.sha256($text+$be_nid+$WVs.join(""));
		
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForTextAndBe", hash);
		
		if ($bMUSTPULL==false) lastBlock = phObj.lastBlock;

		await ANNE.IDB.queryForTextAndBe(phObj, $text, $be_nid, $WVs);

		var response = await this.apiSimpleQueryWithParams("queryForTextAndBe", {
				text:$text, 
				class_nid:$be_nid, 
				wvs:$WVs,
				height:lastBlock
			});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForTextAndBe");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
		return [];
	}


	/**
	 * This returns the beline object of props .nid, .text, .be_nid, .owned_by and .sponsored_by 
	 * 
	 * Here we don't know where a possible match may fall in a be line that has multiple sub-classes.
	 * Primary use case here is for ec_action.   Actions will subclass and their parent and can be more than 2 levels deep.
	 * When a user is search for an existing actions we can't look for the simple parent because the parent may 
	 * be another action.  (and will not match)
	 * @param {*} $text 
	 * @param {*} $be_nid 
	 * @param {*} $WVs 
	 * @returns 
	 */
	queryForText_BEM = async($text, $class_nid, $wvs=[])=> {
		console.log("(API.queryForText_BEM) with text:"+$text);

		//do we need to do this here?
		//var hash = ANNE.other.sha256($text+$class_nid+$wvs.join(""));
		//var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForText_BEM", hash);
		//await ANNE.IDB.queryForText_BEM(phObj, $text, $class_nid, $wvs);
		
		var response = await this.apiSimpleQueryWithParams("queryForText_BEM", {
				text:$text, 
				class_nid:$class_nid, 
				wvs:$wvs
			});

		var beline_objs = [];
		if (response.status=="good") {
			beline_objs = response.result;
			if (beline_objs.length) {
				//if we do this again in any other function, we can build a processor function in the MEMP.
				for (var b=0; b < beline_objs.length; b++) {
					var bemObj = beline_objs[b];
					var nOBJ = MEMG.getNeuronByNid(bemObj.nid.trim());
					if (nOBJ==null) {
						var nOBJ = new anne_neuron(bemObj.text, parseInt(bemObj.ntype), bemObj.nid);
						if (bemObj.owned_by!="") {
							nOBJ.addOwnedBy(bemObj.owned_by);
						}
						if (bemObj.sponsored_by!="") {
							nOBJ.addSponsoredBy(bemObj.sponsored_by);
						}
					}
					else {
						nOBJ.text = bemObj.text;
						bemObj.ntype = parseInt(bemObj.ntype);
					}
				}
			}
		}
		return beline_objs;
	}


	//because i have a hard time remembering things.
	queryNidBasic = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	queryNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	lookupNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNeuronByNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNeuron = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	getNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	queryNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		return await this.lookupNid($nid, $types, $bDOPUBKEY);
	}
	

	/**
	 * Use to find get relons of to neurons a from neuron connects to.
	 * ie, service terms.
	 * @param {*} $nid 
	 * @param {*} $reln 
	 * @returns 
	 */
	getAllToNeuronRelonsGeneric = async($from_nid, $reln)=> {
		var fromNOBJ = ANNE.MEMG.getNeuronByNid($from_nid);
		var relnNOBJ = ANNE.MEMG.getNeuronByNid($reln);
		console.log("(getAllToNeuronRelonsGeneric) from "+fromNOBJ.text+" with reln:"+relnNOBJ.text);
		var response = await this.apiSimpleQueryWithParams("getAllToNeuronRelonsGeneric", {
			from_nid:$from_nid, 
			reln:$reln
		});
		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllToNeuronRelonsGeneric");
		}
		return [];
	}

	getAllToNeuronRelonsGeneric2 = async($nids, $reln)=> {
		if ($nids.length==1) return await this.getAllToNeuronRelonsGeneric($nids[0], $reln);
		var response = await this.apiSimpleQueryWithParams("getAllToNeuronRelonsGeneric2", {
			nids:$nids, 
			reln:$reln
		});
		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllToNeuronRelonsGeneric");
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $reln 
	 * @param {*} $to_s 
	 * @returns 
	 */
	getAllFromNeuronRelonsGeneric = async($reln, $to_s)=> {
		 console.log("(API.getAllFromNeuronRelonsGeneric) with to_s:"+$to_s+" reln:"+$reln);
		var response = await this.apiSimpleQueryWithParams("getAllFromNeuronRelonsGeneric", {
			reln:$reln,
			to_nid:$to_s
		});

		if (response.status=="good" && response.result.length) {
			return await this.returnNOBJS(response.result, "relons", "API.getAllFromNeuronRelonsGeneric");
		}
		return [];
	}


	/**
	 * 
	 * @param {*} $relons 
	 * @param {*} $tbl 
	 * @param {*} $msg 
	 * @returns 
	 */
	returnNOBJS=async($relons, $tbl="relons", $msg="")=> {
		if ($relons==undefined || $relons==null || $relons.length==0) return [];
		var froms = [];
		for (var r=0; r < $relons.length; r++) {
			var rel = $relons[r];
			if (froms.indexOf(rel.from_nid)==-1) froms.push(rel.from_nid);
		}
		await MEMP.processNeuronsFromRelons($relons, true, true, $tbl, $msg);
		return ANNE.MEMG.getNObjsFromList(froms);
	}


	/**
	 * Process these ones as they come in..
	 * We can monitor if it is better to concat the relons and process into our local MEM in one batch
	 * or to do them as they come in the shards...
	 * @param {*} $to_nids 
	 * @returns 
	 */
	getAllInsToTheseNids = async($to_nids, $SHARD_SIZE=30)=> {
		 
		var SHARD=$SHARD_SIZE;
		var handled_nids = [];
		var to_nids = [];

		console.log("(API.getAllInsToTheseNids) with $to_nids:"+$to_nids+" shard size:"+$SHARD_SIZE);
		console.time("getAllInsToTheseNids");

		//loop into shards to reduce strain on annode w/ too many nids in list search.
		for (var n=0; n<$to_nids.length; n++) {
			var nid = $to_nids[n];
			if (handled_nids.indexOf(nid)==-1) {
				handled_nids.push(nid);
				to_nids.push(nid);
			}
			else continue;

			if (to_nids.length==SHARD) {
				var response = await this.apiSimpleQueryWithParams("getAllInsToTheseNids", {
					to_nids:to_nids
				});
				if (response.status=="good" && response.result.length) {
					await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllInsToTheseNids");
				}
				to_nids = [];
			}
		}

		//get any left overs that didn't hit a shard limit (the last bits)
		if (to_nids.length) {
			var response = await this.apiSimpleQueryWithParams("getAllFromNeuronRelonsGeneric", {
				to_nids:to_nids
			});
			if (response.status=="good" && response.result.length) {
				await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.getAllFromNeuronRelonsGeneric");
			}
			to_nids = [];
		}

		console.timeEnd("getAllFromNeuronRelonsGeneric");

		return;
	}



	/**
	 * Lookup a nid (neuron id) to get relons of types.	
	 * no height filter on this??? why??
	 * 
	 * @param {string} - The longstring of the neuron id.
	 * @param {array} - Array of types to filter to
	 */
	 lookupNid = async($nid, $types=[0,1,2], $bDOPUBKEY=false)=> {
		 console.log("(API.lookupNid) with nid:"+$nid);
		
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNid", $nid+$types.join(""));
		await ANNE.IDB.lookupNid(phObj, $nid, $types);

		var response = await this.apiSimpleQueryWithParams("lookupNid", {
			nid:$nid, 
			types:$types,
			dopubkey:$bDOPUBKEY
		});

		if (response.status=="good" && response.result.length) {
			return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupNid :"+$nid+" types:"+$types.join("/"));
		}

		return [];
	}


	/**
	 * 
	 * @param {*} $nids 
	 * @param {*} $types 
	 * @param {*} $sysMsgRelns 
	 * @returns 
	 */
	lookupForNidsAndSysMsgRelns = async($nids, $types=[0,1,2], $sysMsgRelns=[])=> {
		
		var hash = ANNE.other.sha256($nids+$types.join()+$sysMsgRelns.join(""));
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupForNidsAndSysMsgRelns", hash);
		
		await ANNE.IDB.lookupForNidsAndSysMsgRelns(phObj, $nids, $types, $sysMsgRelns);
		
		var response = await this.apiSimpleQueryWithParams("lookupNidsAndSysMsgRelns", {nids:$nids, types:$types, relns:$sysMsgRelns});
		console.log(response);

		if (response.status=="good" && response.result.length>0) {
			return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupForNidsAndSysMsgRelns");
		}
		return [];
	}

	getNames =async($nids)=>{
		return this.lookupNids($nids, [0,1], true, false);
	}

	/**
	 * Lookup a list of ANNE Ids. 
	 * We have missing info and need to get the minimums to display a prop	
	 * or neuron.  If it is ever clicked on we will chase it out a bit more.
	 * Once this is done once, it
	 * will not need to be done again as long as
	 * the user has the data in their indexDB.
	 * @param {Array}
	 * @param {Array}
	 * @author Mr.Scatman								
	 */
	lookupNids = async($nids, $types=[0,1,2], $bDOQUERY=true, $bDOLOADER=true, $from="", $fixAfterIDB=false)=> {
		console.log("%clookupNids nid cnt:"+$nids.length+" %c"+$from, ANNE.TALK.funct, ANNE.TALK.bgred, ANNE.TALK.msg)
		const SHARD = 30;

		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNids", $nids.join("")+$types.join(""));
		
		await ANNE.IDB.lookupNids(phObj, $nids, $types);

		if ($bDOQUERY==true) {
			var nids = [];
			var rezRelons = [];
			for (var n=0; n < $nids.length; n++) {
				
				if ($fixAfterIDB==true) {
					var NOBJ = ANNE.MEMG.getNeuronByNid($nids[n]);
					if (NOBJ.text!="" && NOBJ.bes.length>0) continue;
				}
				
				nids.push($nids[n]);	
				if (nids.length==SHARD) {
					var response = await this.apiSimpleQueryWithParams("lookupNids", {nids:nids, types:$types});
					if (response.status=="good" && response.result.length) {
						if (rezRelons.length==0) rezRelons = response.result;
						else rezRelons = rezRelons.concat(response.result);
					}
					nids = [];
				}
			}

			if (nids.length) {
				var response = await this.apiSimpleQueryWithParams("lookupNids", {nids:nids, types:$types}, $bDOLOADER);
				if (response.status=="good" && response.result.length) {
					if (rezRelons.length==0) rezRelons = response.result;
					else rezRelons = rezRelons.concat(response.result);
				}
				nids = [];
			}
			
			if (rezRelons.length) {
				return await MEMP.processNeuronsFromRelons(rezRelons, true, true, "relons", "API.lookupNids");
			}	
		}

		return [];
	}



	/**
	 * limit is a node param used with the offset. (where did it go?  LOLZ) 
	 * @param {*} $nids 
	 * @param {*} $offset 
	 * @returns 
	 */
	lookupNidsGetAllOuts = async($nids, $offset=0, $bDOIDB=true)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.lookupNidsGetAllOuts ", $nids);
		lastBlock = phObj.lastBlock;
	
		//we get them all on offset=0
		if ($offset==0 && $bDOIDB==true) {
			await ANNE.IDB.lookupNidsGetAllOuts(phObj, $nids);
		}
		
		var response = await this.apiSimpleQueryWithParams("lookupNidsGetAllOuts", {
			nids:$nids,
			height:lastBlock,
			offset:$offset
		});
		
		if (response.status=="good") {
			if (response.result.length) {
				var aNewRelons = await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.lookupNidsGetAllOuts");
				await HISTORY.processResultsForNewPullHistory(aNewRelons, phObj);
				return aNewRelons;
			}
		}
	}


	/**
	 * lookup max 30 txids at a time.
	 */
	lookupTxids = async ($txids)=> {

		await IDB.lookupTxids($txids);
		
		var txids = [];
		var shard=100;	//is this too big?  (100 * int(20)+ a few extras?)
		var results = [];
		for (var t=0; t < $txids.length; t++) {
			txids.push($txids[t]);
			if (txids.length==shard) {
				var response = await this.apiSimpleQueryWithParams("lookupTxids", 
					{
						txids:txids
					});
				if (response.status=="good") {
					results = results.concat(response.result);
				}
				txids=[];
			}
		}

		//left overs
		if (txids.length>0) {
			var response = await this.apiSimpleQueryWithParams("lookupTxids", 
				{
					txids:txids
				});
			if (response.status=="good") {
				results = results.concat(response.result);
			}
			txids=[];
		}

		var aRelons = [];
		if (results.length) {
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.lookupTxids");
		}
		return aRelons;					
	}
	

	/**
	 * 
	 * @param {*} $nids 
	 * @param {*} $types 
	 * @param {*} $relns 
	 * @param {*} $height 
	 * @returns 
	 */
	participantChase = async($nids, $types=[0,1,2], $relns=[], $height=0)=> {
		var response = await this.apiSimpleQueryWithParams("participantChase", {nids:$nids, types:$types, relns:$relns, height:$height});
		var aRelons = [];
		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.participantChase");
		}
		return aRelons;
	}


	javscriptErrorReporting = async($message, $source, $lineno, $colno, $error)=> {
		var response = await this.apiSimpleQueryWithParams("javscriptErrorReporting", 
			{
				message:$message, 
				source:$source, 
				lineno:$lineno, 
				colno:$colno,
				error:$error
			});
		console.log(response);
		return;
	}

		
	/**
	 * NOTE: not in use yet in the EXT.
	 * 
	 * Query for delete
	 * Do a super query on every spot we see this in from, reln, to.
	 * This helps us understand the impact of a delete (necrosis).
	 * 
	 * @param {string} - nid we want to potentially dlete
	 * @author Mr.Scatman							
	 */
	 queryForDelete = async($nid)=> {
		var lastBlock = 0;
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForDelete ", $nid);
		lastBlock = phObj.lastBlock;
		var response = await this.apiSimpleQueryWithParams("queryForDelete", {nid:$nid, height:lastBlock});

		if (response.status=="good") {
			var results = response.result;
			aRelons = await MEMP.processNeuronsFromRelons(results, true, true, "relons", "API.queryForDelete");
		}
		return aRelons;
	}


	/**			
	 * @param {int}
	 * @param {string}
	 * @param {string}
	 * @param {int} 
	 * @param {int}
	 * 
	 * @author Mr.Scatman								
	 */
	 queryForInsWithTo = async($type, $reln, $to, $to_type, $offset=0)=> {
		var phObj = HISTORY.getLastBlockFromPullHistory("ANNODE", "API.queryForInsWithTo ", ANNE.other.sha256($type+$reln+$to+$to_type+$offset));
		var lastBlock = phObj.lastBlock;
		
		if ($offset==0) {
			await ANNE.IDB.queryForInsWithTo(phObj, $type, $reln, $to, $to_type);
		}
		
		var response = await this.apiSimpleQueryWithParams("queryForInsWithTo", 
			{
				type:$type, 
				reln:$reln,
				to:$to,
				to_type:$to_type,
				offset:$offset,
				height:lastBlock

			});
		
		if (response.status=="good") {
			if (aRelons.length) {
				return await MEMP.processNeuronsFromRelons(response.result, true, true, "relons", "API.queryForInsWithTo");
			}
		}
		return [];
	}


	/**
	 * Delete an lnot.
	 * We must prove who we are to the ANNODE in order for the ANNODE to honor our request.
	 * @param {*} $txid 
	 */
	deleteLNOT=async($txid, $el_error=null)=> {
		var encrypted_msg = ANNE.USER.genGeneralProveIt($txid);
		console.log(encrypted_msg);
		var response = await this.apiSimpleQueryWithParams("sendMsgForSesh", {
			claimed_sender_nid:USER._userNObj.nid,
			encrypted_msg:encrypted_msg
		});
		console.log(response);
		if (response.status=="good") {
			var next_proof_id = response.next_proof_id;
			if (next_proof_id && next_proof_id!=-1) {
				await this.updateAndSaveNextProofId(next_proof_id);
			}
			return;
        }
		else {
			if (response.reason!=undefined) {
				console.log("%c "+response.reason, TALK.bgbigred);
				if ($el_error!=null) ANNE.A.setHTML($el_error, response.reason);
			}
			else if ($el_error!=null) ANNE.A.setHTML($el_error, "unknown error");
			return [];
		}
	}



	/**
	 * TODO: get this to work after timeouts or disconnects.
	 * 
	 * Get life info from node.
	 * 
	 * We will use this when we get to registering public ANNODES which are 
	 * competing for public use and service fees.
	 * 
	 * Client-side apps will ping the registry of ANNODES to see if the one they
	 * want to handshake with is alive.
	 */
	ping = async($node_uri_override="")=> {
		try {
			var response = await this.query("ping", null, $node_uri_override);
			if (response.status!=undefined && response.status!="good") {
				console.log(response);
			}
			return response;
		} 
		catch (e) {
			console.log('node not alive', e);
			return {status:"fail", reason:e};
		}
	}


	/**
	 * A willy-wonka CORS workaround, but hey, I like chocolate.  (actually, I don't)
	 * @param {*} $formatted_name 
	 * @returns 
	 */
	avoidCorsWIKISPOOKS = async($formatted_name="")=> {
		if ($formatted_name=="" || $formatted_name.indexOf(" ")!==-1) {
			return {status:"fail", reason:"must provide a wikispooks formatted name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsWIKISPOOKS", {formatted_name:$formatted_name});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	/**
	 * 
	 * @param {*} $formatted_name 
	 * @returns 
	 */
	avoidCorsDBPEDIA = async($formatted_name="")=> {
		if ($formatted_name=="" || $formatted_name.indexOf(" ")!==-1) {
			return {status:"fail", reason:"must provide a dbpedia formatted name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsDBPEDIA", {formatted_name:$formatted_name});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}


	/**
	 * 
	 * @param {*} $relative 
	 * @returns 
	 */
	loadLocalOnlyFile=async($relative)=> {
		let FILERESP =  await fetch($relative).catch(function($e){
			console.log("error:"+$e);
			return;
		});
		//console.log(FILERESP);
		if (FILERESP==undefined) return null; 
		let blob = await FILERESP.blob();
		var response = {};
		response.mimetype2 = blob.type;
		
		return new Promise((resolve, reject) => {
			let reader = new FileReader();
			reader.onloadend = function() {
				let b64data = reader.result;
				response.b64data = b64data;
				return resolve(response);
			};
			reader.readAsDataURL(blob);
		});
	}


	/**
	 * Extract actual src from page..
	 * get image.. store on annode, for later upgrade to real folder.
	 * @param {*} $url 
	 * @param {*} $ext
	 * @param {*} $baseurl  
	 * @param {*} $elid 
	 * @returns 
	 */
	getAndCacheFileFromElement = async($url, $ext, $baseurl="", $elid="")=>{
		if ($url=="") {
			return {status:"fail", reason:"must provide a url"};
		}
		$url = $url.replaceAll(" ", "_");
		try {
			var response = await this.apiSimpleQueryWithParams("getAndCacheFileFromElement",
						 {url:$url, ext:$ext, baseurl:$baseurl, elid:$elid});
			
			if (response.status=="good" && response.finger!=undefined) {
				
				//the files are in the html/ui/  a2a_cache... || a2a_neurons...  
				//let's go get it directly from our folder.
				var relativepath = response.relativepath;
				//ANNE.API.base_uri+
				let FILERESP =  await fetch(relativepath+"/"+response.filename).catch(function($e){
					console.log("error:"+e);
					return;
				}
				);
		
				 console.log(FILERESP);
				if (FILERESP==undefined) return response; 
				
				let blob = await FILERESP.blob();
				response.mimetype2 = blob.type;

				return new Promise((resolve, reject) => {
					let reader = new FileReader();
					reader.onloadend = function() {
						let b64data = reader.result;
						response.b64data = b64data;
						return resolve(response);
					};
					reader.readAsDataURL(blob);
				});
				
			}
			else {
				console.log("bad??  why?");
				console.log(response);
				return response;
			}
		} 

		catch (e) {
			return {status:"fail", reason:e};
		}
	}



	/**
	 * Get file by nid.
	 * If annode permits, an antor will be triggered if this file is not yet on THIS annode.
	 */
	getFileByNid = async($nid="")=>{
		if ($nid=="") {
			return {status:"fail", reason:"must provide a nid"};
		}
		try {
			var fileserve_uri = ANNE.API.base_uri+"/fileserve/nid/"+$nid;
			let FILERESP =  await fetch(fileserve_uri).catch(function($e) {
				console.log("catch error:"+e);
				return {status:FILEDATASTATUS_REQUEST_ERROR};
			}
			);
	
			if (FILERESP==undefined) {
				console.lpg("no FILERESPONSE")
				return {status:FILEDATASTATUS_REQUEST_ERROR}; 
			}

			if (FILERESP.status==FILEDATASTATUS_ANTOR_OFF) {
				console.log("ANTOR IS OFF");
				return FILERESP;
			}

			else if (FILERESP.status==FILEDATASTATUS_PENDING_ANTOR) {
				console.log("pending antor!");
				return FILERESP;
			}

			else if (FILERESP.status==FILEDATASTATUS_CANNOT_INIT_ANTOR) {
				console.log("cannot init antor");
				return FILERESP;
			}

			console.log(FILERESP);

			let blob = await FILERESP.blob();
			var response = {};
			response.mimetype2 = blob.type;
			
			return new Promise((resolve, reject) => {
				let reader = new FileReader();
				reader.onloadend = function() {
					let b64data = reader.result;
					response.b64data = b64data;
					response.status=FILEDATASTATUS_ACQUIRED;
					return resolve(response);
				};
				reader.readAsDataURL(blob);
			});
		} 
		catch (e) {
			return {status:FILEDATASTATUS_REQUEST_ERROR, reason:e};
		}
	}


	/**
	 * @param {*} $url 
	 * @param {*} $ext 
	 * @param {*} $nid 
	 * @param {*} $bDOTHUMB - return a small img thumb instead of the full one
	 * @returns 
	 */
	getAndANNODECacheTheFile = async($url, $ext="", $nid="", $bDOTHUMB=true)=>{
		if ($url=="") {
			return {status:"fail", reason:"must provide a url"};
		}

		if ($ext==null || $ext=="") {
			alert("can't work with this!  MUST HAVE EXTENSION!");
			return;
		}

		try {
			var response = await this.apiSimpleQueryWithParams("getAndCacheFile", {url:$url, ext:$ext, nid:$nid});
			if (response.status=="good" && response.finger!=undefined) {
				
				var relativepath = response.relativepath;
				//console.log("relative path:"+relativepath);

				var fileserve_uri = "";
				if ($bDOTHUMB==true) {
					var parts = response.filename.split(".");
					parts[ parts.length-2] = parts[ parts.length-2]+"_thumb";
					var filename = parts.join(".");
					fileserve_uri = ANNE.API.base_uri+"/fileserve/"+relativepath+"/"+filename;
				}
				else fileserve_uri = ANNE.API.base_uri+"/fileserve/"+relativepath+"/"+response.filename;

				//console.log("using fileserve_uri: "+fileserve_uri);

				let FILERESP =  await fetch(fileserve_uri).catch(function($e){
					console.log("error:"+e);
					return;
				}
				);
		
				//console.log(FILERESP);
				if (FILERESP==undefined) return response; 
				
				let blob = await FILERESP.blob();
				response.mimetype2 = blob.type;
				
				return new Promise((resolve, reject) => {
					let reader = new FileReader();
					reader.onloadend = function() {
						let b64data = reader.result;
						response.b64data = b64data;
						return resolve(response);
					};
					reader.readAsDataURL(blob);
				});
				
			}
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	avoidCorsWikimediaImg = async($name="bill_clinton.png", $width=200)=> {
		if ($name=="") {
			return {status:"fail", reason:"must provide a wikimedia name"};
		}
		try {
			var response = await this.apiSimpleQueryWithParams("avoidCorsWikimediaImg", {img_name:$name, img_width:$width});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}
	

	/**
	 * 
	 * @param {*} $txid 
	 * @param {*} $nid 
	 * @param {*} $type 
	 * @param {*} $reln 
	 * @param {*} $to_type 
	 * @param {*} $to_s 
	 * @returns 
	 */
	saveLocalRelon = async($txid, $nid, $type, $reln, $to_type, $to_s)=> {
		if ($txid=="" || $nid=="" || $type=="" || $reln=="" || $to_type=="" || $to_s=="") {
			return;
		}
		try {
			var response = await this.apiSimpleQueryWithParams("saveLocalRelon", 
				{txid:$txid, from:$nid, type:$type, reln:$reln, to_type:$to_type, to:$to_s});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}


	/**
	 * 
	 * @returns 
	 */
	getPrivateFolderAbsolute = async()=>{
		try {
			var response = await this.apiSimpleQueryWithParams("getPrivateFolderAbsolute");
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}



	/**
	 * 
	 * @param {*} $nid 
	 * @param {*} $b64data 
	 * @param {*} $ext 
	 * @returns 
	 */
	sendB64Data = async($nid, $b64data, $ext)=> {
		try {
			var response = await this.apiSimpleQueryWithParams("sendB64Data", 
					{nid:$nid, b64data:$b64data, ext:$ext});
			return response;
		} 
		catch (e) {
			return {status:"fail", reason:e};
		}
	}

}


/************** (from file ./js/network/API.js)**/ 


//
class generic_websocket {
    constructor($ws_uri, $bKEEP_ALIVE=true) {
        this.uri = $ws_uri;
        this.bKEEP_ALIVE = $bKEEP_ALIVE;
        this.ws = null;
        this.cb_data = null;
    }
    set_data_cb = function($cb) {
        this.cb_data = $cb;
    }
    connect = async()=> {
        return new Promise((resolve, reject) => {
            this.ws = new WebSocket(this.uri); 
            this.ws.onerror = function(err) {
                console.warn("on error");
                console.log("ERROR!", err );
            };
            this.ws.onopen = function() {
                resolve(true);
            };
            this.ws.onmessage = function ($evt) {
                if (this.cb_data!=null) this.cb_data($evt);
            };
            this.ws.onclose = function() {
                if (this.bKEEP_ALIVE==true) {
                    this.connect();
                }
            };
        });
    }

    sendJSON=function($json) {
        if (this.ws==null) return;
        var msg = JSON.stringify($json);
        this.ws.send(msg);
    }
}



var SUBSCRIBER = {}

SUBSCRIBER.bANNEX_ON = false;
SUBSCRIBER.bBOOSTIES_ON = false;
SUBSCRIBER.bBLOCKS_ON = false;
SUBSCRIBER.bANTOR_ON = false;

//the two sockets we use
SUBSCRIBER.bCHAINSOCKET_ON = false;
SUBSCRIBER.bALTDATASOCKET_ON = false;

SUBSCRIBER._cb_new_block = null;
SUBSCRIBER._cb_new_boosties = null;
SUBSCRIBER._cb_new_annex = null;
SUBSCRIBER._cb_new_transaction = null;
SUBSCRIBER._cb_new_mempool = null;
SUBSCRIBER._cb_antor = null;

SUBSCRIBER.chain_ws = null;
SUBSCRIBER.altdata_ws = null;

SUBSCRIBER.el_log = null;

//an assoc array of assoc array of callbacks by [type][subtpe] for alt data.
SUBSCRIBER._cb_alt_data = [];
SUBSCRIBER._cb_alt_data_cdxfilters = [];

SUBSCRIBER.el_cd = "socket_chaindata";
SUBSCRIBER.el_ad = "socket_altdata";

SUBSCRIBER._altdata_closed_cb = null;
SUBSCRIBER._chaindata_closed_cb = null;

/**
 * subscribe to websocket for new blocks. 
 */
SUBSCRIBER.subscribeToBlocks=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bBLOCKS_ON==false) {
            SUBSCRIBER.sendJSONToServer("onblock", null);
            SUBSCRIBER.bBLOCKS_ON = true;
        }
        resolve(true);
    });
}





SUBSCRIBER.subscribeToAntor=async($cb=null, $file_nid, $elid=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_block = $cb;
        if (SUBSCRIBER.bANTOR_ON==false) {
            var params = {filter_nid:$file_nid, filter_nid_prop:"nid"};
            
            // the socket will pass back to us the el id when the antor is 
            // ready.  We can then map to the elid we want and display the goodies
            // directly from the socket return.  We can also store this in our antor file map.
            if ($elid!=null) params.elid = $elid;

            SUBSCRIBER.sendJSONToServer("onantor", params);
            SUBSCRIBER.bANTOR_ON = true;
        }
        resolve(true);
    });
}

/**
 * We are subscribing to mempool TXs that have us as the recipient.
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToOurMempool=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_mempool = $cb;
        SUBSCRIBER.sendJSONToServer("onmempool", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}

SUBSCRIBER.subscribeToOurMinedTXs=async($cb=null)=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_transaction = $cb;
        SUBSCRIBER.sendJSONToServer("ontransaction", {filter_nid:ANNE.USER._userNObj.nid, filter_nid_prop:"recipient"});
        resolve(true);
    });
}


/**
 * subscribe to websocket for annexes (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToAnnex=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_annex = $cb;
        if (SUBSCRIBER.bANNEX_ON==false) {
            SUBSCRIBER.sendJSONToServer("onannex", null);
            SUBSCRIBER.bANNEX_ON = true;  
        }
        resolve(true);
    });
}


/**
 * subscribe to websocket for boosties (mempool or mined)
 * @param {*} $cb 
 */
SUBSCRIBER.subscribeToBoosties=async($cb=null)=> {  
    if (SUBSCRIBER.bCHAINSOCKET_ON==false) {
        await SUBSCRIBER._initChainSocket();
    }  
    return new Promise((resolve, reject) => {
        if ($cb!=null) SUBSCRIBER._cb_new_boosties = $cb;
        if (SUBSCRIBER.bBOOSTIES_ON==false) {
            SUBSCRIBER.sendJSONToServer("onboosties", null);
            SUBSCRIBER.bBOOSTIES_ON = true;
        }
        resolve(true);
    });
}


/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initChainSocket = async()=> {
    if (SUBSCRIBER.bCHAINSOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.chain_ws = new WebSocket(ANNE._annodesLS.last_socket+"chaindata"); 
        SUBSCRIBER.chain_ws.onerror = function(err) {
            console.warn("on error");
            console.log("ERROR!", err );
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
        };
        SUBSCRIBER.chain_ws.onopen = function() {
            //console.log("on open");
            SUBSCRIBER.bCHAINSOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to chaindata service!");
            }
            resolve(true);
        };
        
        SUBSCRIBER.chain_ws.onmessage = function (evt) {
            var json = JSON.parse(evt.data);
            var logmsghtml = SUBSCRIBER.formatLogData(json);
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog(logmsghtml);
            }
            var sub = json.subscription;
            if (sub==undefined || sub==null) {
                console.warn("Must have sub on return from socket");
                return;
            } 

            //ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on chain socket msg", ANNE.TALK.GREEN, sub);
            //console.log("chaindata socket", json);
            
            if (sub=="onblock") {
                if (SUBSCRIBER._cb_new_block!=null) {
                    SUBSCRIBER._cb_new_block(json);
                }
            }
            else if (sub=="onannex") {
                if (SUBSCRIBER._cb_new_annex!=null) {
                    SUBSCRIBER._cb_new_annex(json);
                }
            }
            else if (sub=="onboosties") {
                if (SUBSCRIBER._cb_new_boosties!=null) {
                    SUBSCRIBER._cb_new_boosties(json);
                }
            }
            else if (sub=="ontransaction") {
                console.log("new ontransaction");
                if (SUBSCRIBER._cb_new_transaction!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_transaction(json);
                }
            }

            //subscribes to specific mempool transactions (typically sends we receive.)
            else if (sub=="onmempool") {
                //console.log("new onmempool");
                if (SUBSCRIBER._cb_new_mempool!=null) {
                    console.log("send to callback");
                    SUBSCRIBER._cb_new_mempool(json);
                }
                else console.log("no callback set, do nothing.");
            }
        };

        SUBSCRIBER.chain_ws.onclose = function() {
            console.log("on close");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Disconnected from chaindata service!");
            }
            SUBSCRIBER.bCHAINSOCKET_ON=false;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_cd, "socket-off", "socket-on");
            if (SUBSCRIBER._chaindata_closed_cb!=null) SUBSCRIBER._chaindata_closed_cb();
        };
        });
}




/**
 * if the altdatasocket is not on (never turned on, errored, or turend of)
 * then reconnect it before sending subscribe.
 * @param {*} $type 
 * @param {*} $subtype
 * @param {*} $extraJson 
 * @param {*} $cb 
 * @returns 
 */
SUBSCRIBER.subscribeToAltData=async($type, $subtype, $extraJson, $cb=null)=> {  
    if (SUBSCRIBER.bALTDATASOCKET_ON==false) {
        await SUBSCRIBER._initAltDataSocket();
    }
    return new Promise((resolve, reject) => {
        if ($cb!=null) {
            if (SUBSCRIBER._cb_alt_data[$type]==undefined) {
                SUBSCRIBER._cb_alt_data[$type]=[];
            }
            if ($subtype!="") SUBSCRIBER._cb_alt_data[$type][$subtype]=$cb;
            else SUBSCRIBER._cb_alt_data[$type]["*"]=$cb;
        }
        SUBSCRIBER.sendJSONToServerAlt($type, $subtype, $extraJson);
        resolve(true);
    });
}


/**
 * 
 * @returns 
 */
SUBSCRIBER._initAltDataSocket = async()=> {
    if (SUBSCRIBER.bALTDATASOCKET_ON==true) {
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
        return true;
    }
    return new Promise((resolve, reject) => {
        SUBSCRIBER.altdata_ws = new WebSocket(ANNE._annodesLS.last_socket+"altdata");
        SUBSCRIBER.altdata_ws.onopen = function() {
            console.log("alt data socket opened");
            SUBSCRIBER.bALTDATASOCKET_ON=true;
            ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-on", "socket-off");
            if (SUBSCRIBER.el_log!=null) {
                SUBSCRIBER.appendLog("Connected to altdata service!");
            }
            resolve(true);
        };
        SUBSCRIBER.finalAltDataInit();
    });
}


/**
 * 
 * @param {*} $type 
 * @param {*} $subtype 
 * @param {*} $cdxfilter 
 */
SUBSCRIBER.setAltDataCallbackManual = function($type, $subtype, $cdxfilter) {
    if (SUBSCRIBER._cb_alt_data_cdxfilters[$type]==undefined) SUBSCRIBER._cb_alt_data_cdxfilters[$type]=new Object();
    SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype]=$cdxfilter;
}

SUBSCRIBER.removeAltDataCDXFilter = function($type,$subtype,$cb) {
    delete SUBSCRIBER._cb_alt_data_cdxfilters[$type][$subtype];
}


/**
 * 
 */
SUBSCRIBER.finalAltDataInit=function() {
    SUBSCRIBER.altdata_ws.onmessage = function (evt) {
        var json = JSON.parse(evt.data);

        var logmsghtml = SUBSCRIBER.formatLogData(json);
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog(logmsghtml);
        }
        
        var adtype = json.adtype;
        if (adtype==undefined || adtype==null) {
            console.warn("Must have adtype on return from socket", json);
            return;
        } 

        var adsubtype = json.adsubtype;
        
        ANNE.TALK.printf_console(ANNE.TALK.BLUE, "on altdata socket msg", ANNE.TALK.GREEN, adtype, ANNE.TALK.YELLOW, adsubtype);
        console.log("altdata socket", json);
            
        //filter cb (one-offs)
        if (SUBSCRIBER._cb_alt_data_cdxfilters[adtype]!=undefined 
            && adsubtype!=undefined 
            && SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype]!=undefined) {
            var cdxfilter = SUBSCRIBER._cb_alt_data_cdxfilters[adtype][adsubtype];
            var filter = cdxfilter.filter;
            var cb = cdxfilter.cb;
            var results = json.data.results;
            for (var r=0; r < results.length; r++) {
                var obj = results[r];
                var propkey = filter.propkey;
                var propval = filter.propvalue;
                if (obj[propkey]==undefined) {
                    break;
                }
                if (obj[propkey]!=propval) {
                    break;
                }

                //take first one.  return all, then first match.
                cb(json, obj);
                return;
            }
        }
    
        if (adsubtype!=undefined && SUBSCRIBER._cb_alt_data[adtype]!=undefined) {
            if (SUBSCRIBER._cb_alt_data[adtype][adsubtype]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype][adsubtype](json);
            }
            else if (SUBSCRIBER._cb_alt_data[adtype]["*"]!=undefined) {
                SUBSCRIBER._cb_alt_data[adtype]["*"](json);
            }
        }
        else {
            console.log("*** no finalAltDataInit calback ***");
        }
    };

    SUBSCRIBER.altdata_ws.onclose = function() {
        console.log("on close");
        if (SUBSCRIBER.el_log!=null) {
            SUBSCRIBER.appendLog("Disconnected from chaindata service!");
        }
        SUBSCRIBER.bALTDATASOCKET_ON=false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
        if (SUBSCRIBER._altdata_closed_cb!=null) SUBSCRIBER._altdata_closed_cb();

    };

    SUBSCRIBER.altdata_ws.onerror = function(err) {
        console.warn("on error");
        console.log("ERROR!", err )
        SUBSCRIBER.bALTDATASOCKET_ON = false;
        ANNE.A.addRemoveClass(SUBSCRIBER.el_ad, "socket-off", "socket-on");
    };
}







/**
 * 
 * @param {*} $subscribe 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServer=function($subscribe, $json) {
    //console.log("SendToServer subscribe:"+$subscribe);
    if ($json!=null) $json.subscribe = $subscribe;
    else $json = {subscribe:$subscribe};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.chain_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.chain_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.chain_ws.readyState);
    }
}


/**
 * 
 * @param {*} $adtype 
 * @param {*} $adsubtype 
 * @param {*} $json 
 */
SUBSCRIBER.sendJSONToServerAlt=function($adtype, $adsubtype, $json) {
    console.log("sendJSONToServerAlt adtype:"+$adtype, "adsubtype:"+$adsubtype);
    if ($json!=null) {
        $json.adtype = $adtype;
        $json.adsubtype = $adsubtype;
    }
    else $json = {adtype:$adtype, adsubtype:$adsubtype};
    const msg = JSON.stringify($json);
    if (SUBSCRIBER.altdata_ws.readyState === WebSocket.OPEN) {
        SUBSCRIBER.altdata_ws.send(msg);
    } 
    else {
        SUBSCRIBER.appendLog("ERROR: WebSocket is not open. Ready state: " + SUBSCRIBER.altdata_ws.readyState);
    }
}



/**
 * 
 * @param {*} $logText 
 */
SUBSCRIBER.appendLog=function($logText) {
    if (SUBSCRIBER.el_log==null) return; 
    var log = document.getElementById(SUBSCRIBER.el_log);
    if (log==null) return;
    log.innerHTML = log.innerHTML + $logText + "<BR>>";
}


/**
 * 
 * @param {*} $d 
 * @returns 
 */
SUBSCRIBER.formatLogData = function($d) {
    const good_status = "<span class='good'>good</span>";
    const bad_status = "<span class='bad'>bad</span>";

    var html = "";
    if ($d.status=="good") html = good_status;
    else html = bad_status;
    html = html + "<span class='subscription'>"+$d.subscription +"</span>";
    if ($d.message!=undefined) html = html + "<span class='message'>"+$d.message+"</span>";
    if ($d.data!=undefined) {
        html = html + SUBSCRIBER.jsonSyntaxHighlight($d.data);
    }
    return html;
}


/**
 * 
 * @param {*} $json 
 * @returns 
 */
SUBSCRIBER.jsonSyntaxHighlight = function ($json) {
    if ($json == null) return "";
    if (typeof $json != 'string') {
        $json = JSON.stringify($json, undefined, 2);
    }
    $json = $json.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
    return $json.replace(/("(\\u[a-zA-Z0-9]{4}|\\[^u]|[^\\"])*"(\s*:)?|\b(true|false|null)\b|-?\d+(?:\.\d*)?(?:[eE][+\-]?\d+)?)/g, function (match) {
        var cls = 'number';
        if (/^"/.test(match)) {
            if (/:$/.test(match)) {
                cls = 'key';
            } else {
                cls = 'string';
            }
        } else if (/true|false/.test(match)) {
            cls = 'boolean';
        } else if (/null/.test(match)) {
            cls = 'null';
        }
        return '<span class="' + cls + '">' + match + '</span>';
    });
}

/**
 * Working for blocks, testing for annex/boosties.
 * @returns 
 */
SUBSCRIBER._initGeneric = async($ws_uri, $KEEP_ALIVE=true)=> {
    var ws_generic = new generic_websocket($ws_uri, $bKEEP_ALIVE)
    ws_generic.connect();
}

/************** (from file ./js/network/subscriber.js)**/ 

class user_master  {
    constructor(){
        this.public_nid = "";
        this.pubKey = "";
        this.privKey = "";
        this.seedObj = null;                    //switching from passing seedObj around.. ot just using it here when needed
        this.unlock_pw = "";
        this.balanceObj = null;                 //onoly set with the logged in user's balance
        this.onLoginCallback = null;
        this._bNEWUSER = false;  
        this._userNObj = {nid:0};
        this._userLS = null;                    //user localStorage obj
        this._bWORLDVIEWBIPASS = false;
        this._badPWAttempts = 0;
        this._settingsLS = {};
        this.APP_KEY="ANNE";
    }

    getPrivKeyHex=async()=> {
        //if (this.privKey!=null && this.privKey!="") return this.privKey;
        if (this.seedObj==null) {
            if (this.unlock_pw!="" && this.unlock_pw!=null) {
                this.seedObj= await this.getSeedWithPW(this.unlock_pw);
            }
            else {
                return null;
            }
        }
        var aO = this.prepANNEAuthObjFromSeed(this.seedObj.seed);
        this.privKey = aO.priv_key;
        return this.privKey;
    }

    getSeedObj=async()=> {
        if (this.seedObj!=null) return this.seedObj;
    }



    getLastUser=async()=>{
        var user = null;
        var temp = localStorage.getItem(this.APP_KEY);
        var ANNE_LS = null;

        //if we have a local object, but it is bad, set bDEAD to true.
        if (temp==undefined || temp==null) return null;
        
        var ANNE_LS = JSON.parse(temp);
        if (ANNE_LS==undefined || ANNE_LS==null) return null;
        else {
            if (ANNE_LS.encrypted!=undefined && ANNE_LS.public_nid!=undefined) user = ANNE_LS;
            else user = ANNE_LS.user;

            if (user==undefined || user==null || user.public_nid=="") return null;
            else return user;
        }
        return null;
    }

    
    /**
     * Check localStorage for user info.
     * 
     * first check the localStorage of THIS app --  if we don't find any  
     * goodies, check the ANNE SCOPE IFRAME.  Nifty --  if we find the       
     * goodies there, store them in localStorage here too!                
     *                                                                       
     * example use case:                                                     
     * We want on onboarding app ANNEON to be able to save the hash/address  
     * and count -- so the user can get their key out on another ANNE app    
     * without having to do it all again.                                    
     * @author Mr.scatman
     */
     checkLoginFromStorage=async($bRESTORE=false, $key="ANNE")=> {
   
        //CHECK ANNE SCOPE FIRST
        //THE USER MANAGES ALL THEIR SETTINGS THERE!
        var bDONE = false;
        var user = null;
        var temp = localStorage.getItem($key);
        var ANNE_LS = null;
        var user = null;
        
        //if we have a local object, but it is bad, set bDEAD to true.
        if (temp!=undefined && temp!=null) {

            var ANNE_LS = JSON.parse(temp);
            if (ANNE_LS==undefined || ANNE_LS==null) bDEAD=true
            else {
                if (ANNE_LS.encrypted!=undefined && ANNE_LS.public_nid!=undefined) {
                    user = ANNE_LS;
                }
                else user = ANNE_LS.user;

                if (user==undefined || user==null || user.public_nid=="") {
                    bDEAD=true;
                }
                else {
                    bDONE=true;
                    if (ANNE.API._AAUTH!=null) {
                        user.aauth = ANNE.API._AAUTH;
                        await this.__saveANNE_JO(user);
                    }
                    else {
                        if (user.aauth!=undefined) {
                            ANNE.API.setAAUTHFromLS(user.aauth);
                            //var ok = await this.refreshAAUTH(seedObj.seed, this._userLS);
                            //if we are missing vars, let's refresh this.
                            //await this.__saveANNE_JO(user);
                        }
                    }

                }
            }
        }

        else {
            console.warn("No "+$key+" obj in localStorage found.");
        }


        //if no "ANNE" local storage object,
        //we will make one from the user array.
        if (bDONE==false && $bRESTORE==true) {
            var users = await this.checkLoginFromANNE_history();
            /*******************************/
            //take the first good one we have.
            /*******************************/
            if (users.length) {
                for (var u=0; u < users.length; u++) {
                    var user = users[u];
                    this.public_nid = user.public_nid;
                    this.pubKey = user.pubKey;

                    if (user.encrypted.length && user.count!=-1) {

                        //creates an aauth from stored infos.
                        if (user.aauth!=undefined) {
                            ANNE.API.setAAUTHFromLS(user.aauth);
                        }
                        //update this aauth, set ANNE.
                        localStorage.setItem($key, JSON.stringify(user)); 
                        bDONE=true;
                        break;
                    }   
                }   
            }  
        }


        //no local storage of any users exist.
        //(no ANNE or ANNE_history)
        if (bDONE==false) {
            console.warn("nothing can be done");
            return false;
        }
        
        this.public_nid = user.public_nid;
        this.pubKey = user.pubKey;

        if (ANNE.MEM!=null) {
            this._userNObj = ANNE.MEM.addInUserNeuron(user.public_nid, true);
        }
        else {
            this._userNObj.nid = user.public_nid;
        }

        this._userNObj.my_sponsored=[];
        this._userNObj.mine=true;
        this._userLS = user; 

        this._settingsLS = await this._getSettings();
        if (ANNE.SETTINGS!=null) {
            ANNE.SETTINGS.init();
        }

        if (ANNE.MEMP) await ANNE.MEMP.__postLoginCleanup();

        this.updateLSNickname(this._userNObj);

        return true;
    }



    /**
     * Make a new random seed for an ANNE address and corresponding keys.
     * Populate it into the $el
     * 
     * @param {*} $el 
     * @param {*} $qrdivid 
     * @returns 
     */
    makeSeed=async()=> {
        var neuron = await this.getNewAnneUserAccount();
        return neuron;
    }

    
    makeSeedSOLO=async($seed)=> {
        var neuron = await this.getNewAnneAccount($seed);
        return neuron;
    }

    
    /** 
     * Generate a random new seed.
     * @author Mr.Scatman
     */
    getNewAnneUserAccount=async()=> {
        var seedseed = ANNE.UTILS.randomHex(32)+ANNE.UTILS.randomHex(16)+ANNE.UTILS.randomHex(32);
        var aSeeds = [this.getNewAnneAccount(seedseed)];

        var seed = "";
        if (aSeeds.length) seed = aSeeds[0].seed;
        else {
            seed = ANNE.UTILS.randomHex(32)+ANNE.UTILS.randomHex(16)+ANNE.UTILS.randomHex(32);
        }
        var neuron = ANNE.SIGNA.crypto.generateMasterKeys(seed);
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(neuron.publicKey, "A");
        neuron.accountId = addressObj._numericId;
        neuron.seed = seed;
        neuron.address = addressObj._rs;
        neuron.nid = neuron.accountId;
        //console.log(neuron);
        return neuron;
    }


    getNidFromSeed=function($seedObj) {
        var neuron = ANNE.SIGNA.crypto.generateMasterKeys($seedObj.seed);
        if (neuron!=null) {
            var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(neuron.publicKey, "A");
            return addressObj._numericId;
        }
        return null;
    }


    /*** 
     * 
     * Note: make sure we add acct into RAM ANNE._userAcctsAllJO
     * @param {string}
     * @author Mr.Scatman 
     */
    setSeed=async($seed)=> {
        ANNE.TALK.printf_console(TALK.CYAN, "(setSeed)");
        var seed_text = $seed.trim();
        
        //this is safe to set here in the extension.
        //for this session we will have access to it without
        //requiring a pw check.
        this.seedObj = {seed:seed_text, status:"success"};
        var pw_text = this.unlock_pw;
        var pubKey = ANNE.SIGNA.crypto.generateMasterKeys(seed_text).publicKey;

        var user = await this.setUserSeed(seed_text, pw_text, pubKey);
        if (user==null) {
            console.log("This is not a valid ANNE seed.");
            return false;
        }
        return;
    }



    /**                      
    * Encrypt the passphrase in local storage with a password.
    * note: make sure to add acct into ANNE._userAcctsAllJO
    * @param {string} 
    * @param {string}
    * @param {string}
    * @author Mr.Scatman                                           
    */
    setUserSeed = async($seed, $pw_text, $pubKey="")=> {
        ANNE.TALK.printf(TALK.CYAN, "(this.setUserSeed)");

        var count = Math.floor(Math.random()*(9-5+1)+5);
        var pw_hash = await this.hashPw($pw_text);
        var encrypted = this.encrypt_it($seed, pw_hash);  //using the pw_hash to encrypt the passphrase(seed)
        var public_nid = this.getAccountIdFromSeed($seed);
        
        if (public_nid=="") {
            ANNE.TALK.printf(TALK.RED, "(this.setUserSeed) error: no nid from this seed.");

            this.public_nid = "";
            this._userNObj = {nid:0, my_sponsored:[]};
            this.pubKey = "";
            return null;
        }
        this.public_nid = public_nid;
        
        var user = {
                    "encrypted":encrypted,
                    "count":count, 
                    "activated":false,
                    "activate_testnet":false,
                    "activate_mainnet":false,
                    "public_nid":public_nid,
                    "pubKey":$pubKey, 
                    "annode_uri":ANNE._annodesLS.last_uri,
                    "backed_up":false,
                    "backed_up_on":"",
                    "last_known_annies":0,
                    "last_login":Date.now(),
                    "nickname":""
                } 

        //when the app locks it wipes the user_pw.
        //user.user_pw = $pw_text;

        this.pubKey = $pubKey;
        
        var ok = await this.refreshAAUTH($seed, user);
        if (ok==true) {
            if (ANNE.MEM) {
                this._userNObj = ANNE.MEM.addInUserNeuron(user.public_nid, true);
            }
            			//not many apps will not use the ANNE.MEM, but for those that do...
			if (ANNE.MEM) {
                this._userNObj = ANNE.MEM.addInUserNeuron(user.public_nid, true);
            }
			else {
				if (this._userNObj==null) this._userNObj = {nid:public_nid, my_sponsored:[]};
				else this._userNObj.nid = public_nid;
			}
            this._userNObj.my_sponsored=[];
            this._userLS = user;
        
            await this.__saveANNE_JO(user);

            if (ANNE.BRAIN!=null) ANNE.BRAIN.setPrimarySpeaker(user.public_nid);
            
            ANNE.addNewUser(user);

            return user;
        }

        return null;
    }


    /**                      
    * @param {string} - seed
    * @param {Object} - user LS object
    * @author Mr.Scatman                                           
    */
    refreshAAUTH=async($seed, $userLSOBJ)=> {
        var aO = this.prepANNEAuthObjFromSeed($seed);
        var randuniquerequestid = UTILS.randHex(4);

        var result = await ANNE.API.ANNODEHANDSHAKE(aO, randuniquerequestid);
        if (result && result.status=="good") {
            var result = result.result;
            if (result.logins!=undefined) $userLSOBJ.logins = result.logins;
            if (result.encrypted_proof==undefined || result.nonce2==undefined) {
                console.log("ANNODE must supply proof of annode");
                return false;
            }

            var nonce2 = result.nonce2;

            //this is from the ANNENODE -- we need to decyrpt it
            //then do our standard hash encrypt FOR US, to store it back into localStorage.
            var proof_decrypted = await this.decryptMsgEncryptedFromANNODE(nonce2, result.encrypted_proof, ANNE.API._AAUTH.pubKey, aO.priv_key);
            var parts = proof_decrypted.split(":::");
            if (parts.length==2) {
                var proof_nid = parts[0];
                if (proof_nid.trim()!=$userLSOBJ.public_nid) {
                    console.log("nid of this user must match decrypted nid coming back");
                    return false;
                }
                var uid = parts[1];
                if (uid!=randuniquerequestid) {
                    console.log("uid sent must match decrypted uid coming back");
                    return false;
                }
            }
            else return false;
            
            //we must have the encrypted_sk and the nonce.
            //this will help with various annode services.
            if (result.encrypted_sk!=undefined && result.nonce!=undefined) {
                var nonce = result.nonce;

                //this is from the ANNENODE -- we need to decyrpt it
                //then do our standard hash encrypt FOR US, to store it back into localStorage.
                var shared_secret = await this.decryptMsgEncryptedFromANNODE(nonce, result.encrypted_sk, ANNE.API._AAUTH.pubKey, aO.priv_key);
                var ss_text = shared_secret;
                console.log("audit: "+ss_text);
                
                //hijack this function and use the hashedkey w/ salt as the pw, and encrypt the shared secret
                var encrypted_sk2 = await this.hashSeedWithRecoveryPW("x"+$seed+"y", shared_secret);
                if (encrypted_sk2==null) return false;

                //$user.encyrpted_sk_fromannode = result.encrypted_sk;
                $userLSOBJ.encrypted_sk = encrypted_sk2;    //now it is encrypted by US.
                
                if (ANNE.API._AAUTH!=null) {
                    ANNE.API._AAUTH.encrypted_sk = result.encrypted_sk;
                    ANNE.API._AAUTH.nonce = result.nonce;
                    ANNE.API._AAUTH.encrypted_proof = result.encrypted_proof;
                    ANNE.API._AAUTH.nonce2 = result.nonce2;
                    ANNE.API._AAUTH.next_proof_id = result.next_proof_id;

                    //add in existing extra props from our annode info pull (if we did it)
                    if ($userLSOBJ.aauth!=undefined && $userLSOBJ.aauth!=null) {
                        ANNE.API._AAUTH.ANNDLE_active = $userLSOBJ.aauth.ANNDLE_active;
                        ANNE.API._AAUTH.ASDF_active = $userLSOBJ.aauth.ASDF_active;
                        ANNE.API._AAUTH.BOUNTY_active = $userLSOBJ.aauth.BOUNTY_active;
                        ANNE.API._AAUTH.GOLIVE_active = $userLSOBJ.aauth.GOLIVE_active;
                        ANNE.API._AAUTH.PDNS_active = $userLSOBJ.aauth.PDNS_active;
                    }
                    
                    $userLSOBJ.aauth = ANNE.API._AAUTH;
                }

                
                //store this in ANNE scope memory too, if this is on.
                //unless we're told explicity not to.  Then don't.  :)
                //never store no_pw mode in anne scope.
                ANNE.API.setAAUTHFromLS($userLSOBJ.aauth);
                await this.__saveANNE_JO($userLSOBJ);

                //extract the extras (relons from this annode aimed at us
                //using the relon=user syntax).
                //we need to extend the mem handler to get SYS_REF_TXID
                //in the relon.  We don't need to store it in the attachment.. (or do we??)
                if (result.extras!=undefined) {
                    console.log("WE HAVE EXTRAS");
                    console.log(result.extras);
                    for (var e=0; result.extras.length; e++) {
                        var relon = results.extras[e];      //the relon sent by the annode.
                        var ref_txid = relon.sys_ref_txid; 
                        if (ref_txid==undefined || ref_txid==null) continue;
                        
                        //this will be the relon WE SENT to the annode to trigger a service.
                        var ref_relon = MEMG.lookupRelonByTXID(ref_txid);
                        if (ref_relon==null) continue;

                        //this relon could have been sent in the childRelon pattern
                        //which means we were not in the "from_nid".
                        //the from_nid is the neuron we took action on...
                        //the reln is the action..
                        ref_relon.annode_response = relon;  //back link for mem use
                        
                        var temp_relon = MEM.makeTempRelon(nOBJ, TYPE_HAS, ref_relon.reln, ref_relon.to_type, ref_relon.to_s, true);
                        if (temp_relon) await DB.___update(temp_relon);
                    }
                }

                return true;
            }
        }
        return false;
    }


    /**
     * 
     * @author Mr.scatman
     */
    isUs=function($nid) {
        if (this._userNObj==null) return false;
        if ($nid==this._userNObj.nid) return true;
        return false;
    }


    /**
     * Is this neuron us or owned by us? 
     * @param {string}
     * @author Mr.scatman
     */
    isUsOrOwnedByUs=function($nid) {
        if (typeof $nid ==='object') {
            console.log("FIX THIS CALLER");
            if ($nid.nid!=undefined) $nid=$nid.nid;
            else return false;
        }
        if (this._userNObj==null) return false;
        if ($nid==this._userNObj.nid) return true;
        var nObj = MEMG.getNeuronByNid($nid);
        if (nObj==null) return false;
        if (nObj.owned_by!=undefined && nObj.owned_by==this._userNObj.nid) return true;
        return false;
    }



    /**
    * AES encrypt a string with a key.
    **/
    encrypt_it = function ($stringToEncrypt, $key) {
        if (ANNE.SIGNA.crypto==null==null) {
            console.error("must include SIGNA.crypto");
            return;
        }
        return ANNE.SIGNA.crypto.encryptAES($stringToEncrypt, $key);
    }


    decryptMsgFromANNEUser = function($encryptedMsg, $senderPubKeyHex) {
        //var privKey = this.
        //decryptMessage(encryptedMessage, senderPublicKeyHex, recipientPrivateKeyHex)
    }

    /**
    * AES decrypt a string with a key.
    **/
    decrypt_it = function ($string, $key) {
        if (ANNE.SIGNA.crypto==null) {
            console.error("must include SIGNA.crypto");
            return;
        }
        return ANNE.SIGNA.crypto.decryptAES($string, $key);
    }


    /**
     * 
     * @param {*} $seed 
     * @returns 
     */
    isValidSeedSOLO = function($seed) {
        var len = $seed.length;
        if (len < 64) {
            return {status:"bad", reason:"Seed must be atleast 64 chars."};
        }
        var hits = [];
        var last = -1;
        var consecs = 0;
        for (var i=0; i < len; i++) {
            var c = $seed.charCodeAt(i);
            if (c==last) consecs+=1;
            else consecs=0;
            last = c;
            if (consecs > 5) {
                return {status:"bad", reason:"insecure. Too man consectutive chars."};
            }

            if (hits[c]==undefined) hits[c]=1;
            else {
                hits[c]+=1;
                if (hits[c] > (len/4)) {
                    return {status:"bad", reason:"over 1/4th ur chars are the same!  Need more variance."};
                }
            }
        }

        return {status:"good"};
    }
    

    /**
    * update this for signa.
    */
    isValidPK = function($pk) {
        true;
    }


    /** 
    * from a passphrase, get the accountId
    */
    getAccountIdFromSeed=function($seed) {
        if(ANNE.SIGNA.crypto==null || ANNE.SIGNA.sig==null) return "";
        var neuron = ANNE.SIGNA.crypto.generateMasterKeys($seed);
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(neuron.publicKey);
        return addressObj._numericId;
    }


    /**
    */
    getPrivateKeyStringFromSeed=function($seed) {
        if(ANNE.SIGNA.crypto==null || ANNE.SIGNA.sig==null) return "";
        var masterNeuron = ANNE.SIGNA.crypto.generateMasterKeys($seed);
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(masterNeuron.publicKey);
        return addressObj.signPrivateKey;
    }


    /**
    */
    prepANNEAuthObjFromSeed=function($seed=null) {
        if ($seed==null || $seed==null) { 
            console.log("must pass seed here"); 
            return {sign_key:'', priv_key:'', pub_key:'', rs_addr:''};
        }
        var neuron = this.getNewAnneAccount($seed);
        return {sign_key:neuron.signPrivateKey, priv_key:neuron.agreementPrivateKey, pub_key:neuron.publicKey, rs_addr:neuron.address};
    }


    testKeys = function() {
        var masterNeuron = ANNE.SIGNA.crypto.generateMasterKeys(this.seedObj.seed);
        console.log(masterNeuron);
    }

    /**
    */
    getPubKeyStringFromSeed = function($seed) { //
        if(ANNE.SIGNA.crypto==null || ANNE.SIGNA.sig==null) return "";
        var masterNeuron = ANNE.SIGNA.crypto.generateMasterKeys($seed);
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(masterNeuron.publicKey);
        return addressObj.publicKey;
    }

    auditAcctIdFromPubKey = function($acctId, $pubKey) {
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey($pubKey);
        if (addressObj._numericId==$acctId) return true;
        return false;
    }


    /** 
     * Generate a random new seed via SIGNA passphrase gen or a poor man's random algo.
     * If optionsal seed string is passed it will use it, otherwise it is random.
     * @author Mr.Scatman
     */
    getNewAnneAccount=function($seed="") {
        var seed = $seed;
        if (seed=="" && ANNE.UTILS!=null) {  //we need a random seed!
            if (ANNE.UTILS!=null) {
                seed = ANNE.UTILS.randomHex(32)+ANNE.UTILS.randomHex(16)+ANNE.UTILS.randomHex(32);
                console.log(seed);
            }
            else ""+Math.ceil(Math.random()*1000000)+Math.ceil(Math.random()*7777)+Math.ceil(Math.random()*1000000)
        }
        var neuron = ANNE.SIGNA.crypto.generateMasterKeys(seed);
        var addressObj = ANNE.SIGNA.sig.Address.fromPublicKey(neuron.publicKey, "A");
        neuron.accountId = addressObj._numericId;
        neuron.seed = seed;
        neuron.address = addressObj._rs;
        neuron.nid = neuron.accountId;
        return neuron;
    }

    genKeyPopupMsg=function(){
        var parts = [];
        parts.push("anne.nodeAccountId = "+this._userNObj.nid);
        parts.push("anne.nodePubKey = "+this.pubKey);
        if (this.privKey=="" || this.privKey==undefined) {
            var aO = this.prepANNEAuthObjFromSeed(this.seedObj.seed);
            this.privKey = aO.priv_key;

            if (this.RS_ADDR=="") this.setRSADDR();
        }
        
        parts.push("anne.nodePrivKey = "+this.privKey);
        parts.push("anne.nodeSecret = "+this.seedObj.seed);
        return parts.join("<BR>");
    }

    
    /**
     * 
     * @returns 
     */
    genStandardKeyPopupMsg=async()=>{
        var parts = [];
        parts.push("<span style='display:block'>nid <b>"+this._userNObj.nid+"</b></span>");
        parts.push("<span style='display:block'>public key <b>"+this.pubKey+"</b></span>");
        var seedObj = await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
        if (seedObj==null) {
            return "We could not extract your key with this session.  Sorry.  Must investigate / try again after re-login.";
        }
        if (this.privKey=="" || this.privKey==undefined) {
            var aO = this.prepANNEAuthObjFromSeed(seedObj);
            this.privKey = aO.priv_key;
        }
        parts.push("<span style='display:block'>private key <b>"+this.privKey+"</b></span>");
        parts.push("<span style='display:block;padding-bottom:30px'>secret <b>"+seedObj.seed+"</b></span>");
        return parts.join("<BR>");
    }
    

    /**
     * 
     * @returns 
     */
    genMiningConfig=async()=>{
        var seedObj = await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
        if (seedObj==null) {
            return "We could not extract your key with this session.  Sorry.  Must investigate / try again after re-login.";
        }
        if (this.privKey=="" || this.privKey==undefined) {
            var aO = this.prepANNEAuthObjFromSeed(seedObj);
            this.privKey = aO.priv_key;
        }

        var parts = [];
        parts.push("anne.nodeSecret = "+seedObj.seed);
        parts.push("anne.nodeAccountId = "+this._userNObj.nid);
        return parts.join("<BR>");
    }


    /**
     * Get a real accountId for use in emulation.
     * We are emulating a neuron and want to use an address that conforms 
     * so we don't break any downstream checkers.     
     * @author Mr.Scatman                              
     */
    getRandomEmulatedANNEAccountId=function() {
        return this.getNewAnneAccount().accountId;
    }

    haveLSJustPing = async()=> {
        var result = await ANNE.API.ANNODEPING(aO);
        return;
    }

    /**
     * critical for our next non-TX annode interaction that requires proof of user.
     */
    updateNextProofId=async($id)=>{
        console.log("%c (updateNextProofId) id:"+$id, TALK.bgbigred);
        if (this._userLS.aauth.next_proof_id!==$id){
            this._userLS.aauth.next_proof_id=$id;
            alert("FIX THIS NOW JACKY");
        };
        await this.__saveANNE_JO(this._userLS);
    }


    /**
     * 
     * @returns 
     */
    updateLastSignin=async()=>{
        if (this._userLS==null) return;
        this._userLS.last_login=Date.now()/1000;
        await this.__saveANNE_JO(this._userLS);
    }


    /**
     */ 
    updateLastBalance=async()=> {
        if (this._userNObj.balanceObj==null) {
            console.log("no balance obj to update from.")
            return;
        }
        if (this._userLS.last_known_annies != undefined && this._userLS.last_known_annies!=null
            && this._userLS.last_known_annies==this._userNObj.balanceObj.annies) return;
        this._userLS.last_known_annies = this._userNObj.balanceObj.annies;
        await this.__saveANNE_JO(this._userLS);
        var anncoin = UTILS.displayAnneValue(this._userLS.last_known_annies);
        ANNE.A.quickBalanceUpdate(".bal-"+this._userNObj.temp_nid, anncoin+"<span style='font-size:1.6rem;'>&#x1d110;</span>");
    }


    /**
     * LN Nick will only update if it isn't set.
     */ 
     updateLSNickname=async($nObj, $override=null)=> {
        if ($nObj==null) return;
        var nick = $override;
        if ($override==null) nick = $nObj.text;
        if (this._userLS.nickname!=nick) {
            this._userLS.nickname = $nObj.text;
            await this.__saveANNE_JO(this._userLS);
        }
    }


    /**
     * TODO:  update all user objects
     */ 
     updateBackupFlag=async()=> {
        var history = await this._getHistory();
        var thisAnnodeHistory = history[ANNE._annodesLS.last_uri];
        var users = thisAnnodeHistory.users;
        if (users!=undefined) {
            var keys = Object.keys(users);
            for (var k=0; k < keys.length; k++) {
                var key = keys[k];
                var user = users[key];
                user.backed_up = 1;
                user.backed_up_on = new Date().getTime();

                if (key==this._userLS.public_nid) {
                    this._userLS.backed_up = 1;
                    this._userLS.backed_up_on = user.backed_up_on;
                }
                await this.__saveANNE_JO(user);
            }
        }
    }


    /**
     * 
     * @author Mr.scatman
     */
    updateLSSS=async($ess)=>{
        this._userLS.encrypted_ss = $ess;
        await this.__saveANNE_JO(this._userLS);
    }


    /**
     * 
     * @author Mr.scatman
     */
    updateLSActivated=async($activated=true)=>{
        //alert("updateLSActivated: "+$activated);
        this._userLS['activate_'+ANNE.network] = $activated;
        await this.__saveANNE_JO(this._userLS);
    }



    /**
     * helps with gui stuff before we go in too far to the processes
     * @author Mr.Scatman   
     */
    auditPw=async($pw) =>{
        var result = await this.getSeedWithPW($pw);
        if (result.status=="success") return true;
        return false;
    }

    getPrivateKeyHex=async()=> {
        if (this.unlock_pw!=undefined && this.unlock_pw!=null && this.unlock_pw=="") {
            return {status:"bad", reason:"no user pw in RAM"};
        }
        var result = await this.getSeedWithPW(this.unlock_pw);
        if (result.status=="success") {
            var aO = this.prepANNEAuthObjFromSeed(result.seed);
            aO.status="success";
            return aO;
        }
        return result;
    }


    /**
    * Note: Anyone with access to a user's browser AND/OR any rogue apps    
    * that use the ANNE login frame can take advantage of this method if    
    * fool a user who has selected this "auto password" option.  We must    
    * thus lock it down by domain and require that the user is redirected to
    * the ANNE.network wallet to control their wallet type (mode) w/ ANNE   
    * on a domain-by-domain basis.                
    * @author Mr.Scatman                                                               
    */
    getSeedWithDefaultPW=async()=> {
        alert("(getSeedWithDefaultPW) do we do this still?");
        var nid = this._userNObj.nid;
        if (ANNE.other.sha256==null) {
            console.error("you must include sha256.js");
            return;
        }
        var pw_text = ANNE.other.sha256(nid).substring(16); 
        return await this.getSeedWithPW(pw_text, "encrypted_oneclick", this._userLS.oneclick_count);
    }


    /**                                     
    */
    getSeedWithPW=async($pw, $hash_prop="encrypted", $bGENNID=false)=> {
        var user = null;
        var temp = localStorage.getItem(this.APP_KEY);
        if (temp==undefined || temp==null) return {status:"failure", reason:"no stored ANNE localStorage object."};;
        var user = JSON.parse(temp);
        if (user==undefined || user==null || user=="") {
            return {status:"failure", reason:"no stored user object."};
        }
        var encrypted_hash = user[$hash_prop];
        if (encrypted_hash==undefined || encrypted_hash==null) {
            return {status:"failure", reason:"no stored '"+$hash_prop+"'"};
        }

        var pw_hash = await this.hashPw($pw);
        var decrypted = this.decrypt_it(encrypted_hash, pw_hash);
        if (decrypted==undefined || decrypted=="") {
            return {status:"failure", reason:"invalid pw"};
        }
        var isValid = this.isValidSeedSOLO(decrypted);
        if (isValid==false) return {status:"failure", reason:"invalid pw"};
        
        var nid = "";
        if ($bGENNID==true && decrypted!=null && decrypted!="") {
             nid = this.getNidFromSeed({seed:decrypted});
        }
        else nid = user.public_nid;
        return {status:"success", seed:decrypted, nid:nid};
    }


    /*** hashPw ********************************************** 0x/xx/2019 ***
    * user seed encrypted in browser                                        *
    * not too critical here to use strong encryption.                       *
    * their pw will encrypt the wif so it doesn't sit in browser memory in  *
    * plaintext. pw must be used for every on-chain TX action. (or bundle)  *
    ******************************************************** Mr.Scatman ****/
    hashPw=async($pw)=> {
        var salt_post = "aSk3aCnDDqFm1zVd2hyk1tpxxrp4b24!x*d";
        var hash = $pw + salt_post;
        return hash;
    }


    /** 
     * take a salted seed (user can regen via their key or pw) and 
     * encrypted something.
     */
    hashSeedWithRecoveryPW=async($saltedKeyString, $something_to_encrypt)=> {
        if (ANNE.other.sha256==null) {
            console.error("you must include sha256.js");
            return;
        }
        var hash_as_key = ANNE.other.sha256($saltedKeyString);
        return this.encrypt_it($something_to_encrypt, hash_as_key)
    }


    /*** Recover the seed with the recovery hash and recovery password.
    * With a recovery hash and recovery pw, recover the key!              
    * This is presently used by onboarders who were onboarded via an email 
    * and they had a code.  The code + their email ties them to their
    * recovery hash (which they can retrieve by request).
    */
    recoverSeedWithRecoveryHashAndPW=async($saltedKeyString, $encryptedString)=> {
        if (ANNE.other.sha256==null) {
            console.error("you must include sha256.js");
            return;
        }
        var encryption_key = ANNE.other.sha256($saltedKeyString);
        return this.decrypt_it($encryptedString, encryption_key)
    }



    /**
     * We will do this run-time to encrypt PRIVATE DATA NEURONS.            
     * Only ANNE can and the user submitting them can decrypt them.          
     * THIS IS NOT the only encryption method.  User's can encrypt-for-self  
     * and they and only they can see that data unencrypted.  Private data   
     * neurons are a service used for selective disclosure.                                                                    
    */
    getSharedSecret=async()=>{
        if (this._userLS.encrypted_sk!=undefined) {
            var seedObj = this.seedObj;
            if (seedObj.status=="success") {
                var es = this._userLS.encrypted_sk;
                var ss = await this.recoverSeedWithRecoveryHashAndPW("x"+seedObj.seed+"y", es);
                if (ss!=undefined && ss!=null && ss.length) {
                    seedObj = null;
                    return {status:"success", seed:ss}; //seed here is the shared secret
                }
                else return  {status:"fail", reason:"We could not access your shared secret."}
            }
            return {status:"fail", reason:"The seed Obj used to get the shared secret is not set."}
        }

        //consider asking for the _getKey here, then retrieving the shared secret from
        //ANNE mem (restore localStorage goodies w/ existing wif pasted in)
        return {status:"fail", reason:"You do not have a shared secret.  This should be automatically made when you AAUTH with an ANNODE.  Contact Mr.Scatman."}
    }


    //encrypt msg here, this is what the annode stores on in the non-chain tbl.
    //on-demand this will be decrypted then re-encrpted for the recipient.
    genACHATEncryptMsg=async($sso, $aMsgParts, $sesh_id)=> {
        return $aMsgParts.join("|||");
        //var ephemeral_key = ANNE.other.sha256(shared_secretObj.seed+$sesh_id);
        //var encrypted = await this.encryptForANNODE(ephemeral_key, msgStr);
        //console.log("encrypted:"+encrypted);
        //return encrypted;
    }


    /**
     * 
     * @param {*} $msg 
     * @returns 
     */
    genGeneralProveIt = async($msg)=>{
        var shared_secretObj = await this.getSharedSecret();
        var nextProofId = ANNE.API.getNextProofId();
        var msgStr = this._userNObj.nid+":::"+nextProofId+":::"+$msg;
        var ephemeral_key = ANNE.other.sha256(shared_secretObj.seed+this._userNObj.nid+nextProofId);
        var encrypted = await this.encryptForANNODE(ephemeral_key, msgStr);
        console.log("encrypted:"+encrypted);
        return encrypted;
    }
    

    /**
     * Send a msg to the handshook ANNODE that is not a TX.
     * 0 - sender nid
     * 1 - next_msg_id
     * 2 - the msg (when decrypted parts are separated by |||)
     */
    genACHATProveIt =async($aMsgParts=null, $sesh_id=null)=> {
        var shared_secretObj = await this.getSharedSecret();
        var nextProofId = ANNE.API.getNextProofId();
        var msgStr = this._userNObj.nid+":::"+nextProofId+":::"+"";
        
        if ($aMsgParts!=null && $aMsgParts.length>0 && $sesh_id!=null) {
            msgStr += await this.genACHATEncryptMsg(shared_secretObj, $aMsgParts, $sesh_id)
        }
        console.log(msgStr);
        var ephemeral_key = ANNE.other.sha256(shared_secretObj.seed+this._userNObj.nid+nextProofId);
        var encrypted = await this.encryptForANNODE(ephemeral_key, msgStr);
        console.log("encrypted:"+encrypted);
        return encrypted;
    }

    

    /**
     * 
     */
    encryptRevealRequestQ = async($shared_secret, $q)=> {
        if ($shared_secret.length<32) {
            return "";
        }
        if ($q.length==0) return "";
        console.log("ss:"+$shared_secret+" from_nid:"+$relon.from_nid+" reln:"+$relon.reln);
        var ephemeral_key = ANNE.other.sha256($shared_secret+$relon.from_nid+$relon.reln);
        console.log("ephemeral_key:"+ephemeral_key);
        var encrypted = await this.encryptForANNODE(ephemeral_key, $q);
        console.log("encrypted:"+encrypted);
        return encrypted;
    }

    /**
     * 
     */
    encryptRelonToWithSS = async($shared_secret, $relon, $bUPDATERELON=true)=> {
        if ($relon.to_type < TO_TYPE_ENCR_ANNE) {
            $relon.to_s = "wrong totype. wiping the to_s for safety.";
            console.log("(this.encryptToWithSS)");
            return false;
        }
        if ($shared_secret.length<32) {
            return false;
        }
        console.log("ss:"+$shared_secret+" from_nid:"+$relon.from_nid+" reln:"+$relon.reln);
        var ephemeral_key = ANNE.other.sha256($shared_secret+$relon.from_nid+$relon.reln);
        console.log("ephemeral_key:"+ephemeral_key);
        var string_to_encrypt = $relon.to_s;
        var encrypted_to_s = await this.encryptForANNODE(ephemeral_key, string_to_encrypt);
        if ($bUPDATERELON==true) $relon.to_s = encrypted_to_s;
        console.log("encrypted:"+encrypted_to_s);
        return true;
    }

    //requester_shared_secret+from_nid_unsigned+next_msg_id_unsigned;
    encryptSKChatComs = async($msg, $participant_nids)=> {
        var shared_secretObj = await this.getSharedSecret();
        console.log(shared_secretObj);
        var ephemeral_key = ANNE.other.sha256($shared_secret+this._userNObj.nid+ANNE.API._AAUTH.next_proof_id);
        console.log("ephemeral_key:"+ephemeral_key);
        return await this.encryptForANNODE(ephemeral_key, $msg);
    }

    decryptSKChatComs = async($msg)=>{
    }


    /** 
     * We encrypted this FOR the annode, decrypt it out. */
     decryptSELFToSWithSS = async($shared_secret, $relon, $encryptedMsg)=> {
        if ($shared_secret.length<64) {
            console.log("secret too short");
            return false;
        }
        var ephemeral_key = ANNE.other.sha256($shared_secret+$relon.from_nid+$relon.reln)
        console.log("ephemeral_key:"+ephemeral_key);
        var decrypted_msg = await this.decryptFromSelfForANNODE(ephemeral_key, $encryptedMsg);
        console.log("decrypted:"+decrypted_msg);
        return decrypted_msg;
    }

    /**
     * we have a relon that has an encrypted to type.
     * Try to decrypt it best we can.
     */
    ascertainSelfEncryptionAndDecrypt=async($rObj)=> {
        if ($rObj.decrypted!=undefined && $rObj.decrypted!=null) {
            return {status:"success", decrypted:$rObj.decrypted};
        }

        //encrypted by us for the annode
        if ($rObj.r==MEM.PDN_SECRET.temp_nid) {
            var ssObj = await this.getSharedSecret();
            if (ssObj.status=="success") {
                var decrypted = await this.decryptSELFToSWithSS(ssObj.seed, $rObj, $rObj.to_s);
                $rObj.decrypted=decrypted;   
                return {status:"success", decrypted:decrypted}
            }
            return {status:"fail", reason:""};
        }

        //encrypted by the annode, for us, using the request to reveal
        else if ($rObj.r==MEM.PDN_ANSWER.temp_nid) {
            var ssObj = await this.getSharedSecret();
            if (ssObj.status=="success") {
                if ($rObj.sys_ref_txid==undefined) {
                    return {status:"fail", reason:"must have sys ref txid to decrypt here"};
                }
                var seed_relon = MEMG.lookupByTXID($rObj.sys_ref_txid);
                if (seed_relon==null) {
                    await CHASE.lookupTxid($rObj.sys_ref_txid);
                    seed_relon = MEMG.lookupByTXID($rObj.sys_ref_txid);
                    if (seed_relon==null) return {status:"fail", reason:"must have sys ref txid to decrypt here"};
                }
                if (seed_relon.decrypted!=undefined) return {status:"success", decrypted:seed_relon.decrypted};
                var decrypted = await this.decryptToSWithSS(ssObj.seed, seed_relon, $rObj.to_s);
                $rObj.decrypted=decrypted;   
                return {status:"success", decrypted:decrypted}
            }
            return {status:"fail", reason:""};
        }

        else {
            console.log("unhandled local decryption");
        }
    }


    /**
     * Using the current seedObj, decrypt our rels for this PDN.
     * 
     * NOTE: we assume we've already chased the sys_ref_relon at this point.
     * For primary use of this function we have done it in MEMG.getUserRevealsFor.
     * 
     * TODO: if the PDN annode doesn't match our current ANNODE,
     * notify the user.  We need to get a different shared secret.
     * We can do this auto if we have that object stored.
     * 
     * THIS WILL BE MORE EASY TO TEST WHEN WE ARE IN TESTNET WITH MULTIPLE ANNODES
     * TO HANDSHAKE WITH.
     */
    decryptPDNAnswerRelons = async($rels)=> {
        console.log(this._userLS);
        var ssObj = await this.getSharedSecret();
        if (ssObj.status=="fail") return $rels;
        for (var r=0; r < $rels.length; r++) {
            var rel = $rels[r];
            var sys_ref_relon = rel.sys_ref_relon;
            if (sys_ref_relon==undefined && rel.sys_ref_txid!=undefined) sys_ref_relon = MEMG.lookupByTXID(rel.sys_ref_txid);
            if (sys_ref_relon==null) continue;
            var decrypted = await this.decryptToSWithSS(ssObj.seed, sys_ref_relon, rel.to_s);
            rel.decrypted=decrypted;  
        }
        return $rels;
    }


    //inbound from annode use ephemeral pat of ss+reln+from_nid
    decryptToSWithSS = async($shared_secret, $relon, $encryptedMsg)=> {
        if ($shared_secret.length<64) {
            console.log("secret too short");
            return false;
        }
        console.log("ss:"+$shared_secret+" reln:"+$relon.reln+" from_nid:"+$relon.from_nid);
        var ephemeral_key = ANNE.other.sha256($shared_secret+$relon.reln+$relon.from_nid);
        console.log("ephemeral_key:"+ephemeral_key);
        console.log("encrypted msg:"+$encryptedMsg);
        var decrypted_msg = await this.decryptFromANNODE(ephemeral_key, $encryptedMsg);
        console.log("decrypted:"+decrypted_msg);
        return decrypted_msg;
    }

    decryptACHATmsg = async($msg, $shared_secret, $next_proof_id)=>{
        if ($shared_secret.length<64) {
            console.log("secret too short");
            return false;
        }
        var ephemeral_key = ANNE.other.sha256($shared_secret+$next_proof_id);
        console.log("ephemeral_key:"+ephemeral_key);
        console.log("encrypted msg:"+$msg);
        var decrypted_msg = await this.decryptFromANNODE(ephemeral_key, $msg);
        console.log("decrypted:"+decrypted_msg);
        return decrypted_msg;
    }


    PBKDF2 = async ($password, $salt, $iterations, $length, $hash, $algorithm ='AES-CBC') => {
        const keyMaterial = await window.crypto.subtle.importKey(
            'raw',
            encoder.encode($password),
            {name: 'PBKDF2'},
            false,
            ['deriveKey']
        );
        return await window.crypto.subtle.deriveKey(
            {name: 'PBKDF2', salt: encoder.encode($salt), $iterations, $hash},
            keyMaterial,
            {name: $algorithm, $length },
            false,
            ['encrypt', 'decrypt']
        );
    };


    /**
     * We are doing an ephemeral key that is ONE-TIME-USE only -- 
     * calculated via a hash of the shared
     * secret and relon pieces for a relon that will be encrypting the to_s.
     * 
     * The shared secret was randomly generated and is NOT tied to the user's
     * ANNE keys.
     */
    encryptForANNODE = async($ephemeralKey, $msg, $salt="xxxxxxxxxxxxxxxx")=> {
        const encoder = new TextEncoder();
        const toBase64 = buffer =>btoa(String.fromCharCode(...new Uint8Array(buffer)));
        var iv = new Uint8Array(16);
        const plain_text = encoder.encode($msg);//65536
        const key = await this.PBKDF2($ephemeralKey, $salt, 1024, 256, 'SHA-256');
        const encrypted = await window.crypto.subtle.encrypt(
            {name: "AES-CBC", iv },
            key,
            plain_text
        );
        var enc = toBase64(encrypted);
        console.log(enc);
        return enc;
    }

    //same as decryptFromANNODE but the ephemeral key is and msg are different.
    //we pass in here what we broadcast to the ANNODE.  this is just for us to get
    //our own stuff back out.
    decryptFromSelfForANNODE = async($ephemeralKey, $msg, $salt="xxxxxxxxxxxxxxxx")=> {
        return await this.decryptFromANNODE($ephemeralKey, $msg, $salt);
    }

    decryptFromANNODE = async($ephemeralKey, $encryptedMsg, $salt="xxxxxxxxxxxxxxxx")=> {
        const decoder = new TextDecoder();
        const fromBase64 = buffer =>Uint8Array.from(atob(buffer), c => c.charCodeAt(0));
        const encrypted = fromBase64($encryptedMsg);
        const salt = $salt;                 //i know, looks insecure.
        const iv = new Uint8Array(16);      //i know, looks insecure. //65536
        try {
            const key = await this.PBKDF2($ephemeralKey, salt, 1024, 256, 'SHA-256');
            var decrypted = await window.crypto.subtle.decrypt({name: "AES-CBC", iv }, key, encrypted);
            var result = decoder.decode(decrypted);
            console.log(result);
            return result;
          }
          catch(err) {
            console.log(err);
          }
        return "[** decrypt ERROR **]";
    }


    /**
     * 
     */
    PBKDF2 = async (password, salt, iterations, length, hash, algorithm ='AES-CBC') => {
        const encoder = new TextEncoder();
        const keyMaterial = await window.crypto.subtle.importKey(
            'raw',
            encoder.encode(password),
            {name: 'PBKDF2'},
            false,
            ['deriveKey']
        );
        return await window.crypto.subtle.deriveKey(
            {
            name: 'PBKDF2',
            salt: encoder.encode(salt),
            iterations,
            hash
            },
            keyMaterial,
            { name: algorithm, length },
            false, // we don't need to export our key!!!
            ['encrypt', 'decrypt']
        );
    };


    /**
     * 
     */
    signOwnershipMsg=async($childPrivateKey, $parentNID)=> {
        console.log("pre sign msg: "+$parentNID+"::ownsme", $childPrivateKey);
        var sigmsg = await this.signMessage($parentNID+"::ownsme", $childPrivateKey);
        console.log("sigmsg");
        return sigmsg;
    }


    /**
     * 
     */
    signMessage=async($msgRaw, $signKeyHex=null)=> {
        if ($msgRaw==null || $msgRaw=="") {
            console.warn("no msgRaw provided");
            return null;
        }
        if ($signKeyHex==null || $signKeyHex=="") {
            console.warn("no signKeyHex provided");
            return null;
        }
        var msgHex = ANNE.SIGNA.util.convertStringToHexString($msgRaw);
        return ANNE.SIGNA.crypto.generateSignature(msgHex, $signKeyHex);
    }

    //AnneCrypto ac = AnneCrypto.getInstance();
	//String jsonstring = JSON.toJsonString(payloadJO);
	//byte[]msgBytes= Convert.toBytes(jsonstring);
	//byte[]sig = ac.sign(privKey, msgBytes);
    //return Conversions.bytesToHex(sig);

    verifySignedJson = async($json, $sig, $pubkeyhex)=> {
        var msgRaw = JSON.stringify($json);
        console.log("sig:"+$sig);
        var msgHex = ANNE.SIGNA.util.convertStringToHexString(msgRaw);
        return ANNE.SIGNA.crypto.verifySignature($sig, msgHex, $pubkeyhex);
        //console.log("test1:"+test);
        //return await this.verifyADSig($json, $sig, $pubkeyhex)
        //return test2;
    }

    auditSignVerify = async($msgRaw, $signKeyHex, $pubKeySigner)=>{
        var msgHex = ANNE.SIGNA.util.convertStringToHexString($msgRaw);
        var sig = ANNE.SIGNA.crypto.generateSignature(msgHex, $signKeyHex);
        return ANNE.SIGNA.crypto.verifySignature(sig, msgHex, $pubKeySigner);
    }

    signHexMessage = async($hexMsg, $signKeyHex)=>{
        return ANNE.SIGNA.crypto.generateSignature($hexMsg, $signKeyHex);
    }

    auditTest = async()=>{
        //var randseed = UTILS.randHex(64);
        //var neuron = ANNE.SIGNA.crypto.generateMasterKeys(randseed);        
        //var aO = this.prepANNEAuthObjFromSeed(randseed);
        //var json = {"somevar":"test", "yourmom":"goestocollege"};
        //var msgRaw = JSON.stringify(json);
        //var verified = await this.auditSignVerify(msgRaw, aO.sign_key, neuron.publicKey);
        //console.log(verified);

        //var pubkeyhex="7444e77e865970388bc20009cc8f571bfe466367319b3db5435d4fabb807e37f";
        //var sig ="0fd1c596cab113cae2a3bdb257b1000426e4dba0b1d0f3905b1c9ccd983aa1074b927d58590a78e44c738fafbd7809e9ddc09aeb76c0c9d6908e992dcf6c424a";
        //var json = {"sender_nid":"7770353147211057540","ret_target_node_nid":"14270667173691107670","results":[{"swap_id":60,"swap_adid":"bcbf5c8ae061","swap_target_nid":"7770353147211057540","swap_ticker":"BCH","annies_per_alt":10000000000,"annies_available":4500000,"anne_height_last_status":154669,"swap_status":"awaiting_funds","take_id":55,"take_status":"awaiting_funds","take_requested_at_height":154669,"take_last_status_change_height":154669,"take_nid_to_credit":"8021634444898763749","take_altaddr":"bitcoincash:qpsxl2ueqha8lwt0pnz6mlggcmfwqznndghpgeengj","take_pubkey":"4413c698dbf7afc25ba3dd27b37cd189fb7edb309a5643b240c1e274cff9c83c","ispaystream":0,"moonths":0,"limitedofferuntil":0,"onboarding":0}],"txs":[],"msg":"no txids for found for bitcoincash:qpsxl2ueqha8lwt0pnz6mlggcmfwqznndghpgeengj","adid":"90ea7ecc2ead"};
        //var msgraw = JSON.stringify(json);
        //var msgHex = ANNE.SIGNA.util.convertStringToHexString(msgraw);
        return await this.verifyADSig(json, sig, pubkeyhex)
        //var msgBytes = ANNE.SIGNA.util.convertHexStringToByteArray();
        //var test = ANNE.SIGNA.crypto.verifySignature(sig, msgHex, pubkeyhex);
        //console.log("test1:"+test);
    }


    verifyADSig=async($payload, $sig, $pubkeyhex)=> {
        const jsonString = JSON.stringify($payload);
        //const encoder = new TextEncoder();
        //const msgBytes = encoder.encode(jsonString);
        //const msgHex = Array.from(msgBytes).map(byte => ('0' + byte.toString(16)).slice(-2)).join('');
        var test = ANNE.SIGNA.crypto.verifySignature($sig, jsonString, $pubkeyhex);
        console.log("test1:"+test);
        return test;
    }


    /**
     * 
     */
    verifyMsg=async($sig, $msgHex, $pubKeySignerHex)=> {
        return ANNE.SIGNA.crypto.verifySignature($sig, $msgHex, $pubKeySignerHex);
    }


    md5 = function($stringtohash) {
        if (typeof CryptoJS=="undefined") {
            if (md5!=undefined) {
                return md5($stringtohash);
            }
            console.warn("no md5 hashing algo found");
            return null;
        }
        const hash = CryptoJS.MD5($stringtohash);
        return hash.toString(CryptoJS.enc.Hex).toUpperCase();
    }


    /**
     * Quickly encrypt a message for anothre user 
     * This will return nonce and data
     */
    encryptMsgForPUBKEY=async($msg, $pubKeyViewer, $privKeySigner)=>{
        if ($msg==undefined || $msg==null) return null;
        if ($pubKeyViewer==undefined || $pubKeyViewer==null) return null;
        if ($privKeySigner==undefined || $privKeySigner==null) return null;
        
        var eObj = ANNE.SIGNA.crypto.encryptData($msg, $pubKeyViewer, $privKeySigner);
        eObj.data_string = ANNE.SIGNA.util.convertByteArrayToHexString(eObj.data);
        eObj.nonce_string = ANNE.SIGNA.util.convertByteArrayToHexString(eObj.nonce);
        return eObj;
    }
    

    /**
     * Decrypt a message sent from the annode.
     */
    decryptMsgEncryptedFromANNODE=async($nonce, $encryptedString, $pubKeySignerHex, $privKeyViewerHex=null)=>{
        var encryptedData = {};
        //use our privkey.
        if ($privKeyViewerHex==null) {
            $privKeyViewerHex = this.prepANNEAuthObjFromSeed(this.seedObj.seed).priv_key;
        }
        encryptedData.nonce = ANNE.SIGNA.util.convertHexStringToByteArray($nonce);
        encryptedData.data = ANNE.SIGNA.util.convertHexStringToByteArray($encryptedString);
        var data_buf = await ANNE.SIGNA.crypto.decryptData(encryptedData, $pubKeySignerHex, $privKeyViewerHex);
        var decrypted_msg = ANNE.SIGNA.util.convertByteArrayToString(data_buf);
        return decrypted_msg;
    }

    decryptJSONEncryptedFromANNODE=async($nonce, $encryptedString, $pubKeySignerHex, $privKeyViewerHex=null)=>{
        var encryptedData = {};
        //use our privkey.
        if ($privKeyViewerHex==null) {
            $privKeyViewerHex = this.prepANNEAuthObjFromSeed(this.seedObj.seed).priv_key;
        }
        encryptedData.nonce = ANNE.SIGNA.util.convertHexStringToByteArray($nonce);
        encryptedData.data = ANNE.SIGNA.util.convertHexStringToByteArray($encryptedString);
        var data_buf = await ANNE.SIGNA.crypto.decryptData(encryptedData, $pubKeySignerHex, $privKeyViewerHex);
        var decrypted_msg = ANNE.SIGNA.util.convertByteArrayToString(data_buf);
        if (decrypted_msg==null) return null;
        return JSON.parse(decrypted_msg);
    }


    /**
     * 
     * @param {*} $json 
     * @returns 
     */
    decryptRawTxJsonAttachmentMsg = async($json)=> {
        if ($json.recipient!=this._userNObj.nid) {
            //msg was sent from us.
            if ($json.sender!=USE._userNObj.nid) return "";
        }
        if ($json.attachment==undefined) return "";
        if ($json.senderPublicKey==undefined) return "";
        if ($json.attachment.encryptedMessage==undefined) return "";
        var em = $json.attachment.encryptedMessage;
        var msg = await this.decryptMsgEncryptedFromANNODE(em.nonce, em.data, $json.senderPublicKey);
        console.log(msg);
        return msg;
    }


    /**
     * altFundBoost
     */
    altFundBoost = function($index, $annies, $to_s) {
        return;
    }


    /**
     * return account in json
     * .annies,
     * .unconfirmedBalanceNQT,
     * .forgedBalanceNQT,
     * .guaranteedBalanceNQT
     * 
     * moving away from balanceNQT instead for annies.
     * should also rename other vars as ANNIES not NQT.
     * 
     * TO DO: UPDATE THIS WITH NEW SCAT BALANCE FEATURE I BUILT ON L1.
     * MEH.
     *
     * @param {string}
     * @param {boolean}
     * @author {Mr.Scatman}
     */
    checkBalance = async ($nid)=> {
        if ($nid==null || $nid==0 || $nid==undefined) {
            console.log("no nid, skip");
            return null;
        }

        //if no MEMG we can only use the ANNE.USER fake nObj (manually made) here.
        var nObj = null;
        if (ANNE.MEMG!=null) nObj = ANNE.MEMG.getOrMakeIfNew($nid);
        else if ($nid!=ANNE.USER._userNObj.nid) return null;
        else nObj = ANNE.USER._userNObj; 

        //if we checked this within the last 10 seconds, don't run it again.
        //wasteful.
        if (nObj.balanceObj!=undefined && nObj.balanceObj!=null) {
            if (nObj.balanceObj.lastchecked!=undefined) {
                let currentTime = Math.floor(Date.now() / 1000);
                var diff = currentTime - nObj.balanceObj.lastchecked; 
                if (diff < 10) {
                    console.log("skip, "+diff+"s is too soon. Find out who is calling this.");
                    return nObj.balanceObj; 
                }
            }
        }
        
        //ANNE.MEM_UDF.addSubHunt("checkBalance");  //this is always very fast.
        var json = await ANNE.API.getBalance($nid);

        if (json) {

            //set a new balanceObj if we don't yet have one.
            if (nObj.balanceObj==undefined || nObj.balanceObj==null) {
                nObj.balanceObj = {balance:0, annies:0, annecoin:0};
            }

            //set the last checked.
            nObj.balanceObj.lastChecked = Math.floor(Date.now() / 1000);
            json.lastChecked = nObj.balanceObj.lastChecked;

            //keyless or not keyless
            if (json.keyless!=undefined) {
                if (json.keyless=="y") nObj.keyless = true;
                else nObj.keyless=false; 
            }

            //we have results.
            if (json.annies!=undefined) {   //we should be using .annies now.
                json.annecoin = json.annies/100;
                nObj.balanceObj = json;
            }
            else {
                if (ANNE.USER.isActivated()==false) {
                    //save activated record as false.
                    await ANNE.USER.updateLSActivated(false); //make sure it is false.
                }
                return nObj.balanceObj;
            }
            
            //note,  balance hack already snatches api calls and updates .activate
            //for our LS.
            if (json.activated=="y") {
                if (ANNE.USER.isActivated()==false) {
                    //save activated record.
                    await ANNE.USER.updateLSActivated(true);
                }
            }

            //not activated.
            else {

                //our record says it is activated.
                if (ANNE.USER.isActivated()==true) {
                    //save activated record to false. (when would this ever happen???).
                    await ANNE.USER.updateLSActivated(false);
                }
            }

            if (nObj.pubKey==undefined) {
                nObj.pubKey=json.publicKey;
            }
            
            //updateUserBalance in localStorage (and dds/fields with .bal-TEMPNID)
            if ($nid==this._userNObj.nid) {
                await this.updateLastBalance()
            }

            //ANNE.MEM_UDF.closeSubHunt();
            return nObj.balanceObj;
        }

        nObj.balanceObj = {balance:0, annies:0, annecoin:0};
        return {balance:0, annies:0, annecoin:0};
    }
    

    /**
     * Get balance from an annode.
     * 
     * @author {Mr.Scatman}
     */ 
    getOurBalance = async()=> {
        this._userNObj.balanceObj = await this.checkBalance(this._userNObj.nid);
        await this.updateLastBalance(); //wallet handling
        if (this._userNObj.balanceObj.annies!=undefined) {
            var newspan = ANNE.GUI_STUFF.formatANNIES_to_ANNECoin(this._userNObj.balanceObj.annies);
            document.querySelectorAll(".balance-updatable").forEach(el=> {
                el.innerHtml=newspan;
            });
        }
        return this._userNObj.balanceObj;
    }


    /**
     * 
     * @param {*} $uri 
     * @param {*} $nid 
     * @returns 
     */
    getAcctByUri=async($uri, $nid)=> {
        var anne_history = await this._getHistory();
        var keys = Object.keys(anne_history);
        for (var k=0; k < keys.length; k++) {
            var uri = keys[k];
            if (uri!=$uri) continue;
            
            var uriUserObj = anne_history[uri];
            var keys2 = Object.keys(uriUserObj.users);
            for (var k2=0; k2 < keys2.length; k2++) {
                var k2nid = keys2[k2];
                if (k2nid!=$nid) continue;
                return uriUserObj;
            }
        }
        return null;
    }


    /**
     * Get all the user accounts stored in this EXT.
     * The user has chosen to manage verious accounts in ANNE..
     * 
     * @returns {Array}
     */
    getAccounts=async()=>{
        var anne_history = await this._getHistory();
        var accts = [];
        var keys = Object.keys(anne_history);
        var anne_user_nids = [];
        for (var k=0; k < keys.length; k++) {
            var urikey = keys[k];
            if (urikey=="default_unlock_pw") continue;
            var uriUserObj = anne_history[urikey];
            if (uriUserObj.users==undefined) continue;
            //looking for users within an uri.
            var keys2 = Object.keys(uriUserObj.users);
            for (var k2=0; k2 < keys2.length; k2++) {
                var k2nid = keys2[k2];  //nid are key of user acct.
                //we just need one user acct
                //same user acct could be in multiple annodes.
                if (anne_user_nids.indexOf(k2nid)!=-1) continue;
                var k2uo = uriUserObj.users[k2nid];
                accts.push(k2uo);
                anne_user_nids.push(k2nid);
            }
        }

        return accts;
    }


    /**
     * 
     * @param {*} $accts 
     */
    updateAllAcctsOfAccts=async($accts) =>{
        var anne_history = await this._getHistory();
        for (var a=0; a < $accts.length; a++) {
            var acct = $accts[a];
            var keys = Object.keys(anne_history);
            for (var k=0; k < keys.length; k++) {
                var urikey = keys[k];
                if (urikey=="default_unlock_pw") continue;
                var uriUserObj = anne_history[urikey];
                if (uriUserObj.users==undefined) continue;
                
                //looking for users within an uri.
                var keys2 = Object.keys(uriUserObj.users);

                for (var k2=0; k2 < keys2.length; k2++) {
                    var k2nid = keys2[k2];
                    //we just need one user acct
                    //same user acct could be in multiple annodes.
                    if (k2nid!=acct.public_nid) continue;

                    var userJO = uriUserObj.users[k2nid];

                    var keys3 = Object.keys(acct);
                    for (var k3=0; k3 < keys3.length; k3++) {
                        var k33 = keys3[k3];
                        if (k33=="annode_uri") continue;
                        if (k33=="encrypted") continue;
                        if (k33=="encrypted_sk") continue;
                        userJO[k33] = acct[k33];
                    }
                }
            }
        }
        localStorage.setItem('ANNE_history', JSON.stringify(anne_history));
    }


    /**
     * Just update an acct in all uris it has been used.
     * @param {*} $acct 
     * @returns 
     */
    updateAllAcctsOfAcct=async($acct) =>{
        var anne_history = await this._getHistory();
        var keys = Object.keys(anne_history);
        for (var k=0; k < keys.length; k++) {
            var urikey = keys[k];
            if (urikey=="default_unlock_pw") continue;
            var uriUserObj = anne_history[urikey];
            if (uriUserObj.users==undefined) continue;
            
            //looking for users within an uri.
            var keys2 = Object.keys(uriUserObj.users);

            for (var k2=0; k2 < keys2.length; k2++) {
                var k2nid = keys2[k2];
                //we just need one user acct
                //same user acct could be in multiple annodes.
                if (k2nid!=$acct.public_nid) continue;

                var userJO = uriUserObj.users[k2nid];

                var keys3 = Object.keys($acct);
                for (var k3=0; k3 < keys3.length; k3++) {
                    var k33 = keys3[k3];
                    if (k33=="annode_uri") continue;
                    if (k33=="encrypted") continue;
                    if (k33=="encrypted_sk") continue;
                    
                    userJO[k33] = $acct[k33];
                }
            }
        }

        localStorage.setItem('ANNE_history', JSON.stringify(anne_history));
        return;
    }



    /**
     * Get or make a new Alt balance object.  
     * If know alt_balances is passed in, check for and create it too.
     * @param {Object} $alt_balances 
     * @param {String} $ticker 
     * @param {String} $for_anne_user - nidlong string 
     * @param {String} $uniqueID 
     * @param {String} $pref_uri 
     * @returns {Object}
     */
    getOrMakeNewAltBalanceObj=async($alt_balances, $ticker, $for_anne_user, $uniqueID, $pref_uri="")=>{
        var nObj = MEMG.getNeuronByNid($for_anne_user);
        var alt_balances = $alt_balances;
        if ($alt_balances==null) alt_balances = {};

        if (ANNE.UTILS.isset(alt_balances)==false) {
            alt_balances =  await this.getAltBalances($for_anne_user);
            if (ANNE.UTILS.isset(alt_balances)==false) alt_balances = {};
        }

        if (ANNE.UTILS.isset(alt_balances[$ticker])==true) {
            var tickerObj = alt_balances[$ticker];
            nObj.alt_balances = alt_balances;
            if (tickerObj[$uniqueID]!=undefined) return tickerObj[$uniqueID];
        }

        var altBalObj = {
            balance:0, 
            aoe_balance:0, 
            for_anne_user:$for_anne_user,
            uniqueID:$uniqueID, 
            ticker:$ticker, 
            pref_uri:$pref_uri, 
            nickname:"",
            last_block_checked:0,
            activated:false
        };
        
        //within each nid then organized by ticker.
        var tickerObj = {};
        alt_balances[$ticker] = tickerObj;
        tickerObj[$uniqueID] = altBalObj;

        await this.setAltBalances(alt_balances);
        nObj.alt_balances = alt_balances;
        return alt_balances;
    }



    /**
     * get a specific alt balance by ticker and unique ID
     * @param {*} $alt_balances 
     * @param {*} $ticker 
     * @param {*} $uniqueID 
     */
    getAltBalance=function($alt_balances, $ticker, $uniqueID) {
        if (ANNE.UTILS.isset($alt_balances)==false) return null;
        if (ANNE.UTILS.isset($alt_balances[$ticker])==false) return null;
        var tickerObj = $alt_balances[$ticker];
        if (ANNE.UTILS.isset(tickerObj[$uniqueID])==false) return null;
        return tickerObj[$uniqueID];
    }


    /**
     * An alt balance updated.
     * Save it in localStorage.
     * @param {*} $nid 
     * @returns 
     */
    updateAltBalanceObj=async($nid)=>{
        var nObj = MEMG.getNeuronByNid($nid);
        if (ANNE.UTILS.isset(nObj.alt_balances)==false) return;
        return await this.setAltBalances(nObj.alt_balances);
    }
    
    setAltBalances=async($altBalances)=>{
        console.log($altBalances);
        var alt_balances = {};
        alt_balances.alt_balances = $altBalances;
        localStorage.setItem('alt_balances', JSON.stringify(alt_balances));
        return;
    }

    getAltBalances=async($nid)=>{
        var nObj = MEMG.getNeuronByNid($nid);
        if (ANNE.UTILS.isset(nObj.alt_balances)==true) return nObj.alt_balances;
        var alt_balances = localStorage.getItem('alt_balances');
        if (ANNE.UTILS.isset(alt_balances)==false) return null;
        return JSON.parse(alt_balances).alt_balances;
    }


    /**
     * @author {Mr.Scatman}
     */
    _getHistory=async()=> {
        var history = localStorage.getItem('ANNE_history');
        if (history!=undefined && history!=null) {
            var history_LS = JSON.parse(history);
            if (history_LS!=undefined & history_LS!=null) return history_LS;
        }
        return {};
    }


    /**
     * @author {Mr.Scatman}
     */
    _getSettings=async()=> {
        if (this._userNObj==null) return {};
        
        var settings = localStorage.getItem('ANNE_ext_settings_'+this._userNObj.nid);
        if (settings!=undefined && settings!=null) {
            var settingsLS = JSON.parse(settings);
            if (settingsLS!=undefined & settingsLS!=null) return settingsLS;
        }
        return {};
    }
    
    getPref=function($pref) {
        if (this._settingsLS==null) return null;
        if (this._settingsLS[$pref]!=undefined) return this._settingsLS[$pref];
        return null;
    }
    setPref=function($pref, $value) {
        this._settingsLS[$pref] = $value;
        localStorage.setItem('ANNE_ext_settings_'+this._userNObj.nid, JSON.stringify(this._settingsLS));
    }
    hasPrefTrue=function($pref) {
        if (this._settingsLS==null) return false;
        var val = this._settingsLS[$pref];
        if (val=="1" || val=="true" || val==true) return true;
        return false;
    }


    /**                                            
     * ANNE_history.  Key = last_uri.
     * {users:, annode_uri:, loggedout:}
     * @author {Mr.Scatman}
     */
    checkLoginFromANNE_history=async()=> {
        var history = localStorage.getItem('ANNE_history');
        if (history!=undefined && history!=null) {
            var hObj = JSON.parse(history);
            if (hObj!=undefined && hObj==null) hObj = {};
            var uObj = hObj[ANNE._annodesLS.last_uri];
            if (uObj!=undefined && uObj!=null && uObj.users!=undefined && uObj.users!=null) {
                var users = uObj.users;
                var loggedOut = uObj.loggedOut;
                
                //bail out with an array of users.
                if (loggedOut!=undefined && loggedOut==true) {
                    var keys = Object.keys(users);
                    var users_arr = [];
                    for (var k=0; k < keys.length; k++) {
                        var key = keys[k];
                        var user = users[key];
                        users_arr.push(user);
                    }
                    return users_arr;
                }

                var keys = Object.keys(users);
                var winner = null;
                var max_lt = -1;
                for (var k=0; k < keys.length; k++) {
                    var key = keys[k];
                    var user = users[key];
                    delete user.user_pw;
                    var last_time = user.last_time;
                    if (last_time==undefined) last_time=0;
                    if (last_time > max_lt) {
                        max_lt = last_time;
                        winner = user;
                    }
                }
                if (winner) {
                    return [winner];
                }
            }
        }
        return [];
    }


    /**
     * @author {Mr.Scatman}
     */
     __saveANNE_JO=async($user)=> {
        var keys = Object.keys($user);
        delete $user.user_pw;
        if (keys.indexOf("encrypted")==-1 || $user.public_nid==undefined || $user.public_nid=="") {
            return;
        }
        var ANNE_LS = {};
        ANNE_LS.user = $user;

        //we will sort by time in the default app.
        $user.last_time = new Date().getTime();

        //update the history of used logins on this annode.
        if (ANNE._annodesLS!=undefined && this.APP_KEY=="ANNE") {
            await this.updateHistory($user, ANNE._annodesLS.last_uri);
        }

        localStorage.setItem(this.APP_KEY, JSON.stringify($user));
        
        return;
    }


    /**
     * 
     * @param {*} $user 
     * @param {*} $bDEFAULTNID 
     * @returns 
     */
    getBestUserAcctName=function($user, $bDEFAULTNID=false) {
        if ($user==null) return "";
        if ($user.public_nid==undefined) {
            console.warn("THIS IS VERY BAD!!!!");
            if ($user.uniqueID!=undefined) {
                console.log($user.uniqueID+" has no public nid set in user obj???");
            }
            return "";
        }
        var nObj = MEMG.getOrMakeIfNew($user.public_nid);
        if (nObj.text!="") return nObj.text;
        if ($user.nickname!=undefined && $user.nickname!="") return $user.nickname;
        if ($bDEFAULTNID==true) return nObj.nid;
        return "";
    }


    /**
     */
    getDefaultName=function($default="") {
        if (this._userLS.nickname!=undefined && this._userLS.nickname!="") return this._userLS.nickname;
        if (this._userNObj.text!="") return this._userNObj.text;
        if ($default!="") return $default;
        return this._userNObj.nid;
    }


    /**
     * 
     * @param {*} $nid 
     * @returns 
     */
    getUserAcct=async($nid)=> {
        var users = await ANNE.getAcctsArray();
        for (var u=0; u < users.length; u++) {
            var user = users[u];
            if (user.public_nid!=$nid) continue;
            return user;
        }
        return null;
    }


    /**
     * user is flipping between accounts.
     * we have their pw, but we need to reload the ANNE localStorage & their seedObj.
     * @author {Mr.Scatman}
     */
    loadAccount=async($nid, $bLOCK=true)=>{
        this._settingsLS = await this._getSettings();
        this.seedObj = null;
        this.privKey = null;
        this.pubKey = null;

        var user = await this.getUserAcct($nid);
        if (user==null) return null;

        this.public_nid = user.public_nid;
        this.pubKey = user.pubKey;
        
        if (user.encrypted.length && user.count!=-1) {
            if (ANNE.API._AAUTH!=null) user.aauth = ANNE.API._AAUTH;
            localStorage.setItem(this.APP_KEY, JSON.stringify(user));  
            if (user.aauth!=undefined) ANNE.API.setAAUTHFromLS(user.aauth);
        }   

        if (ANNE.MEM!=null) {
            this._userNObj = ANNE.MEM.addInUserNeuron(user.public_nid, true);
        }

        this._userNObj.my_sponsored=[];
        this._userNObj.mine=true;
        this._userLS = user; 

        if (ANNE.BRAIN!=null) ANNE.BRAIN.setPrimarySpeaker(user.public_nid);

        await ANNE.MEMP.__postLoginCleanup();

        this.updateLSNickname(this._userNObj);

        if ($bLOCK==true) {
            this.lockApp(null);
            this.seedObj = null;
            this.unlock_pw = "";
        }

        return user;
    }
    


    /**
     * Get users array for annode endpoint,
     * not key/associated based array.
     * @author {Mr.Scatman}
     */
    getUserArrByAEP=async($aep)=> {
        var history = await this._getHistory();
        var thisAnnodeHistory = history[$aep];
        var users = thisAnnodeHistory.users;
        var arrUsers = [];
        if (users!=undefined) {
            var keys = Object.keys(users);
            for (var k=0; k < keys.length; k++) {
                var key = keys[k];
                var user = users[key];
                arrUsers.push(user);
            }
        }
        return arrUsers;
    }

    
    //only use this for self
    validSelfIdentity=function() {
        if (ANNE.USER._userNObj.balanceObj!=null) {
            if (ANNE.USER._userNObj.balanceObj.valid_real_pub_key==undefined) return false;
            if (ANNE.USER._userNObj.balanceObj.valid_real_pub_key==true) return true;
        }
        return false;
    }


    /**
     * Users have unique entries per node due to diff shared keys.
     * This function is just trying to extract all total nids from users 
     * to help us build easy dds for sends.
     * @returns 
     */
    getAllUsers=async()=> {
        var users_return = [];
        var history = await this._getHistory();
        var keys = Object.keys(history);
        var unique_nids = [];
        for (var k=0; k < keys.length; k++) {
            var uri= keys[k];
            var histObj = history[uri];
            var users = histObj.users;
            if (users==null || users==undefined) continue;
            var keys2 = Object.keys(users);
            for (var k2=0; k2 < keys2.length; k2++) {
                var user = users[keys2[k2]];
                if (unique_nids.indexOf(user.public_nid)!=-1) continue;
                //var obj = {public_nid:user.public_nid, pubkey:user.pubkey, nickname:this.getBestUserAcctName(user)};
                if (user.nickname==undefined || user.nickname=="") {
                    user.nickname = this.getBestUserAcctName(user)
                }
                users_return.push(user);
                unique_nids.push(user.public_nid);
            }
        }
        return users_return;
    }


    /**
     * Update history in local storage.
     * 
     * This will support the user's ability to swap between identities
     * and use the default ANNE wallet with ease.
     *
     * @author {Mr.Scatman}
     */
     updateHistory = async ($user, $annode_uri)=> {
        if ($user==null) return;
        if ($user==undefined) return;
        if ($annode_uri==null) return;
        if ($annode_uri==undefined) return;

        var history = localStorage.getItem('ANNE_history');
        var hObj = JSON.parse(history);
        delete $user.user_pw;
        
        //no history object make it
        if (hObj==undefined || hObj==null) {
            hObj = {};
            var users = {};
            users[$user.public_nid]=$user;
            hObj[$annode_uri]={'users':users,'loggedOut':false};
            localStorage.setItem('ANNE_history', JSON.stringify(hObj));
            return;
        }

        var thisAnnodeHistory = hObj[$annode_uri];
        
        //we have history, but not this annode_uri yet
        if (thisAnnodeHistory==undefined || thisAnnodeHistory==null) {
            var users = {};
            users[$user.public_nid]=$user;
            hObj[$annode_uri] = {'users':users, 'loggedOut':false};
            localStorage.setItem('ANNE_history', JSON.stringify(hObj));
            return;
        }

        var users = thisAnnodeHistory.users;

        //we have this annode_uri but no users..
        if (users==undefined || users==null) {
            var users = {};
            users[$user.public_nid]=$user;
            hObj[$annode_uri] = {'users':users, 'loggedOut':false};
            localStorage.setItem('ANNE_history', JSON.stringify(hObj));
            return;
        }

        var user = users[$user.public_nid];

        for (var u=0; u < users.length; u++) {
            delete users[u].user_pw;
        }

        //we have users but not THIS user.
        if (user==null || user==undefined) {
            users[$user.public_nid]=$user;
            hObj[$annode_uri]={users:users, 'loggedOut':false};
            localStorage.setItem('ANNE_history', JSON.stringify(hObj));
            return;
        }
         
        //we have this user, but update it to the most recent value.
        else {
            users[$user.public_nid]=$user;
            hObj[$annode_uri]={users:users, 'loggedOut':false};
            localStorage.setItem('ANNE_history', JSON.stringify(hObj));
            return;
        }
    }


    /**
     * Destruction of current account in data wallet.
     * @author Mr.scatman
     */
     destroyAccount=async($nid)=> {
        localStorage.removeItem("ANNE");        //log us out of this account.
        if ($nid!=this._userNObj.nid) {
            console.log("u can only destroy an acct you are logged into to"); 
            return;
        }
        var history = localStorage.getItem('ANNE_history');
        var hObj = JSON.parse(history);
        if (hObj==undefined) return;
        var thisAnnodeHistory = hObj[ANNE._annodesLS.last_uri];
        if (thisAnnodeHistory==undefined) return;
        var users = thisAnnodeHistory.users;
        if (users==undefined) return;
        var user = users[this._userNObj.nid];
        if (user==undefined) return;
        delete users[this._userNObj.nid]; //delete the user out.
        this.seedObj = null;
        this._userLS = null;
        this._userNObj = null;

        //update the hObj.
        localStorage.setItem('ANNE_history', JSON.stringify(hObj));
        return;
    }


    /**
     * Even if a rogue actor comes to a computer with the ext open,
     * and it is locked...
     * and they alter JS code in chrome to enact unlocking and use of the ext..
     * The seedObj is not set nor is the unlock_pw ---- so no seeds will be exposed
     * or able to be decrypted.  The user account is safe.
     * @author Mr.scatman
     */
    lockApp = async()=> {
        this.seedObj = null;
        this.unlock_pw = "";
        ANNE.bLOCKED=true;
        ANNE.bJUSTLOGGEDIN=false;
        ANNE.MEM_UDF.loadUnlock(true);
    }


    /**
     * Is this a legit neuron on ANNE chain?
     * @returns 
     */
    isActivated=function() {
        if (this._userNObj==null) return false;
        if (this._userNObj.ghost==true) return false;
        if (this._userLS==null) return false;
        if (this._userNObj.activated_in_mempool!=undefined) return true;
        if (this._userLS['activate_'+ANNE.network]==undefined) return false;
        var activated = this._userLS['activate_'+ANNE.network];
        return activated;
    }
    

    /**
     * Total destruction of data wallet.
     * @author Mr.scatman
     */
    nukeAllDataFromBrowserMem = async()=>{
        localStorage.removeItem("ANNE");
        localStorage.removeItem("ANNE_history");
        return;
    }



    //WHAT IS THIS>>>>>>>>>>>>>>?????

    
    /**
     * 
     * @author Mr.scatman
     * //nukeAllDataFromBrowserMem
     */
    logoutThisAccount = async()=> {
        localStorage.removeItem("ANNE");
        await this.logoutAccountByAnnode(ANNE._annodesLS.last_uri);
        location.reload();  // refresh page/app
    }

    logoutAccountByAnnode = async($annode_uri)=> {
        var history = localStorage.getItem('ANNE_history');
        var hObj = JSON.parse(history);
        if (hObj==undefined || hObj==null) return;
        var thisAnnodeHistory = hObj[$annode_uri];
        if (thisAnnodeHistory==undefined || thisAnnodeHistory==null) return;
        thisAnnodeHistory.loggedOut=true;
        localStorage.setItem('ANNE_history', JSON.stringify(hObj));
    }

}

const USER = new user_master ();

/************** (from file ./js/base/user.js)**/ 

const LINGO_TYPE_ANNE = 0;
const LINGO_TYPE_GRAPH = 1;
const LINGO_TYPE_COMMON = 2;

const TENSE_PRESENT = 1;
const TENSE_PAST = 2;
const TENSE_PAST_COMPLETE = 3;


//This is not semantic langauge via clumps.  This is just a basic output
//class for debugging or quick english display of things.
class anne_talk extends anne_singleton {
	constructor(){
		super();
		
		//THIS IS A SIMPLE WAY TO PROGRAM GRAMMAR utilizing base roles as storage.
		//different languages can have different grammar patterns.
		//one could spend a very long time in here building a robust langauge handler..
		//it can be done for AI/AGI use cases.
		
		//WE CAN STORE THE ROLES PROPERLY.
		//WE CAN VERBALIZE THEM.
		//In time, (with a fuller vocab) we can fully parse and conceptualize into the ANNE semantics.

		//ANNE training is done by experience/use..  The very basics are put in at inception to have something 
		//to work with.  The idea is parts-of-speachless (we use a higher concept of part of anne)
		//Language is a map into a higher-order brain protocol.
		//The Language may change, and with it the ways to communicate, but the concepts are stored using the 
		//higher-order unified brain protocol.  The same brain can speak 4 different languages and doesn't rqeuire
		//4 unique concepts repeating what a "dog" is defined as.  There are 4 differeent lang word maps to the 
		//same concepts (if it is indeed the same concept) 

		//clump definition pattern.
		this.def = new Object();
		this.def.infini = {pat_part_name:"infini", roles:[{role:"role_action", p:"to"}]};	//no use ot do doInfinitive w/ rule as it is just a "to"+
		this.def.mode = {pat_part_name:"mode", roles:["role_mode"]};
		this.def.obj = {pat_part_name:"obj", roles:[{role:'role_affected_thing', mod:'role_affected_thing_mod', det_func:this.___doThingDet}, 
													{role:'role_resulting_new_thing', mod:'role_resulting_new_thing_mod', det_func:this.___doThingDet},
													{role:'role_theme_thing', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_instrument_mod',
													{role:'role_instrument', mod:'role_theme_thing_mod', det_func:this.___doThingDet},
													'role_state', 'role_feelz', 'role_thought', 'role_fact']};

		this.def.manner = {pat_part_name:"manner", roles:[{role:"role_manner", modfunc:this.___doMannerEngShifter}]}
													
		this.def.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"to"}, 
				{role:'role_when', mod:'role_when_mod'},
				'role_start','role_end']
		};

		this.eng_pat_def = [
			this.def.infini,
			this.def.mode,
			this.def.obj,
			this.def.manner,
			this.def.wherewhen2
		]


		//action inst pattern
		this.actionInst = new Object();
		this.actionInst.subj = {pat_part_name:"subj", roles:['role_actor']};
		this.actionInst.mode = {pat_part_name:"mode", roles:['role_manner', 'role_mode']};
		this.actionInst.pred = {pat_part_name:"pred", roles:['role_action'], mod:'role_tense', modfunc:this.englishTensify}; //'role_reason'
		this.actionInst.obj = {pat_part_name:"obj", roles:['role_affected_thing', 'role_resulting_new_thing', 
										'role_state', 'role_feelz', 'role_thought', 'role_fact',
										'role_instrument_mod','role_instrument']};

		this.actionInst.theme_thing = {pat_part_name:"theme_thing", roles:['role_theme_thing_mod', 'role_theme_thing']};
		
		//to use role_source/path/goal in the same clump you will likely want 
		//more than one role_where_mod.  if so, tuff beans.  Create a nested neuron.
		this.actionInst.wherewhen = {
				pat_part_name:"wherewhen",
				roles:['role_source', 'role_path', 'role_goal',
				'role_where_mod', 'role_where', 'role_when_mod', 'role_when', 
				'role_start', 'role_end']
			};
		this.actionInst.wherewhen2 = {
			pat_part_name:"wherewhen2",
			roles:['role_source', 'role_path', 'role_goal',
				{role:'role_where', mod:'role_where_mod', p_default:"at", det_func:this.___doWhereDet}, 
				{role:'role_when', mod:'role_when_mod', p:"on"}
				   ,'role_start', 'role_end']
		}
		this.actionInst.pre_connectors = {
			pat_part_name:"pre_connectors",
			recurse:true,
			roles:['role_cause', 'role_cdx', 'role_prev']};
		this.actionInst.post_connectors = {
			recurse:true,
			pat_part_name:"post_connectors",
			roles:['role_effect', 'role_result', 'role_next']};

		//this.actionInst.tense,
		this.eng_pat1 = [
			this.actionInst.subj, 
			this.actionInst.mode,
			this.actionInst.pred, 
			this.actionInst.theme_thing, 
			this.actionInst.obj, 
			this.actionInst.wherewhen2,
			this.actionInst.pre_connectors,
			this.actionInst.post_connectors
		];


		this.type_pre = ["sys", "be", "param", "has", "awareness", "exper", "clump", "sys msg", "brain", "role", "episodic"];

		this.vowel=["a", "e", "i", "o", "u"];  //and sometimes y? lolz.
		
		this.init();
	}


	init = function() {
		this.bgbigred = this.redbig;
		this.bgblack = 'background:black; color:white';
		this.bgblackbold = 'background:black; color:white; font-weight:bold';
		this.bgblackyellow = 'background:#171717; color: #fefe17';
		this.bgblue = 'background:blue; color: white';
		this.bgbluebig = 'background:blue; color: white; font-size:1.1rem';
		this.bgerror = 'background: red; color: white; font-weight:bold; font-size:1.1rem;';
		this.bgfire = 'background:#F7C24A; color: #171717';
		this.bggray = 'background:#ccc; color: #171717';
		this.bggreen = 'background: green; color: white';
		this.bggreenbig = 'background: green; color: white; font-size:1.1rem;';
		this.bgivoryblue = 'background:ivory; color: blue';
		this.bglightblue = 'background:lightblue; color: #171717';
		this.bglightbluebold = 'background:lightblue; color: #171717; font-weight:bold';
		this.bglightgreen = 'background:lightgreen; color: #171717';
		this.bglightred = 'background:lightgreen; color: #171717';
		this.bgneg = this.bgred;
		this.bgneu = this.bgyellow;
		this.bgorange = 'background: orange; color: #171717';
		this.bgpink = 'background:pink; color: #171717';
		this.bgcyan = 'background:cyan; color: #171717';
		this.bgpos = 'background: #00ff00; color: #171717';
		this.bgpurple = 'background:purple; color: white';
		this.bgred = 'background: red; color: white';
		this.bgredbig = 'background: red; color: white; font-size:1.1rem;';
		this.bgyellow = 'background:#fefe17; color: black';
		
		this.bgheading_yellow = 'background:black; color: #fefe17; font-size:1.1rem';
		this.bgheading_black = 'background:whitesmoke; color: black; font-size:1.1rem';
		this.bgheading = 'background:ivory; color: #171717; font-size:1.4; font-weight:bold';

		this.fromMsg = 'background:ivory; color: #171717; font-weight:bold';
		this.funct = 'background:pink; color: #171717; font-style:italic';
		this.ghosts = 'background:white; color: blue';
		this.msg = 'background: green; color: white';
		this.msgNote = 'background:ivory; color:green; font-style:italic';
		this.pooltitle = 'background:ivory; color: black; font-weigt:bold; font-size:1.2rem;';
		this.title = this.pooltitle;
		
		this.bold_black = 'background:ivory; color: #171717; font-weight:bold';
		this.black = 'background:ivory; color: #171717';
		this.green = 'background: ivory; color: green';
		this.orange = 'background: ivory; color: orange;';
		this.redbig = 'background: red; color: white';
		this.yellow = 'background: ivory; color: yellow';
		
		this.part_cnt = this.bgpink;
		this.part_depth = this.bgyellow;
		this.part_id = this.bglightblue;
		this.part_full = this.bggreen;
		this.part_hex = this.bglightblue;
		this.part_mod = this.bglightred;
		this.part_mod_ph = this.bgorange;
		this.part_part = this.bglightgreen;
		this.part_sig = this.bglightblue;
		
		this.action = 'background:red;color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.aux = 'background:salmon; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.connector = 'background:purple; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.manner = 'background:indianred; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
		this.mode = 'background:firebrick; color: white; border-width:1px. border-style:solid; border-color:whitesmoke';
			
		this.tab0 ='';
		this.tab1 ='██';
		this.tab2 ='████';
		this.tab3 ='██████';
		this.tab4 ='████████';
		this.tab5 ='██████████';
		this.tab6 ='████████████';
		this.tab7 ='██████████████';
		this.tab8 ='████████████████';
		this.tab9 ='██████████████████';
		this.tab10='████████████████████';


		this.BLACK = 30;

		this.RED = 31;
		this.TRADE_RED = 311;
		
		this.GREEN = 32;
		this.TRADE_GREEN = 321;
		
		this.YELLOW = 33;
		this.YELLOW_i = 330;
		this.TRADE_YELLOW = 331;
		
		this.BLUE = 34;
		this.BLUE_i = 340;
		
		this.PINK = 35;
		this.CYAN = 36;
		this.WHITE = 37;

	}


	printf=function(...$args) {
		var lastColor = TALK.WHITE;
		var html = "";
		$args.forEach($arg => {
			if(typeof $arg==="string") html+="<span class='copr-"+lastColor+"'>"+$arg+"&nbsp;</span>";
			else {
				lastColor = parseInt($arg);
				if (lastColor==undefined) lastColor = TALK.WHITE;	
			}
		});
		return html;
	}

	/**
	 * 
	 * @param {*} $int 
	 * @returns 
	 */
	console_color_map = function($int) {
		$int = parseInt($int);
		if ($int==ANNE.TALK.BLACK) return this.bgblack;
		else if ($int==ANNE.TALK.RED) return this.bgred;
		else if ($int==ANNE.TALK.GREEN) return this.bggreen;
		else if ($int==ANNE.TALK.YELLOW) return this.bgyellow;
		else if ($int==ANNE.TALK.BLUE) return this.bgblue;
		else if ($int==ANNE.TALK.PINK) return this.bgpink;
		else if ($int==ANNE.TALK.CYAN) return this.bgcyan;
		else if ($int==ANNE.TALK.WHITE) return this.bgwhite;
		return this.bgblack;
	}
	
	printf_console=function(...$args) {
		var lastColor = this.bgblack;
		var strs = [];
		var colors = [];
		$args.forEach($arg => {
			if(typeof $arg==="string") {
				strs.push("%c"+$arg);
				colors.push(lastColor);
			}
			else {
				lastColor = this.console_color_map(parseInt($arg));
				if (lastColor==undefined) lastColor = this.bgwhite;	
			}
		});
		console.log(strs.join(""),...colors);
	}



	/**
	 * 
	 * @param {*} $vb 
	 * @param {*} $tense 
	 * @returns 
	 */
	 englishTensify = function($vb, $tense) {
		if ($tense=="past tense") {
			var rev = $vb.split("").reverse().join("");
			var c0 = rev[0];
			var c1 = "";
			if (rev.length>1) c1 = rev[1];
			if (c0=="e") return $vb+"d";
			if (c0=="y" && $text.length>1) return $vb.substring(0, $vb.length-1)+"ied";
			if (this.vowel.indexOf(c1)!=-1 && this.vowel.indexOf(c0)==-1) return $vb+c0+"ed";
			if (c0=="c") return $vb+"ked"; 
			return $vb+"ed";
		}
		else return $vb;
	}


	/**
	 * For a given role, what is the EC type we expect to fill this role?
	 * @param {String} $role_nid - nid long string 
	 * @returns 
	 */
	__getTopMatchRoleFromRole=function($role_nid) {
        if ($role_nid==MEM.role_action_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_theme_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_action.temp_nid) return MEM.ec_action.temp_nid;
        else if ($role_nid==MEM.role_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_actor.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_affected_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_posn_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_be_thing.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_state.temp_nid) return MEM.ec_state.temp_nid;
        else if ($role_nid==MEM.role_when.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_where.temp_nid) return MEM.ec_thing.temp_nid;
        else if ($role_nid==MEM.role_manner.temp_nid) return MEM.ec_manner.temp_nid;
        else if ($role_nid==MEM.role_mode.temp_nid) return MEM.ec_mode.temp_nid;
        else if ($role_nid==MEM.role_connector.temp_nid) return MEM.ec_expect_more_of_prev.temp_nid;
        else if ($role_nid==MEM.role_aux_action.temp_nid) return MEM.ec_expect_action.temp_nid;
        else if ($role_nid==MEM.role_thing_mod.temp_nid) return EM.ec_expect_thing_mod.temp_nid;
        else if ($role_nid==MEM.role_question.temp_nid) return MEM.role_question.temp_nid;
        else {
            var nOBJ = MEMG.getNeuronByTempNid($temp_nid);
            console.log("%c unhandled (__getTopMatchRoleFromRole) %c "+nOBJ.text, TALK.bgred, TALK.green);
        }
    }

	/**
	 * 
	 * @param {*} $balanceObj 
	 */
	printBalanceObj=function($balanceObj) {
		console.log($balanceObj);
		console.log("%c annecoin "+ANNE.UTILS.commas($balanceObj.annecoin), TALK.bgblackyellow);
		console.log("%c annies "+ANNE.UTILS.commas($balanceObj.annies), TALK.bglightblue);

		if ($balanceObj.avail_now_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_now_annies/100), TALK.bglightblue);
		}
		if ($balanceObj.avail_soon_annies!=undefined) {
			console.log("%c anncoin available now "+ANNE.UTILS.commas($balanceObj.avail_soon_annies/100), TALK.bglightblue);
		}
	}

	/**
	 * Shift an english word from the stored EC type to the desired EC type.
	 * 
	 * @param {anne_neuron} $roleNOBJ - the role we are shifting from. 
	 * @param {*} $valueNOBJ - the value of the role we are shiting from.
	 * @returns 
	 */
	__engShifter = function($roleNOBJ, $valueNOBJ) {
		console.log("(TALK.__engShifter)", $roleNOBJ, $valueNOBJ);

		//the top EC type of the role we are shifting from.
		var ec_top_desired_tnid = ANNE.TALK.__getTopMatchRoleFromRole($roleNOBJ.temp_nid);
		if (ec_top_desired_tnid==null) return;

		//the ec type we want to shift to.
		var ec_top_observed_tnid = $valueNOBJ.setECConceptualizingParent();
			
		//if the type is already right, return the text of the from.  No shift is needed.
		if (ec_top_desired_tnid==ec_top_observed_tnid) return $valueNOBJ.text;
		
		//we need to find the rule that shifts this concept from ec_top_observed_nid to the ec_top_desired_nid!
		for (var g=0; g < MEM.eng_grammar_rule.children.length; g++) {
			var ruleNOBJ = MEMG.getNeuronByTempNid( MEM.eng_grammar_rule.children[g]);
			var role_src = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_source.temp_nid);
			if (role_src==null) continue;
			if (role_src.t!=ec_top_observed_tnid) continue;	

			//we have a rule that starts from where our concept starts.
			//now we need to see if it converts to our desired
			var role_goal = MEMG.getRoleRelon(ruleNOBJ.outs, MEM.role_goal.temp_nid);
			if (role_goal==null) continue;
			if (role_goal.t!=ec_top_desired_tnid) continue;

			//we found a rule!
			//we need to now look for the suffixes that use this rule.
			for (var i=0; i < ruleNOBJ.ins.length; i++) {
				var in_rel = ruleNOBJ.ins[i];
				if (in_rel.type!=TYPE_HAS) continue;
				if (in_rel.r!=MEM.eng_grammar_rule.temp_nid) continue;

				//we found a suffix that uses this rule.
				var suffixNOBJ = MEMG.getNeuronByTempNid(in_rel.f);
				if (suffixNOBJ==null) continue;
				return TALK.__applyOutEngSuffix(suffixNOBJ, $valueNOBJ);
			}
		}

		console.log("%c could not find rule.");
	}


	/**
	 * 
	 * @param {anne_neuron} $suffixNOBJ 
	 * @param {anne_neuron} $valueNOBJ 
	 * @returns 
	 */
	__applyOutEngSuffix=function($suffixNOBJ, $valueNOBJ) {
		console.log("(TALK.__applyOutEngSuffix)");
		var suffixOBJ  = ANNE.BRAIN.NLP.LC._suffixMap.getByText($suffixNOBJ.text);
		if (suffixOBJ==null || suffixOBJ.cdx_rule_out==undefined) {
			//console.log("%c NO KNOWN OUTBOUND ENGLISH RULE FOR SUFFIX "+$suffixNOBJ.text, TALK.bgredbig);
			return $valueNOBJ.text+$suffixNOBJ.text;
		}
		return $suffixNOBJ.cdx_rule_out($valueNOBJ.text);
	}



	/**
	 * Take a manner role or roles and convert them to be proper adverbs if they are not adverbs.
	 * @param {Array} $manner_roles 
	 * @returns 
	 */
	___doMannerEngShifter = function($manner_roles) {
		console.log("(TALK.___doMannerEngShifter)", $manner_roles);
		var manner_text = [];
		for (var w=0; w < $manner_roles.length; w++) {
			var rel_manner = $manner_roles[w];
			var toNOBJ = MEMG.getNeuronByTempNid(rel_manner.t);
			var manner = ANNE.TALK.__engShifter(MEM.role_manner, toNOBJ);
			manner_text.push(manner);
		}
		return TALK.listifyStrArr(manner_text);
	}

	___doThingDet = function($thing_roles) {
		console.log("(TALK.___doWhereDet)", $thing_roles);
		var thing_text = [];
		for (var w=0; w < $thing_roles.length; w++) {
			var rel_where = $thing_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				thing_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					thing_text.push("the "+toNObj.text);
				}
				else thing_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(thing_text);
	}

	___doWhereDet = function($where_roles) {
		//console.log("(TALK.___doWhereDet)", $where_roles);
		var where_text = [];
		for (var w=0; w < $where_roles.length; w++) {
			var rel_where = $where_roles[w];
			var toNObj = MEMG.getNeuronByTempNid(rel_where.t);
			if (toNObj.ntype==NTYPE_CLASS) {
				where_text.push(ANNE.TALK.getDet(toNObj.text)+" "+toNObj.text);
			}
			else {
				if (toNObj.bes.indexOf(MEM.site.temp_nid)!=-1 && toNObj.text.toLowerCase().indexOf("the")!=0) {
					where_text.push("the "+toNObj.text);
				}
				else where_text.push(toNObj.text);
			}
		}
		return TALK.listifyStrArr(where_text);
	}

	/**
	 * This print out the meaning (infinitive form definition) of an action neuron.
	 * Whatever ROLES are stored in the clump will define the clump.
	 * 
	 * We can also show connotations via TYPE_BRAIN and assocs but it will be 
	 * better to do that in the dossier file for this concept where we can display them
	 * more cleanly.
	 * 
	 * @param {anne_neuron} $clumpNOBJ 
	 * @param {Object} $gpat 
	 * @param {Boolean} $bDEBUG 
	 * @returns 
	 */
	verbalizeClassActionDef = function($clumpNOBJ, $gpat=null, $bDEBUG=true, $tenseOverride="", $lvl=0) {
		if  ($gpat==null) $gpat = this.eng_pat_def;
		if ($clumpNOBJ==null) return;
		var output = "";
		var parts = [];
		var order = [];

        if ($bDEBUG==true) {
			console.log("%c (verbalizeClassActionDef) lvl:"+$lvl+" for clump"+$clumpNOBJ.text, TALK.bggreenbig);
		}

		var part_of_anne = null;
		var role_pre_var = "";

		//if this is a class and not present tense, rewire to verbalize the parent action instead.
		if ($clumpNOBJ.ntype==NTYPE_CLASS && $clumpNOBJ.tense!="present tense" && $tenseOverride=="") {
			role_pre_var = "_meaning_";
		}
		
		console.log("role_pre_var: "+role_pre_var);

		//will have to fix roles here for EC TYPE shifts.. hmm.
		var doModNest=function($mods, $lvl=0, $det_func=null, $bDETFIRSTONLY=false) {
			console.log("(doModNest) lvl"+$lvl);
			var modTxt = "";
			var keys = Object.keys($mods)
			for (var k=0; k < keys.length; k++) {
				var rObj = $mods[ keys[k] ];
				if (rObj.mods!=undefined) {
					modTxt += doModNest(rObj.mods, $lvl+1, $det_func, true)+" ";
					$det_func = null;
				}
				if (rObj.nameless!=undefined) continue;
				var this_text =  "";
				if (rObj.r!=MEM.role_connector.temp_nid && $det_func!=null) {
					if ($bDETFIRSTONLY==true && k==0) this_text = $det_func([rObj]);
					else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				} 
				else this_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
				modTxt+=this_text+" ";
			}
			return modTxt;
		}

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var pre_default = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre = "";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				var arr = $clumpNOBJ[role_pre_var+role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null) {
					mod_arr = $clumpNOBJ[role_pre_var+mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";
				order = order.concat(arr);


				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {

						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}

					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				//mod mods the WHOLE phrase if there are multiples.
				else if (role=="role_manner" && modfunc!=null) {
					to = modfunc(arr);
				}

				//{role:'role_where', mod:'role_where_mod', p:"to"},
				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}


				//new method of mods..
				else if (arr[0].mods!=undefined) {

					console.log("mods in role "+role, arr[0].mods);
					mod_arr = arr[0].mods;
					var modTxt = doModNest(mod_arr, 0, det_func);

					//the base role that has these mods (in eng pat we put the mod BEFORE)

					//note, this should not be an array, only one with our mods method.
					//will pinch it off here -- (will write a single verbalizeRole in a bit)
					to = modTxt+TALK.verbalizeRolesArr([ arr[0] ]);	

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);

					//if a sig, don't add the default if we have it
					if (pre_default!=null && (role=="role_where" || role=="role_when")) {
						//var modNOBJ = MEMG.getNeuronByTempNid(mod_arr[0].t); //should also be role_text already..
						//console.log(modNOBJ);
						var sig = ANNE.BRAIN.NLP.LC._getPhSignals(mod_arr[0].role_text);
						if (sig==null) pre = pre_default;
						else if (sig.mod !=true) pre = pre_default; 
					}
				}

				else {
					if (det_func!=null) to = det_func(arr);
					else to = TALK.verbalizeRolesArr(arr);
				}
				

				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		console.log(output);
		$clumpNOBJ.eval = output;
		if ($lvl!=0) return output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		var tense = $clumpNOBJ.tense;
		if ($tenseOverride!="") tense = $tenseOverride;

		return new anne_babytalk(output, order, part_of_anne, $clumpNOBJ.tense);
	}

	

    /**
     * Rudimentary verbalizer of _roles via a gpat (for lang or maodality specific orderings)
	 *
	 * @param {anne_neuron} $clumpNOBJ - the clump we want to print 
	 * @param {Object} $gpat - the language pattern we want to use
	 * @param {Array} $skipRoleNames - any roles we want to skip
     * @param {Boolean} $bDOSENTENCE - print in sentence mode
	 * @param {Boolean} $bDEBUG - degug the printing.
	 *
	 * @returns {anne_babytalk} - returns .output of sentence and .order of role tnids in the order they were printed from the gpat.
	 */
	 verbalizeInstActionClump=function($clumpNOBJ, $gpat=null, $skipRoleNames=[], $bDOSENTENCE=true, $bDEBUG=true){
        if ($gpat==null) $gpat = this.eng_pat1;
        var output = "";
		var parts = [];
		var order = [];
        if ($bDEBUG==true) console.log("%c print for clump"+$clumpNOBJ.text+" %cdosentence:"+$bDOSENTENCE, TALK.bggreenbig, TALK.bgblackyellow);
			
		var part_of_anne = null;	//EC type

		for (var p = 0; p < $gpat.length; p++) {
            var part = $gpat[p];
            var roles = part.roles;
			if ($bDEBUG==true) console.log("%c"+part.pat_part_name, TALK.bgblack);

			//loop the roles in the part roles array.
			//if the role is a string, it is just simple.
			//if it is an object, we have .p, .mod, .modfunct, and .role props.
            for (var r=0; r < roles.length; r++) {
                var role = roles[r];
				var pre = "";
				var mod = "";
				var modfunc = null;
				var det_func = null;
				var pre_default = "";
				if (typeof role !=="string") {
					pre = role.p;
					if (pre==undefined) pre="";
					mod = role.mod;
					if (mod==undefined) mod = null;
					modfunc = role.modfunc;
					if (modfunc==undefined) modfunc =null;
					det_func = role.det_func;
					if (det_func==undefined) det_func = null;
					pre_default = role.p_default;
					if (pre_default == undefined) pre_default = null;
					else {
						if (mod==null) pre = pre_default;	
					}
					role = role.role;	//set this last or we lose our others.
				}

				if ($skipRoleNames.length && $skipRoleNames.indexOf(role)!=-1) continue;

				var arr = $clumpNOBJ[role];
                if (ANNE.UTILS.isset(arr)==false) continue;

				var mod_arr = null;
				if (mod!=null && mod!="") {
					mod_arr = $clumpNOBJ[mod];
					if (ANNE.UTILS.isset(mod_arr)==false) {
						mod=null;
						mod_arr=null;
					}
				}

                var to = "";

				order = order.concat(arr);

				//if this is a role_action.. and the the tense is present.
				if (role=="role_action") {
					var actionNOBJ = MEMG.getNeuronByNid(arr[0].to_s);
					if ($clumpNOBJ.shift_lang==true) {
						//yet we have a role_tense shifter..  apply the tense.
						to = TALK.englishTensify(actionNOBJ.text, $clumpNOBJ.tense);
					}
					else to = TALK.verbalizeRolesArr(arr);

					part_of_anne = actionNOBJ.getECTop();
				}

				else if (modfunc!=null) {
					to = modfunc(arr, mod);
				}

				else if (mod_arr!=null) {
					var newmod = TALK.verbalizeRolesArr(mod_arr);
					to = newmod+" "+TALK.verbalizeRolesArr(arr);
				}

				else if (det_func!=null) {
					to = det_func(arr);
				}
		 
				else {
					to = TALK.verbalizeRolesArr(arr);
				}
				
				if (to=="") continue;	//do period still. whatever.
				if (pre!="" && pre!=undefined && pre!=null && pre.length>0) {
					parts.push({pre:pre, str:to, output:pre+" "+to, role:role, roles:arr});
					to = pre+" "+to;
				}
				else parts.push({pre:pre, str:to, output:to, role:role, roles:arr});
			
			
				if ($bDEBUG==true) console.log("%c"+role+" %cpre:"+pre+" %cto:"+to, TALK.bgblackyellow, TALK.bgred, TALK.bggreen);
			}
        }

		for (var p=0; p < parts.length; p++) {
			if (p == (parts.length-1)) output=output+parts[p].output+".";
			else output=output+parts[p].output+" ";
		}

		//make a sentence out of it?
		if ($bDOSENTENCE==true) {
			output = output.charAt(0).toUpperCase() + output.slice(1);
		}	
		console.log(output);
		$clumpNOBJ.eval = output;

		if ($bDEBUG==true) console.log("%c"+output, TALK.bggreen);

		return new anne_babytalk(output, order, part_of_anne);
    }


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	verbalizeRolesArr = function($arr) {
		if ($arr==undefined || $arr==null || $arr.length==0) return "[ERROR]";
		if ($arr.length==1) {
			var rObj = $arr[0];
			if (rObj.nameless!=undefined) return "";
			if (rObj.role_text!=undefined) return rObj.role_text;
			rObj.role_text = MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			return rObj.role_text;
		}
		var str = "";
		var froms = [];
		var conj = "and";

		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			if (froms.length==0) froms.push(rObj.f);
			else if (froms.indexOf(rObj.f)!=-1) froms.push(rObj.f);
			else conj="or";
		}
		for (var r=0; r < $arr.length; r++) {
			var rObj = $arr[r];
			var role_text =  MEMG.__MASTER_DISPLAY_RELON_TO(rObj);
			str += role_text;
			rObj.role_text = role_text;
			if (r==$arr.length-2) str=str+" "+conj+" ";
			else if (r==$arr.length-1) ;
			else str=str+", ";
		}
		return str;
	}


	/**
	 * 
	 * @param {*} $arr 
	 * @returns 
	 */
	listifyStrArr = function($arr) {
		if (ANNE.UTILS.isset($arr)==false) return "";
		if ($arr.length==1) return $arr[0];
		if ($arr.length==2) return $arr[0]+" and "+$arr[1];
		var finalstr = $arr[0];
		for (var a=1; a <$arr.length; a++) {
			if (a==$arr.length-1) finalstr = finalstr+", "+$arr[a];
			else finalstr = finalstr+" and "+$arr[a];
		}
		return finalstr;
	}


	
	/** 
	 * print simple relon tags.
	 * 
	 * This TALK class is an utter disaster.. changed a gazillion times.
	 * Use this to do basic printing.  We have lots of rework to do just
	 * for this basic talk.  Would be better to get full PAS understanding.
	 */ 
	 simpleRelonTags = function ($rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}
		if ($rObj.type==TYPE_AWARENESS) {
			var tt = TARGET_NID;
			if ($rObj.to_type==TO_TYPE_TXID) tt = TARGET_TXID;
			return this.verbalizeForAwareness($rObj, true);
		}

		var typeText = this.getTypeName($rObj.type, true, "rel-type");

		//willy wonka from name
		var fromNOBJ = null;
		var fromText = "";
		if ($rObj.from_nid!=null) {
			fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		}
		else if ($rObj.pending_from!=null) {
			fromNOBJ = MEMG.getNeuronByPendingNid($rObj.pending_from);
			if ($rObj.pending_from_debug!=undefined && $rObj.pending_from_debug.length>0) fromText = $rObj.pending_from_debug;//+"("+$rObj.pending_from+")";
			else fromText = $rObj.pending_from;
		}
		if (fromNOBJ==null) {
			console.log("NO FROM OBJ??");
			console.log($rObj);
			$rObj.error = true;
			return "[[error, bad relon]]";
		}
		else {
			if (fromNOBJ.emulated) {
				console.log("NO RELONS FROM EMULATED NEURONS -- they are not real!");
				return;
			}
			fromText = fromNOBJ.bestNameHTML();
		}
		var fromClass = this.getNeuronDisplayClass(fromNOBJ);

		//we want the name of the here even tho the from is actually
		//our nid (in the cmd relon)
		if ($rObj.debug_name!=undefined) {
			fromText = "(MAKE) "+$rObj.debug_name;
			fromClass = "neuron-ghost";
		}

		//reln text
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		if (relnNOBJ==null) {
		    console.log("problem with the reln. It must exist in local mem");
			$rObj.error = true;
		    return "[[ERROR in RELN "+$rObj.reln+"]]"
		}
		var relonText = relnNOBJ.text;
		var relnClass = "";
		if (relnNOBJ.nid==MEM.hanai_in.nid || relnNOBJ.nid==MEM.hanai_in_all.nid) relnClass = "neuron-hanai";
		else if (relnNOBJ.nid==MEM.hanai_out.nid)  relnClass = "neuron-hanai";
		else relnClass = this.getNeuronDisplayClass(relnNOBJ);
		

		if ($showfromNid==true) {
			fromText+="("+fromNOBJ.nid+")";
		}

		//to type
		var toText = this.verbalizeRelonToByToType($rObj, $bCOLORIZEBYTYPE);
		var ttypemsg = "";
		if ($ttype!=-1) ttypemsg = TXB.printType($ttype);
		if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" "+relonText+"&nbsp;<b>"+toText+"</b></span>";
		}
		else if ($ttype==ANNE.TXB.TXTYPE_MAKE_GLOBAL_LOCATION) {
			return "<span>&nbsp;"+typeText +" "+ttypemsg+" MAKE LOCATION <b>&nbsp;"+toText+"</b>&nbsp;in&nbsp;"+fromText+"</span>";
		}

		//to stuff
		var toClass = "rel-to";
		var toNOBJ = null;
		if ($rObj.to_type==TO_TYPE_ANNE) {
			if ($rObj.pending_to_s!=undefined) {
				var parts = $rObj.pending_to_s.split("_");
				if (parts.length>1) {
					toNOBJ = MEMG.getNeuronByNid(parts[1].trim());
					if (toNOBJ!=null) {
						toClass = this.getNeuronDisplayClass(toNOBJ);
					}
				}
			}
			else { 
				toNOBJ = MEMG.getNeuronByNid($rObj.to_s);
				if (toNOBJ!=null) {
					toClass = this.getNeuronDisplayClass(toNOBJ);
				}
			}
		}

		return "<span class='q-part "+fromClass+"' title='"+fromNOBJ.nid+"'>"+fromText+"</span>"+typeText+"<span class='q-part "+relnClass+"'>"+relonText+"</span><span class='q-part "+toClass+"'>"+toText+"</span><span class='q-part anne-tx-type'>*"+ttypemsg+"TX*</span>";
	}

	/**
	 * Helpful display to colorize things consistently thruout the app.
	 * This took a bit of time to arrive at -- for some time I tried to colorize
	 * class neurons red, instances blue, text green...
	 * but have now settled on ghost blue, anne yellow, text green, self red (maybe orange later), and hanai pink.
	 */
	getNeuronDisplayClass=function($nObj) {
		if ($nObj==null) {
			return "neuron-ghost";
		}
		
		if ($nObj.nid==MEM.USER._userNObj.nid) return "neuron-self";
		if ($nObj==undefined || $nObj==null) {
			return "neuron-ghost";
		}
		if ($nObj.ghost==true) return "neuron-ghost";
		else if (ANNE.isANNE($nObj.owned_by)==true) return "neuron-anne";
		else {
			if ($nObj.owned_by==ANNE.USER._userNObj.nid) {
				return "neuron-self";
			}
			else {
				return "neuron-other"  //black for someone else's neuron (not ANNE's).
			}
		}
	}



	/**
	 * 
	 * @param {*} $realTXID 
	 * @param {*} $relon 
	 * @param {*} $sep 
	 * @param {*} $bCOLORIZE 
	 * @returns 
	 */
	genShortRelonTextByTXID=function($realTXID, $relon=null, $sep="&nbsp;»&nbsp;", $bCOLORIZE=false) {
		var relon = $relon;
		if ($relon==null) relon = MEMG.lookupRelonByTXID($realTXID);
		if (relon==null) return "";
		var fNObj = MEMG.getNeuronByNid(relon.from_nid);
		var rNObj = MEMG.getNeuronByNid(relon.reln);
		var tNOBJ = null;
		if (relon.t) tNOBJ = MEMG.getNeuronByNid(relon.to_s);
		var to = "";
		if (tNOBJ!=null) to = tNOBJ.bestNameHTML();
		else if ($relon.to_type==TO_TYPE_TXID) to="{url}";
		else if ($relon.to_type==TO_TYPE_BLOB) to="{blob}";
		else to = relon.to_s;
		var from = fNObj.text;
		if (from=="") from = fNObj.nid;
		if ($bCOLORIZE==true) {
			return "<b>"+from+"</b>"+$sep+"<em class='text-red'>"+rNObj.text+"</em>"+$sep+"<b class='text-green'>"+to+"</b>";
		} 
		return "<b>"+from+"</b>"+$sep+"<em>"+rNObj.text+"</em>"+$sep+"<b>"+to+"</b>";
	}


	/**
	 * Uses the Q to print a friendly gui.
	 */
	printSanitizedAlertCmds=function($updateElID, $aQ, $bPRINTBALANCE=true) {
		var msgDiv = ANNE.A.setHTML($updateElID, ""); //qrb_txs_list_div
		
		console.log($aQ);
		
		//append bar to qrb_progress div
		var progDiv = document.createElement("div");
		progDiv.classList = "bg-light tx-proggy";
		progDiv.setAttribute("id", "tx_totals");
		ANNE.A.append('qrb_progress', progDiv);

		var progressBar = document.createElement("div");
		progressBar.setAttribute("id", "tx_progress_bar");
		progressBar.classList = "proggy-bar positive";
		progressBar.style.display = "inline-block!important";

		ANNE.A.append(progDiv, progressBar, true);

		for (var c=0; c < $aQ.length; c++) {
			var cmdObj = $aQ[c];
			var txtlbl = "";

			if (cmdObj.cmds_special.length && cmdObj.cmds_special.indexOf(TXB.CMD_POPULATE_WITH_VARS_ONLY_NO_SEND)!=-1) {
				continue;//no need to display this in the gui.
			}
			
			if (cmdObj.relon!=undefined && cmdObj.relon.warning!=undefined) {
				ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q bg-warning", "tx", c+"x_warn",  cmdObj.relon.warning);
			}

			var msg = "";
			if (cmdObj.msg!="" && cmdObj.relon!=null && cmdObj.relon.debug_name==undefined) {
				msg = cmdObj.msg;
			}
			else if (cmdObj.relon!=undefined) {
				//$rObj, $ttype=0, $bCOLORIZEBYTYPE=true, $showfromNid=false
				msg = ANNE.TALK.simpleRelonTags(cmdObj.relon, cmdObj.TXTYPE, false);
			}

			
			if (cmdObj.relon!=undefined && cmdObj.relon!==null) {
				if (cmdObj.relon.type==TYPE_PARAM) {
					msg = msg +" <span class='tiny-text'>PARAM setter</span>";
				}
			}

			if (cmdObj.relon!=undefined && cmdObj.relon.sys_ref_txid!=undefined) {
				msg = msg +" <span class='tiny-text'>&nbsp;depends on "+cmdObj.relon.sys_ref_txid+"</span>";
			}

			if (cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_CHILD_TO_CMD)!=-1 
			|| cmdObj.cmds_special.indexOf(TXB.CMD_ENCRYPT_SELF_TO)!=-1) {
				msg = msg +"<span class='tiny-text'>&nbsp;ENCRYPTED AFTER NID IS ACQUIRED</span>";
			}

			var ico = "";
			ico = UTILS.formatNumber(cmdObj.totalAnnies/100)+ANNE.GUI_STUFF.getAnneImgHtml('16px');

			//friendly ico display in the q
			var ico3 = "";
			if (cmdObj.relon!=undefined) {
				ico3 = ANNE.TXB.getTxICOByType(cmdObj.TXTYPE, cmdObj.relon, "16");
			}

			var txtlbl = "<span class='q-relon-link'>tx #"+(c+1)+"</span>"+ico3;
			var total = "<span class='q-flexright' style='text-align:right'>"+ico+"</span>";

			var finalmsg = txtlbl+msg+total;

			ANNE.A.ezAlert(msgDiv, "bg-txid-alert-q", "tx", c+"x",  finalmsg);
		}
	}


	/**
	*/
	printNids = function($nids, $title="") {
		if ($title!="") console.log("%c"+$title, ANNE.TALK.bgbigred);
		
		for (var n=0; n < $nids.length; n++) {
			var nObj = MEMG.getNeuronByNid($nids[n]);
			console.log("n="+n+" "+nObj.text);
		}	
	}

	
	/**
	 *
	 */
	updateSantizedAlertCmd=function($updateElID, $cmdObj, $current_index, $total, $badgeEl="") {
		var txid = $cmdObj.broadcast_txid;
		var percentFloat = ($current_index+1)/$total;
		var bFINAL = false;
		var remain = $total-($current_index+1);
		
		if (remain>0) ANNE.A.setHTML($badgeEl, remain);
		else ANNE.A.setHTML($badgeEl, "");
		
		if ($current_index+1==$total) bFINAL=true;

		var percent = Math.ceil(percentFloat*100);
		ANNE.A.setStyle("tx_progress_bar", {width:percent+"%"}, true);
		var msgDiv = document.getElementById($updateElID);

		//try to track with the element getting updated.
		
		if (($cmdObj.skipped==undefined || $cmdObj.skipped==false) && txid!=undefined && txid.length>0) {
			var msg = "";
			if ($cmdObj.msg=="") msg = "success! ";
			else msg = $cmdObj.msg;
		
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-green-themeDARK", "bg-txid-alert-q");
			
			//update the txid #n with the txid ico.
			var txidcol = tx_alert.querySelector(".q-relon-link");
			if (txidcol!=null && txid!="") {
				ANNE.MEM_UDF.makeRelonViewLink(txidcol, txid, true, 12, true);
			}

			//if this was the last $cmdObj, print the complete transactions.  :)
			if (bFINAL==true) {
				var ezal = ANNE.A.ezAlert(msgDiv, "bg-green-themeDARK", "", "",  "<b>"+$total+"</b>&nbsp;Transactions Complete");
				ezal.classList.add("txs-complete");
				
			}
			return;
		}

		else if ($cmdObj.skipped!=undefined && $cmdObj.skipped==true) {
			var msg = "skipped::";
			if ($cmdObj.skipped_reason!=undefined) msg += $cmdObj.skipped_reason;
			var el = document.getElementById('tx_'+$current_index+'x');
			if (el==null) return;

			msgDiv.scrollTop = el.offsetTop-400; //msgDiv.scrollHeight;
		
			var tx_alert = ANNE.A.addRemoveClass(el, "bg-red", "bg-txid-alert-q");
		}
	}



	/*** printRelonByID ************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/ 
	printRelonByID=function($id) {
		var rObj = MEMG.lookupRelonByTXID($id);
		if (rObj) this.printRelon(rObj);
	}

	//here we just want the names.
	printNeurons = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log($nObjs[n].bestNameHTML());
		}
	}

	printNObjsList = function($nObjs) {
		for (var n=0; n < $nObjs.length; n++) { console.log($nObjs[n].bestName());}
	}

	/*** printNObjs ****************************************** xx/xx/202x ***
	******************************************************** Mr.Scatman ****/
	printNObjs = function ($nObjs) {
		for (var n=0; n < $nObjs.length; n++) {
			console.log("***********************************");
			var nObj = $nObjs[n];
			console.log("for neuron: "+nObj.nid);
			if (nObj.mine==true) console.log("mine=true!");
			else console.log("not mine");

			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printFromTempNidArray=function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log("for neuron: "+nObj.nid);
			if (nObj.temp==true) console.log("temp neuron");
			if (nObj.ghost==true) console.log("local ghost neuron");
			if (nObj.preloaded==true) console.log("preloaded");
			if (nObj.emulated==true) console.log("emulated neuron");
			if (nObj.owned_by!=undefined) console.log("owned by "+nObj.owned_by);
			if (nObj._gt!=undefined) {
				console.log("gt "+MEMG.getNeuronByTempNid(nObj._gt.t).text);
				this.printRelon(nObj._gt);
			}
			this.printRelons(nObj.outs);
		}
	}


	printOnlyRelons = function($TYPE, $aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		for (var r = 0; r < $aRelons.length; r++) {
			if ($aRelons[r].type!=$TYPE) continue;
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
	}

	/**
	 * Print an array of relons.
	 * 
	 * @param {Array} $aRelons 
	 * @param {Boolean} $plainText 
	 * @param {Boolean} $bDOTEMPNIDS 
	 * @returns 
	 */
	printRelons = function ($aRelons, $plainText=true, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if (ANNE.UTILS.isset($aRelons)==false) {
			console.log("no relons to print");
			return;
		} 

		console.log($aRelons);

		for (var r = 0; r < $aRelons.length; r++) {
			this.printRelon($aRelons[r], $plainText, $bDOTEMPNIDS, $consoleColor);
		}
		return;
	}


	printTNIDS = function($tnids){
		var texts=[];
		for (var t=0; t < $tnids.length; t++) {
			var nOBJ = MEMG.getNeuronByTempNid($tnids[t]);
			console.log(nOBJ.text);
			texts.push(nOBJ.text);
		}
		console.log(texts.join(","));
		return texts.join(",");
	}



	/*** selfRelon ******************************************* xx/xx/2020 ***
	******************************************************** Mr.Scatman ****/ 
	selfRelon = function ($relon, $bYOU=true) {
		var relnNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var toObj = MEMG.getNeuronByTempNid($relon.t);
		var totxt="";
		if (toObj!=null) totxt = toObj.text;
		else totxt = $relon.to_s;

		var your = "Your";
		if ($bYOU==false) your = "Their";
		var you = "You";
		if ($bYOU==false) you = "They";

		var msg = "<span class='text-primary'> "+you+"</span>";
		if ($relon.type==TYPE_AWARENESS) {
			
			var toTxt = "";
			if ($relon.to_type==TO_TYPE_ANNE) {
				toTxt = this.verbalizeForAwareness($relon);
			}
			else toTxt = $relon.to_s;
			return your + " "+relnTxt + " "+toTxt;
		}

		else if ($relon.type==TYPE_BE) {
			if ($relon.to_type==TO_TYPE_STRING) {
				var your = "<span class='text-primary'> "+your+"</span>";
				if ($relon.r==MEM.name.temp_nid) return your+" <span class='text-danger'>name </span>is <span class='text-success-lbl'>"+totxt+"</span>";
				else if ($relon.r==MEM.aka.temp_nid) return msg+" are <span class='text-danger'>also known as </span>is <span class='text-success-lbl'>"+totxt+"</span>";
			}
			msg = msg +" are ";
			var determiner = this.getDet(totxt);
			return msg +determiner+" <span class='text-danger'>"+totxt+"</span>";
		}
		else if ($relon.type==TYPE_SYS) {
			if (toObj!=null) {
				if ($relon.r==MEM.necrosis.temp_nid) {
					"<span class='text-danger-lbl'>"+relnNOBJ.text+" (delete) </span> <span class='text-success'>"+totxt+"</span>";;
				}
				else return relnNOBJ.text+" = "+totxt;
			}
			else return relnNOBJ.text+" = "+$relon.to_s; 
		}
		else if ($relon.type==TYPE_HAS) {
			msg = msg+" have ";	
			var determiner = this.getDet(relnNOBJ.text);
			var reln_txt = "<span class='text-danger'>"+relnNOBJ.text+"</span>";
			msg = msg+determiner+" "+reln_txt+" of ";

			var toStr = "";
			if (toObj!=null) toStr = totxt;
			else toStr = $relon.to_s;

			var to_txt = ""
			if (toObj!=null && toObj.bTERMINATING_TYPE==true) to_txt = "<span class='text-success'>"+toStr+"</span>";
			else to_txt = "<span class='text-danger'>"+toStr+"</span>";
			return msg + to_txt;
		}
	}


	printListByReln = function($outs, $r, $excludeTypes=[TYPE_PARAM], $sep=", ") {
		var relons = MEMG.getRelonsByTempR($outs, $r, $excludeTypes);
		if (relons.length==0) return "";
		return MEMG.getTextListByRelonTos(relons, $sep);
	}


	colorizeByPol=function($nObj) {
		if ($nObj.polarity==undefined) MEM.polarity.set_value($nObj);
		var fclass = "";
		if ($nObj.polarity==1) fclass = "text-success-lbl";
		else if ($nObj.polarity==-1) fclass = "text-danger-lbl";
		else if ($nObj.polarity==0) fclass = "text-warning-lbl";
		return "<span class='q-part "+fclass+"'>"+$nObj.text+"</span>";
	}


	printEzName = function($nObj) {
		var text = $nObj.text;
		var be = $nObj.be;
		if (text=="") {
			return "<em>(unnamed) "+be+"</em> <span class='raw-address'>"+$nObj.nid+"</span>";
		}

		return "<em>"+be+"</em> <span style='color:black; font-weight:bold'>"+text+"</span>&nbsp;(<span class='raw-address'>"+$nObj.nid+"</span>)";
	}



	/**
	 * Get a determiner for a given string.
	 * 
	 * Have to go to phonemes to get accurate with this.
	 * Not doing that now.  going the cheap and inaccurate road.
	 * 
	 * @author {Mr.Scatman}
	 */ 
	getDet=function ($str) {
		var str =$str.toLowerCase();
		var char = str.substring(0,1);
		var aeiou = "aeiou";
		if (aeiou.indexOf(char)!=-1) return "an";
		return "a";	
	}

	/**
	 * Another rudimentary det function in lieu of using NLP in this app.
	 * (that comes later with a more robust vocab and understanding of lang in the circuits..
	 * ...tho.. technically... we can use the NLP classes to parse now)
	 * @param {*} $nObj 
	 * @returns 
	 */
	getDet2=function ($nObj) {
		var str = $nObj.text.toLowerCase();

		if ($nObj.ntype==NTYPE_INST) {
			if ($nObj.bes.indexOf(MEM.person.temp_nid)!=-1 || $nObj.bes.indexOf(MEM.location.temp_nid)!=-1) return $nObj.text;
			return "the "+$nObj.text;
		}

		//an elephant, an apple, a peach, a car, a dog.
		else {
			var char = str.substring(0,1);
			var aeiou = "aeiou";
			if (aeiou.indexOf(char)!=-1) return "an "+$nObj.text;
			return "a "+$nObj.text;		
		}
	
	}
	


	/** for later.. */
	toSoundEncoding(word) {
		return word.toLowerCase().split('')
		.map(function(c) { return encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */ 
	soundsLike=function($w1, $w2) {
		return this.toSoundEncoding($w2) === this.toSoundEncoding($w2);
	}


	/** for later.. */ 
	toSoundEncoding=function($w) {
		return $w.toLowerCase().split('')
		.map(function(c) {return this.encodeCharacter(c);})
		.join('')
		.replace('ng', 'n')
		.replace('nn', 'n')
		.replace('oo', 'o')
		.replace('oo', 'o')
		.replace('tt', 't')
		.replace('ol', 'l')
		.slice(0,4);
	}
	

	/** for later.. */
	encodeCharacter=function(c) {
		if ('kgq'.indexOf(c) >= 0) return 'k'; // velar stops
		if ('bpfv'.indexOf(c) >= 0) return 'p'; // labials
		if('td'.indexOf(c) >= 0) return 't'; // alveolars
		if ('szjc'.indexOf(c) >= 0) return 's'; // shushes
		if ('mn'.indexOf(c) >= 0) return 'm'; // nasals
		if ('lr'.indexOf(c) >= 0) return 'l'; // glides
		if ('aeiou'.indexOf(c) >= 0) return 'o'; // vowels
		if ('yhw'.indexOf(c) >= 0) return '';
		return '';
	}
	

	/**
	 * Quick n dirty ending printer for frenly day printing of dates.
	 */
	lazyDayIntEnding = function($int) {
		if ($int==1 || $int==21 || $int==31) return $int+"st";
		else if ($int==2 || $int==22) return $int+"nd";
		else if ($int==3 || $int==23) return $int+"rd";
		else return $int+"th";
	}


	/**
	 * poor man's printing -- no intelligence involved.	
	 * quite rudimentary -- but for now, it works.
	 */
	humanReadableRelon = function ($relon, $bNOFONT=false) {
		//redo this.. for now, just print relon.
		return this.printRelon($relon, true, false);
	}


	/**
	 */
	whoOwnsPhrase = function($who, $nObj=null) {
		if ($who=="anne") return "A.N.N.E. owns this neuron";
		else if ($who=="app") return "This app owns this neuron";
		else if ($who=="me") {
			if ($nObj.temp_nid==USER._userNObj.temp_nid) return "<span class='text-primary'>This IS you!</span>";
			else return "<span class='text-primary'>You owns this neuron!</span>";
		}
		else if ($who=="worldview") return "This is in your <span class='text-primary'>worldview</span>";
		
		return "**This neuron is not in your worldview**";
	}


	
	/**
	 */ 
	 verbalizeForAwareness = function ($relon, $bTAGIFY=true) {
		if ($relon.to_type!=TO_TYPE_ANNE) return " [[[don't use this w/ non neuron tos]]] ";
		var toNOBJ = MEMG.getNeuronByTempNid($relon.t);
		var rNOBJ = MEMG.getNeuronByTempNid($relon.r);
		var relnTxt = this.colorizeByPol(rNOBJ);

		//opinionz are just keywords..
		//if (rNOBJ.bes[0]==MEM.symbol.temp_nid) {
			//var determiner = this.getDet(relnTxt);
			//if ($bTAGIFY==false) return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+relnTxt;	
			//return "<span class='type-aware'>AWARE</span><span class='q-part rel-from'>I</span> <span class='q-part'>think </span><span class='q-part'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>is</span> "+relnTxt;	
		//}

		//else 
		if (rNOBJ.nid==MEM.rememberz.nid) {
			if ($bTAGIFY==false) return "rememberz "+toNOBJ.bestNameHTML();
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-reln'>rememberz</span><span class='q-part reln-to'>"+toNOBJ.bestNameHTML()+"</span>";
		}

		//the rest are pos/neg/neu feelings (buckets) or feelz (most refined expression of feeling)
		else {
			if ($bTAGIFY==false) return ""+toNOBJ.bestNameHTML()+" makes me FEELZ "+relnTxt;
			return "<span class='type-aware'>AWARE</span><span class='q-part rel-to'>"+toNOBJ.bestNameHTML()+"</span><span class='q-part'>makes</span><span class='q-part rel-from'>me</span>"+relnTxt;
		}
	}


	/**
	 */
	 printAwarenessForDisplay = function($relnNOBJ, $tnid) {
		var tnid = parseInt($tnid);
		var targetNOBJ = MEMG.getNeuronByTempNid(tnid);
		if (targetNOBJ==null) return "fix this";
		if (targetNOBJ.bes[0]==MEM.opinionz.temp_nid) {
			var determiner = this.getDet($relnNOBJ.text);
			return "I think "+toNOBJ.bestNameHTML()+" is "+determiner+" "+$relnNOBJ.text;	
		}
		else if ($relnNOBJ.nid==MEM.rememberz.nid) return "rememberz "+targetNOBJ.bestNameHTML();	
		return targetNOBJ.bestNameHTML()+" makes me FEELZ "+$relnNOBJ.bestNameHTML();
	}


	//tells us what something is from a text search.
	//this function is used as a priority when searching for things
	//we prioritize YOU first, then ANNE, then temp-worldview-pass (review this), then outside the network.
	verbalizePriorityFromTallySummary=function($tally, $textVal, $parentNObj=null, $prop="name") {
		var who = ""
		var has = "has";
		if ($tally.me) {
			who = "YOU";
			has = "have";
		}
		else if ($tally.anne) who = "A.N.N.E.";
		else if ($tally.app) who = "THIS '"+MEM_UDF.THIS_APP.text.toUpperCase()+"' APP";
		else if ($tally.worldview) who = "An adopted neuron in your network";
		else if ($tally.wv_pass) who = "A temporary-worldview-pass neuron"; 
		else if ($tally.other) who = "Someone outside of your network";
		who = "<span style='color:white'>"+who+"</span>"+"&nbsp;"+has;
		var search_msg = who+"&nbsp;already created a neuron with the same '"+$prop+"' <span class='text-success'>&nbsp;'"+$textVal+"'</span>";
		if ($parentNObj != null) search_msg=search_msg+"&nbsp;with parent class <span class='text-success'>&nbsp;'"+$parentNObj.text+"'</span>";
		msg = "As a friendly reminder;&nbsp;"+search_msg+"!  Are you sure you want to create another one?"; 
		return msg;
	}


	/**
	 */
	 typeTalk=function($type, $bTEXTONLY=false, $invertBG=false) {
		if ($bTEXTONLY) return this.type_pre[parseInt($type)];
		var invert = "";
		if ($invertBG==true) invert="-invert";
        var type_string="";
        if ($type==TYPE_SYS_MSG) type_string="<span class='type-sys"+invert+"'>sys msg</span>";
        else if ($type==TYPE_SYS) type_string="<span class='type-sys"+invert+"'>sys</span>";
        else if ($type==TYPE_BE) type_string="<span class='type-be"+invert+"'>be</span>";
        else if ($type==TYPE_PARAM) type_string="<span class='type-param"+invert+"'>param</span>";
        else if ($type==TYPE_HAS) type_string="<span class='type-has"+invert+"'>has</span>";
        else if ($type==TYPE_AWARENESS) type_string="<span class='type-aware"+invert+"'>aware</span>";
        else if ($type==TYPE_CLUMP) type_string="<span class='type-clump"+invert+"'>clump</span>";
		else if ($type==TYPE_ROLE) type_string="<span class='type-role"+invert+"'>role</span>";
		else if ($type==TYPE_BRAIN) type_string="<span class='type-brain"+invert+"'>brain</span>";
		else if ($type==TYPE_EPISODIC) type_string="<span class='type-episodic"+invert+"'>episodic</span>";
		else if ($type==TYPE_EXPERIENCE) type_string="<span class='type-experience"+invert+"'>experience</span>";
		else if ($type==TYPE_INTERNAL_PAY) type_string="<span class='type-pay"+invert+"'>internal payment</span>";
		else if ($type==TYPE_BROADCAST_PAY) type_string="<span class='type-pay"+invert+"'>broadcast payment</span>";
		else if ($type==TYPE_BRAIN120) type_string="<span class='type-brain120"+invert+"'>brain120 backlink</span>";
		else type_string = "unhandled ("+$type+")";
		return type_string;
    }


	/**
	 * 
	 * @param {*} $type 
	 * @param {*} $fromNOBJ 
	 * @param {*} $relnNOBJ 
	 * @param {*} $toType 
	 * @param {*} $toText 
	 */
	quickRelonLinePrint=function($type=null, $fromNOBJ=null, $relnNOBJ=null, $toType=null, $toText=null) {
		var parts = [];
		if ($type!=null) parts.push("["+ANNE.TALK.typeTalk($type)+"]");
		if ($fromNOBJ!=null) parts.push($fromNOBJ.bestNameHTML()+" ->");
		if ($relnNOBJ!=null) parts.push($relnNOBJ.bestNameHTML());
		if ($toType!=null) parts.push("["+TALK.toTypeTalk($toType)+"] ->");
		if ($toText!=null) parts.push($toText);
        return parts.join(" ");
	}

	
	/**
	 */
    toTypeTalk=function($to_type, $bI=false) {
        var to_type_string="";
        if($to_type==TO_TYPE_SEED) to_type_string="encrypted seed";
        else if($to_type==TO_TYPE_ANNE) to_type_string="address";
        else if($to_type==TO_TYPE_TXID) to_type_string="txid";
        else if($to_type==TO_TYPE_URL) to_type_string="url";
        else if($to_type==TO_TYPE_NUMBER) to_type_string="number";
        else if($to_type==TO_TYPE_DATE) to_type_string="date";
        else if($to_type==TO_TYPE_DATETIME) to_type_string="datetime";
        else if($to_type==TO_TYPE_STRING) to_type_string="string";
        else if($to_type==TO_TYPE_BLOB) to_type_string="blob";
        else if($to_type==TO_TYPE_TIME) to_type_string="time";
        else if($to_type==TO_TYPE_BOOLEAN) to_type_string="boolean";
        else if($to_type==TO_TYPE_ENCR_ANNE) to_type_string="encr addr";
        else if($to_type==TO_TYPE_ENCR_TXID) to_type_string="encr txid";
        else if($to_type==TO_TYPE_ENCR_URL) to_type_string="encr url";
        else if($to_type==TO_TYPE_ENCR_NUMBER) to_type_string="encr #";
        else if($to_type==TO_TYPE_ENCR_DATE) to_type_string="encr date";
        else if($to_type==TO_TYPE_ENCR_DATETIME) to_type_string="encr dtime";
        else if($to_type==TO_TYPE_ENCR_STRING) to_type_string="encr str";
        else if($to_type==TO_TYPE_ENCR_BLOB) to_type_string="encr blob";
        else if($to_type==TO_TYPE_ENCR_TIME) to_type_string="encr time";
        else if($to_type==TO_TYPE_ENCR_BOOLEAN) to_type_string="encr bool";
        

        if($bI==true) return "<i>"+to_type_string+"</i>";
        else return to_type_string;
    }


	/**
	 */
	use_d_pat=function($d_pat, $nObj, $bDOLINKS=true, $linkType=LINKS_NEURON_ANNET) {
		if ($d_pat==undefined || $d_pat==null || $d_pat.length==0) return;
		var msg = "";
		for (var d=0; d < $d_pat.length; d++) {
			var pat = $d_pat[d];
			var prop = pat.prop;
			var pre = pat.pre;
			var post = pat.post;
			var bALT = false;
			if (prop==undefined) continue;
			var val = $nObj[prop];
			if (val==undefined || val==null) {
				if (pat.alt_lbl!=undefined) {
					val = pat.alt_lbl;
					bALT = true;
				}
				else continue;
			}
			if ($bDOLINKS==true && bALT==false) {
				var valNOBJ = MEMG.getNeuronByTempNid(val);
				if (valNOBJ==null) {
					if (pat.cancel_prev!=undefined) return "";
					continue;
				}
				val = LINKS.makeNeuronLink(valNOBJ, $linkType);
				if (pat.show_aid!=undefined && pat.show_aid==true) {
					val = val +"&nbsp;("+valNOBJ.nid+")";
				}
			}

			if (pre!="") {
				if (msg!="") msg = msg+"&nbsp;"+pre;
				else msg = pre;
			}
			if (msg=="") msg = val;
			else msg = msg+"&nbsp;"+val;
			
			if (post!="") msg = msg+"&nbsp;"+post;
		}

		return msg.trim();
	}

	printTempNids = function($tnids) {
		for (var t=0; t < $tnids.length; t++) {
			var nObj = MEMG.getNeuronByTempNid($tnids[t]);
			if (nObj==null) continue;
			console.log(nObj.text, nObj);
		}
	}

	/**
	 * plaintext no longer does anything.
	 * this function is a DISASTER, fay.  -dlm
	 */
	printRelon = function ($rObj, $plainText=false, $bDOTEMPNIDS=false, $consoleColor=-1) {
		if ($rObj==null) {
			console.log("silly wabbit, you have to have a relon to print a relon");
			return;
		}

		//console.log($rObj);
		var typeText = this.getTypeName($rObj.type);
		var fromNOBJ = MEMG.getNeuronByNid($rObj.from_nid);	//all should be indexed
		if (fromNOBJ==null) {
			return;
		}

		var emulated = "";
		var fromText = fromNOBJ.text;
		if (fromText=="") fromText = fromNOBJ.nid;
		var relnNOBJ = MEMG.getNeuronByNid($rObj.reln);
		var relonText = relnNOBJ.text;
		if ( fromNOBJ.emulated) emulated = "(emulated) ";
		var relonStr = emulated+typeText + " from:" + fromText+ "  reln:"+relonText;
		var toText = this.verbalizeRelonToByToType($rObj);
		relonStr += " to:"+toText;

		var toSPrintOnly = this.toSPrint($rObj);

		var freeze = fromNOBJ.freeze;
		if (freeze==1) console.log("%cfrom nid froze", TALK.bgivoryblue);

		var ghost = "";
		if ($rObj.ghost==true) ghost="(gh)";

		var sysreftxid=""
		if ($rObj.sys_ref_txid!=undefined) sysreftxid=" sysref:"+$rObj.sys_ref_txid;

		var annies=""
		if ($rObj.annies!=undefined) annies=" annies:"+$rObj.annies;
		
		var fhi="";
		if ($rObj.flathexint!=undefined) fhi=" fhi:"+$rObj.flathexint;
		
		if ($consoleColor!=-1) {
			console.log("%cg"+ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi, $consoleColor);
		}
		else {
			console.log(ghost+"txid:"+$rObj.txid+"["+typeText+"]"+"["+fromText+"]"+"["+relonText+"]"+"["+toSPrintOnly+"] "+sysreftxid+annies+fhi);
		}

		if ($plainText==true) return emulated+"[" + typeText+"]"+fromText+relonText+toSPrintOnly;
		return relonStr;
	}


	toSPrint=function($rObj){
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, false);
		return toText;
	}


	/**
	 */
	verbalizeRelonToByToType = function ($rObj, $bCOLORIZEBYTYPE=true) {
		if ($rObj.pending_to_s!=undefined && $rObj.to_s==null) {
			if ($rObj.pending_to_debug!=undefined) return "*pending* "+$rObj.pending_to_debug;
			else return "pending nid/txid";
		}
		var toText = this.getToTextByToType($rObj.to_type, $rObj.to_s, $bCOLORIZEBYTYPE);
		return toText;
	}


	/**
	 */ 
	getToTextByToType = function ($to_type, $to_s, $colorize=true) {
		var text = $to_s;
		if ($to_type == TO_TYPE_ANNE) {
			var nObj = MEMG.getNeuronByNid($to_s);
			if (nObj) {
				if ($colorize==true) text = nObj.bestNameHTML();
				else text = nObj.bestName();
			}	
		}
		if ($colorize) return "<span class='to-type tt-"+$to_type+"'>"+text+"</span>";
		else return text;
	}


	/**
	 */
	getTypeName = function ($type, $bDOCLASSHTML=false, $addClass="") {
		if ($bDOCLASSHTML==false) {
			if ($type == TYPE_SYS || $type=="sys") return 'SYS';
			if ($type == TYPE_BE || $type=="be") return 'BE';
			if ($type == TYPE_PARAM || $type=="para") return 'PARAM';
			if ($type == TYPE_HAS || $type=="has") return 'HAS';
			if ($type == TYPE_AWARENESS || $type=="awareness") return 'AWARENESS';
			if ($type == TYPE_SYS_MSG || $type=="msg") return  'SYS MSG';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return  'EXPER';
			if ($type == TYPE_CLUMP || $type=="clump") return  'CLUMP';
			if ($type == TYPE_ROLE || $type=="role") return  'ROLE';
			if ($type == TYPE_BRAIN || $type=="brain") return  'BRAIN';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return  'INTERNAL PAYMENT';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return  'BROADCAST PAYMENT';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return  'BRAIN120';
			
			
		}
		else {
			if ($type == TYPE_SYS || $type=="sys") return '<span class="type-sys '+$addClass+'">SYS</span>';
			if ($type == TYPE_BE || $type=="be") return '<span class="type-be '+$addClass+'">BE</span>';
			if ($type == TYPE_PARAM || $type=="param") return '<span class="type-param '+$addClass+'">PARAM</span>';
			if ($type == TYPE_HAS || $type=="has") return '<span class="type-has '+$addClass+'">HAS</span>';
			if ($type == TYPE_AWARENESS || $type=="awareness") return '<span class="type-awareness '+$addClass+'">AWARE</span>';
			if ($type == TYPE_SYS_MSG || $type=="msg") return '<span class="type-sys-msg '+$addClass+'">SYS MSG</span>';
			if ($type == TYPE_EXPERIENCE || $type=="exper") return '<span class="type-experience '+$addClass+'">EXPER</span>';
			if ($type == TYPE_CLUMP || $type=="clump") return '<span class="type-clump '+$addClass+'">CLUMP</span>';
			if ($type == TYPE_ROLE || $type=="role") return '<span class="type-role '+$addClass+'">ROLE</span>';
			if ($type == TYPE_BRAIN || $type=="brain") return '<span class="type-brain '+$addClass+'">BRAIN</span>';
			if ($type == TYPE_INTERNAL_PAY || $type=="internal payment") return '<span class="type-brain '+$addClass+'">INTERNAL PAYMENT</span>';
			if ($type == TYPE_BROADCAST_PAY || $type=="broadcast payment") return '<span class="type-brain '+$addClass+'">BROADCAST PAYMENT</span>';
			if ($type == TYPE_BRAIN120 || $type=="brain120") return '<span class="type-brain '+$addClass+'">BRAIN120</span>';
			
		}

		//anything 100 and over is typebrain 0, 1, 2 etc
		if ($type>=20 && $type<120) return "BRAIN"+(parseInt($type));
		return 'unknown '+$type;
	}


	/**
	 */
	pluralizeNounIf = function($mixed, $baseWord, $plural, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $baseWord+$plural;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $baseWord+$plural;
		return $baseWord+$singularPostFix;
	}

	pluralizeNounIf2 = function($mixed, $baseWord, $pluralwd, $singularPostFix="") {
		if ((Array.isArray($mixed)==true) && ($mixed.length>1 || $mixed.length==0)) return $pluralwd;
		else if (typeof $mixed==="number" && ($mixed>1 || $mixed==0)) return $pluralwd;
		return $baseWord+$singularPostFix;
	}

	/**
	 */
	pluralizeNP = function ($arr, $baseWord, $plural, $conj="or", $nestedProp="content") {
		if ($arr.length>1) return $baseWord+$plural+" "+this.listify($arr, $conj, $nestedProp);
		else  {
			if ($nestedProp!="") return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0][$nestedProp].text+"</span>";
			else return $baseWord+"&nbsp;<span class='text-success-lbl'>"+$arr[0].text+"</span>";
		}
	}


	/**
	 */
	listify=function($arr, $conj, $nestedProp="content"){
		var list = "";
		for (var a=0; a < $arr.length; a++) {
			var sep = "";
			if ((a+1)==$arr.length) sep = "</span>&nbsp"+$conj+"&nbsp;<span class='text-success-lbl'>";
			else if (a>=1) sep = ",&nbsp;";
			else list = "<span class='text-success-lbl'>";
			
			if ($nestedProp) list += sep+""+$arr[a][$nestedProp].text+"";
			else list += sep+"<b>"+$arr[a].text+"</b>";
		}
		return list+"</span>";
	}


	/**
	 */
	prounounPick=function($arr, $pr, $prpl) {
		if ($arr.length>1) return $prpl;
		else return $pr;
	}


	//To make regular nouns plural, add ‑s to the end. ...
	//If the singular noun ends in ‑s, -ss, -sh, -ch, -x, or -z, add ‑es to the end to make it plural. ...
	//In some cases, singular nouns ending in -s or -z, require that you double the -s or -z prior to adding the -es for pluralization.
	pluralizeNoun=function($text) {
		var rev = $text.split("").reverse().join("");
		var c0 = rev[0];
		var c1 = "";
		if (rev.length>1) c1 = rev[1];
		if (c0=="s" || c0=="x" || c0=="z") {
			if (c0=="s" && c1=="s") return $text+"s";
			if (c0=="s") return $text+"ses";
			if (c0=="z") return $text+"zes";
			return $text+"es";
		}
		if (c0=="h" && (c1=="s" || c1=="c")) return $text+"es";
		return $text + "s";
	}


	/**
	 */
	beState = function($array) {
		if ($array.length==0) return "are";
		else if ($array.length==1) return "is";
		return "are";
	}

	/**
	 * make an english list phrase of "x, y, z and a" from an array of [x, y, z, a]
	 */
	listifyArray=function($arr) {
		if ($arr.length==0) return "";
		if ($arr.length==1) return $arr[0];
		var str = $arr[0];
		for (var a=1; a < $arr.length-1; a++) {str = str+", "+$arr[a];}
		str = str+ " and "+$arr[($arr.length-1)];
		return str;
	}


	/**
	 * Just some helpful dumb msgs popped in here for certain relons.
	 * We can display these in the relons page to inform the user a bit more
	 * about what this relons is/was for.
	 */
	basicExplain = async($rObj)=> {
		var rel_sponsor_msg="";

		if ($rObj.relon_sponsor!=undefined && $rObj.relon_sponsor!=null && $rObj.relon_sponsor!="") {
			rel_sponsor_msg = "  This was originally a collab";
		}

		if ($rObj.from_nid==MEM.___SELF___ && $rObj.reln==MEM.BLOCK_WINNER.nid) {
			return "This is a block reward winner.";
		}
		if ($rObj.type==TYPE_BROADCAST_PAY && $rObj.reln==MEM.annies.nid && $rObj.to_type==TO_TYPE_ANNE) {
			return "This was a broadcast payment relon.";
		}
		if ($rObj.ghost==true && $rObj.to_type>=TO_TYPE_ENCR_ANNE) {
			var fromNOBJ = MEMG.getNeuronByTempNid($rObj.f);
			if (fromNOBJ.ghost==true) return "This to value will be encrypted at run-time when broadcast.";
		}
		if ($rObj.r==MEM.anj_accepted.temp_nid) return "This is an accepted anjucator invite.";
		if ($rObj.r==MEM.anj_rejected.temp_nid) return "This is an declined anjucator invite.";
		if ($rObj.reln==MEM.anjucator.nid) {
			if ($rObj.sys_ref_txid!=undefined && $rObj.sys_ref_txid.length) {
				if ($rObj.to_type==TO_TYPE_ANNE) {
					var srtRel=MEMG.lookupByTXID($rObj.sys_ref_txid);
					if (srtRel!=null) {
						if (srtRel.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to add an anjucator to a neuron.";
						else if (srtRel.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to add an anjucator to a neuron.";
					}
				}
				else if ($rObj.to_type==TO_TYPE_TXID) {
					if ($rObj.r==MEM.anj_rejected.temp_nid) return "This was a rejected request to be added as an anjucator to a neuron.";
					else if ($rObj.r==MEM.anj_accepted.temp_nid) return "This was an accepted request to be added as an anjucator to a neuron.";
				}
			}
			else return "This is a request to add an anjucator to a neuron.";
		}
		if ($rObj.reln==MEM.winner_payout.nid) return "This is a winning payout for the boosties bucket draw."
		if ($rObj.reln==MEM.hanai_in.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is approval of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is approval of a HANAI request.";
		}
		if ($rObj.reln==MEM.hanai_in_reject.nid) {
			if ($rObj.sys_ref_txid==undefined || $rObj.sys_ref_txid==null || $rObj.sys_ref_txid=="") {
				return "This is rejection of a HANAI request.  This TX should be mined shortly and become official.";
			}
			return "This is rejection of a HANAI request.";
		}
        if ($rObj.r==MEM.andy_collab_accepted.temp_nid) return "This is an accepted collab request.";
        if ($rObj.r==MEM.andy_collab_rejected.temp_nid) return "This is a rejected collab request.";
        if ($rObj.r==MEM.hanai_out.temp_nid) {
            var whoNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			return "This is a request to HANAI out a neuron into the <b>"+whoNOBJ.bestNameHTML()+"</b> worldview.";
        }
		if ($rObj.r==MEM.anndle_offer.temp_nid) {
			var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
			var pending = "is a pending";
			if (ANNE.UTILS.isset($rObj.sys_ref_txid)==true) pending = "was an ";
			return "This "+pending+" ANNDLE offer that paid "+UTILS.formatNumber($rObj.annies)+" annies ("+UTILS.formatNumber($rObj.annies/100)+" ANNE coin) to control the '"+toNOBJ.text+"' ANNDLE.";
		}

        if ($rObj.type==TYPE_SYS_MSG) {
			if ($rObj.r==MEM.annexed_by.temp_nid) {

				var arel = null;
				var ANNEXED_THING = null;

				if ($rObj.sys_ref_txid!=null) arel = ANNE.MEMG.lookupByTXID($rObj.sys_ref_txid);
				if (arel!=null) {
					ANNEXED_THING = MEMG.getNeuronByTempNid(arel.t);
				}

				if ($rObj.f==USER._userNObj.nid) {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Your neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "Your neuron was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
				else {
					if (ANNEXED_THING!=null) {
						$rObj.msg = "Neuron <b>"+ANNEXED_THING.bestName()+"</b> was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
					else {
						$rObj.msg = "A neuron from user "+MEMG.getOrMakeIfNew($rObj.to_s)+" was <b>annexed by</b> "+MEMG.getOrMakeIfNew($rObj.to_s).bestName(); 
					}
				}
			}

            //if ($rObj.r==MEM.aon.temp_nid) return "This is an onboarding relon with "+$rObj.annies+" annies given to the recipient.";
            return "This is a relon sent as a <b>message</b> to an ANNODE for instructions to take action.  It can also be generated by the annode on behalf of a keyless neuron and serve as a status update or receipt.";
        }

        else if ($rObj.type==TYPE_SYS) return "This is a <b>system</b> relon used to engage with ANNE protocol rules.";
        else if ($rObj.type==TYPE_BE) {
            if ($rObj.r==MEM.class.temp_nid || $rObj.r==MEM.inst.temp_nid) {
                return "This is a relon that specifies be-hierarchical parent information."+rel_sponsor_msg;
            }
            else if ($rObj.r==MEM.name.temp_nid) return "This is a relon that sets the neuron's name."+rel_sponsor_msg;
            else if ($rObj.r==MEM.aka.temp_nid) return "This is a relon that sets an <b>also known as</b> for this neuron."+rel_sponsor_msg;
            else if ($rObj.r==MEM.fka.temp_nid) return "This is a relon that sets a <b>formerly known as</b> for this neuron."+rel_sponsor_msg;
        	return "This is a relon that specifies additional be-defining information."+rel_sponsor_msg; 
        }
        else if ($rObj.type==TYPE_PARAM) {
            return "This is a <b>parameter</b> relon defining a relationship that instances can provide."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_HAS) {
            if (this.bCOLLAB==true) {
                if (this.bISPENDING==true || ($rObj.status_text!=undefined && $rObj.status_text=="p")) {
					return "This is a collab relon request.";
                }
                return "This is a collab relon request.";
            }
            return "This is a relon which specifies a parameter value a neuron <b>has</b>."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_AWARENESS) {
            return "This is a relon for run-time feelz and opinionz of an identity using ANNE.";
        }
        else if ($rObj.type==TYPE_BRAIN) {
            return "This is a relon used for wiring neurons together in the non-semantic <b>brain</b> dimension."+rel_sponsor_msg;
        }
        else if ($rObj.type==TYPE_CLUMP) return "This is a relon connecting to a more clump neuron."+rel_sponsor_msg;
        else if ($rObj.type==TYPE_ROLE) {
            return "This is a semantic role relon that exists in a clump neuron and defines a semantic role within the clump."+rel_sponsor_msg;
        }
		else if ($rObj.type==TYPE_INTERNAL_PAY) {
			if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block reward won via solo mining by "+toNOBJ.text;
			}
			else if($rObj.f==ANNE.MEM.___SELF___.temp_nid && $rObj.r==MEM.BLOCK_SHARE_WINNER.temp_nid) {
				var toNOBJ = MEMG.getNeuronByTempNid($rObj.t);
				return "This is a block share reward won via share mining by "+toNOBJ.text;
			}
			else if ($rObj.r==ANNE.MEM.boosties.temp_nid) {
				return "This added boosties to this neuron.";
			}
            else return "This is an internal payment from a keyless neuron made by L1 protocol rules.";
        }
		else if ($rObj.type>=TYPE_BRAIN0 && $rObj.type<TYPE_BRAIN120) {
			return "This is a special type to indicate sorting order #"+$rObj.type;
		}
		else if($rObj.type==TYPE_BRAIN120) {
			return "This is a special BRAIN120 relon created by the ANNODE to document backlinks and provide helpful summaries.";
		}
		else if ($rObj.type==TYPE_BRAIN121) {
			return "myelination";
		}
	}


	logFunc=function($funct, $str="", $str2="", $str3="") {
		if ($str3!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2+" %c"+$str3, TALK.funct, TALK.bggreen, TALK.bgblue, TALK.bgyellow);
		else if ($str2!="") console.log("%c("+$funct+") %c"+$str+" %c"+$str2, TALK.funct, TALK.bggreen, TALK.bgblue);
		else if ($str!="") console.log("%c("+$funct+") %c"+$str, TALK.funct, TALK.bggreen);
	}
        
       
        

}


const TALK = new anne_talk ();
TALK.set_init(TALK);




/************** (from file ./js/base/TALK.js)**/ 

var DB_local = {};
DB_local.db = null;
DB_local.dbName = "ANNODE_local";
DB_local.bSUPPORTED = false;
DB_local.version = 6;

DB_local.editName = function($POST) {
    if (DB_local.dbName.indexOf("_"+$POST)!=-1) return;
    DB_local.dbName = DB_local.dbName+"_"+$POST;
}

/**
 * Do we even use full pull history at all anymore in DB LOCAL????
 * @author {Mr.Scatman}
 */
 DB_local.___init=async()=> {
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback();
        return;
    }
    var done = await DB_local.openConnection(DB_local.dbName, DB_local.version);
    return;
}

DB_local.___initMMOnly=async()=>{
    if (DB_local.db!=null) return true;
    if (ANNE.network!=null && ANNE.network=="testnet") {
        DB_local.editName("testnet");
    }
    return new Promise((resolve, reject) => {
        var request = indexedDB.open(DB_local.dbName, DB_local.version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;
            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }
            return resolve(false);  //no load needed
        };
    });
}


DB_local.getDBSize=async()=> {
    var size = await DB_local.getDBTblSize("collab_relons");
    size+= await DB_local.getDBTblSize("fullPullHistory");
    size+= await DB_local.getDBTblSize("ghost_relons");
    size+= await DB_local.getDBTblSize("traditional_txs");
    size+= await DB_local.getDBTblSize("external_requests");
    size+= await DB_local.getDBTblSize("adf_otc_swaps");
    size+= await DB_local.getDBTblSize("bot_trade_history");
    size+= await DB_local.getDBTblSize("otc_commish");
    return size;
}

DB_local.getDBTblSize=async($tbl, $method="readwrite")=> {
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                var arr = event.target.result;
                var json = JSON.stringify(arr);
                size = json.length;
                return resolve(size);
            };
        } 
    });
}

/**
 * We wait until we know what user we are dealing with to get these ones
 */
DB_local.__getGhostsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("ghost_relons");
    return aGhostRelons;
}


/**
 * Table for exensions to store a history of the requests they saw.
 * 
 * Use.  subtype, msg, exernaldata and datestamp.
 * @param {*} $onLoadCallback 
 * @returns 
 */
DB_local.__getExternalRequestsOnly=async($onLoadCallback=null)=>{
    if (DB_local.bSUPPORTED==false) {
        if ($onLoadCallback!=null) $onLoadCallback(null, null, $callbackCallback);
        return [];
    }
    var aGhostRelons = await DB_local.loadAll("external_requests");
    return aGhostRelons;
}


/**
 */
DB_local.__getStoredTxs=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("traditional_txs");
}


DB_local.loadADFSwaps=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aSwaps = await DB_local.loadAll("adf_otc_swaps");
    var aSwapsFinal = [];
    for (var s=0; s < aSwaps.length; s++) {
        if (aSwaps[s].network==$network) aSwapsFinal.push(aSwaps[s]);
    }
    return aSwapsFinal;
}


/**
 * 
 * @param {*} $network 
 * @returns 
 */
DB_local.loadCommishHistory=async($network="mainnet")=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCommishHistory = await DB_local.loadAll("otc_commish");
    var aComisHistoryFinal = [];
    for (var s=0; s < aCommishHistory.length; s++) {
        if (aCommishHistory[s].network==$network) aComisHistoryFinal.push(aSwaps[s]);
    }
    return aComisHistoryFinal;
}



/**
 * 
 * @returns 
 */
DB_local.__getFullPullHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("fullPullHistory");
}


DB_local.__getBotTradeHistory=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    return await DB_local.loadAll("botcoinbalance");
}

/**
 */
 DB_local.__getCollabRelons=async()=>{
    if (DB_local.bSUPPORTED==false) return [];
    var aCollabs = await DB_local.loadAll("collab_relons");
    var ours = [];
    for (var c=0; c < aCollabs.length; c++) {
        if(aCollabs[c].from_nid==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].to_s==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else if(aCollabs[c].collaborator==ANNE.USER._userNObj.nid) ours.push(aCollabs[c]);
        else {
            //do we have right to approve this?
        }
    }
    return ours;
}


/**
 * @author {Mr.Scatman}
 */
DB_local.___deleteEntireDB=async()=> {
    if (DB_local.bSUPPORTED==false) return;
    if (DB_local.db==null) return;
    return new Promise((resolve, reject) => {
        DB_local.db.close();
        var request = indexedDB.deleteDatabase(DB_local.dbName, 2);
        request.onblocked = function () {
            console.log("Delete blocked.");
            return resolve(false);
        }
        request.onerror = function () {
            console.log("Error deleting the DB");
            return resolve(false);
        };
        request.onsuccess = function () {
            console.log("Deleted OK.");
            return resolve(true);
        }   
    });             
}

/**
 * @author {Mr.Scatman}
 */
DB_local.openConnection=async($dbName, $version)=> {
    if (DB_local.db!=null) return true;
    return new Promise((resolve, reject) => {
        var request = indexedDB.open($dbName, $version);
        request.onerror = function(event) {
            console.log("openConnection request.Error");
            return resolve(false);
        };

        request.onsuccess = function(event) {
            DB_local.db = event.target.result;
            return resolve(true);
        };
        
        request.onupgradeneeded = function(event) {
            console.log("openConnection request.onUpradeneeded");
            DB_local.db = event.target.result;

            if (DB_local.db.objectStoreNames.contains('ghost_relons')==false) {
                var ghostObjStore = DB_local.db.createObjectStore("ghost_relons", { keyPath:"txid" });
                ghostObjStore.createIndex("txid", "txid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('fullPullHistory')==false) {
                var fullPullHistoryStore = DB_local.db.createObjectStore("fullPullHistory", { keyPath:"pullid" });
                fullPullHistoryStore.createIndex("pullid", "pullid", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('collab_relons')==false) {
                var collabRelonStore = DB_local.db.createObjectStore("collab_relons", { keyPath:"txid" });
                collabRelonStore.createIndex("txid", "txid", { unique: true });
            }
            
            if (DB_local.db.objectStoreNames.contains('traditional_txs')==false) {
                var ttStore = DB_local.db.createObjectStore("traditional_txs", { keyPath:["txid", "anne_app_user"] });
                ttStore.createIndex("usertxid", ["txid", "anne_app_user", "confirmed"], { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('external_requests')==false) {
                var extReqStore = DB_local.db.createObjectStore("external_requests", { keyPath:"req_id" });
                extReqStore.createIndex("req_id", "req_id", { unique: true });
            }

            if (DB_local.db.objectStoreNames.contains('adf_otc_swaps')==false) {
                var swaps = DB_local.db.createObjectStore("adf_otc_swaps", { keyPath: ["swap_id", "network"] });
                swaps.createIndex("swap_id", "swap_id", { unique: false });
                swaps.createIndex("network", "network", { unique: false });
            }

            //ticker + currentTime = unique key
            if (DB_local.db.objectStoreNames.contains('botcoinbalance')==false) {
                var bottradehistory = DB_local.db.createObjectStore("botcoinbalance", {keyPath:["ticker", "currentTime"]});
                bottradehistory.createIndex("ticker", "ticker", { unique: false });
                bottradehistory.createIndex("currentTime", "currentTime", { unique: false });
            }

            if (DB_local.db.objectStoreNames.contains('otc_commish')==false) {
                var commish = DB_local.db.createObjectStore("otc_commish", { keyPath:"midmanne_commishid" });
                commish.createIndex("midmanne_commishid", "midmanne_commishid", { unique: true });
            }

            return resolve(false);  //no load needed
        };
    });
}




DB_local.loadAll=async($tbl, $method="readwrite")=> {
    if (DB_local.db==null) return;
    if (DB_local.db.transaction==undefined) {
        return;
    }
    
    console.log("DB.loadAll "+$tbl);
    
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {
        //FULL PULL HISTORY RETREIVAL
        if ('getAll' in objectStore) {
            objectStore.getAll().onsuccess = function(event) {
                return resolve(event.target.result);
            };
        } 
        else {
            var arr = [];

            // Fallback to the traditional cursor approach if getAll isn't supported.
            objectStore.openCursor().onsuccess = function(event) {
                var cursor = event.target.result;
                if (cursor) {
                    arr.push(cursor.value);
                    cursor.continue();
                } 
                else return resolve(arr);
            };
        }
    });
}


DB_local.___insert = async($objs, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    if ($objs.length==0) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;

    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;

    return new Promise((resolve, reject) => {    
        $objs.forEach(function($obj) {
            var request = null;

            if ($tbl=="ghost_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                if ($obj.txid==undefined || $obj.txid==null || $obj.txid=="") return;

                //used in the gui but never broadcast.
                var TEMP_ONLY = false;
                if ($obj.TEMP_ONLY==true) {
                    TEMP_ONLY=true;
                }

                var sanitized_obj = {
                    from_nid: $obj.from_nid,
                    reln: $obj.reln,
                    rorigin: $obj.rorigin,
                    time:$obj.time,
                    to_s: $obj.to_s,
                    to_type: $obj.to_type,
                    txid: $obj.txid,
                    type: $obj.type,
                    ghost_order: $obj.ghost_order,
                    ghost_owner: $obj.ghost_owner,
                    owned_by:$obj.ghost_owner,
                    TEMP_ONLY: TEMP_ONLY
                }
                if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;

                request = objectStore.add(sanitized_obj);
            }
   
            //will need collab sanitized stuff too. (TO DO)
            else if ($tbl=="collab_relons") {
                //dont save this relon, EVER
                if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;
                request = objectStore.add($obj);
            }
            
            else {
                console.log("adding INDEXDB:"+$obj);
                request = objectStore.add($obj);
            }
        });

        transaction.onsuccess = function(event) {
            return resolve(true);
        };
        transaction.onerror = function(event) {
            return resolve(false);
        }
        return resolve(true);
    });
}


DB_local.___insert_single = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;

    //dont save this relon, EVER
    if ($obj.reln==MEM.INTRA_BLOCK_PENDING.nid) return;

    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    return objectStore.add($obj);    
}

DB_local.___updateAll = async($objs, $tbl="collab_relons", $method="readwrite")=> {
    for (var r=0; r < $objs.length; r++) {
        await DB_local.___update($objs[r], $tbl, $method);
    }
}

DB_local.___update = async($obj, $tbl="collab_relons", $method="readwrite", $callaback=null)=> {
    if (DB_local.bSUPPORTED==false) return;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    var sanitized_obj  = {};

    //for ghost relons we santitize the obj first.
    if ($tbl=="ghost_relons") {
        var TEMP_ONLY = false;
        if ($obj.TEMP_ONLY==true) {
            TEMP_ONLY=true;
        }
        sanitized_obj = {
            from_nid: $obj.from_nid,
            reln: $obj.reln,
            rorigin: $obj.rorigin,
            time:$obj.time,
            to_s: $obj.to_s,
            to_type: $obj.to_type,
            txid: $obj.txid,
            type: $obj.type,
            ghost_order: $obj.ghost_order,
            ghost_owner: $obj.ghost_owner,
            owned_by:$obj.ghost_owner,
            TEMP_ONLY:TEMP_ONLY
        }
        if ($obj.flathexint!=undefined) sanitized_obj=$obj.fathexint;
        if ($obj.from_text!=undefined) sanitized_obj.from_text = $obj.from_text;
    }

    //all other tbl objects are fine as-is.
    else sanitized_obj = $obj;

    return new Promise((resolve, reject) => {
        var request = objectStore.put(sanitized_obj);
        request.onsuccess = function(event) {
            return resolve(true);
        };
        request.onerror = function(event) {
            console.log(event);
            return resolve(false);
        }
    });
}


DB_local.___delete = async ($ids, $tbl="collab_relons", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($ids.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $ids.length; t++) {
        if ($ids[t]==undefined || (typeof $ids[t]!=='string' && typeof $ids[t]!=='number')) continue;
        await DB_local.__deleteFinal(objectStore, $ids[t]);
    }
}

DB_local.___deleteWithMultiKey = async ($multikey, $tbl="adf_otc_swaps", $method="readwrite", $callback=null)=> {
    if (DB_local.bSUPPORTED==false) return false;
    if ($multikey.length==0) return false;
    var transaction  = DB_local.db.transaction([$tbl], $method);
    if (transaction==null) return;
    var objectStore = transaction.objectStore($tbl);
    if (objectStore==null) return;
    for (var t=0; t < $multikey.length; t++) {
        await DB_local.__deleteFinal(objectStore, $multikey[t]);
    }
}


DB_local.__deleteFinal = async($objectStore, $key)=>{
    return new Promise((resolve, reject) => {
        var request = $objectStore.delete($key);
        request.onsuccess = function(event) { 
            return resolve(true); 
        }
        request.onerror = function(event) { 
            return resolve(false); 
        } 
    });
}

/************** (from file ./js/db/db_local.js)**/ 


/**************************************************************************** */
/**************************************************************************** */
/**************************************************************************** */

var LOCKSET = {};
LOCKSET.init=function() {
    const el_btnYes = document.getElementById("ll_btnYes");
    const el_pin = document.getElementById("ll_pin");
    ANNE.A.hide(el_btnYes);
    el_pin.addEventListener("input", e=>{
        if (el_pin.value.length < 6) {
            ANNE.A.setHTML("ll_warning", "pin must be atleast 6 chars long.", true);
            ANNE.A.hide(el_btnYes);
        }
        else {
            var el_warn = ANNE.A.setHTML("ll_warning", "");
            ANNE.A.hide(el_warn);
            ANNE.A.show(el_btnYes);
        }
    });
    el_btnYes.addEventListener("click", $e=>{
        $e.preventDefault();

        ANNE.USER.unlock_pw = el_pin.value;

        //this tells us we have pin.
        var pinhash = ANNE.USER.md5("mypin_"+ANNE.USER.unlock_pw+"_ishere");

        if (ANNE.USER.APP_KEY=="ANNE") {
            localStorage.setItem("pinhash", pinhash);
            localStorage.setItem('has_lock_pin', "1");
        }
        else {  //custom pinhash for custom one-off app
            localStorage.setItem("pinhash_"+ANNE.USER.APP_KEY, pinhash);
            localStorage.setItem('has_lock_pin_'+ANNE.USER.APP_KEY, "1");
        }
        
        var pinhashbackup = localStorage.getItem("pinhash_backup");
        if (pinhashbackup==null) {
            localStorage.setItem('pinhash_backup', pinhash);
        }

        return ANNE.nextStepHandler("set lock PIN");
    });
}


/**************************************************************************** */
/**************************************************************************** */
/**************************************************************************** */



var UNLOCK = {};
UNLOCK.afterUnlock = null;
UNLOCK._cb = null;

UNLOCK.init = async($lskeytouse)=> {
    //console.log("UNLOCK.init"); 
    const el_pin = document.getElementById("u_pin");
    const btn_unlock = document.getElementById("u_btnYes");
    const unlockBg = document.getElementById("unlockbg");

    btn_unlock.addEventListener('mousedown', function() {
        unlockBg.classList.remove('yellow');
        unlockBg.classList.add('teal');
    });

    btn_unlock.addEventListener('mouseup', function() {
        unlockBg.classList.remove('teal');
        unlockBg.classList.add('yellow');
    });

    btn_unlock.addEventListener('mouseleave', function() {
        unlockBg.classList.remove('teal');
        unlockBg.classList.add('yellow');
    });


    el_pin.addEventListener("input", e=>{
        if (el_pin.value.length>0) {
            btn_unlock.classList.remove("disabled");
        }
        else {
            btn_unlock.classList.add("disabled");
        }
    });

    btn_unlock.addEventListener("click", (e)=>{
        UNLOCK.unlockANNEID(e);
    });
    btn_unlock.classList.add("disabled");

    ANNE.extractURLVars();

    if (ANNE.loadVars!=null && ANNE.loadVars.nid!=undefined) {
        console.log(ANNE.loadVars);
        ANNE.TALK.printf_console(ANNE.TALK.BLUE, "we have a nid to load post login ", ANNE.TALK.GREEN, ANNE.loadVars.nid);
        //nid: "123445"
        //page: "dossier"
        //userfrenly: "dossier nid=123445"
        ANNE.A.setHTML("u_pending_passthru", ANNE.loadVars.userfrenly);
        ANNE.A.hide("u_origin_passthru");
        ANNE.A.hide("u_olbl");
    }

    //legacy method.. no APP_VARS..
    else if (ANNE.MEM_UDF!=null && ANNE.MEM_UDF.loadPassThrus!=null){
        ANNE.A.setHTML("u_pending_passthru", ANNE.MEM_UDF.loadPassThrus.subtype);
        ANNE.A.setHTML("u_origin_passthru", ANNE.MEM_UDF.loadPassThrus.inputdata.origin);
    }
    else if (ANNE.MEM_UDF!=null && ANNE.MEM_UDF.preload_request!=null) {
        console.log(ANNE.MEM_UDF.preload_request);
        const inputdata = ANNE.MEM_UDF.preload_request.inputdata;
        ANNE.A.setHTML("u_pending_passthru", ANNE.preload_request.subtype);
        ANNE.A.setHTML("u_origin_passthru", inputdata.origin);
    }


    else if(ANNE.MEM_UDF!==null && ANNE.MEM_UDF.APP_VARS!=undefined && ANNE.MEM_UDF.APP_VARS!=null 
            && ANNE.MEM_UDF!=null && ANNE.MEM_UDF.APP_VARS.loadPassThrus!=null){
        ANNE.A.setHTML("u_pending_passthru", ANNE.MEM_UDF.APP_VARS.loadPassThrus.subtype);
        ANNE.A.setHTML("u_origin_passthru", ANNE.MEM_UDF.APP_VARS.loadPassThrus.inputdata.origin);
    }
    else if (ANNE.MEM_UDF!==null && ANNE.MEM_UDF.APP_VARS!=undefined && ANNE.MEM_UDF.APP_VARS!=null 
            && ANNE.MEM_UDF!=null && ANNE.MEM_UDF.preload_request!=null) {
        console.log(ANNE.MEM_UDF.preload_request);
        const inputdata = ANNE.MEM_UDF.preload_request.inputdata;
        ANNE.A.setHTML("u_pending_passthru", ANNE.MEM_UDF.preload_request.subtype);
        ANNE.A.setHTML("u_origin_passthru", inputdata.origin);
    }
    else {
        ANNE.A.hide("u_section_passthru");
    }

    //if first open (this session), or first time ever, we have no USER._userNObj yet
    //if (ANNE.USER._userNObj==null || ANNE.USER._userNObj.nid==0) {
        //console.log("no _userNObj set yet");
        //await ANNE.USER.checkLoginFromStorage(false);
    //}


    var user = await ANNE.USER.getLastUser();
    
    //no user, then no name.
    if (user!=null) {
        console.log(USER._userLS);
        if (user.nickname!=undefined && user.nickname!=null && user.nickname!="") {
             ANNE.A.setHTML("u_current_user", user.nickname);
             ANNE.A.addRemoveClass("u_current_user", "name", "nid");
        }
        else {
            ANNE.A.setHTML("u_current_user", user.public_nid);
        }
        return true;
    }
    else {
        ANNE.A.setHTML("u_current_user", "no user currently set");
        return false;
    }
}




/**
 * 
 * @returns 
 */
UNLOCK.unlockANNEID = function($e) {
    $e.preventDefault();
    var el_pin = document.getElementById("u_pin");
    var pw = el_pin.value;

    //there may be no seed in mem.. just use the pinhash.
    var pinhash_test = ANNE.USER.md5("mypin_"+pw+"_ishere");

    var pinhash = "";
    if (ANNE.USER.APP_KEY=="ANNE") pinhash = localStorage.getItem("pinhash");
    else pinhash = localStorage.getItem("pinhash_"+ANNE.USER.APP_KEY);

    if (pinhash==null) {
        console.error("must have pinhash here.");
        return;
    }


    if (pinhash.toUpperCase()==pinhash_test.toUpperCase()) {
        const unlockBg = document.getElementById("unlockbg");
        unlockBg.classList.remove("yellow");
        unlockBg.classList.add("teal");
        ANNE.A.playSound("neonbuzz");

        ANNE.USER._badPWAttempts=0;
        ANNE.USER.unlock_pw = pw;    //we will delete this when we lock, but while ext is open we can use it.
        ANNE.bLOCKED=false;
        if (ANNE.bJUSTLOGGEDIN==true) {
            ANNE.bJUSTLOGGEDIN = false;
        }

        //pretty cheesy to do this here..
        if (ANNE.EXT!=null && ANNE.MEM_UDF!=null) {
            ANNE.MEM_UDF.showLoader("extension unlocked");
        }
        else if (ANNE.MEM_UDF!=null && ANNE.MEM_UDF!=null) {
            ANNE.MEM_UDF.showLoader("App unlocked");
        }
        
        if (UNLOCK._cb!=null) {
            return UNLOCK._cb();
        }
        else if (UNLOCK.afterUnlock!=null 
            && ANNE.MEM_UDF!=null 
            && ANNE.MEM_UDF.handlePageDoneDefault!=undefined) {
            ANNE.A.hide("login_loader");
            ANNE.MEM_UDF.handlePageDoneDefault(UNLOCK.afterUnlock);
            UNLOCK.afterUnlock = null;
            return;
        }
        else ANNE.nextStepHandler("ext unlocked");
    }

    else {
        ANNE.A.playSound("neonbuzz_bad");
        ANNE.bLOCKED=true;
        console.log("no dice");
        ANNE.USER.seedObj = null;
        ANNE.USER.unlock_pw = "";
        ANNE.USER._badPWAttempts+=1;
        el_pin.classList.add("error-shake");

        const unlockBg = document.getElementById("unlockbg");
        unlockBg.classList.remove("yellow");
        unlockBg.classList.add("orange");
        
        el_pin.style.animation = 'none';
        el_pin.offsetHeight; /* trigger reflow */
        el_pin.style.animation = null; 
        return false;
    }
}


/**************************************************************************** */
/**************************************************************************** */
/**************************************************************************** */

var LOGIN_SOLO = {};

/**
 */
LOGIN_SOLO.init = async()=> {
    const el_sa_seed = document.getElementById("sa_seed")
    el_sa_seed.addEventListener("input", LOGIN_SOLO.loginkeytype);
    el_sa_seed.focus();
    
    const btn_make_new_key = document.getElementById("sa_btnNew");
    btn_make_new_key.addEventListener("click", e=>LOGIN_SOLO.genRandSeed(e));
    const btn_l_btnSet = document.getElementById("sa_btnSet");

    //once we have account we can run 
    btn_l_btnSet.addEventListener("click", async($e)=>{
        ANNE.A.hide(btn_l_btnSet);
        if ($e && $e.preventDefault!=undefined) $e.preventDefault();
        var seed = ANNE.A.getValue("sa_seed");
        await ANNE.USER.setSeed(seed);  //make sure we add acct into  ANNE._userAcctsAllJO
        if (ANNE.MEM_UDF!=null) ANNE.MEM_UDF.hasANNEID(true);
        else if (ANNE.hasANNEID!=null) ANNE.hasANNEID(true);
    });
    ANNE.A.hide(btn_l_btnSet);

    var users = await ANNE.getAcctsArray();
    if (users.length>0) {
        ANNE.A.makeHTMLDD_fromUserObjs("240px", "sa_div_acctsdd", users, {defaultLbl:"relogin w/ existing", cb:LOGIN_SOLO.accountDDChanged});
    }
    else {
        ANNE.A.hide("sa_div_acctsdd");
    }

    //allow for quick jump to backup/restore?
    var elbackup = document.getElementById("sa_backup");
    elbackup.addEventListener("click", (e)=>{
        e.preventDefault();
        e.stopPropagation();
        ANNE.MEM_UDF.loadBackup();
    })
}


/**
 * 
 * @param {*} $nid 
 */
LOGIN_SOLO.accountDDChanged=async($nid)=> {
    console.log("(accountDDChanged)");
    var bGOOD = await ANNE.MEM_UDF.chgAcctRuntime($nid);
    if (bGOOD==true) {
        if (typeof ANNE.MEM_UDF.show_balance_bar === "function") {
            ANNE.MEM_UDF.show_balance_bar();
        }
        ANNE.MEM_UDF.loadProfile();
    }
}


/**
 * 
 */
LOGIN_SOLO.loginkeytype=function() {
    var s = document.getElementById("sa_seed").value.trim();
    ANNE.A.hide("sa_seed_section");
    LOGIN_SOLO.checkSeed(s);
}


/**
 * TO DO:  
 *      The 12 words thing ?
 */ 
LOGIN_SOLO.genRandSeed = async($e)=>{
    $e.preventDefault();
    ANNE.A.show("sa_seed_section");
    ANNE.A.setHTML("sa_btnNew", "gen a different random seed")
    ANNE.USER._bNEWUSER=true;
    ANNE._bFIRST_TIME_LOGIN = true;
    var neuron = await USER.makeSeed();
    ANNE.A.setValue("sa_seed", neuron.seed);
    ANNE.A.setHTML("sa_display_seed", neuron.seed);
    ANNE.A.copyIco("sa_seed_copy", neuron.seed);
    ANNE.A.copyIco("sa_nid_copy", neuron.nid);
    ANNE.A.setHTML("sa_nid", neuron.nid);
    ANNE.A.show("sa_btnSet");
    ANNE.A.show("sa_display_section");

    //we now use just the unlock pw for the entire extension.
    ANNE.A.setHTML("sa_btnSet","&nbsp;create new account&nbsp;");
    return;
}


/**
 * 
 * @param {*} $k 
 */
LOGIN_SOLO.checkSeed = async ($s)=> {
    var rez = USER.isValidSeedSOLO($s);
    if (rez.status=="good") {
        var neuron = await USER.makeSeedSOLO($s);
        ANNE.A.setHTML("sa_nid", neuron.nid);
        ANNE.A.setHTML("sa_btnSet","&nbsp;create new account&nbsp;");
        ANNE.A.hide("sa_warning_msg");
        ANNE.A.show("sa_btnSet");
        ANNE.A.show("sa_display_section");
    }
    else {
        ANNE.A.setHTML("sa_warning_msg", rez.reason, true);
        ANNE.A.hide("sa_display_section");
    }
    ANNE.bFRESHLOGIN = true;
}



/**
 * 
 */
LOGIN_SOLO.destroyAll = function() {    
    ANNE.A.hide('login_form');
    document.getElementById("sa_seed").removeEventListener("input", LOGIN_SOLO.loginkeytype);
}



/************** (from file ./js/base/LOGIN_SOLO.js)**/ 

var ALTDATA = {};


ALTDATA.encryptJSON = async($json, $pubkey)=>{
    if ($string==undefined ||  $string==null) {
        console.warn("fix json");
        return;
    }
    if ($pubkey==undefind || $pubkey==null) {
        console.warn("fix pubkey");
        return;
    }

    if (ANNE.USER.priv_key=="") {
        if (ANNE.USER.seedObj==null) {
            if (ANNE.USER.unlock_pw!="" && ANNE.USER.unlock_pw!=null) {
                ANNE.USER.seedObj= await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
            }
            else {
                alert("we need the unlock_pw, the priv_key, or the seedObj set to continue");
                return null;
            }
        }
    }
    else {
        var eObj = await ANNE.USER.encryptMsgForPUBKEY($json.stringify(), $pubkey, ANNE.USER.priv_key);
    }
    var eObj = await ANNE.USER.encryptMsgForPUBKEY($json.stringify(), $pubkey, aO.priv_key);


    if (eObj!=null) return {data_string:eObj.data_string, nonce_string:eObj.nonce_string};
}


/**
 * returns:
 *      {.data_string .nonce_string}
 * @param {*} $string 
 * @param {*} $pubkey 
 * @returns 
 */
ALTDATA.encryptPlain = async($string, $pubkey)=>{
    if ($string==undefined ||  $string==null) {
        console.warn("fix string");
        return;
    }
    if ($pubkey==undefined || $pubkey==null) {
        console.warn("fix pubkey");
        return;
    }
    
    if (ANNE.USER.privKey==null || ANNE.USER.privKey=="") {
        if (ANNE.USER.seedObj==null) {
            if (ANNE.USER.unlock_pw!="" && ANNE.USER.unlock_pw!=null) {
                ANNE.USER.seedObj= await ANNE.USER.getSeedWithPW(ANNE.USER.unlock_pw);
            }
            else {
                alert("we need the unlock_pw, the privKey, or the seedObj set to continue");
                return null;
            }
        }
    }
    else {
        var eObj = await ANNE.USER.encryptMsgForPUBKEY("thisisomebullshit"+$string, $pubkey, ANNE.USER.privKey);
    }
    
    var aO = ANNE.USER.prepANNEAuthObjFromSeed(ANNE.USER.seedObj.seed);
    if (aO==null) {
        alert("problem with private key of anne user account.");
        return null;
    }
    var eObj = await ANNE.USER.encryptMsgForPUBKEY($string, $pubkey, aO.priv_key);
    if (eObj!=null) return {data_string:eObj.data_string, nonce_string:eObj.nonce_string};
}

/**
 * Sign the payload for any coms that should require signed payload.
 * 
 * NOTE: if you change a com to be signMsg it potentially change the order of params
 * that hit the alt-data bash handler.  
 * @param {*} $json 
 * @returns 
 */
ALTDATA.signMsg=async($json)=>{
    var seedObj = null;
    if (USER.seedObj==null) {
        if (USER.unlock_pw=="" || USER.unlock_pw==null) {
            ANNE.USER.lockApp();
            return null;
        }
        seedObj = await USER.getSeedWithPW(USER.unlock_pw);
    }
    else seedObj = USER.seedObj;

    if (seedObj==null) {
        //this.recapBad("problem extracting key from memory.  Can't encrypt message.");
        return null;
    }
    var aO = USER.prepANNEAuthObjFromSeed(seedObj.seed);
    if (aO==null) {
        //this.recapBad("problem extracting key from memory.  Can't encrypt message.");
        return null;
    }

    if (USER.privKey=="" || USER.privKey==undefined) {
        USER.privKey = aO.sign_key;
    }

    var sig=null;
    

    //if we are signning the payload, make sure the zreq_sig flag is on.
    //this will force the alt-data to verify the signature.
    //in the bash that handled various altdata subtypes, the signature flag can be 
    //looked for an ensured to be on -- thus guaranteeing the values were not falsified.
    if ($json.zreq_sig!=true) {
        $json.zreq_sig=true;
    }


    //slip in zzzzzz_newpubkey for every request now.
    if (ANNE.USER._userLS==null || ANNE.USER.isActivated()==false) {
        $json.zzzzzz_newpubkey=ANNE.USER._userLS.pubKey;
    }

    var payloadString = JSON.stringify($json);

    //gen the signature
    try {
        var aObj = await ANNE.USER.getPrivateKeyHex();
        if (aObj==null) {
            console.warn("no private key hex found");
            return null;
        }
        sig = await ANNE.USER.signMessage(payloadString, aObj.sign_key);
        return {sig:sig, payloadstring:payloadString, pubkey:ANNE.USER._userLS.pubKey};
    }

    catch(e) {
        console.log("problem");
        return null;
    }
}

/************** (from file ./js/altdata/altdata.js)**/